#ifndef MESSAGE_H
#define MESSAGE_H

#include <sys/time.h>
#include <arpa/inet.h>
#include <string.h>

const short tpong_version               = 139;
const unsigned long MSG_SIGNATURE_TMON  = 0x35891246;
const unsigned long MSG_SIGNATURE_PONG  = 0x24851273;

const int MAX_MESSAGE_SIZE          = 800;
const int MAX_TMON_PATHS            = 20;   // The maximum number of paths a node can participate in
const int MAX_FRP_HOST              = MAX_TMON_PATHS;   // Maximum number of hosts for fast-rate probing
const int TIMEOUT                   = 15;   // Wait 15 seconds before resend
const int MAX_RETRY                 = 3;    // Resend 3 times before abort
const int TH_CSKEW_EXTENT           = 20;   // Threshold of clock skew extent above which it will be reported to control node
const int TH_CSKEW_SUPPRESS_EXTENT  = 30;   // Threshold of clock skew extent above which the path will be suppressed unless Pong is running on this path
const float TH_LMS                  = 3.0;  // Threshold of minimum link measurability score
const int PATH_BITMAP_SIZE          = 125;  // 125 Bytes, i.e. 1000 bits
const float CL_EDSUM_DECAY_FACTOR   = 0.98; // Decay factor of congestion level edsum

#define TECH_NA     0
#define TECH_2P     1
#define TECH_FSB    2
#define TECH_FSD    3
#define TECH_4P     4

const char * const tech_code[] = { "n/a", "2P", "fsb", "fsd", "4P" };

// Messages sent but not yet acknowledged
struct PendingMessage
{
    time_t last_try;
    int times_tried;
    int dID;
    int timeout;
    const char * message;
};


// Path digest
struct PathDigest
{
    long digest1;
    long digest2;

    void Compute( int num, in_addr * ip_list );
    bool operator == ( const PathDigest& rhs ) const;
};


// Header of UDP report message
struct MsgHeader
{
    unsigned long signature;    // Message signature
    short nID;      // Node id
    short type;     // Message type code
    unsigned char flag;     // Flag bits
    long id;        // Message id
    short len;      // Payload length
};


#define FLAG_ACK        0x01
#define FLAG_NEED_ACK   0x02
#define FLAG_RESEND     0x04

// Type code of UDP report message

#define MSG_TOPO        1   // Topology update
// Payload format: | version | dest ID | number of nodes | path digest | route change count | route change stop flag | IP of source | IP of hop 1 | IP of hop 2 | ... | IP of destination |
#define MSG_TOPO_ACK    2   // Ack of topology update
// Payload format: | version | dest ID |
#define MSG_E2EC        3   // End-to-end congestion event
// Payload format: | dest ID | start time | duration | forward CI | backward CI | sample count | min_f update count | min_b update count | route change count | congestion level edsum f | congestion level edsum b |
//#define MSG_TMON        4   // Indicate this is a TMon process
// Payload format: empty
#define MSG_EDT_TMON    5   // COMMAND: Edit(add, remove or change) a TMon destination host
// Payload format: | dID to add | dID to remove | TMon host list |
#define MSG_TMON_LIST   6   // List all TMon destination hosts
// Payload format: | number of hosts | host id 1 | host id 2 | ... |
#define MSG_ALIVE       7   // Keep-alive message (also used for clock offset calculation)
// Payload format: | version | timestamp | Pong active | Pong state | path bitmap |
#define MSG_ALIVE_ACK   8   // Ack of keep-alive message
// Payload format: | version | clock offset |
#define MSG_CSKEW       9   // Clock skew event
// Payload format: | dest ID | start time | duration | clock skew extent | min_f update count | min_b update count |
#define MSG_RUN_SND     10  // COMMAND: Run Pong sender
// Payload format: | receiver id |
#define MSG_RUN_RCV     11  // COMMAND: Run Pong receiver
// Payload format: | sender id |
#define MSG_END_PONG    12  // COMMAND: End Pong
// Payload format: empty
#define MSG_PONG_FAIL   13  // Permanent failure of running Pong
// Payload format: | role | peer id | reason |
#define MSG_LINKC       14  // Link congestion event
// Payload format: | ip1 | ip2 | congestion intensity | hop # | total number of hops | sample_count | link measurability score | node1 info | node2 info |

// Header of the payload part of MSG_TOPO
struct Header_MSG_TOPO
{
    time_t version; // Version of this topology update
    short dID;      // Destination host ID
    short num_node; // Number of nodes ( equal to hop number + 1 ), including source and destination nodes
    PathDigest digest;  // Path digest
    float rc_count; // Route change count
    bool rc_stop;   // Indicate route changes stopped
};


// Packet of MSG_TOPO_ACK
struct Packet_MSG_TOPO_ACK
{
    MsgHeader header;
    time_t version; // Version of the topology update to ack
    short dID;      // Destination host ID
};


#define BIN_NUM 30  // Number of bins used to stat queuing delay threshold

// Threshold information
struct ThresholdInfo
{
    short bin_index;
    short pattern;
    float percent_below, percent_above;
    float bin_count[BIN_NUM];
};


// Payload part of MSG_E2EC
struct Payload_MSG_E2EC
{
    short dID;          // Destination host ID 
    time_t start_time;  // Start time of the congestion
    short duration;     // Duration of the congestion
    float intensity_f;  // Congestion intensity on the forward path
    float intensity_b;  // Congestion intensity on the backward path
    float intensity_r;  // Congestion intensity on the round trip path
    short sample_count; // Number of probing samples
    short mfu_count;    // min_f update count
    short mbu_count;    // min_b update count
    float rc_count;     // Route change count
    float clevel_edsum_f;   // Exponentially decayed sum of congestion levels (forward path)
    float clevel_edsum_b;   // Exponentially decayed sum of congestion levels (backward path)
    ThresholdInfo th_f;
    ThresholdInfo th_b;
    ThresholdInfo th_r;
    //bool incomplete;    // Indicate that the congestion event has not yet stopped when this message is sent
};


// Payload part of MSG_EDT_TMON
struct Payload_MSG_EDT_TMON
{
    short dID_add;      // Host id to add to TMon FRP
    short dID_rmv;      // Host id to remove from TMon FRP
    short num_host_frp;
    short frp_hosts[MAX_FRP_HOST];
};


// Header of the payload part of MSG_TMON_LIST
struct Header_MSG_TMON_LIST
{
    short num_host_frp; // Number of FRP hosts
    short dID_add;      // Host id to add to TMon FRP in MSG_EDT_TMON
    short dID_rmv;      // Host id to remove from TMon FRP in MSG_EDT_TMON
};

// If 'dID_add' in Header_MSG_TMON_LIST takes the value -1, then 'dID_rmv' takes the following values
#define TL_RECOVERD     1   // TMon list inconsistency recoverd
#define TL_CORRUPT      2   // TMon list sent from control node is corrupt
#define TL_FRP_CORRUPT  3   // TMon list on the measuring node (i.e., dest_host_frp) is corrupt


struct PathBitmap
{
    short max_id;
    char map[PATH_BITMAP_SIZE];

    // NOTE: bit_id start from 1 to 1000

    void SetPathBit( short bit_id );
    void ClearPathBit( short bit_id );
    bool IssetPathBit( short bit_id ) const;
    void Reset( void );
};


// Payload part of MSG_ALIVE
struct Payload_MSG_ALIVE
{
    short version;      // Version of Tpong system
    time_t timestamp;   // Timestamp used to calculate clock offset between the sending node and the control node
    bool pong_active;   // Whether Pong is still running or trying
    char pong_state;    // State of Pong daemon
    PathBitmap path_bitmap; // Path bitmap that indicates with paths are active
};


struct TMonList
{
    short num_tmon;     // Number of TMon hosts
    short check_sum;    // Checksum of the list
    short dID[MAX_FRP_HOST];    // List of TMon hosts
};


// Packet of MSG_ALIVE_ACK
struct Packet_MSG_ALIVE_ACK
{
    MsgHeader header;
    short version;      // Version of Tpong system
    time_t clock_offset;    // Clock offset between the destination node and the control node
    TMonList tmon_list;
};


// Payload part of MSG_CSKEW
struct Payload_MSG_CSKEW
{
    short dID;          // Destination host ID 
    time_t start_time;  // Start time of the statistic
    short duration;     // Duration of the statistic
    short extent;       // Clock skew extent
    short mfu_count;    // min_f update count
    short mbu_count;    // min_b update count
};


// Payload part of MSG_RUN_SND
struct Payload_MSG_RUN_SND
{
    short receiver_id;  // Host id of receiver
};


// Payload part of MSG_RUN_RCV
struct Payload_MSG_RUN_RCV
{
    short sender_id;    // Host id of sender
};


// Payload part of MSG_PONG_FAIL
struct Payload_MSG_PONG_FAIL
{
    char role;      // Whether it is sender or receiver
    short peer;     // Host id of Pong peer
    char reason;    // Reason of the failure
};


struct ThresholdInfo_Lite
{
    short bin_index;
    short pattern;
    float percent_below, percent_above;
};


struct LinkNodeInfo
{
    bool isset;         // Whether information has been set
    char technique;     // Probing technique used for this node
    float QoM;          // QoM associated with this node
    ThresholdInfo_Lite th_s, th_d, th_f, th_b;  // Information of probing thresholds
};

bool CheckLniIntegrity( const LinkNodeInfo & lni );


// Payload part of MSG_LINKC
struct Payload_MSG_LINKC
{
    in_addr ip1;            // IP address of the first node
    in_addr ip2;            // IP address of the second node
    float intensity;        // Congestion intensity
    char probe_num;         // Number of measurable points
    char hop_num;           // Hop number of this link
    char total_hop_num;     // Total number of hops of this path
    short sample_count;     // Number of samples
    float lms;              // Link measurability score
    LinkNodeInfo lni1;      // Information of the first node
    LinkNodeInfo lni2;      // Information of the second node
};

bool CheckIntegrity( const Payload_MSG_LINKC * cong_event );


// Control codes between Pong_Master and other programs

#define MCC_CLOSE   0   // Close the control channel
#define MCC_END     1


#endif
