#ifndef PONG_H
#define PONG_H

#include "Pong_Common.h"
#include "Records.h"

using std::list;
using std::vector;

#define HAVE_PCAP

#define MIN_HOP         3
#define MAX_HOP         63
#define MIN_PROBE_NUM   3
#define PROBE_INTERVAL  0.01
#define PAIRING_DEV     1       // Pairing deviation 

#define PROBE_PACKET_SIZE   sizeof( ProbeUdpPayload )   // = 28 (plus ip and udp header => 56 bytes)
#define MAX_SEND_BUF_SIZE   100

#define ORIGIN_TTL      255
#define UDP_REPETITION  6

#define MAX_CONFIG_LINE_WIDTH   200

#define TH_TRA_CONVERGE_TIME    30

// Control codes
#define CC_RTT    0    // Estimate RTT
#define CC_HFR    1    // Half record from receiver to sender
#define CC_ATO    2    // Adjust time offset of receiver side (for d and b probes)
#define CC_QUIT   3    // Quit the tool
#define CC_TBL    4    // Time_baseline from receiver to sender
#define CC_CFG    5    // Probe configuration from sender to receiver
#define CC_SYNC   6    // Synchronize sender and receiver
#define CC_EHN    7    // Exchange hop number
#define CC_SDV    8    // Send double value
#define CC_HOI    9    // Hop information from receiver to sender
#define CC_INI    10   // Initialize configuration
#define CC_HIS    11   // Load configuration from history
#define CC_PUD    12   // Pairing updates from sender to receiver
#define CC_PUA    13   // Acknowledge of pairing updates from receiver to sender
#define CC_ATA    14   // ATO acknowledge from receiver to sender
#define CC_TUD    15   // Technique updates from sender to receiver
#define CC_TUA    16   // Acknowledge of technique updates from receiver to sender
#define CC_TDF    17   // Notify the failure of targeted d probe from receiver to sender
#define CC_RHN    18   // Receiver's host name from receiver to sender
#define CC_VER    19   // Check version
#define CC_HID    20   // Exchange host_id
#define CC_CNCL   21   // Cancel measurement
#define CC_TRA    22   // Start tracing route alteration
#define CC_ACC    23   // Send access permission result, deny(0) or accpet(1)

// Probing phases
#define PP_INIT     0
#define PP_PROBE    5
#define PP_TRA      6

const float TH_LINK_CI      = 0.01;     // Threshold of congestion intensity above which MSG_LINKC will be sent

// Linux networking compatability macros.  For some unfathomable
// reason, Linux systems seem to have named many of their networking
// constants differently than those used by virtually every other
// sockets API implementation.  We try to bring it in line with
// the more widely-used sockets API standards here.
#ifdef linux

// UDP
#define uh_sport        source
#define uh_dport        dest
#define uh_ulen         len
#define uh_sum          check

#endif /* linux */

class Pong
{
protected:
    bool is_sender;     // whether this is pong_snd or pong_recv
    int version;        // version of Pong
    int host_id;
    int peer_host_id;

    struct in_addr originAddress;
    struct in_addr targetAddress;
    
    int sock_tcp_listener;      // socket used to establish control channel
    int sock_tcp;               // socket of control channel
    int sock_monitor_listener;  // socket used to establish monitor channel
    int sock_monitor;           // socket of monitor channel
    
    pthread_mutex_t ctr_mutex;  // the sending lock for sock_tcp
    struct sockaddr_in target_tcp_addr;
    int sock_udp;   // socket for f/b probe
    struct sockaddr_in target_udp_addr;
    struct sockaddr_in origin_udp_addr;

    bool done;
    bool phase_done;
    bool time_offset_adjusted;
    union
    {
        bool pairing_updated;           // for receiver
        bool pairing_update_pending;    // for sender
    };
    union
    {
        bool technique_updated;         // for receiver
        bool technique_update_pending;  // for sender
    };

    int socketOut;  // output socket (RAW) for s/d probe
    int socketIn;   // input socket (ICMP) for s/d probe

    char *originName;   // origin hostname
    char *targetName;   // target hostname
    char probe_config_file[40];   // configuration file name (actual length is expected to be no more than 21 bytes)
    timeval syscallTime; // estimated overhead for gettimeofday()
    pthread_t tid_sink;

    timeval current_time, time_baseline, time_baseline_receiver;
    int round_id;
    double e2e_rtt;
    MinimumValue min_f, min_b;
    SkewMonitor skew_monitor_f, skew_monitor_b;
    double last_Df, last_Db;

    double route_alteration_count;
    int altered_hop_count;
    bool * route_altered;
    int route_converge_count;
    int TRA_start_time, TRA_stop_time;
    vector<int> * hop_ips;
    char * quit_reason;

    bool auto_tpt;
    timeval last_tuning_paring_time;
    vector<PairingUpdate> pairing_updates;
    vector<TechniqueUpdate> technique_updates;
    pthread_mutex_t tpt_mutex;      // the lock for pairing_updates, technique_udpates, pairing_update_pending and technique_update_pending
                                    // tpt - tuning pairing and technique
    int PUA_round_id;
    int TUA_round_id;

    // Thresholds of time measured by f and b probes
    ProbeThreshold th_f, th_b;
    
    int hop_num;            // number of hops of the end-to-end path
    HopInfo *hop_info;      // information of hops
    int hop_num_2;          // 'hop_num' of the remote side
    HopInfo *hop_info_2;    // 'hop_info' of the remote side
    TargetedDprobeInfo *tdp_info;   // information of targeted d probes

    double probe_period;  // in seconds
    int probe_num;
    ProbeInfo *probe_info;      // NOTE: Order of 'id' must be consistent with that of 'ttl'
    HalfRecordBuffer *hrbuffer; 
    FullRecordBuffer *frbuffer; // only applicable for sender
    pthread_mutex_t hrb_mutex;  // the lock for hrbuffer
    pthread_mutex_t frb_mutex;  // the lock for frbuffer
    ProbeHistory *phistory;     // only applicable for sender
    pthread_mutex_t pih_mutex;  // the lock for probe_info and phistory
    
    ProbeRecord *last_pr;       // only applicable for sender
    double *link_cc;            // link congestion count
    double *link_cc_max;
    double *link_cc_10sec[6];
    double *link_cc_1min[10];
    double *link_cc_10min[6];
    double *link_cc_1hour[6];
    double *link_cc_3hour[8];
    double *LCD_30sec, *LCD_1min, *LCD_3min, *LCD_10min, *LCD_30min, *LCD_1hour, *LCD_3hour, *LCD_6hour, *LCD_12hour, *LCD_24hour;
    timeval last_link_cc_update_time;
    int heart_beat_update_time;
    int link_cc_update_count;
    double *sum_link_ci;            // sum of link congestion intensity
    double *average_link_ci;        // average link congestion intensity
    double *average_link_ci_max;    // maximum value of average link congestion intensity over time
    double max_link_ci;             // maximum value of average link congestion intensity over all links
    double sum_e2e_cc;              // end-to-end congestion count
    double average_e2e_cc_max;
    int sample_count;
    int sum_round_id;
    bool *redundant_link;
    int dominant_bottleneck;

    // ************ Statistic parameters ************
    double link_cc_sum_all;
    double link_ci_sum_all;
    int sample_total;
    QoM_Log *QoM_log;           // QoM log information; indexed by probe_id
    bool *tdp_available;        // state of targeted d probe; used by the sender side; indexed by probe_id
    bool tdpa_updated;          // indicates whether array tdp_available has been updated or not
    int last_update_time;
    // ******** End of statistic parameters *********

    char * GeneratePayload( int size );
    PongActionType GetAction( int icmpType, int icmpCode );
    char * GetName( const void * a );
    char * GetPrintableAddress( const void * a );

    void SendCtrCode( long ctr_code );
    long RecvCtrCode( void );
    inline void SendIntValue( int value ) { SendCtrCode( value ); }
    inline int RecvIntValue( void ) { return RecvCtrCode(); }
    
    int ProcessIcmpPacket( char *ipPacket, ProbeResult &pr, bool silent = false );
    int ProcessUdpPacket( char *ipPacket, ProbeResult &pr );
    int ProcessUdpPayload( char *udp_payload, ProbeResult &pr );

    int SetOriginName( char *o );
    int SetTargetName( char *t );

    int GetSocketIn();
    int GetSocketOut();

    void SetControlSocket_Snd( void );
    void SetControlSocket_Rcv( void );
    
    void SendReceiverName( void );
    void RecvReceiverName( void );

    void ReadVersion( void );
    void CheckVersion( void );
    void CheckAccessResult( void );

    void ReadHostID( void );
    void ExchangeHostID( void );

    // Set udp socket for f/b probe
    void SetUdpSocket( void );

    void WaitForReceiver( void );

    void SetupMonitorServer( void );
    void AcceptMonitorClient( void );
    void TearDownClient( void );

    double EstimateRTT( void );
    void EstimateRTT_Respond( void );

    void SendDoubleValue( double value );
    double RecvDoubleValue( void );

    double GetProbeTime( double time_offset );

    int SendUdpProbe( int round_id, int probe_id, double probe_time_scheduled );
    int SendRouterProbe( int round_id, int probe_id, int ttl, bool hop_probe = false );

    int RecvUdpProbe( ProbeResult &pr );
    int RecvRouterProbe( ProbeResult &pr );

    void SetupPacketSink();

    void ProcessControlMessage();

    void Quit( void );
    void Cancel( void );
    void StartTraceRA( void );

    void ResetRaCount( void );
    void InitializeTRA( void );
    bool DumpRouteAlterationInfo( bool converged );

    void InitializeRouteAlteration( void );
    void InitializeLinkInfo( void );
    void InitializeStatistics( void );

    void InitializeProbeInfo( void );

    void ResetQoMLog( void );
    void UpdateQoMLog( ProbeInfo * pi, ProbeHistory * ph );
    void FinishQoMLog( void );
    bool OutputQoMLog( void );
    bool OutputErrorMetrics( void );
    bool OutputMTLog( void );
    bool OutputQDBin( void );
    bool OutputSummary( void );
    void CheckTDPA( void );
    bool OutputProbeConfiguration( void );
    bool OutputProbeLoss( void );
    bool OutputRouteAlteration( void );

    ProbeInfo * FindProbeInfo( int probe_id );
    ProbeInfo * FindProbeInfoByTTL( int ttl );
    ProbeHistory * FindProbeHistory( int probe_id );
    HalfRecordBuffer * FindHrBuffer( int probe_id );
    FullRecordBuffer * FindFrBuffer( int probe_id );
    TargetedDprobeInfo * FindTdpInfo( int probe_id );
    HopInfo * FindHopInfo( int ttl, bool remote_side = false );

    void UpdateFullRecord( HalfRecord &hr, bool is_sender_record );
    void ProcessRecord( ProbeInfo *pi, ProbeHistory *ph, FullRecord &fr );
    void UpdateThresholds( ProbeHistory *ph, FullRecord &fr );
    void UpdateQoM( ProbeInfo *pi, ProbeHistory *ph, FullRecord &fr );
   
    void UpdateCongestionStatus( ProbeInfo *pi, ProbeHistory *ph, FullRecord &fr );
    void UpdateLCD( const ProbeRecord &pr1, const ProbeRecord &pr2 );
    void UpdateLCD( const ProbeRecord &pr );
    void ShowLCD( void );
    
    void UpdateICS( const ProbeRecord &pr );
    double LCD_PositiveRate( int link_id );
    void UpdateDominantBottleneck( void );
    double LinkCC_30sec( int link_id );
    
    void SendHalfRecord( HalfRecord &hr );
    void RecvHalfRecord( HalfRecord &hr );
    
    void SendATO( int probe_id, double time_offset_increment );
    void RecvATO( int &probe_id, double &time_offset_increment );
    void AdjustTimeOffset( int probe_id, double time_offset_increment );
    void SendATA( int probe_id, int round_id );
    void RecvATA( int &probe_id, int &round_id );
    void SendTDF( int probe_id );

    void SendTimeBaseline( void );
    void RecvTimeBaseline( void );

    int RestoreRoundId( int truncated_round_id );

    bool LoadProbeConfiguration( void );
    void SaveProbeConfiguration( void );
    void WriteProbeConfiguration( FILE * fp );
    void InitializeProbeConfiguration( void );
    void SendProbeConfiguration( void );
    void RecvProbeConfiguration( void );
    void SendPairingUpdates( void );
    void RecvPairingUpdates( void );
    void SendTechniqueUpdates( void );
    void RecvTechniqueUpdates( void );

    void SortProbes( void );

    void ProcessInteractiveCommand( void );
    void HelpInteractiveCommand( void );
    void HelpDebug();
    void ProbeInformation( bool compact );
    void ProbeInformation_Half( void );
    void HopInformation( bool remote_side = false );
    void HopInformation_Both( void );
    bool OutputHopInformation( void );  // Output hop information in HTML format
    void OutputPairing( FILE * fp, const char * title );
    bool OutputCurrentPairing( void );
    void QoM_Information( void );
    void ThresholdInformation( void );
    void MT_Information( void );
    
    bool ChangePairings( int hop_s_start, int hop_s_end, int hop_d_start, bool verbose = true );
    void UpdatePairings( void );
    bool TuningPairings( void );
    bool ChangeTechniques( int hop_s_start, int hop_s_end, int technique, bool verbose = true );
    void UpdateTechniques( void );
    bool EnqueueTechniqueUpdate( int probe_id, int technique );
    
    void GetPairingRange( ProbeInfo * pi, int &lower_bound, int &upper_bound );
    void InitializePairingHistory( void );
    void PrintPairingHistory( bool compact );
    
    bool TptPending( void );
    void ToggleATPT( void );

    void InitializeRecords( void );
    bool OutputGlobalInfo( void );
    void OutputStartTime( void );
    void OutputDataID( void );
    void PrintPongStartTime( void );
    void RemoveDataFile( const char * filename );
    void RemoveLinkCCdata( const char * filename );
    void ResetLinkCC( double * link_cc );
    void CopyLinkCC( double * to, const double * from );
    void AddLinkCC( double * to, const double * from );
    bool OutputLCD( const char * filename, double * link_cc );
    bool OutputLinkCC( void );
    void UpdateLinkCCmax( void );
    void UpdateLinkCC_10sec( void );
    void UpdateLinkCC_1min( void );
    void UpdateLinkCC_10min( void );
    void UpdateLinkCC_1hour( void );
    void UpdateLinkCC_3hour( void );
    bool OutputLinkCI( void );
    void UpdateAverageLinkCI( void );

    // Synchronize sender and receiver
    void Synchronize( void );

    void MeasureHopnum( void );
    bool CheckHopNum( void );
    void SendHopnum( void );
    void RecvHopnum( void );
    void ExchangeHopnum( void );
    void ExploreHopInfo( void );
    void SendHopInformation( bool remote );
    void RecvHopInformation( bool remote );

    void ReportLinkCI( void );
    void SendLinkCI( in_addr ip1, in_addr ip2, float intensity, char probe_num, char hop_num, char total_hop_num, short sample_count, float lms, const LinkNodeInfo& lni1, const LinkNodeInfo& lni2 );
    float GetLinkNodeInfo( int ttl, LinkNodeInfo& lni );
    void CopyThresholdInfo( ThresholdInfo_Lite& dest, const ProbeThreshold& source );

#ifdef HAVE_PCAP

    pthread_t tid_pcap;

    void SetupPCAP();

#endif // HAVE_PCAP

public:

    int phase;      // phase of probing
    char pn_fb_out, pn_fb_in, pn_sd;    // probe name

    Pong( bool is_sender );

    void InitializeSender( void );
    void InitializeReceiver( void );
    void Initialize_Common_Pre( void );
    void Initialize_Common( void );
    
    void Finalize( void );

    void HeartBeat( const char * status, const char * comment = NULL );

    void SendProbes( void );

    void TraceRouterAlteration( void );
    
    void ReceiveMessages( void );

    int ParsePacket( char *ipPacket, ProbeResult &pr );

    void UpdateHalfRecord( ProbeResult &pr );

    void UpdateRouteAlteration( ProbeResult &pr );
    bool UpdateHopIP( int ttl, int ip );

    void SetProbeName( char pn_fb_out, char pn_fb_in, char pn_sd )
    {
        this->pn_fb_out = pn_fb_out;
        this->pn_fb_in = pn_fb_in;
        this->pn_sd = pn_sd;
    }

    void MeasureHopnum_Respond( const ProbeResult &pr );
    void ExploreHopInfo_Respond( const ProbeResult &pr );

#ifdef HAVE_PCAP

    pcap_t *pc; // pcap structure
    unsigned int packet_offset;

#endif // HAVE_PCAP

    friend const char * GetMeasurementTime( void );
};


// Type codes of ProbeUdpPayload

#define PL_UDP      0   // End-to-end udp probe (i.e., f/b probe )
#define PL_HOP      1   // Measuring hop_num
#define PL_ROUTER   2   // Router probe (i.e., s/d probe)

struct ProbeUdpPayload
{
    int round_id;
    int probe_id;
    timeval timestamp;
    char signature[4];
    int type_code;
    char padding[4];    // padding up to 28 bytes (plus ip and udp header => 56 bytes)

    ProbeUdpPayload( void );
    void SetRouterProbe( void );
    void SetHopProbe( void );
    void SetUdpProbe( int round_id, int probe_id, timeval timestamp );
    bool IsValid( void );
};


extern char *OriginHost;
extern char *TargetHost;
extern Pong *pong;
extern char *Interface;
extern bool IgnoreConfiguration;
extern const char * mmt_data_dir;
extern bool enable_interactive_command;
extern int sender_control_port;
extern int min_hop, min_probe_num;
extern int receiver_id;

void Usage( char *argv0 );

void ParseCommand( int argc, char **argv );

u_short InCksum( u_short *addr, int len );

void * sink_process( void * arg );

char ReadParameter( FILE * fp, const char * parameter_name, const char * format_string, void * parameter_ptr );

void SwitchDebug( bool turn_on );

void DebugOffWarning( void );

unsigned int Prefix( struct in_addr ip );

void GetCongestionExtent( double &extent, double &confidence, double L, double U, double threshold );

bool Congested( double extent );

void OpenClockSkewFile( void );

void CheckProcess( void );

void PrintPID( void );

void TimeSyscall( timeval &diff );

void Sig_Term( int sig_num );

void ProcessAck( const char * packet );

#ifdef HAVE_PCAP

// pcap thread process
void * pcap_process( void * );

// pcap callback functions
void pcap_callback( u_char *puc, const struct pcap_pkthdr *ph, const u_char *pd );

#endif // HAVE_PCAP

// ************** Functions of Pong Monitor **************

void SetupMonitorClient( void );

int ResolveTargetName( char *t );

void ProcessMonitorClient( void );

// ********** End of Functions of Pong Monitor ***********

#endif
