#include "Records.h"


ProbeResult::ProbeResult( void )
{
    bzero( this, sizeof( ProbeResult ) );
}


void ProbeResult::ShowRecord( void )
{
    fprintf( stderr, "- probe_type=%s, round_id=%d, probe_id=%d, tv=%.3lf(ms)", ptype_code[probe_type], round_id, probe_id, TimeInMsec( tv ) );
    if ( probe_type == probe_fb_out || probe_type == probe_fb_in )
        fprintf( stderr, ", timestamp=%.3lf(ms)", TimeInMsec( timestamp ) );
    else if ( probe_type == probe_sd_in )
        fprintf( stderr, ", icmpSourceAddress=%s", inet_ntoa( icmpSourceAddress ) );
    else if ( probe_type == probe_sd_out )
        fprintf( stderr, ", ttl=%d", ttl );
    else if ( probe_type == probe_icmp_error )
        fprintf( stderr, ", icmpResult=%d", icmpResult );
    fputc( '\n', stderr );
}


// ************** PairingHistory member functions **************

void PairingHistory::Initialize( int ttl_d )
{
    this->ttl_d = ttl_d;
    times_tried = 0;
    average_age = 0;
    average_QoM = 0;
    QoM_JP = QoM_4P = QoM_fsd = 0;
}


double PairingHistory::Priority( void )
{
    if ( times_tried == 0 )
        return 999;
    double k1 = 10 / (double)times_tried;
    double k2 = average_age / ( TH_QoM_OLD_AGE * 5 );
    if ( k2 > 3 ) k2 = 3;
    double k3 = average_QoM;
    return k1 + k2 + k3 * 8;
}

// *********** End of PairingHistory member functions ***********


// ************** ProbeRecord member functions **************

void ProbeRecord::Initialize( const ProbeInfo *pi )
{
    ttl = pi->ttl;
    technique = pi->technique;
    fr.round_id = -1;
}

// *********** End of ProbeRecord member functions ***********


// ************** ProbeHistory member functions **************

void ProbeHistory::Initialize( int probe_id )
{
    this->probe_id = probe_id;
    ttl_d = VALUE_UNDEFINED;
    ATO_pending = false;
    ATA_round_id = 0;

    min_s.Initialize( MIN_SD_AGE_MAX );
    min_d.Initialize( MIN_SD_AGE_MAX );

    last_Ds = VALUE_UNDEFINED;
    last_Dd = VALUE_UNDEFINED;

    th_s.Initialize( TH_INITIAL_SMALL, TH_SD_CNT_MAX );
    th_d.Initialize( TH_INITIAL_SMALL, TH_SD_CNT_MAX );

    InitializeQoMs();
}


void ProbeHistory::InitializePairingHistory( int lower_hop_d, int upper_hop_d, bool * valid_hop_d )
{
    int num = upper_hop_d - lower_hop_d + 2;
    pairing_history.resize( num );

    PairingHistory pah;
    pah.Initialize( TTL_TARGETED_D );
    pairing_history[0] = pah;   // targeted d probe is set to be the first candidate

    int invalid_num = 0;
    for ( int i = 1; i < num; i ++ )
    {
        int hop_d = upper_hop_d - i + 1;
        
        pah.Initialize( hop_d );
        if ( ! valid_hop_d[hop_d] )
        {
            pah.times_tried = 999;
            invalid_num ++;
        }
        pairing_history[i] = pah;
    }
    if ( invalid_num > 0 )
    {
        SortPairingHistory();
        pairing_history.resize( num - invalid_num );
    }
}


bool ProbeHistory::QoM_OldAge( void )
{
    return ( QoM_4P.age > TH_QoM_OLD_AGE && QoM_fsd.age > TH_QoM_OLD_AGE && QoM_fsb.age > TH_QoM_OLD_AGE );
}


void ProbeHistory::InitializeQoMs( void )
{
    ResetQoMs();
    QoM_fsb.Initialize( TH_INIT_FSB, 0.1, 0.03, 10 );
}


void ProbeHistory::ResetQoMs( void )
{
    QoM_JP.Initialize( 0.5, 0.1, 0.03, 10 );
    QoM_4P.Initialize( TH_INIT_4P, 0.1, 0.03, 10 );
    QoM_fsd.Initialize( TH_INIT_FSD, 0.1, 0.03, 10 );
}


// Sort pairing_history by Priority

void ProbeHistory::SortPairingHistory( void )
{
    int num = pairing_history.size();
    for ( int i = 1; i < num; i ++ )
    {
        for ( int j = i - 1; j >= 0; j -- )
        {
            if ( pairing_history[j+1].Priority() > pairing_history[j].Priority() )
            {
                PairingHistory temp = pairing_history[j];
                pairing_history[j] = pairing_history[j+1];
                pairing_history[j+1] = temp;
            }
            else break;
        }
    }
}


void ProbeHistory::UpdatePairingHistory( void )
{
    if ( ttl_d != VALUE_UNDEFINED )
    {
        int num = pairing_history.size();
        int i;
        for ( i = 0; i < num; i ++ )
            if ( ttl_d == pairing_history[i].ttl_d )
                break;
        if ( i >= num )
            // Entry might be removed due to failure of targeted d probe
            // Entry might not exist due to manual pairing
            return;

        pairing_history[i].times_tried ++;
        pairing_history[i].QoM_JP = QoM_JP.value_l;
        pairing_history[i].QoM_4P = QoM_4P.value_l;
        pairing_history[i].QoM_fsd = QoM_fsd.value_l;

        int times_tried = pairing_history[i].times_tried;
        pairing_history[i].average_age = ( pairing_history[i].average_age * ( times_tried - 1 ) + ( QoM_4P.age + QoM_fsd.age ) / 2 ) / times_tried;
        pairing_history[i].average_QoM = ( pairing_history[i].average_QoM * ( times_tried - 1 ) + MAX( QoM_4P.value_l, QoM_fsd.value_l ) ) / times_tried;
        
        // Reset ages of QoMs
        QoM_4P.age = QoM_fsd.age = QoM_JP.age = QoM_fsb.age = 0;

        for ( int j = i + 1; j < num; j ++ )
        {
            if ( pairing_history[j-1].Priority() < pairing_history[j].Priority() )
            {
                PairingHistory temp = pairing_history[j-1];
                pairing_history[j-1] = pairing_history[j];
                pairing_history[j] = temp;
            }
            else break;
        }
    }
}


PairingHistory * ProbeHistory::FindPairingHistory( int ttl_d )
{
    int num = pairing_history.size();
    for ( int i = 0; i < num; i ++ )
        if ( ttl_d == pairing_history[i].ttl_d )
            return &(pairing_history[i]);
    return NULL;
}


bool ProbeHistory::DeletePairingHistory( int ttl_d )
{
    vector<PairingHistory>::iterator i;
    for ( i = pairing_history.begin(); i != pairing_history.end(); i ++ )
        if ( ttl_d == i->ttl_d )
            break;
    if ( i == pairing_history.end() )
        return false;
    pairing_history.erase( i );
    return true;
}


// *********** End of ProbeHistory member functions ***********


// ************** MinmumValue member functions **************

void MinimumValue::Initialize( int age_max )
{
    this->age_max = age_max;
    value = candidate_value = 9E9;
    age = 0;
}


int MinimumValue::Update( double value )
{
    if ( value < this->value )
    {
        this->value = value;
        candidate_value = 9E9;
        age = 0;
        return 1;
    }
    else
    {
        if ( value < candidate_value )
            candidate_value = value;
        if ( ++ age >= age_max )
        {
            this->value = candidate_value;
            candidate_value = 9E9;
            age = 0;
            return 2;
        }
    }
    return 0;
}

// *********** End of MinmumValue member functions ***********


// ************** SkewMonitor member functions **************

void SkewMonitor::Initialize( MinimumValue * mv )
{
    this->mv = mv;
    first_value = VALUE_UNDEFINED;
    average_skew = 9E9;
}


void SkewMonitor::ResetAge( const timeval &current_time )
{
    age = current_time;
}


#define SKEW_THRESHOLD  1E-5    // 1 ms per 100 sec
#define AGE_THRESHOLD   30E6    // 30 sec


const char * skew_extent_text[3] = { "Non-negligible", "Considerable", "Significant" };

int SkewExtent( double skew )
{
    if ( skew < 0 )
        return 2;
    if ( skew < SKEW_THRESHOLD * 10 )
        return 1;
    if ( skew < SKEW_THRESHOLD * 100 )
        return 2;
    return 3;
}


void SkewMonitor::CheckSkew( const timeval &time )
{
    if ( first_value == VALUE_UNDEFINED )
    {
        first_value = mv->value;
        first_time = time;
        return;
    }

    if ( average_skew == 9E9 )
    {
        if ( delta_t( first_time, time ) > 3E6 )
        {
            average_skew = ( mv->value - first_value ) / delta_t( first_time, time );
            last_value = mv->value;
            last_time = time;
            double skew_value = fabs( average_skew );
            if ( skew_value > SKEW_THRESHOLD && delta_t( age, time ) > AGE_THRESHOLD )
            {
                const char * mt = GetMeasurementTime();
                int extent = SkewExtent( skew_value );
                clock_skew_extent += extent;
                clock_skew_indicator = SKEW_SUPPRESS_TIME;
                const char * se = skew_extent_text[ extent - 1 ];
                double sv = skew_value * 1E3;
                fprintf( clock_skew_fp, "%s %s clock skew: skew=%5g(ms/sec)\n", mt, se, sv );
                if ( ! suppress_clock_skew_info )
                    IF_DEBUG( 3, fprintf( stderr, "%s WARNING: %s clock skew between sender and receiver detected: skew=%5g(ms/sec)\n", mt, se, sv ) );
            }
        }
        return;
    }

    double diff_value = mv->value - last_value;
    double diff_time = delta_t( last_time, time );
    double skew = diff_value / diff_time;
    double skew_difference = fabs( skew - average_skew ) / MAX( fabs( average_skew ), fabs( skew ) );

    //fprintf( stderr, "skew=%g, average_skew=%g, skew_difference=%.3lf, skew_count=%d, count=%d\n", skew, average_skew, skew_difference, skew_count, count );
    
    if ( skew_difference < 0.9 )
    {
        average_skew = ( mv->value - first_value ) / delta_t( first_time, time );
        double skew_value = fabs( average_skew );
        if ( skew_value > SKEW_THRESHOLD && delta_t( age, time ) > AGE_THRESHOLD )
        {
            const char * mt = GetMeasurementTime();
            int extent = SkewExtent( skew_value );
            clock_skew_extent += extent;
            clock_skew_indicator = SKEW_SUPPRESS_TIME;
            const char * se = skew_extent_text[ extent - 1 ];
            double sv = skew_value * 1E3;
            fprintf( clock_skew_fp, "%s %s clock skew: skew=%5g(ms/sec)\n", mt, se, sv );
            if ( ! suppress_clock_skew_info )
                IF_DEBUG( 3, fprintf( stderr, "%s WARNING: %s clock skew between sender and receiver detected: skew=%5g(ms/sec)\n", mt, se, sv ) );
        }
    }
    else if ( skew_difference > 0.99 )
    {
        // Seem to be clock jump, reset monitor
        if ( fabs( average_skew ) > SKEW_THRESHOLD && delta_t( age, time ) > AGE_THRESHOLD )
        {
            const char * mt = GetMeasurementTime();
            double dv = fabs( diff_value ) / 1000;
            double dt = fabs( diff_time ) / 1E6;
            clock_skew_extent += SkewExtent( -1 );
            clock_skew_indicator = SKEW_SUPPRESS_TIME;
            fprintf( clock_skew_fp, "%s Clock jump: %5g(ms) in %5g(sec)\n", mt, dv, dt );
            if ( ! suppress_clock_skew_info )
                IF_DEBUG( 3, fprintf( stderr, "%s WARNING: Clock jump detected: jump %5g(ms) in %5g(sec)\n", mt, dv, dt ) );
        }
        average_skew = 9E9;
        first_value = mv->value;
        first_time = time;
    }
    last_value = mv->value;
    last_time = time;
}

// *********** End of SkewMonitor member functions ***********


// ************** EWMA_Value member functions **************

void EWMA_Value::Initialize( double value, double alpha, double alpha_l, int value_l_min_age )
{
    this->alpha = alpha;
    this->alpha_l = alpha_l;
    this->value_l_min_age = value_l_min_age;
    this->value = value;
    value_l = VALUE_UNDEFINED;
    age = 0;
}


void EWMA_Value::Update( double value_in, double confidence )
{
    if ( value == VALUE_UNDEFINED )
        value = value_in;
    else
    {
        age ++;
        double factor = alpha * confidence;
        value = factor * value_in + ( 1 - factor ) * value;
        if ( age >= value_l_min_age )
        {
            if ( value_l == VALUE_UNDEFINED )
                value_l = value;
            else
            {
                factor = alpha_l * confidence;
                value_l = factor * value_in + ( 1 - factor ) * value_l;
            }
        }
    }
}

// *********** End of EWMA_Value member functions ***********


// ************** BinCount member functions **************

BinCount::BinCount( void )
{
    Reset();
}


void BinCount::Reset( void )
{
    for ( int i = 0; i < BIN_NUM; i ++ )
        bin_count[i] = 0;
    total_count = 0;
    percent_below = percent_above = 0;
}


void BinCount::Update( int bin_index )
{
    total_count ++;
    bin_count[bin_index] ++;
    if ( total_count >= 2000 )
        Aging( 0.8 );
}


void BinCount::Aging( float factor )
{
    if ( total_count < 100 )
        return;
    total_count = 0;
    for ( int i = 0; i < BIN_NUM; i ++ )
    {
        bin_count[i] = (int)( (float)bin_count[i] * factor + 0.5 );
        total_count += bin_count[i];
    }
}


void BinCount::Output( FILE *fp )
{
    for ( int i = 0; i < BIN_NUM; i++ )
    {
        int fraction = ( total_count == 0 ? 0 : bin_count[i] * 1000 / total_count );
        fprintf( fp, "%d\t", fraction );
    }
    fprintf( fp, "%d\n", total_count );
}


void BinCount::CheckThreshold( int threshold_index )
{
    int i;
    int count;
    count = 0;
    for ( i = 0; i <= threshold_index; i ++ )
        count += bin_count[i];
    percent_below = count * 100.0 / total_count;
    count = 0;
    for ( i = threshold_index + 1; i < BIN_NUM; i ++ )
        count += bin_count[i];
    percent_above = count * 100.0 / total_count;
}


#ifndef EMULAB

#define ASCEND      1
#define DESCEND     2

bool BinCount::LocateValley( int& bin_index, int& pattern )
{
    if ( total_count < 30 )     // This threshold is different for TMon (set to 100) and Pong (set to 30)
        return false;

    double bin_percent[BIN_NUM];
    int i;
    for ( i = 0; i < BIN_NUM; i ++ )
        bin_percent[i] = bin_count[i] * 100.0 / total_count;

    double bin_percent_0 = bin_percent[0];
    if ( bin_percent_0 > 90 )
    {
        bin_index = 1;
        while ( bin_index <= 4 && bin_percent[bin_index] > 3 )
            bin_index ++;
        bin_index = bin_index - 1;
        pattern = PATTERN_A;
        return true;
    }

    // ************* Find dominant valley (Pattern B) *************
    
    // Find the last bin that has percentage >= 1%
    int last_bin = BIN_NUM - 2;
    while ( last_bin > 1 && bin_percent[last_bin] < 1 )
        last_bin --;
    int first_bin = 1;
    // Find the first bin that has percentage >= 1%
    while ( first_bin < last_bin && bin_percent[first_bin] < 1 )
        first_bin ++;

    // Locate all valley points
    int state = ASCEND;
    int valley_point[BIN_NUM / 2];
    int vp_count = 0;
    bin_index = first_bin + 1;
    while ( bin_index <= last_bin )
    {
        if ( bin_percent[bin_index] < bin_percent[bin_index - 1] )
        {
            if ( state == ASCEND )
                state = DESCEND;
        }
        else if ( bin_percent[bin_index] > bin_percent[bin_index - 1] )
        {
            if ( state == DESCEND )
            {
                valley_point[vp_count ++] = bin_index - 1;
                state = ASCEND;
            }
        }
        bin_index ++;
    }
    if ( vp_count > 0 )
    {
        // Search for the dominant valley point
        int max_valley_length = -1;
        int dominant_valley_point = -1;
        for ( i = 0; i < vp_count; i ++ )
        {
            int valley_index = valley_point[i];
            int max_index = valley_index;
            bin_index = valley_index - 1;
            while ( bin_index > first_bin )
            {
                if ( bin_percent[bin_index] > bin_percent[max_index] )
                    max_index = bin_index;
                bin_index --;
            }
            int valley_length = valley_index - max_index;
            double max_value = bin_percent[max_index];
            
            max_index = valley_index;
            bin_index = valley_index + 1;
            while ( bin_index <= last_bin )
            {
                if ( bin_percent[bin_index] > bin_percent[max_index] )
                    max_index = bin_index;
                bin_index ++;
            }
            if ( max_index - valley_index < valley_length )
                valley_length = max_index - valley_index;
            if ( max_value > bin_percent[max_index] )
                // Get the smaller one of the two maximum values
                max_value = bin_percent[max_index];

            if ( max_value * 0.7 > bin_percent[valley_index] && max_valley_length < valley_length )
            {
                max_valley_length = valley_length;
                dominant_valley_point = valley_index;
            }
        }
        if ( dominant_valley_point != -1 )
        {
            bin_index = dominant_valley_point;
            pattern = PATTERN_B;
            return true;
        }
    }
    // ********** End of Find dominant valley (Pattern B) ********** 

    double bin_threshold;

    if ( bin_percent_0 > 15 && bin_percent[BIN_NUM - 1] < 3 )
    {
        bin_index = 1;
        bin_threshold = bin_percent_0 * 0.1;
        if ( bin_threshold > 3 )
            bin_threshold = 3;
        while ( bin_index <= last_bin && bin_percent[bin_index] > bin_threshold )
            bin_index ++;
        if ( bin_index > 1 )
        {
            bin_threshold = bin_percent[bin_index - 1] * 0.7;
            while ( bin_index <= last_bin && bin_percent[bin_index] > bin_threshold )
                bin_index ++;
        }
        bin_index = bin_index - 1;
        pattern = PATTERN_C;
        return true;
    }
    
    int max_index = 1;
    bin_index = 2;
    while ( bin_index <= last_bin )
    {
        if ( bin_percent[bin_index] > bin_percent[max_index] )
            max_index = bin_index;
        bin_index ++;
    }
    bin_threshold = bin_percent[max_index] * 0.2;
    if ( bin_threshold < 1 )
        bin_threshold = 1;
    bin_index = max_index + 1;
    while ( bin_index <= last_bin && bin_percent[bin_index] > bin_threshold )
        bin_index ++;
    bin_index = bin_index - 1;
    pattern = PATTERN_D;
    return true;
}

#else

// LocateValley -- Emulab version

bool BinCount::LocateValley( int& bin_index, int& pattern )
{
    int bin_sum[BIN_NUM];
    int max_a = 0;
    int sec_a = 0, sec_index;
    bin_index = 0;
    bin_sum[0] = bin_count[0];
    pattern = PATTERN_A;

    // Locate all bins match pattern:
    //   1) bin_count[i-1] > bin_count[i] * 2 and choose the one with the largest bin_count[i-1]
    //   2) bin_sum[i] > 0.4 * total_count
    for ( int i = 1; i < BIN_NUM; i ++ )
    {
        bin_sum[i] = bin_sum[i-1] + bin_count[i];
        if ( bin_sum[i] > 0.4 * total_count && bin_count[i-1] > bin_count[i] * 2 )
        {
            if ( max_a < bin_count[i-1] )
            {
                sec_a = max_a;
                sec_index = bin_index;
                max_a = bin_count[i-1];
                bin_index = i - 1;
            }
            // sec_index must < bin_index, so the following lines are commented
            /*
            else if ( sec_a < bin_count[i-1] )
            {
                sec_a = bin_count[i-1];
                sec_index = i - 1;
            }
            */
        }
    }

    // a must greater than 0.1 * total_count
    if ( max_a > 0.1 * total_count )
    {
        if ( sec_a > 0.2 * total_count || sec_a > 0.1 * total_count &&
           ( sec_index >= 2 && bin_count[sec_index - 2] > bin_count[sec_index] * 2 ||
             sec_index <= BIN_NUM - 2 && sec_a > bin_count[sec_index + 1] * 2 ) )
            bin_index = sec_index;
        return true;
    }
    return false;
}

#endif

// *********** End of BinCount member functions ***********


// ************** ProbeThreshold member functions **************

void ProbeThreshold::Initialize( double small, int count_max )
{
    this->small = small;
    this->big = small * 1.5;
    this->count_max = count_max;
    bin_index = pattern = -1;
    percent_below = percent_between = percent_above = 0;
    count = count_below = count_between = count_above = 0;
    average_below = average_between = average_above = 0;
    deviation_below = deviation_between = deviation_above = 0;
    max_below = min_above = VALUE_UNDEFINED;

    avg_below = avg_betwn = avg_above = 0;
    dev_below = dev_betwn = dev_above = 0;

    /*
    if ( bin_count.total_count != 0 )
        bin_count.Reset();
    */
}

/*
// Quadratic 30 Bin
const double BIN_UPPER[] = { 180,   320,   500,   720,   980,  1280,  1620,  2000,  2420,  2880,
                            3380,  3920,  4500,  5120,  5780,  6480,  7220,  8000,  8820,  9680,
                           10580, 11520, 12500, 13520, 14580, 15680, 16820, 18000, 19220, 20480 };
*/

// Exponential 30 Bin
const double BIN_UPPER[] = { 150, 195, 254, 330, 428, 557, 724, 941, 1224, 1591, 2068, 2688, 3495, 4543, 5906, 7678, 9981, 12976, 16868, 21929, 28507, 37060, 48178, 62631, 81420, 105846, 137600, 178880, 232544, 302307 };


// Look up a corresponding index of bin for the input queuing delay value

int ProbeThreshold::GetBinIndex( double value )
{
    int i;
    for ( i = 0; i < BIN_NUM - 1; i ++ )
        if ( value < BIN_UPPER[i] )
            break;
    return i;
}


void ProbeThreshold::Update( double value )
{
    if ( value < small )
    {
        // Below
        count_below ++;
        average_below += value;
        deviation_below += fabs( value - average_below / (double)count_below );
        if ( max_below == VALUE_UNDEFINED || value > max_below )
            max_below = value;
    }
    else if ( value > big )
    {
        // Above
        count_above ++;
        average_above += value;
        deviation_above += fabs( value - average_above / (double)count_above );
        if ( min_above == VALUE_UNDEFINED || value < min_above )
            min_above = value;
    }
    else
    {
        // Between
        count_between ++;
        average_between += value;
        deviation_between += fabs( value - average_between / (double)count_between );
    }

    // Update per bin statistics
    bin_count.Update( GetBinIndex( value ) );

    if ( ++ count >= count_max )
    {
        //assert( count_below + count_between + count_above == count );
        percent_below = (double)count_below / (double)count;
        percent_above = (double)count_above / (double)count;
        percent_between = (double)count_between / (double)count;
        if ( count_below > 0 )
        {
            average_below /= (double)count_below;
            deviation_below /= (double)count_below;
        }
        else
            average_below = small;
        if ( count_above > 0 )
        {
            average_above /= (double)count_above;
            deviation_above /= (double)count_above;
        }
        else
            average_above = big;
        if ( count_between > 0 )
        {
            average_between /= (double)count_between;
            deviation_between /= (double)count_between;
        }
        else
            average_between = ( small + big ) * 0.5;


        //***********************************************************************
        // Adjust 'small' and 'big' in the following way (Mar. 1, 2007):
        // 1) Initial threshold:
        //    small = 400; big = 600;
        // 2) Use LocateValley Algorithm to get bin_index:
        //    small = BIN_UPPER[bin_index], big = BIN_UPPER[bin_index+1]
        //    bin_index = 0..8
        //***********************************************************************

        if ( bin_count.LocateValley( bin_index, pattern ) )
        {
            bin_count.CheckThreshold( bin_index );
            small = BIN_UPPER[bin_index];
            big = BIN_UPPER[bin_index+1];
        }

        avg_below = average_below;
        avg_betwn = average_between;
        avg_above = average_above;
        dev_below = deviation_below;
        dev_betwn = deviation_between;
        dev_above = deviation_above;

        count = count_below = count_between = count_above = 0;
        average_below = average_between = average_above = 0;
        deviation_below = deviation_between = deviation_above = 0;
        max_below = min_above = VALUE_UNDEFINED;
    }
}

// *********** End of ProbeThreshold member functions ***********


// ************** HalfRecord member functions **************

void HalfRecord::Initialize( int round_id, int probe_id )
{
    this->round_id = round_id;
    this->probe_id = probe_id;
    status = 0;
}

void HalfRecord::SetSendTime_sd( timeval &send_time_sd )
{
    this->send_time_sd = send_time_sd;
    status |= SEND_TIME_S;
    if ( DebugOn( probe_id ) )
        IF_DEBUG( 9, fprintf( stderr, "- send_time_sd, status=%X\n", status ) );
}

void HalfRecord::SetRecvTime_sd( timeval &recv_time_sd )
{
    this->recv_time_sd = recv_time_sd;
    status |= RECV_TIME_S;
    if ( DebugOn( probe_id ) )
        IF_DEBUG( 9, fprintf( stderr, "- recv_time_sd, status=%X\n", status ) );
}

void HalfRecord::SetSendTime_fb( timeval &send_time_fb )
{
    this->send_time_fb = send_time_fb;
    status |= SEND_TIME_F;
    if ( DebugOn( probe_id ) )
        IF_DEBUG( 9, fprintf( stderr, "- send_time_fb, status=%X\n", status ) );
}

void HalfRecord::SetRecvTime_bf( timeval &recv_time_bf )
{
    this->recv_time_bf = recv_time_bf;
    status |= RECV_TIME_B;
    if ( DebugOn( probe_id ) )
        IF_DEBUG( 9, fprintf( stderr, "- recv_time_bf, status=%X\n", status ) );
}

bool HalfRecord::RecordComplete( bool router_probe_on )
{
    if ( router_probe_on )
        return ( ( status & HR_ALL_TIME ) == HR_ALL_TIME ? true : false );
    else
        return ( ( status & ( SEND_TIME_F|RECV_TIME_B ) ) == ( SEND_TIME_F|RECV_TIME_B ) ? true : false );
}

void HalfRecord::Print( void )
{
    fprintf( stderr, "round_id=%d", round_id );
    if ( status & SEND_TIME_S )
        fprintf( stderr, ", send_time_sd=%.3lf", TimeInMsec( send_time_sd ) );
    if ( status & RECV_TIME_S )
        fprintf( stderr, ", recv_time_sd=%.3lf", TimeInMsec( recv_time_sd ) );
    if ( status & SEND_TIME_F )
        fprintf( stderr, ", send_time_fb=%.3lf", TimeInMsec( send_time_fb ) );
    if ( status & RECV_TIME_B )
        fprintf( stderr, ", recv_time_bf=%.3lf", TimeInMsec( recv_time_bf ) );
    fputc( '\n', stderr );
}

// *********** End of HalfRecord member functions ***********


// ************** HalfRecordBuffer member functions **************

void HalfRecordBuffer::Initialize( int probe_id )
{
    this->probe_id = probe_id;
    last_round_id = 0;
    last_rid_bf = 0;
    last_rid_sd = 0;
    ResetLossCount();
}


// Reset counters used for loss rate statistics

void HalfRecordBuffer::ResetLossCount( void )
{
    expected_total_bf = 0;
    expected_total_sd = 0;
    lost_bf = 0;
    lost_sd = 0;
}


// Check and update loss statistics of probes

void HalfRecordBuffer::CheckProbeLoss( ProbeResult &pr )
{
    if ( pr.probe_type == probe_fb_in )
    {
        if ( pr.round_id > last_rid_bf )
        {
            int count = pr.round_id - last_rid_bf;
            expected_total_bf += count;
            lost_bf += count - 1;
            last_rid_bf = pr.round_id;
            last_recv_time_bf = pr.tv;
        }
    }
    else if ( pr.probe_type == probe_sd_in )
    {
        if ( pr.round_id > last_rid_sd )
        {
            int count = pr.round_id - last_rid_sd;
            expected_total_sd += count;
            lost_sd += count - 1;
            last_rid_sd = pr.round_id;
            last_recv_time_sd = pr.tv;
        }
    }
}


// Locate the HalfRecord corresponding to <round_id>.
// If this is a new round_id, create a new HalfRecord.

list<HalfRecord>::iterator HalfRecordBuffer::GetRecord( int round_id )
{
    list<HalfRecord>::iterator i;
    if ( round_id > last_round_id )
    {
        // Append new HalfRecords
        HalfRecord hr;
        for ( int rid = last_round_id + 1; rid <= round_id; rid ++ )
        {
            hr.Initialize( rid, probe_id );
            buffer.push_back( hr );
        }
        last_round_id = round_id;
        i = buffer.end();
        i --;
        return i;
    }
    
    for ( i = buffer.begin(); i != buffer.end(); i ++ )
    {
        if ( i->round_id == round_id )
            break;
    }
    return i;
}


// Search for and remove incomplete(expired) records whose round_id <= <round_id>

void HalfRecordBuffer::CheckIncompleteRecords( int round_id )
{
    list<HalfRecord>::iterator i = buffer.begin();
    while ( i != buffer.end() )
    {
        if ( i->round_id <= round_id  )
        {
            if ( DebugOn( probe_id ) )
                IF_DEBUG( 6, fprintf( stderr, "Incomplete half record found, probe_id=%d, round_id=%d, status=%X\n", probe_id, i->round_id, i->status ) );
            i = buffer.erase( i );
        }
        else
            i ++;
    }
}

// *********** End of HalfRecordBuffer member functions ***********


// ************** FullRecord member functions **************

void FullRecord::Initialize( int round_id, int probe_id )
{
    this->round_id = round_id;
    this->probe_id = probe_id;
    status = 0;
}


bool FullRecord::Isset_s( void )
{
    return ( ( status & SEND_TIME_S ) && ( status & RECV_TIME_S ) );
}


bool FullRecord::Isset_d( void )
{
    return ( ( status & SEND_TIME_D ) && ( status & RECV_TIME_D ) );
}


bool FullRecord::Isset_f( void )
{
    return ( ( status & SEND_TIME_F ) && ( status & RECV_TIME_F ) );
}


bool FullRecord::Isset_b( void )
{
    return ( ( status & SEND_TIME_B ) && ( status & RECV_TIME_B ) );
}


bool FullRecord::RecordComplete( void )
{
    return ( ( status & BOTH_DONE ) == BOTH_DONE ? true : false );
}


void FullRecord::Print( void )
{
    fprintf( stderr, "round_id=%d, status=%03X", round_id, status );
    if ( Isset_s() )
        fprintf( stderr, ", s=%.3lf(%.3lf)", s / 1000, Ds / 1000 );
    if ( Isset_d() )
        fprintf( stderr, ", d=%.3lf(%.3lf)", d / 1000, Dd / 1000 );
    if ( Isset_f() )
        fprintf( stderr, ", f=%.3lf(%.3lf)", f / 1000, Df / 1000 );
    if ( Isset_b() )
        fprintf( stderr, ", b=%.3lf(%.3lf)", b / 1000, Db / 1000 );
    double time_difference;
    if ( Isset_f() && Isset_d() )
    {
        time_difference = delta_t( recv_time_d, recv_time_f );
        fprintf( stderr, ", drt_fd=%.3lf", time_difference / 1000 );
    }
    if ( Isset_s() && Isset_b() )
    {
        time_difference = delta_t( recv_time_b, recv_time_s );
        fprintf( stderr, ", drt_sb=%.3lf", time_difference / 1000 );
    }
    // Comment: drt - difference of receiving time
    fputc( '\n', stderr );
}


double FullRecord::QoM_JP( void )
{
    double denominator = MAX( f + b, s + d );
    if ( denominator < 1 )
        return VALUE_UNDEFINED;
    else
        return 1 - fabs( ( f + b) - ( s + d ) ) / denominator;
}


double FullRecord::QoM_4P( void )
{
    double denominator = MAX( Df + Db, Ds + Dd );
    if ( denominator < 1 )
        return VALUE_UNDEFINED;
    else
        return 1 - fabs( ( Df + Db ) - ( Ds + Dd ) ) / denominator;
}


double FullRecord::QoM_fsd( void )
{
    double denominator = MAX( Df, Ds + Dd );
    if ( denominator < 1 )
        return VALUE_UNDEFINED;
    else
        return 1 - fabs( Df - ( Ds + Dd ) ) / denominator;
}


double FullRecord::QoM_fsb( void )
{
    double denominator = MAX( Ds, Df + Db );
    if ( denominator < 1 )
        return VALUE_UNDEFINED;
    else
        return 1 - fabs( Ds - ( Df + Db ) ) / denominator;
}


// *********** End of FullRecord member functions ***********


// ************** FullRecordBuffer member functions **************

void FullRecordBuffer::Initialize( int probe_id )
{
    this->probe_id = probe_id;
    last_round_id = 0;
}


// Locate the FullRecord corresponding to <round_id>.
// If this is a new round_id, create a new FullRecord.

list<FullRecord>::iterator FullRecordBuffer::GetRecord( int round_id )
{
    list<FullRecord>::iterator i;
    if ( round_id > last_round_id )
    {
        // Append new FullRecords
        FullRecord fr;
        for ( int rid = last_round_id + 1; rid <= round_id; rid ++ )
        {
            fr.Initialize( rid, probe_id );
            buffer.push_back( fr );
        }
        last_round_id = round_id;
        i = buffer.end();
        i --;
        return i;
    }
    
    for ( i = buffer.begin(); i != buffer.end(); i ++ )
    {
        if ( i->round_id == round_id )
            break;
    }
    return i;
}


// Search for and remove incomplete(expired) records whose round_id <= <round_id>

void FullRecordBuffer::CheckIncompleteRecords( int round_id )
{
    list<FullRecord>::iterator i = buffer.begin();
    while ( i != buffer.end() )
    {
        if ( i->round_id <= round_id  )
        {
            if ( DebugOn( probe_id ) )
                IF_DEBUG( 6, fprintf( stderr, "Incomplete full record found, probe_id=%d, round_id=%d, status=%X\n", probe_id, i->round_id, i->status ) );
            i = buffer.erase( i );
        }
        else
            i ++;
    }
}

// *********** End of FullRecordBuffer member functions ***********
