#ifndef RECORDS_H
#define RECORDS_H

#include <arpa/inet.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>
#include <list>
#include <vector>
#include "Globals.h"
#include "Message.h"

using std::list;
using std::vector;


// Action codes. ICMPv4 and ICMPv6 have different values for their type
// and code fields. The PongActionType abstracts these differences.
typedef enum {
    PongActionReserved = 0,         // reserved code
    PongActionValid = 1,            // store valid measurement (e.g. ICMP time exceeded)
    PongActionValidLasthop = 2,     // store valid measurement, this is last hop (e.g. ICMP port unreachable)
    PongActionFiltered = 3,         // packets filtered, give up (e.g.ICMP prohibited)
    PongActionTimeout = 4,          // Timeout
    PongActionAbort = 255           // huh? we haven't a clue
} PongActionType;


typedef enum {
    probe_fb_out = 1,
    probe_fb_in = 2,
    probe_sd_out = 3,
    probe_sd_in = 4,
    probe_icmp_error = 5,
    probe_hop_in = 6,
    probe_hop_out = 7
} ProbeType;

extern const char * ptype_code[];

struct ProbeResult
{
    timeval tv, timestamp;      // timestamp recorded
    struct in_addr icmpSourceAddress;   // source address of ICMP packet
    PongActionType icmpResult;  // icmp result
    unsigned int packet_size;   // bytes in the ip packet
    int round_id;
    int probe_id;
    ProbeType probe_type;
    int ttl;

    ProbeResult( void );
    void ShowRecord( void );
};


#define VALUE_UNDEFINED     -1

struct EWMA_Value
{
    double value, value_l;
    double alpha, alpha_l;  // alpha for value and value_l
    int value_l_min_age;    // the age above which value_l will be initialized
    int age;
    
    void Initialize( double value, double alpha, double alpha_l, int value_l_min_age );
    void Update( double value_in, double confidence );
};


#define MIN_SD_AGE_MAX 200

struct MinimumValue
{
    double value, candidate_value;
    int age_max;
    int age;

    void Initialize( int age_max );
    int Update( double value );
};

#define SKEW_SUPPRESS_TIME  5

struct SkewMonitor
{
    MinimumValue * mv;
    double first_value, last_value;
    timeval first_time, last_time;
    double average_skew;
    timeval age;

    void Initialize( MinimumValue * mv );
    void CheckSkew( const timeval &time );
    void ResetAge( const timeval &current_time );
};


#define TDP_ICMP    0
#define TDP_UDP     1
#define TDP_NONE    2

struct ProbeInfo
{
    int id;
    int ttl;
    double time_offset;     // in seconds
    int technique;          // probe technique
    struct in_addr ip_fwd;  // ip address of the intermediate node on the forward path; used by the receiver side
    int tdp_state;          // state of targeted d probe; used by the receiver side
};


#define SEND_TIME_S     0x01
#define RECV_TIME_S     0x02
#define SEND_TIME_F     0x04
#define RECV_TIME_B     0x08
#define HR_ALL_TIME     0x0F

#define SEND_TIME_D     0x10
#define RECV_TIME_D     0x20
#define SEND_TIME_B     0x40
#define RECV_TIME_F     0x80

#define SENDER_DONE     0x100
#define RECEIVER_DONE   0x200
#define BOTH_DONE       0x300

struct HalfRecord
{
    int round_id;
    int probe_id;
    int status;     // A bitmap reflects which timestamps are set
    timeval send_time_sd;
    timeval recv_time_sd;
    timeval send_time_fb;
    timeval recv_time_bf;
    timeval send_time_fb_scheduled;
    double time_offset_d;   // Used by the receiver side to inform the sender side of time_offset of d probe
    int ttl_d;      // Used by the receiver side to inform the sender side of ttl of d probe
    
    void Initialize( int round_id, int probe_id );
    void SetSendTime_sd( timeval &send_time_sd );
    void SetRecvTime_sd( timeval &recv_time_sd );
    void SetSendTime_fb( timeval &send_time_fb );
    void SetRecvTime_bf( timeval &recv_time_bf );
    bool RecordComplete( bool router_probe_on );
    void Print( void );
};


struct FullRecord
{
    int round_id;
    int probe_id;
    int status;     // A bitmap reflects which timestamps are set
    timeval send_time_s, recv_time_s;
    timeval send_time_d, recv_time_d;
    timeval send_time_f, recv_time_f;
    timeval send_time_b, recv_time_b;
    timeval send_time_f_scheduled; 
    timeval send_time_b_scheduled;

    double s, d, f, b;
    double Ds, Dd, Df, Db;
    double es, ed, ef, eb;  // Measurement error
    double Lfs, Ufs, Lfd, Ufd;  // Dfs = [Lfs, Ufs], Dfd = [Lfd, Ufd]
    double Efs, Efd;    // Extent of congestion
    double Cfs, Cfd;    // Confidence of extent
    /*
    double Lbs, Ubs, Lbd, Ubd;  // Dbs = [Lbs, Ubs], Dbd = [Lbd, Ubd]
    double Ebs, Ebd;    // Extent of congestion
    double Cbs, Cbd;    // Confidence of extent
    */
    double Q_JP, Q_4P, Q_fsd, Q_fsb;    // Instantaneous QoM

    void Initialize( int round_id, int probe_id );
    bool Isset_s( void );
    bool Isset_d( void );
    bool Isset_f( void );
    bool Isset_b( void );
    bool RecordComplete( void );
    void Print( void );

    double QoM_JP( void );
    double QoM_4P( void );
    double QoM_fsd( void );
    double QoM_fsb( void );
};


struct ProbeRecord
{
    int ttl;
    int technique;
    FullRecord fr;

    void Initialize( const ProbeInfo *pi );
};


// Threshold constants
#define TH_INITIAL_SMALL    9999999
//#define TH_MIN_SMALL      150
//#define TH_MAX_SMALL      1500
#define TH_SD_CNT_MAX       20      // Threshold of s/d probes is updated every 20 probes
#define TH_FB_CNT_MAX       100     // Threshold of f/b probes is updated every 100 probes

#define TH_INIT_4P          0.4
#define TH_INIT_FSD         0.4
#define TH_INIT_FSB         0.3
#define TH_DEMOTE_4P        0.5
#define TH_DEMOTE_FSD       0.5
#define TH_DEMOTE_FSB       0.4
#define TH_PROMOTE_4P       0.6
#define TH_PROMOTE_FSD      0.6
#define TH_PROMOTE_FSB      0.5

#define TH_QoM_AGE_MIN      20      // Corresponding QoM must be updated for at least 20 times before promotion or demotion being applicable
#define TH_QoM_OLD_AGE      30      // A candidate pairing can be released if age of QoMs are larger than 30 (times)
#define TP_PERIOD           5       // Automatically tuning pairings every 5 seconds

//#define DELAY_THRESHOLD     2000    // Threshold of segment congestion
#define I_THRESHOLD_FACTOR  0.5     // I_threshold = I_THRESHOLD_FACTOR * DELAY_THRESHOLD

#define BIN_NUM             30

#define PATTERN_A           1
#define PATTERN_B           2
#define PATTERN_C           3
#define PATTERN_D           4

struct BinCount
{
    int bin_count[BIN_NUM];
    int total_count;
    float percent_below, percent_above;

    BinCount( void );
    void Reset( void );
    void Update( int bin_index );
    void Aging( float factor );
    void Output( FILE *fp );
    bool LocateValley( int& bin_index, int& pattern );
    void CheckThreshold( int threshold_index );
};


struct ProbeThreshold
{
    double small;   // the threshold below which a value is considerred to be small
    double big;     // the threshold above which a value is considerred to be big
    int count_max;  // statistics is performed on every count_max samples
    int bin_index, pattern;

    // ***************************************************
    // Statistic variables
    int count, count_below, count_between, count_above;
    double percent_below, percent_between, percent_above;
    double average_below, average_between, average_above;
    double deviation_below, deviation_between, deviation_above;
    double max_below, min_above;

    double avg_below, avg_betwn, avg_above;
    double dev_below, dev_betwn, dev_above;
    
    BinCount bin_count;
    // ***************************************************

    void Initialize( double small, int count_max );
    void Update( double value );
    static int GetBinIndex( double value );
};


#define TTL_TARGETED_D  64  // Pseudo TTL value used to identify a targeted d probe

struct PairingHistory
{
    int ttl_d;
    int times_tried;
    double average_age;
    double average_QoM;
    double QoM_JP, QoM_4P, QoM_fsd;

    void Initialize( int ttl_d );
    double Priority( void );
};


struct ProbeHistory
{
    int probe_id;
    double time_offset_d;
    int ttl_d;
    MinimumValue min_s, min_d;
    double last_Ds, last_Dd;
    EWMA_Value QoM_JP, QoM_4P, QoM_fsd, QoM_fsb;

    // Thresholds of time measured by s and d probes
    ProbeThreshold th_s, th_d;

    // ATO status
    bool ATO_pending;
    int ATA_round_id;
    vector<PairingHistory> pairing_history;

    void Initialize( int probe_id );
    void InitializeQoMs( void );
    void ResetQoMs( void );
    void InitializePairingHistory( int lower_hop_d, int upper_hop_d, bool * valid_hop_d );
    bool QoM_OldAge( void );
    void SortPairingHistory( void );
    void UpdatePairingHistory( void );
    PairingHistory * FindPairingHistory( int ttl_d );
    bool DeletePairingHistory( int ttl_d );
};


struct HalfRecordBuffer
{
    int probe_id;
    int last_round_id;
    list<HalfRecord> buffer;

    // ************ Parameters used to estimate loss rate ************
    int last_rid_bf;    // round id of the last received b or f probe (b probe for sender side and f probe for receiver side)
    int last_rid_sd;    // round id of the last received s or d probe (b probe for sender side and f probe for receiver side)
    timeval last_recv_time_bf;  // receive time of the last b or f probe
    timeval last_recv_time_sd;  // receive time of the last s or d probe
    int expected_total_bf;  // Expected number of b or f probes to have been received
    int expected_total_sd;  // Expected number of s or d probes to have been received
    int lost_bf;    // Number of b or f probes lost (or out of order)
    int lost_sd;    // Number of s or d probes lost (or out of order)
    // ********* End of parameters used to estimate loss rate ********
    
    void Initialize( int probe_id );

    // Locate the HalfRecord corresponding to <round_id>.
    // If this is a new round_id, create a new HalfRecord.
    list<HalfRecord>::iterator GetRecord( int round_id );
    void CheckIncompleteRecords( int round_id );
    void ResetLossCount( void );
    void CheckProbeLoss( ProbeResult &pr );
};


struct FullRecordBuffer
{
    int probe_id;
    int last_round_id;
    list<FullRecord> buffer;
    
    void Initialize( int probe_id );

    // Locate the FullRecord corresponding to <round_id>.
    // If this is a new round_id, create a new FullRecord.
    list<FullRecord>::iterator GetRecord( int round_id );
    void CheckIncompleteRecords( int round_id );
};


struct ProbeConfigGlobal
{
    double probe_period;  // in seconds
    int probe_num;
};


struct ProbeConfig
{
    int id;
    int ttl;
    double time_offset;     // in seconds
    int technique;          // probe technique
    struct in_addr ip_fwd;  // ip address of the intermediate node on the forward path
};


struct TargetedDprobeInfo
{
    int probe_id;
    double min_d;
    double th_d_small;
    BinCount th_d_bin;

    TargetedDprobeInfo( void ) :
        min_d( 9E9 ),
        th_d_small( VALUE_UNDEFINED )
    {}
};


struct HopInfo
{
    int ttl;
    struct in_addr ip;
    const char *host_name;
    double min_d;
    double th_d_small;
    BinCount th_d_bin;

    HopInfo( void ) :
        min_d( 9E9 ),
        th_d_small( VALUE_UNDEFINED )
    {}
};


struct HopInfoNet
{
    int ttl;
    struct in_addr ip;
};


struct PairingUpdate
{
    int probe_id;
    int ttl_d;
    double time_offset_d;
};


struct TechniqueUpdate
{
    int probe_id;
    int technique;
};


struct QoM_Log
{
    bool isset;
    int ttl_d;
    int technique;
    double QoM_JP, QoM_4P, QoM_fsd, QoM_fsb;
    ThresholdInfo_Lite th_s, th_d, th_f, th_b;
};


extern bool suppress_clock_skew_info;
extern FILE * clock_skew_fp;
extern int clock_skew_extent;
extern int clock_skew_indicator;

const char * GetMeasurementTime( void );

#endif
