#include "TMon.h"
#include "Pong_Daemon_v2.h"
#include "Pong_Report.h"

const char * ptype_code[] = { "n/a", "f_out", "f_in", "b_out", "b_in", "i_out", "i_in", "s_out", "s_in", "icmp_msg" };

int ProbeToDebug = -1;
TMon *tmon = NULL;
pthread_t tid_tmon;
unsigned int Mtu = 1500;
char *OriginHost = NULL;
char *TargetHostList = NULL;
char *Interface = NULL;
double probe_period = DEFAULT_PROBE_PERIOD;
int my_host_id;
FILE * clock_skew_fp;
unsigned long msg_signature = MSG_SIGNATURE_TMON;
long last_run_pong_id = 0;  // Message id of the last MSG_RUN_SND or MSG_RUN_RCV

TMon::TMon( void )
{
    GetCurrentTime( current_time );
    time_baseline.tv_sec = 0;
    TimeSyscall( syscallTime );
    pthread_mutex_init( &hl_mutex, NULL );
    pthread_mutex_init( &hlfrp_mutex, NULL );
    pthread_mutex_init( &log_mutex, NULL );
    dest_host_frp.resize( MAX_FRP_HOST );   // NOTE: Calling ProbeRecord::Reset() within constructor is problematic due to resize() call
    num_host_frp = 0;
    clock_offset = 0;
    srand( my_host_id + time( NULL ) );
}


void TMon::InitializeTMon( void )
{
    OpenLogFile();
    OpenClockSkewFile();
    // Do the necessary name resolution to determine the source and
    // destination of our probe packets.
    if ( SetTargetName( FAKE_TARGET ) < 0 )
    {
        exit(1);
    }
    // SetOriginName() must be after SetTargetName()
    if ( SetOriginName( OriginHost ) < 0 )
    {
        exit(1);
    }

    SetUdpSocket();

    Initialize_Common();
}


void TMon::Finalize( void )
{
#ifdef HAVE_PCAP
    pcap_breakloop( pc );
    pcap_close( pc );
    //pthread_join( tid_pcap, NULL );
#endif // HAVE_PCAP
    //pthread_join( tid_sink, NULL );
    //pthread_join( tid_frp, NULL );
    close( sock_udp );
    close( socketOut );
    close( socketIn );
    fclose( log_fp );
    fclose( clock_skew_fp );
    pthread_mutex_destroy( &hl_mutex );
    pthread_mutex_destroy( &hlfrp_mutex );
    pthread_mutex_destroy( &log_mutex );
}


void TMon::BuildHostList( void )
{
    if ( ! TargetHostList )
        return;

    FILE *fp;
    if ( ( fp = fopen( TargetHostList, "r" ) ) == NULL )
    {
        fprintf( stderr, "Fail to read file '%s'\n", TargetHostList );
        exit(1);
    }

    IF_DEBUG( 3, fprintf( stderr, "Resolving target host names ... " ) );

    char buffer[MAX_LINE_WIDTH];
    int line_no = 1;
    while ( fgets( buffer, MAX_LINE_WIDTH, fp ) )
    {
        if ( buffer[0] != '#' ) // Not a comment line
        {
            int host_id;
            char host_name[MAX_LINE_WIDTH ];
            if ( sscanf( buffer, "%d\t%s", &host_id, host_name ) != 2 )
            {
                fprintf( stderr, "Corrupt file '%s'(line %d)\n", TargetHostList, line_no );
                exit(1);
            }
            HostInfo *dest = new HostInfo( host_id, strdup( host_name ) );
            dest_host[host_id] = dest;
            // Resolve IP address in parallel using threads
            ResolveIpParallel( dest );
        }
        line_no ++;
    }
    fclose( fp );

    // Waiting for all threads to finish
    while ( ! pending_host_list.empty() )
        usleep( 100000 );   // Sleep 100 ms

    IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
}


void TMon::ResolveIpParallel( HostInfo * host )
{
    while ( pending_host_list.size() > 30 )
        usleep( 10000 );    // Sleep 10 ms
    pending_host_list.push_back( host->host_id );
    pthread_t tid;
    if ( pthread_create( &tid, NULL, resolve_ip_process, host ) != 0 )
    {
        fprintf( stderr, "pthread_create(resove_ip) failed\n" );
        exit(1);
    }
    else
        pthread_detach( tid );
}


void * resolve_ip_process( void * arg )
{
    HostInfo *host_info = (HostInfo *)arg;

    tmon->ResolveIP( host_info );

    // Remove the host from the pending list
    tmon->pending_host_list.remove( host_info->host_id );

    pthread_exit(NULL);
}


void TMon::ResolveIP( HostInfo * host_info )
{
    struct hostent *host;
    host = gethostbyname( host_info->host_name );

    if ( host == NULL )
    {
        // Resolver failed
#ifdef HAVE_HERROR
        IF_DEBUG( 4, herror( host_info->host_name ) );
#else
        IF_DEBUG( 4, fprintf( stderr, "%s: Host not found\n", host_info->host_name ) );
#endif /* HAVE_HERROR */
        RemoveHost( host_info->host_id );
    }
    else
    {
        // Resolver succeeded
        // Get IP address
        memcpy( &host_info->ip, host->h_addr_list[0], host->h_length );
        IF_DEBUG( 5, fprintf( stderr, "HOST - %3d: %15s %s\n", host_info->host_id, inet_ntoa( host_info->ip ), host->h_name ) );
        if ( host_info->ip.s_addr == originAddress.s_addr )
            // Remove itself from the host list
            RemoveHost( host_info->host_id );
    }
}


void TMon::RemoveHost( int host_id )
{
    assert( pthread_mutex_lock( &hl_mutex ) == 0 );
    dest_host.erase( host_id );
    assert( pthread_mutex_unlock( &hl_mutex ) == 0 );
}


void TMon::SendProbes( void )
{
    probe_id = 1;
    
    GetCurrentTime( time_baseline );

    if ( pthread_create( &tid_frp, NULL, frp_process, NULL ) != 0 )
    {
        fprintf( stderr, "pthread_create(frp_process) failed\n" );
        exit(1);
    }

    char time_string[50];
    time_t gmt_time = time_baseline.tv_sec;
    strcpy( time_string, asctime( gmtime( &gmt_time ) ) );
    time_string[strlen( time_string ) - 1] = 0;
    IF_DEBUG( 1, fprintf( stderr, "Measurement started at %s UTC.\n", time_string ) );
    
    int valid_host_num = dest_host.size();

    while ( ! done )
    {
        double probe_time = GetProbeTime( probe_id, probe_period, 0 );
        GetCurrentTime( current_time );
        double sleep_usec = probe_time - TimeInUsec( current_time ) - TimeInUsec( syscallTime );
        // Sleep waiting
        if ( sleep_usec > 0 )
            usleep( (useconds_t)sleep_usec );

        if ( ( probe_id & 0x1 ) == 1 )
            // Send keep-alive message every 2 rounds
            SendKeepAlive();

        // Recompute probe_interval based on new 'valid_host_num'
        double probe_interval = ( valid_host_num > 0 ? probe_period * 0.95 / valid_host_num : 0 ) * 1000000.0;
        valid_host_num = 0;

        for ( map<int, HostInfo *>::iterator it = dest_host.begin(); it != dest_host.end(); it ++ )
        {
            HostInfo *dest = it->second;
            assert( it->first == dest->host_id );
            if ( dest->hop_count != INVALID_PROBE )
            {
                valid_host_num ++;
                bool send_probe = false;
                if ( probe_id - dest->last_contact < 5 )
                {
                    // Normal probe rate
                    send_probe = true;
                }
                else if ( probe_id - dest->last_contact < 25 )
                {
                    // Backoff: 1/4 probe rate
                    if ( ( probe_id & 0x3 ) == 0 )
                        send_probe = true;
                }
                else
                {
                    // Backoff further: 1/16 probe rate
                    if ( ( probe_id & 0xF ) == 0 )
                        send_probe = true;
                }
                if ( send_probe )
                {
                    dest->probe_id = probe_id;
                    dest->ResetRawTime();
                    if ( DebugOn( dest->host_id ) )
                        IF_DEBUG( 7, fprintf( stderr, "PROBE - host_id=%d, probe_id=%ld\n", dest->host_id, probe_id ) );
                    SendUdpFprobe( dest->ip, dest->host_id, probe_id );
                    if ( dest->hop_count > 0 && probe_id - dest->last_contact < 3 )
                        TraceRouteStart( dest );
                }
                usleep( (unsigned long)probe_interval );
                GetCurrentTime( current_time );
            }
        }

        // Flush log file
        assert( pthread_mutex_lock( &log_mutex ) == 0 );
        fflush( log_fp );
        assert( pthread_mutex_unlock( &log_mutex ) == 0 );

        probe_id ++;
    }
}


void TMon::SendFrProbes( void )
{
    frp_probe_id = 1;
    const double probe_interval = DEFAULT_PROBE_PERIOD_FP / MAX_FRP_HOST;

    while ( ! done )
    {
        bool flag_1sec = false;
        bool flag_30sec = false;
        if ( frp_probe_id % 5 == 0 )
        {
            // Update every 1sec
            flag_1sec = true;
            if ( frp_probe_id % 150 == 0 )
            {
                // Update every 30sec
                flag_30sec = true;
                fflush( clock_skew_fp );
            }
        }
        for ( int i = 0; i < MAX_FRP_HOST; i ++ )
        {
            double probe_time = GetProbeTime( frp_probe_id, DEFAULT_PROBE_PERIOD_FP, i * probe_interval );
            GetCurrentTime( current_time );
            double sleep_usec = probe_time - TimeInUsec( current_time ) - TimeInUsec( syscallTime );
            // Sleep waiting
            if ( sleep_usec > 0 )
                usleep( (useconds_t)sleep_usec );

            ProbeRecord& dest_frp = dest_host_frp[i];
            if ( dest_frp.host_id != -1 )
            {
                if ( flag_1sec )
                {
                    // Aging clock_skew_indicator every 1sec
                    if ( dest_frp.clock_skew_indicator > 0 )
                        dest_frp.clock_skew_indicator --;
                }

                // Try to send fast-rate probe
                map<int, HostInfo *>::iterator it = dest_host.find( dest_frp.host_id );
                if ( it != dest_host.end() )
                {
                    HostInfo *dest = it->second;
                    assert( it->first == dest->host_id );
                    if ( dest->hop_count != INVALID_PROBE )
                    {
                        bool send_probe = false;
                        if ( frp_probe_id - dest_frp.last_contact < 25 )
                        {
                            // Normal probe rate
                            send_probe = true;
                        }
                        else if ( frp_probe_id - dest_frp.last_contact < 50 )
                        {
                            // Backoff: 1/4 probe rate
                            if ( ( frp_probe_id & 0x3 ) == 0 )
                                send_probe = true;
                        }
                        else if ( frp_probe_id - dest_frp.last_contact < 150 )
                        {
                            // Backoff further: 1/16 probe rate
                            if ( ( frp_probe_id & 0xF ) == 0 )
                                send_probe = true;
                        }else
                        {
                            // Backoff further: 1/64 probe rate
                            if ( ( frp_probe_id & 0x3F ) == 0 )
                                send_probe = true;
                        }
                        if ( send_probe )
                        {
                            /*
                            if ( DebugOn( dest->host_id ) )
                                fprintf( stderr, "FR-PROBE - host_id=%d, probe_id=%ld, last_contact=%ld\n", dest->host_id, frp_probe_id, dest_frp.last_contact );
                            */
                            SendUdpFprobe( dest->ip, dest->host_id | FRP_FLAG, frp_probe_id );
                        }
                    }
                }
            }
        }
        frp_probe_id ++;
    }
}


void TimeSyscall( timeval &diff )
{
    timeval t1, t2;

    gettimeofday( &t1, NULL );
    gettimeofday( &t2, NULL );
    
    diff.tv_sec = t2.tv_sec - t1.tv_sec;
    diff.tv_usec = t2.tv_usec - t1.tv_usec;
    if ( diff.tv_usec < 0 )
    {
        diff.tv_usec += 1000000;
        diff.tv_sec--;
    }
}


// usec = GetProbeTime( long probe_id, double time_offset )

double TMon::GetProbeTime( long probe_id, double probe_period, double time_offset )
{
    double usec = TimeInUsec( time_baseline );
    usec += ( ( probe_id - 1 ) * probe_period + time_offset ) * 1000000.0;
    return usec;
}


int TMon::GetSocketOut()
{
    int rc;
    
    // NOTE: Using IPPROTO_UDP, IPPROTO_ICMP or IPPROTO_RAW works the same here
    // We send both UDP packets and ICMP packets through this socket
    socketOut = socket( AF_INET, SOCK_RAW, IPPROTO_RAW );
    if ( socketOut < 0 )
    {
        perror( "socket" );
        return socketOut;
    }
    
    int hdrinclOption = 1;
    rc = setsockopt( socketOut, IPPROTO_IP, IP_HDRINCL, &hdrinclOption, sizeof( hdrinclOption ) );
    if ( rc < 0 )
    {
        perror( "setsockopt( IP_HDRINCL )" );
        return rc;
    }

    // Bind remote side of socket
    rc = connect( socketOut, (struct sockaddr *)&target_udp_addr, sizeof( struct sockaddr_in ) );
    if ( rc < 0 )
    {
        perror( "connect" );
        return rc;
    }

    return socketOut;
}


//
// SetOriginName
//
// Input:  origin hostname
//
// Output:  success code (negative if error)
//
// Attempt to set the origin of probe packets emanating from this host,
// for tests where it is applicable, using protocol family-dependent
// resolution as necessary.  If no origin hostname is given, 
// then use the target address to compute the correct outgoing interface.
// originName and originAddress are set.
//

int TMon::SetOriginName( const char *o )
{
    // If nothing was passed in, then bind a dummy socket to try
    // to figure out where packets exit this host.
    if (o == NULL)
    {
        int dummySock;
        struct sockaddr_in dummyAddr, localAddr;
#ifdef HAVE_SOCKLEN_T
        socklen_t localAddrLength;
#else /* HAVE_SOCKLEN_T */
#ifdef NEED_GETSOCKNAME_HACK
        int localAddrLength;
#else /* NEED_GETSOCKNAME_HACK */
        size_t localAddrLength;
#endif /* NEED_GETSOCKNAME_HACK */
#endif /* HAVE_SOCKLEN_T */

        // Create socket, then connect to it, and then read out
        // the local socket address with a getsockname call.
        dummySock = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
        if ( dummySock < 0 )
        {
            perror( "socket" );
            return -1;
        }

        memset( (void *)&dummyAddr, 0, sizeof( struct sockaddr_in ) );
#ifdef HAVE_SOCKADDR_SA_LEN
        dummyAddr.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
        dummyAddr.sin_family = AF_INET;
        dummyAddr.sin_port = htons( 32768 ); // port number irrelevant
        memcpy( &dummyAddr.sin_addr, &targetAddress, sizeof( struct in_addr ) );

        if ( connect( dummySock, (sockaddr *)&dummyAddr, sizeof( struct sockaddr_in ) ) < 0 )
        {
            perror( "connect" );
            return -1;
        }

        localAddrLength = sizeof( sockaddr_in );
        memset( (void *)&localAddr, 0, sizeof( struct sockaddr_in ) );
        if ( getsockname( dummySock, (struct sockaddr *)&localAddr, &localAddrLength ) < 0)
        {
            perror( "getsockname" );
            return -1;
        }

        // Got the local address, now do a reverse DNS looup
        memcpy( &originAddress, &localAddr.sin_addr, sizeof( struct in_addr ) );
        originName = strdup( GetName( &originAddress ) );
        if ( originName == NULL )
        {
            fprintf( stderr, "Couldn't allocate memory for origin hostname.\n" );
            return -1;
        }

        close( dummySock );

        IF_DEBUG( 3, fprintf( stderr, "originName = %s\n", originName ) );
        IF_DEBUG( 3, fprintf( stderr, "originAddress = %s\n", inet_ntoa( originAddress ) ) );
    }

    // User gave a source name/address.  Attempt to resolve, if possible.
    else {
        struct hostent *host;    // resolver host entry
        host = gethostbyname( o );

        // Resolver failed
        if ( host == NULL )
        {

#ifdef HAVE_HERROR
            herror( o );
#else
            fprintf( stderr, "%s: Host not found\n", o );
#endif /* HAVE_HERROR */

            memset( (void *)&originAddress, 0, sizeof( struct in_addr ) );
            originName = strdup( o );

            if ( originName == NULL )
            {
                fprintf( stderr, "Couldn't allocate space for origin hostname.\n" );
                return -1;
            }
            return -1;
        }

        IF_DEBUG( 3, fprintf( stderr, "originName = %s\n", host->h_name ) );
        //IF_DEBUG( 3, fprintf( stderr, "h_length = %d\n", host->h_length) );
        IF_DEBUG( 3, fprintf( stderr, "originAddress = %s\n", inet_ntoa( *((struct in_addr *)(host->h_addr_list[0])) ) ) );
             
        // Get IP address
        memcpy( &originAddress, host->h_addr_list[0], host->h_length );

        // Make a copy of the canonical hostname
        originName = strdup( host->h_name );
        if ( originName == NULL )
        {
            fprintf( stderr, "Couldn't allocate memory for origin hostname.\n" );
            return -1;
        }
    }

    return 0;
}


//
// SetTargetName
//
// Input:  target hostname (target)
//
// Output:  success code (negative if error)
//
// Set target name and do protocol-dependent resolving to get a 
// network address.  In case of an error, we're responsible for
// printing some error message.
//

int TMon::SetTargetName( const char *t )
{
    struct hostent *host;    // resolver host entry

    // Attempt to resolve, if possible
    host = gethostbyname( t );

    // Resolver failed
    if ( host == NULL )
    {

        // Some systems don't have herror (non-BSD?), so for those,
        // we'll cobble together an error message.
#ifdef HAVE_HERROR
        herror( t );
#else
        fprintf( stderr, "%s: Host not found\n", t );
#endif /* HAVE_HERROR */

        memset( (void *)&targetAddress, 0, sizeof( struct in_addr ) );
        targetName = strdup( t );
        if ( targetName == NULL )
        {
            fprintf( stderr, "Couldn't allocate memory for target hostname.\n" );
            return -1;
        }
        return -1;
    }

    //IF_DEBUG( 3, fprintf( stderr, "targetName = %s\n", host->h_name ) );
    //IF_DEBUG( 3, fprintf( stderr, "targetAddress = %s\n", inet_ntoa( *((struct in_addr *)(host->h_addr_list[0])) ) ) );
         
    // Get IP address
    memcpy( &targetAddress, host->h_addr_list[0], host->h_length );

    memset( (void *)&target_udp_addr, 0, sizeof( struct sockaddr_in ) );
    // Note:  Only BSD4.3Reno and later have sin_len in struct
    // sockaddr_in, so we need to test for it.
#ifdef HAVE_SOCKADDR_SA_LEN
    target_udp_addr.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
    target_udp_addr.sin_family = AF_INET;
    target_udp_addr.sin_port = htons( UDPTMON_PORT );

    memcpy( &target_udp_addr.sin_addr, host->h_addr_list[0], host->h_length );

    // Make a copy of the canonical hostname
    targetName = strdup( host->h_name );
    if ( targetName == NULL )
    {
        fprintf( stderr, "Couldn't allocate memory for target hostname.\n" );
        return -1;
    }

    return 0;
}


//
// GetSocketIn
//
// Input:  None
//
// Output:  In return value, returns socket number.
//
// Get input socket of an appropriate type.
//

int TMon::GetSocketIn()
{
#ifdef _PLANETLAB_H_
    socketIn = socket( PF_INET, SOCK_RAW, IPPROTO_ICMP_UDP );
#else
    socketIn = socket( PF_INET, SOCK_RAW, IPPROTO_ICMP );
#endif /* _PLANETLAB_H_ */

    if ( socketIn < 0 )
    {
        perror( "socket" );
        return socketIn;
    }

#ifdef _PLANETLAB_H_
    struct sockaddr_in sin;
    memset( &sin, 0, sizeof( sin ) );
    sin.sin_port = htons( UDPTMON_PORT );
  
    bind( socketIn, (struct sockaddr *)&sin, sizeof( sin ) );
#endif /* _PLANETLAB_H_ */

    return socketIn;
}


//
// GetName
//
// Input:  Pointer to address structure
//
// Output:  Pointer to ASCII representation of name (in return
// value)
//

char * TMon::GetName( const void *a )
{
    struct hostent *host;
    host = gethostbyaddr( a, sizeof( struct in_addr ), AF_INET );

    if ( host )
    {
        return ( host->h_name );
    }
    else
    {
        return ( GetPrintableAddress( a ) );
    }
}


//
// GetPrintableAddress
//
// Input:  Pointer to address structure
//
// Output:  Pointer to ASCII representation of address (in return
// value)
//

char * TMon::GetPrintableAddress( const void * a )
{
    return ( inet_ntoa( *((struct in_addr *)a) ) );
}


//
// GetAction
//
// Input:  a test record
//
// Output:  action code
//
// Figure out the meaning of a particular combination of ICMPv4 type
// and code values.
//

TMonActionType TMon::GetAction( int icmpType, int icmpCode ) 
{
    if ( icmpType == ICMP_TIMXCEED )
    {
        return TMonActionValid;
    }
    else if ( ( icmpType == ICMP_UNREACH ) && ( icmpCode == ICMP_UNREACH_PORT ) )
    {
        return TMonActionValidLasthop;
    }
    else if ( ( icmpType == ICMP_UNREACH ) && ( icmpCode == ICMP_UNREACH_FILTER_PROHIB ) )
    {
        return TMonActionFiltered;
    }
    else
    {
        return TMonActionAbort;
    }
}


void TMon::SetUdpSocket( void )
{
    /* 
    Data stream: UDP socket 
    */
    if ( ( sock_udp = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_DGRAM, 0 )" );
        exit(1);
    }

#ifdef linux
    int bsdcompatOption;
    bsdcompatOption = 1;
    if ( setsockopt( sock_udp, SOL_SOCKET, SO_BSDCOMPAT, &bsdcompatOption, sizeof( bsdcompatOption ) ) < 0 )
    {
        perror( "setsockopt( SO_BSDCOMPAT )" );
        exit(1);
    }
#endif /* linux */

    memset( (void *) &origin_udp_addr, 0, sizeof( struct sockaddr_in ) );
#ifdef HAVE_SOCKADDR_SA_LEN
    origin_udp_addr.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
    origin_udp_addr.sin_family = AF_INET;
    origin_udp_addr.sin_port = htons( UDPTMON_PORT );
    memcpy( &(origin_udp_addr.sin_addr), &originAddress, sizeof( struct in_addr ) );
    
    if ( bind( sock_udp, (struct sockaddr *)&origin_udp_addr, sizeof( origin_udp_addr ) ) < 0 )
    {
        perror( "bind( sock_udp )" );
        exit(1);
    }
}


void TMon::Initialize_Common( void )
{
    // Get sockets
    if ( GetSocketOut() < 0 )
    {
        exit(1);
    }
    if ( GetSocketIn() < 0 )
    {
        exit(1);
    }

    IF_DEBUG( 3, fprintf( stderr, "Traceroute period = %g sec\n", probe_period ) );

    BuildHostList();

    // Test
    //AddFrpHost( 364 );
    //RandomFrpHost();

#ifdef HAVE_PCAP
    SetupPCAP();
#endif // HAVE_PCAP

    SetupPacketSink();
}


void TMon::SetupPacketSink()
{
    if ( pthread_create( &tid_sink, NULL, sink_process, this ) != 0 )
    {
        fprintf( stderr, "pthread_create(sink) failed\n" );
        exit(1);
    }
    IF_DEBUG( 3, fprintf( stderr, "Packet sink process installed\n" ) );
}


void * sink_process( void * arg )
{
    TMon *obj = (TMon *)arg;
    
    obj->ReceiveMessages();

    pthread_exit(NULL);
    
    return NULL;
}


// Receive probe packets

void TMon::ReceiveMessages( void )
{
    int max_fd = MAX( socketIn, sock_udp );

    while ( ! done )
    {
        fd_set read_fds;

        FD_ZERO( &read_fds );
        FD_SET( socketIn, &read_fds );
        FD_SET( sock_udp, &read_fds );

        int rc;
        rc = select( max_fd + 1, &read_fds, NULL, NULL, NULL );
        if ( rc < 0 )
        {
            perror( "select" );
            exit(1);
        }

        ProbeResult pr;
        
        GetCurrentTime( pr.tv );
        
        if ( FD_ISSET( socketIn, &read_fds ) )
        {
            IF_DEBUG( 9, fprintf( stderr, "Router probe received\n" ) );
            RecvRouterProbe( pr );
        }
        if ( FD_ISSET( sock_udp, &read_fds ) )
        {
            IF_DEBUG( 9, fprintf( stderr, "End-to-end probe received\n" ) );
            RecvUdpProbe( pr );
        }
    }
}


double TimeInMsec( timeval &tv )
{
    return tv.tv_sec * 1000.0 + tv.tv_usec / 1000.0;
}

double delta_t( const timeval &t1, const timeval &t2 )
{
    double dt;
    dt = (double)( t2.tv_sec - t1.tv_sec) * 1000000.0
       + (double)( t2.tv_usec - t1.tv_usec);
    return dt;
}


timeval Timeval( double usec )
{
    timeval tv;
    tv.tv_sec = (int)( usec / 1E6 );
    tv.tv_usec = (int)( usec - tv.tv_sec * 1E6 );
    return tv;
}


double TimeInUsec( timeval &tv )
{
    return tv.tv_sec * 1000000.0 + tv.tv_usec;
}


void GetCurrentTime( timeval &time )
{
    gettimeofday( &time, (struct timezone *)NULL );
}


ProbeResult::ProbeResult( void )
{
    bzero( this, sizeof( ProbeResult ) );
}


void ProbeResult::ShowRecord( void )
{
    if ( probe_type == probe_icmp_error )
    {
        fprintf( stderr, "- probe_type=%s, icmpResult=%d", ptype_code[probe_type], icmpResult );
    }
    else
    {
        fprintf( stderr, "- probe_type=%s, host_id=%d", ptype_code[probe_type], host_id & ~FRP_FLAG );
        if ( probe_type == probe_s_in || probe_type == probe_s_out )
            fprintf( stderr, ", hop_id=%d", this->hop_id );
        else
            fprintf( stderr, ", probe_id=%ld", probe_id );
        if ( probe_type == probe_f_in || probe_type == probe_b_in || probe_type == probe_i_in )
            fprintf( stderr, ", remote_address=%s", inet_ntoa( remote_address ) );
        if ( probe_type == probe_i_out || probe_type == probe_i_in )
            fprintf( stderr, ", b_send_time=%.3lf(ms)", TimeInMsec( time_value ) );
        else if ( probe_type == probe_b_out || probe_type == probe_b_in )
            fprintf( stderr, ", hop_count=%d, f_recv_time=%.3lf(ms)", hop_count, TimeInMsec( time_value ) );
        else if ( probe_type == probe_s_in )
            fprintf( stderr, ", icmpSourceAddress=%s", inet_ntoa( icmpSourceAddress ) );
        else if ( probe_type == probe_s_out )
            fprintf( stderr, ", ttl=%d", ttl );
        //fprintf( stderr, ", tv=%.3lf(ms)", TimeInMsec( tv ) );
    }
    fputc( '\n', stderr );
}


ProbeUdpPayload::ProbeUdpPayload( void )
{
    // Set signature
    strncpy( signature, "TMon", 4 );
    bzero( padding, PAD_NUM );
    host_id = 0;
    hop_count = 0;
    probe_id = 0;
    time_value.tv_sec = 0;
    time_value.tv_usec = 0;
}


bool ProbeUdpPayload::IsValid( void )
{
    if ( strncmp( signature, "TMon", 4 ) != 0 )
    {
        fprintf( stderr, "signature\n" );
        return false;
    }
    if ( type_code < PL_UDP_F || type_code > PL_ROUTER )
        return false;
    for ( int i = 0; i < PAD_NUM; i ++ )
        if ( padding[i] )
            return false;
    return true;
}


void ProbeUdpPayload::SetRouterProbe( void )
{
    type_code = PL_ROUTER;
}


void ProbeUdpPayload::SetUdpFprobe( int host_id, long probe_id )
{
    type_code = PL_UDP_F;
    this->host_id = (short)host_id;
    this->probe_id = probe_id;
}


void ProbeUdpPayload::SetUdpBprobe( int host_id, long probe_id, int hop_count, timeval f_recv_time )
{
    type_code = PL_UDP_B;
    this->host_id = (short)host_id;
    this->hop_count = (char)hop_count;
    this->probe_id = probe_id;
    time_value = f_recv_time;
}


void ProbeUdpPayload::SetUdpIprobe( int host_id, long probe_id, timeval b_send_time )
{
    type_code = PL_UDP_I;
    this->host_id = (short)host_id;
    this->probe_id = probe_id;
    time_value = b_send_time;
}


// ************** PCAP related functions **************

#ifdef HAVE_PCAP

void TMon::SetupPCAP()
{
    int fileno;
    char PcapErrBuf[PCAP_ERRBUF_SIZE];

    if ( Interface == NULL )
    {
        Interface = pcap_lookupdev( PcapErrBuf );
        if ( Interface == NULL )
        {
            fprintf( stderr, "%s\n", PcapErrBuf );
            exit(1);
        }
    }

    IF_DEBUG( 3, fprintf( stderr, "PCAP interface = %s\n", Interface ) );

    pc = pcap_open_live( Interface, Mtu, 0, 3000, PcapErrBuf );
    if ( pc == NULL )
    {
        fprintf( stderr, "%s\n", PcapErrBuf );
        assert(0);
    }

    bpf_u_int32 netp, maskp;    // net and mask parameters

    if ( pcap_lookupnet( Interface, &netp, &maskp, PcapErrBuf ) < 0 )
    {
        fprintf( stderr, "%s\n", PcapErrBuf );
        assert(0);
    }

    fileno = pcap_fileno( pc );
    if ( fileno < 0 )
    {
        fprintf( stderr, "pcap_fileno() failed\n" );
        exit(1);
    }

#ifdef HAVE_BPF
    // Set "immediate" mode in BPF.  We need this to avoid libpcap
    // waiting the entire timeout period before passing any received
    // packets to us.
    u_int immediate = 1;
    if ( ioctl( fileno, BIOCIMMEDIATE, &immediate ) < 0 )
    {
        perror( "ioctl( BIOCIMMEDIATE )" );
        exit(1);
    }
#endif // HAVE_BPF

    // Initialize packet filter
    // 
    // Note that we will see both inbound and outbound packets
    // with this filter rule.  We *could* write up a new rule 
    // to add a src host predicate.

    char filter[200];
    sprintf( filter, "(ip proto \\icmp) or (udp port %d)", UDPTMON_PORT );
    struct bpf_program fp;      // filter program
    if ( pcap_compile( pc, &fp, filter, 1, maskp ) < 0 )
    {
        fprintf( stderr, "pcap_compile failed\n" );
        exit(1);
    }
    
    if ( pcap_setfilter( pc, &fp ) < 0 )
    {
        fprintf( stderr, "pcap_setfilter failed\n" );
        exit(1);
    }

    // Set packet offset based on datalink type
    //
    // So far we only deals with Ethernet-like datalinks.  We need
    // something like what tcpdump's packet-printing does.
    packet_offset = 0;
    int link_type = pcap_datalink( pc );
    switch ( link_type )
    {
        case DLT_NULL:
            break;

        case DLT_EN10MB:
            packet_offset = sizeof( struct ether_header );
            break;

        default:
            fprintf( stderr, "Unknown datalink layer %d\n", link_type );
            exit(1);
    }

    if ( pthread_create( &tid_pcap, NULL, pcap_process, this ) != 0 )
    {
        fprintf( stderr, "pthread_create(pcap) failed\n" );
        exit(1);
    }

    IF_DEBUG( 3, fprintf( stderr, "Packet capture process installed\n" ) );
}


void * pcap_process( void * arg )
{
    TMon *obj = (TMon *)arg;

    // Run the capture process
    if ( pcap_loop( obj->pc, -1, pcap_callback, (u_char *)arg ) == -1 )
    // NOTE: Returning -2 is not an error. It is due to pcap_breakloop call. 
    {
        fprintf( stderr, "pcap_loop failed\n" );
        exit(1);
    }

    pthread_exit( NULL );

    return NULL;
}

//
// TMon::pcap_callback
//
// Inputs:  Object pointer, pcap_pkthdr pointer, pointer to packet
// data.
//
// Outputs:  None.
//
// pcap callback routine.
//
void pcap_callback( u_char *puc, const struct pcap_pkthdr *ph, const u_char *pd )
{
    TMon *obj = (TMon *)puc;
    unsigned int packetLength = ph->len - obj->packet_offset;
    char *packet = (char *)pd + obj->packet_offset;

    ProbeResult pr;
    pr.tv = ph->ts;
    pr.packet_size = packetLength;

    if ( obj->ParsePacket( packet, pr ) == 0 )
    {
        bool is_probe = false;
        int real_host_id = ( pr.host_id & ~FRP_FLAG );
        switch ( pr.probe_type )
        {
            case probe_f_out:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "f probe (outbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_f_in:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "f probe (inbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_b_out:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "b probe (outbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_b_in:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "b probe (inbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_i_out:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "i packet (outbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_i_in:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "i packet (inbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_s_out:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "s probe (outbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_s_in:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "s probe (inbound) captured\n" ) );
                is_probe = true;
                break;
            case probe_icmp_error:
                if ( DebugOn( real_host_id ) )
                    IF_DEBUG( 8, fprintf( stderr, "ICMP message from the other side captured\n" ) );
                break;
            default:;
        }
        if ( DebugOn( real_host_id ) )
            IF_DEBUG( 8, pr.ShowRecord() );
        if ( is_probe )
        {
            obj->ProcessProbes( pr );
        }
    }
}

#endif // HAVE_PCAP

// *********** End of PCAP related functions ***********


void TMon::ProcessProbes( ProbeResult &pr )
{
    int real_host_id = ( pr.host_id & ~FRP_FLAG );

    switch ( pr.probe_type )
    {
        case probe_f_out:
            {
                if ( pr.host_id & FRP_FLAG )
                {
                    vector<ProbeRecord>::iterator it = FindFrpHost( real_host_id );
                    if ( it != dest_host_frp.end() )
                    {
                        PendingProbe record( pr.probe_id, pr.tv );
                        it->Enqueue( record );
                    }
                }
                else
                {
                    map<int, HostInfo *>::iterator it = dest_host.find( real_host_id );
                    if ( it != dest_host.end() )
                    {
                        HostInfo *dest = it->second;
                        if ( pr.probe_id == dest->probe_id )
                            dest->send_time_f = pr.tv;
                    }
                }
            }
            break;

        case probe_f_in:
            {
                SendUdpBprobe( pr.remote_address, pr.host_id, pr.probe_id, ORIGIN_TTL - pr.ttl, pr.tv );
            }
            break;

        case probe_b_out:
            {
                SendUdpIprobe( pr.remote_address, pr.host_id, pr.probe_id, pr.tv );
            }
            break;

        case probe_b_in:
            {
                if ( pr.host_id & FRP_FLAG )
                {
                    vector<ProbeRecord>::iterator it = FindFrpHost( real_host_id );
                    if ( it != dest_host_frp.end() )
                    {
                        it->UpdateRecord( pr.probe_id, recv_time_fb, pr );
                    }
                }
                else
                {
                    map<int, HostInfo *>::iterator it = dest_host.find( real_host_id );
                    if ( it != dest_host.end() )
                    {
                        HostInfo *dest = it->second;
                        if ( pr.probe_id == dest->probe_id )
                        {
                            dest->recv_time_b = pr.tv;
                            dest->recv_time_f = pr.time_value;
                            if ( dest->hop_count == 0 && pr.hop_count < MIN_HOP_NUM )
                                dest->hop_count = INVALID_PROBE;
                            else
                                dest->hop_count = pr.hop_count;
                            dest->UpdateContact();
                        }
                    }
                }
            }
            break;

        case probe_i_in:
            {
                if ( pr.host_id & FRP_FLAG )
                {
                    vector<ProbeRecord>::iterator it = FindFrpHost( real_host_id );
                    if ( it != dest_host_frp.end() )
                    {
                        it->UpdateRecord( pr.probe_id, send_time_b, pr );
                    }
                }
                else
                {
                    map<int, HostInfo *>::iterator it = dest_host.find( real_host_id );
                    if ( it != dest_host.end() )
                    {
                        HostInfo *dest = it->second;
                        if ( pr.probe_id == dest->probe_id )
                        {
                            dest->send_time_b = pr.time_value;
                            dest->UpdateContact();
                            if ( DebugOn( real_host_id ) )
                                IF_DEBUG( 8, fprintf( stderr, "- send_f=%.3lf(ms) recv_b=%.3lf(ms) recv_f=%.3lf(ms) send_b=%.3lf(ms)\n", TimeInMsec( dest->send_time_f ), TimeInMsec( dest->recv_time_b ), TimeInMsec( dest->recv_time_f ), TimeInMsec( dest->send_time_b ) ) );
                            double rtt = dest->GetRtt();
                            if ( rtt )
                            {
                                dest->UpdateRttEwma( rtt );
                                double rtt_ms = rtt * 0.001;
                                if ( dest->hop_count != INVALID_PROBE )
                                {
                                    // Output to log file
                                    assert( pthread_mutex_lock( &log_mutex ) == 0 );
                                    fprintf( log_fp, "%ld\tRTT\t%d\t%ld\t%d\t%.3lf\n", current_time.tv_sec, real_host_id, pr.probe_id, dest->hop_count, rtt_ms );
                                    assert( pthread_mutex_unlock( &log_mutex ) == 0 );
                                }
                                if ( DebugOn( real_host_id ) )
                                    IF_DEBUG( 7, fprintf( stderr, "RTT - host_id=%d, probe_id=%ld, hop_count=%d, rtt=%.3lf(ms)\n", real_host_id, pr.probe_id, dest->hop_count, rtt_ms ) );
                                
                            }
                        }
                    }
                }
            }
            break;

        case probe_s_in:
            {
                map<int, HostInfo *>::iterator it = dest_host.find( real_host_id );
                if ( it != dest_host.end() )
                {
                    HostInfo *dest = it->second;
                    if ( dest->tracing && pr.hop_id >= 1 && pr.hop_id <= (int)dest->hop_list.size() )
                        dest->hop_list[pr.hop_id - 1] = pr.icmpSourceAddress;
                }
            }
            break;

        default:;
    }
}


int TMon::ParsePacket( char *ipPacket, ProbeResult &pr )
{
    if ( pr.packet_size < sizeof( ip ) )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete IP packet of %d bytes\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }
    
    ip *ipHeader = (ip *)ipPacket;

    // Check protocol in IP header
    switch ( ipHeader->ip_p )
    {
    case IPPROTO_ICMP:
        return ProcessIcmpPacket( ipPacket, pr );

    case IPPROTO_UDP:
        return ProcessUdpPacket( ipPacket, pr );

    default:
        fprintf( stderr, "Received unknown protocol %d\n", ipHeader->ip_p );
        return -1;
    }

    return 0;
}


int TMon::ProcessIcmpPacket( const char *ipPacket, ProbeResult &pr, bool silent )
{
    // Parse the ICMP packet, doing a little error checking along
    // the way.  By the end, we'll have ipHeader and icmpHeader
    // pointing to valid structures within the packet, and
    // ipHeader2 pointing to the IP header of the generating
    // IP packet.

    ip *ipHeader, *ipHeader2;
    icmp *icmpHeader;
    udphdr *udpHeader;
    unsigned int ipHeaderLength, ipHeaderLength2;

    ipHeader = (ip *)ipPacket;

#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeaderLength = ( ipHeader->ip_vhl & 0x0f ) << 2;
#else
    ipHeaderLength = ipHeader->ip_hl << 2;
#endif /* __osf__ */

    if ( pr.packet_size - ( 0 + ipHeaderLength ) < ICMP_MINLEN )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete ICMP packet of %d bytes\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }

    icmpHeader = (icmp *)( ((char *)ipHeader) + ipHeaderLength );

    pr.time_value.tv_sec = pr.time_value.tv_usec = 0;
    memcpy( &pr.icmpSourceAddress, &(ipHeader->ip_src), sizeof( struct in_addr ) );
    pr.icmpResult = GetAction( icmpHeader->icmp_type, icmpHeader->icmp_code );

    if ( ( icmpHeader->icmp_type != ICMP_TIMXCEED ) && ( icmpHeader->icmp_type != ICMP_UNREACH ) )
    {
        if ( ! silent )
        {
            IF_DEBUG( 5, fprintf( stderr, "%s Ignoring ICMP packet\n", GetMeasurementTime() ) );
            IF_DEBUG( 5, fprintf( stderr, "ICMP type = %d, code = %d, action = %d, src_ip = %s", icmpHeader->icmp_type, icmpHeader->icmp_code, pr.icmpResult, inet_ntoa( ipHeader->ip_src ) ) );
            IF_DEBUG( 5, fprintf( stderr, ", dest_ip = %s\n", inet_ntoa( ipHeader->ip_dst ) ) );
        }
        return -1;
    }
    
    if ( pr.packet_size - ( 0 + ipHeaderLength + ICMP_MINLEN ) < sizeof( ip ) )
    {
        if ( ! silent )
            IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete inner IP packet, %d bytes total\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }

    // Check for a valid (to us) IP header within the packet.
    // For "time exceeded" or "destination unreachable", this
    // header will be 8 bytes past the ICMP header.
    ipHeader2 = (ip *)( (char *)icmpHeader + ICMP_MINLEN );

#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeaderLength2 = (ipHeader2->ip_vhl & 0x0f) << 2;
#else
    ipHeaderLength2 = ipHeader2->ip_hl << 2;
#endif /* __osf__ */

    // Additional checking here ... must be UDP
    if ( ipHeader2->ip_p != IPPROTO_UDP )
    {
        if ( ! silent )
        {
            IF_DEBUG( 9, fprintf( stderr, "%s Ignoring ICMP packet for unknown non-UDP packet\n", GetMeasurementTime() ) );
            IF_DEBUG( 9, fprintf( stderr, "ICMP type = %d, code = %d, action = %d, src_ip = %s", icmpHeader->icmp_type, icmpHeader->icmp_code, pr.icmpResult, inet_ntoa( ipHeader->ip_src ) ) );
            IF_DEBUG( 9, fprintf( stderr, ", dest_ip = %s\n", inet_ntoa( ipHeader->ip_dst ) ) );
        }
        return -1;
    }

    // Align UDP header template.
    if ( pr.packet_size - ( 0 + ipHeaderLength + ICMP_MINLEN + ipHeaderLength2 ) < sizeof( udphdr ) )
    {
        if ( ! silent )
            IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete inner UDP packet, %d bytes total\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }

    udpHeader = (udphdr *)( (char *)ipHeader2 + ipHeaderLength2 );

    // Check source and destination UDP port number and UDP
    // length (header+payload)
    if ( ( ntohs( udpHeader->uh_sport ) != S_UDPTMON_PORT ) || ( ntohs( udpHeader->uh_dport ) != UDPTMON_PORT ) || ( ntohs( udpHeader->uh_ulen ) != PROBE_PACKET_SIZE + sizeof( udphdr ) ) )
    {
        if ( ! silent )   
            IF_DEBUG( 9, fprintf( stderr, "%s Ignoring ICMP packet for unknown UDP packet\n", GetMeasurementTime() ) );
        return -1;
    }

    if ( pr.icmpResult == TMonActionFiltered )
    {
        if ( ! silent )
        {
            IF_DEBUG( 3, fprintf( stderr, "%s TTL limited UDP probe has been blocked, probe_id = %ld\n", GetMeasurementTime(), pr.probe_id ) );
        }
        return -1;
    }

    // Fill in return fields
    int ip_id = ntohs( ipHeader2->ip_id );
    pr.host_id = ip_id >> 6;
    pr.hop_id = ip_id & 0x3F;
    if ( pr.icmpResult != TMonActionValid )
        pr.probe_type = probe_icmp_error;
    else
        pr.probe_type = probe_s_in;

    return 0;
}


int TMon::ProcessUdpPacket( const char *ipPacket, ProbeResult &pr )
{
    ip *ipHeader;
    udphdr *udpHeader;
    unsigned int ipHeaderLength;

    ipHeader = (ip *)ipPacket;

#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeaderLength = ( ipHeader->ip_vhl & 0x0f ) << 2;
#else
    ipHeaderLength = ipHeader->ip_hl << 2;
#endif /* __osf__ */

    udpHeader = (udphdr *)( (char *)ipHeader + ipHeaderLength );

    // Check source and destination UDP port number and UDP
    // length (header+payload)
    if ( ( ntohs( udpHeader->uh_sport ) != S_UDPTMON_PORT ) || ( ntohs( udpHeader->uh_dport ) != UDPTMON_PORT ) || ( ntohs( udpHeader->uh_ulen ) != PROBE_PACKET_SIZE + sizeof( udphdr ) ) )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Ignoring unknown(port) UDP packet\n", GetMeasurementTime() ) );
        return -1;
    }


    bool e2e = true;
    bool outbound = true;
    // Check source IP address and destination IP address
    if ( ipHeader->ip_src.s_addr == originAddress.s_addr )
    {
        // Outbound UDP traffic
        pr.remote_address = ipHeader->ip_dst;
        if ( ipHeader->ip_ttl < 64 )
        {
            e2e = false;
            pr.probe_type = probe_s_out;
        }
        else
        {
            // pr.probe_type =  probe_f_out or probe_b_out or probe_i_out
        }
    }
    else if ( ipHeader->ip_dst.s_addr == originAddress.s_addr )
    {
        // Inbound UDP traffic
        outbound = false;
        pr.remote_address = ipHeader->ip_src;
        // pr.probe_type = probe_f_in, probe_b_in or probe_i_in
    }
    else
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Ignoring unknown(address) UDP packet\n", GetMeasurementTime() ) );
        return -1;
    }

    int rc = ProcessUdpPayload( (char *)udpHeader + sizeof( udphdr ), pr );
    if ( rc == -1 )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Invalid UDP payload received\n", GetMeasurementTime() ) );
        return -1;
    }
    if ( ! e2e )
    {
        assert( rc == PL_ROUTER );
        int ip_id = ntohs( ipHeader->ip_id );
        pr.host_id = ip_id >> 6;
        pr.hop_id = ip_id & 0x3F;
    }
    else if ( rc == PL_UDP_F )
    {
        pr.probe_type = ( outbound ? probe_f_out : probe_f_in );
    }
    else if ( rc == PL_UDP_B )
    {
        pr.probe_type = ( outbound ? probe_b_out : probe_b_in );
    }
    else if ( rc == PL_UDP_I )
    {
        pr.probe_type = ( outbound ? probe_i_out : probe_i_in );
    }
    else
    {
        if ( rc == PL_ROUTER )
            IF_DEBUG( 3, fprintf( stderr, "%s Invalid UDP probe: this is a router-targeted probe, src_ip=%s, remain_ttl=%d\n", GetMeasurementTime(), inet_ntoa( ipHeader->ip_src ), ipHeader->ip_ttl ) );
        return -1;
    }
    pr.ttl = ipHeader->ip_ttl;

    return 0;
}


int TMon::ProcessUdpPayload( const char *udp_payload, ProbeResult &pr )
{
    ProbeUdpPayload * payload = (ProbeUdpPayload *)udp_payload;

    if ( payload->IsValid() )
    {
        pr.host_id = payload->host_id;
        pr.probe_id = payload->probe_id;
        pr.hop_count = payload->hop_count;
        pr.time_value = payload->time_value;
        if ( payload->type_code == PL_ROUTER )
            return PL_ROUTER;
        else if ( payload->type_code == PL_UDP_F )
            return PL_UDP_F;
        else if ( payload->type_code == PL_UDP_B )
            return PL_UDP_B;
        else if ( payload->type_code == PL_UDP_I )
            return PL_UDP_I;
        else
        {
            assert(0);
            return -1;
        }
    }
    else
        return -1;
}


void SetSockAddr( struct sockaddr_in &sock_address, struct in_addr ip_address )
{
    bzero( &sock_address, sizeof( struct sockaddr_in ) );
#ifdef HAVE_SOCKADDR_SA_LEN
    sock_address.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
    sock_address.sin_family = AF_INET;
    sock_address.sin_port = htons( UDPTMON_PORT );
    sock_address.sin_addr = ip_address;
}


// Send f probe

int TMon::SendUdpFprobe( struct in_addr target_ip, int host_id, long probe_id )
{
    ProbeUdpPayload payload;
    payload.SetUdpFprobe( host_id, probe_id );
    if ( SendProbePacket( originAddress, target_ip, host_id, ORIGIN_TTL, payload ) < 0 )
        return -1;
    return 0;
}


// Send b probe

int TMon::SendUdpBprobe( struct in_addr target_ip, int host_id, long probe_id, int hop_count, timeval f_recv_time )
{
    ProbeUdpPayload payload;
    payload.SetUdpBprobe( host_id, probe_id, hop_count, f_recv_time );
    struct sockaddr_in target;
    SetSockAddr( target, target_ip );
    if ( sendto( sock_udp, &payload, PROBE_PACKET_SIZE, 0, (struct sockaddr *)&target, sizeof( struct sockaddr_in ) ) < 0 )
        return -1;
    return 0;
}


// Send info packet

int TMon::SendUdpIprobe( struct in_addr target_ip, int host_id, long probe_id, timeval b_send_time )
{
    ProbeUdpPayload payload;
    payload.SetUdpIprobe( host_id, probe_id, b_send_time );
    struct sockaddr_in target;
    SetSockAddr( target, target_ip );
    if ( sendto( sock_udp, &payload, PROBE_PACKET_SIZE, 0, (struct sockaddr *)&target, sizeof( struct sockaddr_in ) ) < 0 )
        return -1;
    return 0;
}


// Send router probe (i.e., s probe)

int TMon::SendRouterProbe( struct in_addr target_ip, int host_id, long probe_id, int hop_id )
{
    ProbeUdpPayload payload;
    payload.SetRouterProbe();
    if ( SendProbePacket( originAddress, target_ip, host_id, hop_id, payload ) < 0 )
        return -1;
    return 0;
}


// Receive f, b probe, i packet

int TMon::RecvUdpProbe( ProbeResult &pr )
{
    ProbeUdpPayload payload;
    int rc = recvfrom( sock_udp, &payload, PROBE_PACKET_SIZE, 0, (struct sockaddr *)NULL, (socklen_t *)NULL );
    if ( rc == -1 )
    {
        perror( "readfrom" );
        return -1;
    }
    pr.packet_size = rc;

    rc = ProcessUdpPayload( (char *)&payload, pr );

    bool valid = true;
    if ( rc == PL_UDP_F )
        pr.probe_type = probe_f_in;
    else if ( rc == PL_UDP_B )
        pr.probe_type = probe_b_in;
    else if ( rc == PL_UDP_I )
        pr.probe_type = probe_i_in;
    else
        valid = false;

    if ( valid && DebugOn( pr.host_id & ~FRP_FLAG ) )
        IF_DEBUG( 9, pr.ShowRecord() );

    return rc;
}


// Recevie s probe

int TMon::RecvRouterProbe( ProbeResult &pr )
{
    // NOTE: If RecvRouterProbe() will be called in multiple threads, 'recv_buf' must not be defined as "static"
    static char recv_buf[IP_MAXPACKET];
    int rc = read( socketIn, recv_buf, IP_MAXPACKET );
    if (  rc < 0 )
    {
        perror( "read" );
        return -1;
    }
    pr.packet_size = rc;
    
    rc = ProcessIcmpPacket( recv_buf, pr, true );
    if ( rc == 0 && DebugOn( pr.host_id & ~FRP_FLAG ) )
        IF_DEBUG( 9, pr.ShowRecord() );
    return rc;
}


// Send a probe packet

int TMon::SendProbePacket( struct in_addr origin_ip, struct in_addr target_ip, int host_id, int hop_id, ProbeUdpPayload &udpPayload )
{
    // Make up a UDP packet to send out.  Start with an IP header.
    // See Section 25.3 of UNIX Network Programming, Second Edition,
    // for why we need to twiddle the byte-orders of some fields but
    // not others, depending on what OS we run.
    struct ip ipHeader;
    memset( &ipHeader, 0, sizeof( ipHeader ) );
#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeader.ip_vhl = ( sizeof( ip ) >> 2 ) | ( 4 << 4 );
#else
    ipHeader.ip_hl = sizeof( ip ) >> 2;
    ipHeader.ip_v = 4;
#endif /* __osf__ */
    ipHeader.ip_tos = 0;
    int ipPacketSize = sizeof( ip ) + sizeof( udphdr ) + PROBE_PACKET_SIZE;
    // NOTE: For targeted d probe(ICMP echo packet), the packet size is the same( sizeof( ip ) + ICMP_MINLEN + PROBE_PACKET_SIZE )
#ifdef linux
    ipHeader.ip_len = htons( ipPacketSize );
#else
    ipHeader.ip_len = ipPacketSize;
#endif /* linux */
    // Encode host_id into the high 10 bits, encode hop_id into the low 6 bits
    ipHeader.ip_id = htons( ( ( host_id & 0x3FF ) << 6 ) + ( hop_id & 0x3F ) );
#ifdef linux
    ipHeader.ip_off = htons( IP_DF );
#else
    ipHeader.ip_off = IP_DF;
#endif /* linux */

    ipHeader.ip_sum = 0;
    memcpy( &(ipHeader.ip_src), &origin_ip, sizeof( struct in_addr ) );
    memcpy( &(ipHeader.ip_dst), &target_ip, sizeof( struct in_addr ) );

    ipHeader.ip_ttl = hop_id;
    ipHeader.ip_p = IPPROTO_UDP;

    // Make up UDP header;
    struct udphdr udpHeader;
    memset( &udpHeader, 0, sizeof( udphdr ) );
    udpHeader.uh_sport = htons( S_UDPTMON_PORT );
    udpHeader.uh_dport = htons( UDPTMON_PORT );
    udpHeader.uh_ulen = htons( PROBE_PACKET_SIZE + sizeof( udphdr ) );
    udpHeader.uh_sum = htons( 0 ); // Let the UDP checksum be 0

    // Build the packet now.
    char ipPacket[MAX_SEND_BUF_SIZE];
    memcpy( ipPacket, &ipHeader, sizeof( ipHeader ) );
    memcpy( ipPacket + sizeof( ipHeader ), &udpHeader, sizeof( udphdr ) );
    memcpy( ipPacket + sizeof( ipHeader) + sizeof( udphdr ), &udpPayload, PROBE_PACKET_SIZE );

    // Compute UDP checksum.  We're going to build this up part by part,
    // starting with the different components of the pseudo-header,
    // and then doing the UDP header and payload as a single block.
    // The idea for this comes from the libnet code, basically we build the
    // checksum for the pseudo-header without actually having to
    // store it anywhere.  Not real efficient but it should be OK for
    // our purposes.
    //
    // We note that the UDP checksum routine returns the 1s complement
    // of the 1s complement sum; we need to unroll this to do some
    // other manipulations before actually storing it.
    // We assemble some components to the checksum, and then re-do
    // the checksum carry and complement operations.
    u_int checksum;
    checksum = (u_short)~InCksum( (u_short *)&(ipHeader.ip_src), 2 * sizeof( struct in_addr ) );
    checksum += (u_short)htons( IPPROTO_UDP ) + (u_short)htons( sizeof( udphdr ) + PROBE_PACKET_SIZE );
    checksum += (u_short)~InCksum( (u_short *)( ipPacket + sizeof( ipHeader ) ), sizeof( udphdr ) + PROBE_PACKET_SIZE );
    while ( checksum >> 16 )
    {
        checksum = ( checksum & 0xffff ) + ( checksum >> 16 );
    }
    ( (udphdr *)( ipPacket + sizeof( ipHeader ) ) )->uh_sum = ~checksum;

    // Send the packet
    if ( send( socketOut, ipPacket, ipPacketSize, 0 ) < 0 )
    {
        perror( "send" );
        return -1;
    }

    return 0;
}


//
// InCksum
//
// Input:  addr, len (buffer to checksum)
//
// Output:  IP checksum for this buffer in return value
//
// Compute IP checksum for a buffer.  It's put here because it's fairly
// general-purpose, and both the IPv4 and IPv6 tests could potentially
// make use of it.  RFC 1071 has implementation notes, and we use
// the sample implementation (essentially unmodified) therein.
//
u_short InCksum( u_short *addr, int len )
{
    register int sum = 0;
    u_short checksum;
    
    while ( len > 1 )
    {
        sum += *addr++;
        len -= 2;
    }
    
    // Add left-over byte, if any
    if ( len > 0 )
    {
        sum += *(u_char *)addr;
    }
    
    // fold 32-bit sum to 16 bits
    while ( sum >> 16 )
    {
        sum = (sum & 0xffff) + (sum >> 16);
    }
    
    checksum = ~sum;
    return checksum;
}


void TMon::OpenLogFile( void )
{
    const char * file = "tmon_log";
    if ( ( log_fp = fopen( file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to open file '%s' for writing\n", file );
        exit(1);
    }
}


bool DebugOn( int host_id )
{
    return( ProbeToDebug == -1 || ProbeToDebug == host_id ? true : false );
}


HostInfo::HostInfo( int host_id, const char * host_name )
{
    this->host_id = host_id;
    this->host_name = host_name;
    ip.s_addr = 0;
    probe_id = 0;
    last_contact = 0;
    hop_count = 0;
    tid_trace = 0;
    tracing = false;
    ewma_rtt = 0;
    hop_list_ver = 0;
    acked_hl_ver = 0;
    time_last_topo_update = 0;
    route_change_count = 0;
}


// Reset hop information before each round of traceroute

void HostInfo::ResetHopInfo( void )
{
    if ( (unsigned int)hop_count != hop_list.size() )
        hop_list.resize( hop_count );
    for ( vector<struct in_addr>::iterator i = hop_list.begin(); i != hop_list.end(); i ++ )
        i->s_addr = 0;
}


// Reset RTT related data before each round of RTT measurement

void HostInfo::ResetRawTime( void )
{
     bzero( &send_time_f, sizeof( timeval ) );
     bzero( &recv_time_f, sizeof( timeval ) );
     bzero( &send_time_b, sizeof( timeval ) );
     bzero( &recv_time_f, sizeof( timeval ) );
}


// Update average RTT (exponentially weighted moving average)

void HostInfo::UpdateRttEwma( double rtt )
{
    ewma_rtt = ( ewma_rtt ? ewma_rtt * 0.8 + rtt * 0.2 : rtt );
}


// Update date probe id of last contact each time when the destination is contacted

void HostInfo::UpdateContact( void )
{
    last_contact = probe_id;
}


// Compute RTT value, return 0 if raw data is incomplete

double HostInfo::GetRtt( void )
{
    if ( ( send_time_f.tv_sec || send_time_f.tv_usec ) &&
         ( recv_time_f.tv_sec || recv_time_f.tv_usec ) &&
         ( send_time_b.tv_sec || send_time_b.tv_usec ) &&
         ( recv_time_b.tv_sec || recv_time_b.tv_usec ) )
        return delta_t( send_time_f, recv_time_b ) - delta_t( recv_time_f, send_time_b ); 
    return 0;
}


void HostInfo::PrintHops( FILE * fp, const char * sep )
{
    int hop_num = hop_list.size();
    if ( hop_num )
        fprintf( fp, "%s", inet_ntoa( hop_list[0] ) );
    for ( int i = 1; i < hop_num; i ++ )
        fprintf( fp, "%s%s", sep, inet_ntoa( hop_list[i] ) );
}


// Check whether there are any differences between hop_list and last_hop_list

bool HostInfo::TopologyChanged( void )
{
    if ( hop_list.size() != last_hop_list.size() )
    {
        last_hop_list = hop_list;
        return true;
    }
    bool changed = false;
    bool previous_hop_same = true;
    vector<struct in_addr>::iterator i = hop_list.begin();
    vector<struct in_addr>::iterator j = last_hop_list.begin();
    while ( i != hop_list.end() && j != last_hop_list.end() )
    {
        if ( i->s_addr != 0 )
        {
            if ( i->s_addr != j->s_addr )
            {
                // Current hop is different
                changed = true;
                j->s_addr = i->s_addr;
                previous_hop_same = false;
            }
            else
            {
                // Current hop is the same
                previous_hop_same = true;
            }
        }
        else // i->s_addr == 0
        {
            if ( j->s_addr != 0 && ! previous_hop_same )
            {
                changed = true;
                j->s_addr = i->s_addr;
            }
            // In this case, previous_hop_same keeps value for this iteration
        }
        i ++;
        j ++;
    }
    return changed;
}


// Initialize record data

PendingProbe::PendingProbe( long probe_id, timeval& send_time_f )
{
    this->probe_id = probe_id;
    this->send_time_f = send_time_f;
    bzero( &recv_time_f, sizeof( timeval ) );
    bzero( &send_time_b, sizeof( timeval ) );
    bzero( &recv_time_f, sizeof( timeval ) );
}


// The default constructor

ProbeRecord::ProbeRecord( void )
{
    host_id = -1;   // Not allocated
    // NOTE: Call Reset() within this constructor is problematic because we use resize() to initialize element
    // The problematic codes are: "skew_monitor_f.Initialize( &min_f, this ); skew_monitor_b.Initialize( &min_b, this );"
    // Because min_f and min_b do not have correct addresses now.
}


// Clear current record such that it becomes available for new allocating

void ProbeRecord::Reset( void )
{
    min_f.Initialize( 1000 );
    min_b.Initialize( 1000 );
    min_r.Initialize( 1000 );
    skew_monitor_f.Initialize( &min_f, this );
    skew_monitor_b.Initialize( &min_b, this );
    th_f.Initialize( TH_INITIAL_SMALL, TH_FB_CNT_MAX );
    th_b.Initialize( TH_INITIAL_SMALL, TH_FB_CNT_MAX );
    th_r.Initialize( TH_INITIAL_SMALL, TH_FB_CNT_MAX );
    pending_probes.clear();
    last_contact = 0;
    hop_count = 0;
    clock_skew_extent = 0;
    clock_skew_indicator = 0;
    e2e_ci_f = e2e_ci_b = rtt_ci = 0;
    e2e_cc_f = e2e_cc_b = rtt_cc = 0;
    sample_count = 0;
    min_f_update_count = 0;
    min_b_update_count = 0;
    min_r_update_count = 0;
    clevel_edsum_f = clevel_edsum_b = 0;
}


// Enqueue a new record and dequeue outdated records

void ProbeRecord::Enqueue( PendingProbe& new_record )
{
    pending_probes.push_back( new_record );
    while ( new_record.probe_id - pending_probes.front().probe_id > MAX_PENDING_PROBE )
        pending_probes.pop_front();
}


// Search a record in pending_probes that matchs probe_id

list<PendingProbe>::iterator ProbeRecord::FindRecord( long probe_id )
{
    list<PendingProbe>::iterator it;
    for ( it = pending_probes.begin(); it != pending_probes.end(); it ++ )
        if ( it->probe_id == probe_id )
            break;;
    return it;
}


static float CongestionLevel( float e2e_ci )
{
    if ( e2e_ci > TH_E2E_CI_LEVEL3 )
        return 3;
    if ( e2e_ci > TH_E2E_CI_LEVEL2 )
        return 2;
    if ( e2e_ci > TH_E2E_CI )
        return 1;
    return 0;
}


void ProbeRecord::UpdateRecord( long probe_id, UpdateType type, ProbeResult &pr )
{
    if ( probe_id > last_contact )
        last_contact = probe_id;
    list<PendingProbe>::iterator it = FindRecord( probe_id );
    if ( it == pending_probes.end() )
        return;
    bool record_complete = false;
    if ( type == recv_time_fb )
    {
        it->recv_time_f = pr.time_value;
        it->recv_time_b = pr.tv;
        hop_count = pr.hop_count;
        if ( it->send_time_b.tv_sec || it->send_time_b.tv_usec )
            record_complete = true;
    }
    else if ( type == send_time_b )
    {
        it->send_time_b = pr.time_value;
        if ( ( it->recv_time_f.tv_sec || it->recv_time_f.tv_usec ) && ( it->recv_time_b.tv_sec || it->recv_time_b.tv_usec ) )
            record_complete = true;
    }
    if ( ! record_complete )
        return;

    double f = delta_t( it->send_time_f, it->recv_time_f );
    double b = delta_t( it->send_time_b, it->recv_time_b );
    if ( min_r.Update( f + b ) == 1 )
    {
        min_r_update_count ++;
    }
    if ( min_f.Update( f ) == 1 )
    {
        min_f_update_count ++;
        skew_monitor_f.CheckSkew( it->recv_time_f );
    }
    if ( min_b.Update( b ) == 1 )
    {
        min_b_update_count ++;
        skew_monitor_b.CheckSkew( it->recv_time_b );
    }
    double Df = f - min_f.value;
    double Db = b - min_b.value;
    double Dr = f + b - min_r.value;
    if ( ! clock_skew_indicator )
    {
        th_f.Update( Df );
        th_b.Update( Db );
        th_r.Update( Dr );
        sample_count ++;
        if ( Df > th_f.small )
            e2e_cc_f ++;
        if ( Db > th_b.small )
            e2e_cc_b ++;
        if ( Dr > th_r.small )
            rtt_cc ++;
    }
    if ( tmon->current_time.tv_sec - last_check_time >= TIME_E2E_CHECK )
    {
        e2e_ci_f = e2e_ci_b = rtt_ci = VALUE_UNDEFINED_2;
        float cong_level_f = 0;
        float cong_level_b = 0;
        // Check congestion event
        if ( sample_count > 0 && route_change_count < TH_RC_COUNT )
        {
            if ( th_r.bin_count.percent_above < TH_BIN_ABOVE )
                rtt_ci = (float)rtt_cc / (float)sample_count;
            else
            {
                rtt_ci = VALUE_UNDEFINED_2;         // Threshold tends to be problematic, suppress the generation of CI
                rtt_cc = 0;
            }
            int min_update_count = min_f_update_count + min_b_update_count;
            if ( min_update_count <= 0 || e2e_cc_f <= rtt_cc )
            {
                if ( th_f.bin_count.percent_above < TH_BIN_ABOVE )
                {
                    e2e_ci_f = (float)e2e_cc_f / (float)sample_count;
                    if ( e2e_ci_f > TH_E2E_CI )
                        cong_level_f = CongestionLevel( e2e_ci_f );
                    else
                        e2e_ci_f = VALUE_UNDEFINED_2;
                }
                else
                    e2e_ci_f = VALUE_UNDEFINED_2;   // Threshold tends to be problematic, suppress the generation of CI
            }
            if ( min_update_count <= 0 || e2e_cc_b <= rtt_cc )
            {
                if ( th_b.bin_count.percent_above < TH_BIN_ABOVE )
                {
                    e2e_ci_b = (float)e2e_cc_b / (float)sample_count;
                    if ( e2e_ci_b > TH_E2E_CI )
                        cong_level_b = CongestionLevel( e2e_ci_b );
                    else
                        e2e_ci_b = VALUE_UNDEFINED_2;
                }
                else
                    e2e_ci_b = VALUE_UNDEFINED_2;   // Threshold tends to be problematic, suppress the generation of CI
            }
        }

        // Update exponentially decayed sum of congestion levels
        clevel_edsum_f = clevel_edsum_f * CL_EDSUM_DECAY_FACTOR + cong_level_f;
        clevel_edsum_b = clevel_edsum_b * CL_EDSUM_DECAY_FACTOR + cong_level_b;

        if ( cong_level_f > 0 || cong_level_b > 0 )
        {
            // Report E2E congestion event
            tmon->SendE2ecReport( *this );
        }

        // Check clock skew event
        if ( clock_skew_extent > TH_CSKEW_EXTENT )
        {
            // Report clock skew event
            tmon->SendClockSkewReport( *this );
        }

        // Reset counters
        e2e_cc_f = e2e_cc_b = rtt_cc = 0;
        sample_count = 0;
        clock_skew_extent = 0;
        min_f_update_count = 0;
        min_b_update_count = 0;
        min_r_update_count = 0;
        last_check_time = tmon->current_time.tv_sec;
    }

    tmon->OutputFrpResult( *this, *it );

    pending_probes.erase( it );
}


void TMon::OutputFrpResult( ProbeRecord& host_record, PendingProbe& record )
{
    // Output to log file
    /*
    assert( pthread_mutex_lock( &log_mutex ) == 0 );
    fprintf( log_fp, "%ld\tOWD\t%d\t%ld\t%d\t%.3lf\t%.3lf\n", current_time.tv_sec, host_record.host_id, record.probe_id, host_record.hop_count, host_record.Df * 0.001, host_record.Db * 0.001 );
    assert( pthread_mutex_unlock( &log_mutex ) == 0 );
    */
}


void TMon::CheckFrpHost( const TMonList & tmon_list )
{
    // Compute the checksum of TMon list
    short check_sum = 0;
    int i = 0;
    for ( vector<ProbeRecord>::iterator it = dest_host_frp.begin(); it != dest_host_frp.end(); it ++ )
    {
        if ( it->host_id != -1 )
        {
            if ( i >= num_host_frp )
            {
                fprintf( stderr, "Error: Inconsistent num_host_frp\n" );
                SendTMonList( -1, TL_FRP_CORRUPT );
                return;
            }
            check_sum += it->host_id;
            i ++;
        }
    }
    if ( num_host_frp != tmon_list.num_tmon || check_sum != tmon_list.check_sum )
    {
        IF_DEBUG( 3, fprintf( stderr, "Inconsistency between TMon list of control node and that of this host detected, recover it\n" ) );

        // Check the integrity of tmon_list from control node
        check_sum = tmon_list.check_sum;
        for ( i = 0; i < tmon_list.num_tmon; i ++ )
            check_sum -= tmon_list.dID[i];
        if ( check_sum != 0 )
        {
            IF_DEBUG( 3, fprintf( stderr, "Error: Corrupt tmon_list from control node\n" ) );
            SendTMonList( -1, TL_CORRUPT );
            return;
        }

        // Remove hosts in dest_host_frp that are not in tmon_list of the control node
        for ( i = 0; i < MAX_FRP_HOST; i ++ )
        {
            short dID = dest_host_frp[i].host_id;
            if ( dID != -1 )
            {
                // Search dID in tmon_list.dID
                bool found = false;
                for ( int j = 0; j < tmon_list.num_tmon; j ++ )
                {
                    if ( dID <= tmon_list.dID[j] )
                    {
                        if ( dID == tmon_list.dID[j] )
                            found = true;
                        break;
                    }
                }
                if ( ! found )
                {
                    // Remove the host
                    dest_host_frp[i].host_id = -1;
                    num_host_frp --;
                }
            }
        }

        // Add hosts that are in tmon_list of the control node but are missing in dest_host_frp
        for ( i = 0; i < tmon_list.num_tmon; i ++ )
            AddFrpHost( tmon_list.dID[i] );

        SendTMonList( -1, TL_RECOVERD );
    }
}


vector<ProbeRecord>::iterator TMon::FindFrpHost( int host_id )
{
    vector<ProbeRecord>::iterator it;
    for ( it = dest_host_frp.begin(); it != dest_host_frp.end(); it ++ )
        if ( it->host_id == host_id )
            break;
    return it;
}


bool TMon::AddFrpHost( int host_id )
{
    vector<ProbeRecord>::iterator it = FindFrpHost( host_id );
    if ( it != dest_host_frp.end() )
        return true;
    it = FindFrpHost( -1 );
    if ( it != dest_host_frp.end() )
    {
        it->Reset();
        it->skew_monitor_f.ResetAge( current_time );
        it->skew_monitor_b.ResetAge( current_time );
        it->host_id = host_id;
        num_host_frp ++;
        return true;
    }
    return false;
}


bool TMon::RemoveFrpHost( int host_id )
{
    vector<ProbeRecord>::iterator it = FindFrpHost( host_id );
    if ( it != dest_host_frp.end() )
    {
        it->host_id = -1;   // Mark the record as unallocated
        num_host_frp --;
        return true;
    }
    return false;
}


void TMon::ShowFrpHost( void )
{
    fprintf( stderr, "%d frp hosts", num_host_frp );
    int i = 0;
    for ( vector<ProbeRecord>::iterator it = dest_host_frp.begin(); it != dest_host_frp.end(); it ++ )
        if ( it->host_id != -1 )
            fprintf( stderr, "-(%d)%d", ++ i, it->host_id );
    fputc( '\n', stderr );
}


struct RandHostList
{
    int host_id;
    int rand;
};


void TMon::RandomFrpHost( void )
{
    srand( my_host_id );

    int host_num = dest_host.size();
    RandHostList * rlist = new RandHostList[host_num];

    // Build rlist
    int index = 0;
    for ( map<int, HostInfo *>::iterator it = dest_host.begin(); it != dest_host.end(); it ++ )
    {
        rlist[index].host_id = it->first;
        rlist[index].rand = rand();
        index ++;
    }

    // Sort rlist by field 'rand'
    for ( int i = 1; i < host_num; i ++ )
    {
        for ( int j = i - 1; j >= 0; j -- )
        {
            if ( rlist[j+1].rand < rlist[j].rand )
            {
                RandHostList temp = rlist[j];
                rlist[j] = rlist[j+1];
                rlist[j+1] = temp;
            }
            else break;
        }
    }

    // Select the first MAX_FRP_HOST hosts in rlist
    host_num = MIN( host_num, MAX_FRP_HOST );
    for ( int i = 0; i < host_num; i ++ )
        AddFrpHost( rlist[i].host_id );

    delete [] rlist;
}


void TMon::TraceRouteStart( HostInfo *host ) 
{
    if ( ! host->tid_trace )
    {
        if ( pthread_create( &(host->tid_trace), NULL, traceroute_process, host ) != 0 )
        {
            fprintf( stderr, "pthread_create(traceroute) failed\n" );
            host->tid_trace = 0;
        }
        else
            pthread_detach( host->tid_trace );
    }
}


void * traceroute_process( void * arg )
{
    tmon->TraceRoute( (HostInfo *)arg );

    ((HostInfo *)arg)->tid_trace = 0;

    pthread_exit( NULL );

    return NULL;
}


void TMon::TraceRoute( HostInfo *host )
{
    //fprintf( stderr, "TraceRoute( %d )\n", host->host_id );
    if ( host->tracing )
        return;
    host->ResetHopInfo();
    host->tracing = true;
    // NOTE: Values of host->probe_id and host->hop_count might be changed during the time this thread is running (by main process)
    int hop_num = host->hop_list.size();    // Snapshot hop number
    long probe_id = host->probe_id;         // Snapshot the probe_id
    for ( int i = 0; i < TRACEROUTE_RETRY; i ++ )
    {
        // Send s probes
        for ( int j = 0; j < hop_num; j ++ )
        {
            if ( host->hop_list[j].s_addr == 0 )
                SendRouterProbe( host->ip, host->host_id, probe_id, j + 1 );
        }

        // Wait for 500 milliseconds or 2 RTT, take the larger one
        usleep( MAX( 500000, (unsigned long)( host->ewma_rtt * 2 ) ) );

        // Check exploration response
        int k;
        for ( k = 0; k < hop_num; k ++ )
        {
            if ( host->hop_list[k].s_addr == 0 )
                break;
        }
        if ( k == hop_num )
            break;
    }
    
    // Output to log file
    assert( pthread_mutex_lock( &log_mutex ) == 0 );
    fprintf( log_fp, "%ld\tROUTE\t%d\t%ld\t%d\t", current_time.tv_sec, host->host_id, probe_id, host->hop_list.size() );
    host->PrintHops( log_fp, "," );
    fputc( '\n', log_fp );
    assert( pthread_mutex_unlock( &log_mutex ) == 0 );

    if ( DebugOn( host->host_id ) )
        IF_DEBUG( 7, fprintf( stderr, "ROUTE - host_id=%d, probe_id=%ld, hop_num=%d, host_list: ", host->host_id, probe_id, host->hop_list.size() ); host->PrintHops( stderr, " -> " ); fputc( '\n', stderr ) );

    bool send_topo_update = false;
    bool route_changed = false;
    bool rc_stop = false;
    if ( host->TopologyChanged() )
    {
        // Topology change detected
        // Update counter
        host->route_change_count ++;
        route_changed = true;
        // Debug
        //fprintf( stderr, "Topology changed, rc_count=%.1f, version=%ld\n", host->route_change_count, host->hop_list_ver );
        if ( host->route_change_count > TH_RC_COUNT )
        {
            // Route change is too frequent, don't send topology update any more
            if ( host->route_change_count > MAX_RC_COUNT )
                host->route_change_count = MAX_RC_COUNT;
        }
        else
        {
            // Normal case, send topology update
            host->hop_list_ver = current_time.tv_sec;
            send_topo_update = true;
        }
    }
    else if ( host->acked_hl_ver != host->hop_list_ver && current_time.tv_sec - host->time_last_topo_update > 250 )
    {
        // The latest topology update has not been acknowledge, resend it
        // Debug
        //fprintf( stderr, "The latest topology update has not been acknowledge, resend it, version=%ld\n", host->hop_list_ver );
        send_topo_update = true;
    }
    if ( host->route_change_count > 0 )
    {
        // Aging the counter
        host->route_change_count -= 0.3;
        if ( host->route_change_count < 0 )
        {
            host->route_change_count = 0;
        }
        else if ( ! route_changed && ! send_topo_update && host->route_change_count >= TH_RC_COUNT - 0.3 && host->route_change_count < TH_RC_COUNT )
        {
            // Route change stopped, send the latest topology
            host->hop_list_ver = current_time.tv_sec;
            IF_DEBUG( 3, fprintf( stderr, "%s Route change stopped, send the latest topology, host_id=%d, version=%ld\n", GetMeasurementTime(), host->host_id,host->hop_list_ver ) );
            send_topo_update = true;
            rc_stop = true;
        }
    }
    if ( send_topo_update )
    {
        //fprintf( stderr, "SendTopologyUpdate( %d )\n", host->host_id );
        SendTopologyUpdate( host, rc_stop );
    }

    vector<ProbeRecord>::iterator it = FindFrpHost( host->host_id );
    if ( it != dest_host_frp.end() )
    {
        // Copy route_change_count to FRP host record
        it->route_change_count = host->route_change_count;
        if ( route_changed && host->route_change_count > 1 )
        {
            // Aging queuing delay statistics due to route change
            it->th_f.bin_count.Aging( 0.2 );
            it->th_b.bin_count.Aging( 0.2 );
            it->th_r.bin_count.Aging( 0.2 );
        }
    }

    host->tracing = false;
}


// Process ACK for topology update

void TMon::ProcessTopoAck( const Packet_MSG_TOPO_ACK * ack )
{
    map<int, HostInfo *>::iterator it = dest_host.find( ack->dID );
    if ( it != dest_host.end() )
    {
        HostInfo *dest = it->second;
        if ( dest->acked_hl_ver < ack->version )
            dest->acked_hl_ver = ack->version;
    }
}


// Process a UDP message

void ProcessUdpMessage( const char * packet )
{
    MsgHeader * msg_header = (MsgHeader * )packet;
    switch( msg_header->type )
    {
    case MSG_TOPO_ACK:
        {
            if ( (unsigned int)msg_header->len + sizeof( MsgHeader ) != sizeof( Packet_MSG_TOPO_ACK ) )
            {
                IF_DEBUG( 3, fprintf( stderr, "%s TOPO_ACK: invalid, payload_len=%d\n", GetMeasurementTime(), msg_header->len ) );
                break;
            }
            Packet_MSG_TOPO_ACK * ack = (Packet_MSG_TOPO_ACK *)packet;
            // Debug
            //fprintf( stderr, "MSG_TOPO_ACK received: version=%ld, dID=%d\n", ack->version, ack->dID );
            tmon->ProcessTopoAck( ack );
        }
        break;

    case MSG_ALIVE_ACK:
        {
            if ( (unsigned int)msg_header->len + sizeof( MsgHeader ) != sizeof( Packet_MSG_ALIVE_ACK ) )
            {
                IF_DEBUG( 3, fprintf( stderr, "%s ALIVE_ACK: invalid, payload_len=%d\n", GetMeasurementTime(), msg_header->len ) );
                break;
            }
            Packet_MSG_ALIVE_ACK * ack = (Packet_MSG_ALIVE_ACK *)packet;
            // Debug
            //fprintf( stderr, "MSG_ALIVE_ACK received: version=%d, clock_offset=%ld\n", ack->version, ack->clock_offset );
            if ( ack->version != tpong_version )
            {
                // Version error detected, end program
                fprintf( stderr, "TPong version error\n" );
                exit(1);
            }
            else
            {
                // Update check point time
                checkpoint_time = time( NULL );
                tmon->clock_offset = ack->clock_offset;
                // Check the TMon list for consistency with the control node
                tmon->CheckFrpHost( ack->tmon_list );
            }
        }
        break;

    case MSG_EDT_TMON:
        {
            if ( (unsigned int)msg_header->len != sizeof( Payload_MSG_EDT_TMON ) )
            {
                IF_DEBUG( 3, fprintf( stderr, "%s EDT_TMON: invalid, payload_len=%d\n", GetMeasurementTime(), msg_header->len ) );
                break;
            }
            Payload_MSG_EDT_TMON * payload = (Payload_MSG_EDT_TMON *)( packet + sizeof( MsgHeader ) );
            // Debug
            //fprintf( stderr, "MSG_EDT_TMON received: dID_add=%d, dID_rmv=%d\n", payload->dID_add, payload->dID_rmv );
            bool send_tmon_list = false;
            if ( payload->dID_rmv )
            {
                if ( ! tmon->RemoveFrpHost( payload->dID_rmv ) )
                {
                    /*
                    Fail to remove the target host, most probabily due to synchronization problem.
                    Therefore, send TMon list to control node to recover synchronization.
                    */
                    send_tmon_list = true;
                }
            }
            if ( payload->dID_add )
            {
                if ( ! tmon->AddFrpHost( payload->dID_add ) )
                {
                    /*
                    Fail to add the new target, most probabily due to synchronization problem.
                    Therefore, send TMon list to control node to recover synchronization.
                    */
                    send_tmon_list = true;
                }
            }
            if ( send_tmon_list )
                tmon->SendTMonList( payload->dID_add, payload->dID_rmv );
            // Debug
            //tmon->ShowFrpHost();
        }
        break;

    case MSG_RUN_SND:
        {
            if ( (unsigned int)msg_header->len != sizeof( Payload_MSG_RUN_SND ) )
            {
                IF_DEBUG( 3, fprintf( stderr, "%s RUN_SND: invalid, payload_len=%d\n", GetMeasurementTime(), msg_header->len ) );
                break;
            }
            Payload_MSG_RUN_SND * payload = (Payload_MSG_RUN_SND *)( packet + sizeof( MsgHeader ) );
            EndPong( true );
            char receiver_id[10];
            sprintf( receiver_id, "%d", payload->receiver_id );
            RunSender( receiver_id, true );
            last_run_pong_id = msg_header->id;
            peer_id = payload->receiver_id;
        }
        break;

    case MSG_RUN_RCV:
        {
            if ( (unsigned int)msg_header->len != sizeof( Payload_MSG_RUN_RCV ) )
            {
                IF_DEBUG( 3, fprintf( stderr, "%s RUN_RCV: invalid, payload_len=%d\n", GetMeasurementTime(), msg_header->len ) );
                break;
            }
            Payload_MSG_RUN_RCV * payload = (Payload_MSG_RUN_RCV *)( packet + sizeof( MsgHeader ) );
            const char * sender_host = tmon->GetHostName( payload->sender_id );
            if ( sender_host )
            {
                EndPong( true );
                sleep( 10 );
                RunReceiver( sender_host, true );
                last_run_pong_id = msg_header->id;
                peer_id = payload->sender_id;
            }
        }
        break;

    case MSG_END_PONG:
        {
            if ( (unsigned int)msg_header->len != 0 )
            {
                IF_DEBUG( 3, fprintf( stderr, "%s END_PONG: invalid, payload_len=%d\n", GetMeasurementTime(), msg_header->len ) );
                break;
            }
            if ( msg_header->id - last_run_pong_id > 0 )
                EndPong( true );
        }
        break;
    }
}


const char * TMon::GetHostName( short host_id ) const
{
    map<int, HostInfo *>::const_iterator it = dest_host.find( host_id );
    if ( it != dest_host.end() )
        return it->second->host_name;
    return NULL;
}


// Send topology update message to control node

void TMon::SendTopologyUpdate( HostInfo * host, bool rc_stop )
{
    int hop_num = host->last_hop_list.size();
    int payload_len = sizeof( Header_MSG_TOPO ) +  ( hop_num + 2 ) * sizeof( in_addr );
    char * message = CreateMessage( MSG_TOPO, GetNextMessageID(), payload_len );
    if ( message )
    {
        Header_MSG_TOPO * header = (Header_MSG_TOPO *)( message + sizeof( MsgHeader ) );
        header->version = host->hop_list_ver;
        header->dID = host->host_id;
        header->num_node = hop_num + 2;
        header->rc_count = host->route_change_count;
        header->rc_stop = rc_stop;
        in_addr * ip_list = (in_addr *)( message + sizeof( MsgHeader ) + sizeof( Header_MSG_TOPO ) );
        ip_list[0] = originAddress;
        for ( int j = 0; j < hop_num; j ++ )
            ip_list[j + 1] = host->last_hop_list[j];
        ip_list[hop_num + 1] = host->ip;
        header->digest.Compute( header->num_node, ip_list );
        SendMessage( message );
        host->time_last_topo_update = current_time.tv_sec;
    }
}


// Send FRP host list to control node

void TMon::SendTMonList( short dID_add, short dID_rmv )
{
    int payload_len = sizeof( Header_MSG_TMON_LIST ) +  num_host_frp * sizeof( short );
    char * message = CreateMessage( MSG_TMON_LIST, GetNextMessageID(), payload_len );
    if ( message )
    {
        Header_MSG_TMON_LIST * header = (Header_MSG_TMON_LIST *)( message + sizeof( MsgHeader ) );
        header->num_host_frp = num_host_frp;
        header->dID_add = dID_add;
        header->dID_rmv = dID_rmv;
        short * id_list = (short *)( message + sizeof( MsgHeader ) + sizeof( Header_MSG_TMON_LIST ) );
        int i = 0;
        for ( vector<ProbeRecord>::iterator it = dest_host_frp.begin(); it != dest_host_frp.end(); it ++ )
        {
            if ( it->host_id != -1 )
            {
                if ( i >= num_host_frp )
                {
                    fprintf( stderr, "Error: Inconsistent num_host_frp\n" );
                    break;
                }
                id_list[i] = it->host_id;
                i ++;
            }
        }
        SendMessage( message );
    }
}


// Send end-to-end congestion report message

void TMon::SendE2ecReport( ProbeRecord& host_record )
{
    char * message = CreateMessage( MSG_E2EC, GetNextMessageID(), sizeof( Payload_MSG_E2EC ) );
    if ( message )
    {
        Payload_MSG_E2EC * payload = (Payload_MSG_E2EC *)( message + sizeof( MsgHeader ) );
        payload->dID = host_record.host_id;
        payload->start_time = host_record.last_check_time + clock_offset;
        payload->duration = current_time.tv_sec - host_record.last_check_time;
        payload->intensity_f = host_record.e2e_ci_f;
        payload->intensity_b = host_record.e2e_ci_b;
        payload->intensity_r = host_record.rtt_ci;
        payload->sample_count = host_record.sample_count;
        payload->mfu_count = host_record.min_f_update_count;
        payload->mbu_count = host_record.min_b_update_count;
        payload->rc_count = host_record.route_change_count;
        payload->clevel_edsum_f = host_record.clevel_edsum_f;
        payload->clevel_edsum_b = host_record.clevel_edsum_b;
        payload->th_f.bin_index = host_record.th_f.bin_index;
        payload->th_f.pattern = host_record.th_f.pattern;
        payload->th_f.percent_below = host_record.th_f.bin_count.percent_below;
        payload->th_f.percent_above = host_record.th_f.bin_count.percent_above;
        payload->th_b.bin_index = host_record.th_b.bin_index;
        payload->th_b.pattern = host_record.th_b.pattern;
        payload->th_b.percent_below = host_record.th_b.bin_count.percent_below;
        payload->th_b.percent_above = host_record.th_b.bin_count.percent_above;
        payload->th_r.bin_index = host_record.th_r.bin_index;
        payload->th_r.pattern = host_record.th_r.pattern;
        payload->th_r.percent_below = host_record.th_r.bin_count.percent_below;
        payload->th_r.percent_above = host_record.th_r.bin_count.percent_above;
        float total_count_f = ( host_record.th_f.bin_count.total_count != 0 ? host_record.th_f.bin_count.total_count : 1 );
        float total_count_b = ( host_record.th_b.bin_count.total_count != 0 ? host_record.th_b.bin_count.total_count : 1 );
        float total_count_r = ( host_record.th_r.bin_count.total_count != 0 ? host_record.th_r.bin_count.total_count : 1 );
        for ( int i = 0; i < BIN_NUM; i ++ )
        {
            payload->th_f.bin_count[i] = host_record.th_f.bin_count.bin_count[i] / total_count_f;
            payload->th_b.bin_count[i] = host_record.th_b.bin_count.bin_count[i] / total_count_b;
            payload->th_r.bin_count[i] = host_record.th_r.bin_count.bin_count[i] / total_count_r;
        }
        SendMessage( message );
    }
}


// Send clock skew report message

void TMon::SendClockSkewReport( ProbeRecord& host_record )
{
    char * message = CreateMessage( MSG_CSKEW, GetNextMessageID(), sizeof( Payload_MSG_CSKEW ) );
    if ( message )
    {
        Payload_MSG_CSKEW * payload = (Payload_MSG_CSKEW *)( message + sizeof( MsgHeader ) );
        payload->dID = host_record.host_id;
        payload->start_time = host_record.last_check_time + clock_offset;
        payload->duration = current_time.tv_sec - host_record.last_check_time;
        payload->extent = host_record.clock_skew_extent;
        payload->mfu_count = host_record.min_f_update_count;
        payload->mbu_count = host_record.min_b_update_count;
        SendMessage( message );
    }
}


// Send Pong failure indication

void TMon::SendPongFailure( char role, short peer, char reason )
{
    char * message = CreateMessage( MSG_PONG_FAIL, GetNextMessageID(), sizeof( Payload_MSG_PONG_FAIL ) );
    if ( message )
    {
        Payload_MSG_PONG_FAIL * payload = (Payload_MSG_PONG_FAIL *)( message + sizeof( MsgHeader ) );
        payload->role = role;
        payload->peer = peer;
        payload->reason = reason;
        SendMessage( message );
    }
}


// Create path bitmap based on path reachability

void TMon::CreatePathBitmap( PathBitmap * bitmap )
{
    bitmap->Reset();
    short max_id = 0;
    for ( map<int, HostInfo *>::iterator it = dest_host.begin(); it != dest_host.end(); it ++ )
    {
        HostInfo *dest = it->second;
        if ( dest->hop_count != INVALID_PROBE && probe_id - dest->last_contact < 5 && dest->last_contact != 0 && dest->route_change_count < TH_RC_COUNT )
        {
            bitmap->SetPathBit( dest->host_id );
            if ( max_id < dest->host_id )
                max_id = dest->host_id;
        }
    }
    bitmap->max_id = max_id;
}


// Send keep-alive message

void TMon::SendKeepAlive( void )
{
    char * message = CreateMessage( MSG_ALIVE, GetNextMessageID(), sizeof( Payload_MSG_ALIVE ) );
    if ( message )
    {
        Payload_MSG_ALIVE * payload = (Payload_MSG_ALIVE *)( message + sizeof( MsgHeader ) );
        payload->version = tpong_version;
        payload->timestamp = time( NULL );
        payload->pong_active = IsPongRunning();
        payload->pong_state = state;
        CreatePathBitmap( &(payload->path_bitmap) );
        SendMessage( message );
    }
}


int TMon::GetHostNum( void )
{
    return dest_host.size();
}


void SetupTMonProcess( void )
{
    tmon = new TMon();
    tmon->InitializeTMon();

    if ( pthread_create( &tid_tmon, NULL, tmon_process, NULL ) != 0 )
    {
        fprintf( stderr, "pthread_create(tmon_process) failed\n" );
        exit(1);
    }
    IF_DEBUG( 3, fprintf( stderr, "TMon process installed. " ) );

    int valid_host_num = tmon->GetHostNum();
    if ( ! valid_host_num )
    {
        IF_DEBUG( 3, fprintf( stderr, "Run in pure reponser mode.\n" ) );
    }
    else
    {
        IF_DEBUG( 3, fprintf( stderr, "%d potential destination hosts.\n", valid_host_num ) );
    }
}


void * tmon_process( void * arg )
{
    tmon->SendProbes();

    tmon->Finalize();

    pthread_exit(NULL);
    
    return NULL;
}


void * frp_process( void * arg )
{
    tmon->SendFrProbes();

    pthread_exit(NULL);
}


void ReadHostID( void )
{
    const char * file = "host_id";
    FILE * fp;
    if ( ( fp = fopen( file, "r" ) ) == NULL )
    {
        fprintf( stderr, "Fail to read file '%s'\n", file );
        exit(1);
    }
    if ( fscanf( fp, "%d", &my_host_id ) != 1 )
    {
        fprintf( stderr, "Corrupt file '%s'\n", file );
        exit(1);
    }
    fclose( fp );
}


// ************** MinmumValue member functions **************

void MinimumValue::Initialize( int age_max )
{
    this->age_max = age_max;
    value = candidate_value = 9E9;
    age = 0;
}


int MinimumValue::Update( double value )
{
    if ( value < this->value )
    {
        this->value = value;
        candidate_value = 9E9;
        age = 0;
        return 1;
    }
    else
    {
        if ( value < candidate_value )
            candidate_value = value;
        if ( ++ age >= age_max )
        {
            this->value = candidate_value;
            candidate_value = 9E9;
            age = 0;
            return 2;
        }
    }
    return 0;
}

// *********** End of MinmumValue member functions ***********


// ************** SkewMonitor member functions **************

void SkewMonitor::Initialize( MinimumValue * mv, ProbeRecord * host_record )
{
    this->mv = mv;
    this->host_record = host_record;
    first_value = VALUE_UNDEFINED;
    average_skew = 9E9;
}


void SkewMonitor::ResetAge( const timeval &current_time )
{
    age = current_time;
}


#define SKEW_THRESHOLD  1E-5    // 1 ms per 100 sec
#define AGE_THRESHOLD   30E6    // 30 sec


const char * skew_extent_text[3] = { "Non-negligible", "Considerable", "Significant" };

int SkewExtent( double skew )
{
    if ( skew < 0 )
        return 2;
    if ( skew < SKEW_THRESHOLD * 10 )
        return 1;
    if ( skew < SKEW_THRESHOLD * 100 )
        return 2;
    return 3;
}


void SkewMonitor::CheckSkew( const timeval &time )
{
    if ( first_value == VALUE_UNDEFINED )
    {
        first_value = mv->value;
        first_time = time;
        return;
    }

    if ( average_skew == 9E9 )
    {
        if ( delta_t( first_time, time ) > 3E6 )
        {
            average_skew = ( mv->value - first_value ) / delta_t( first_time, time );
            last_value = mv->value;
            last_time = time;
            double skew_value = fabs( average_skew );
            if ( skew_value > SKEW_THRESHOLD && delta_t( age, time ) > AGE_THRESHOLD )
            {
                const char * mt = GetMeasurementTime();
                int extent = SkewExtent( skew_value );
                host_record->clock_skew_extent += extent;
                host_record->clock_skew_indicator = SKEW_SUPPRESS_TIME;
                const char * se = skew_extent_text[ extent - 1 ];
                double sv = skew_value * 1E3;
                fprintf( clock_skew_fp, "%s %d-%d %s clock skew: skew=%5g(ms/sec)\n", mt, my_host_id, host_record->host_id, se, sv );
            }
        }
        return;
    }

    double diff_value = mv->value - last_value;
    double diff_time = delta_t( last_time, time );
    double skew = diff_value / diff_time;
    double skew_difference = fabs( skew - average_skew ) / MAX( fabs( average_skew ), fabs( skew ) );
    
    if ( skew_difference < 0.9 )
    {
        average_skew = ( mv->value - first_value ) / delta_t( first_time, time );
        double skew_value = fabs( average_skew );
        if ( skew_value > SKEW_THRESHOLD && delta_t( age, time ) > AGE_THRESHOLD )
        {
            const char * mt = GetMeasurementTime();
            int extent = SkewExtent( skew_value );
            host_record->clock_skew_extent += extent;
            host_record->clock_skew_indicator = SKEW_SUPPRESS_TIME;
            const char * se = skew_extent_text[ extent - 1 ];
            double sv = skew_value * 1E3;
            fprintf( clock_skew_fp, "%s %d-%d %s clock skew: skew=%5g(ms/sec)\n", mt, my_host_id, host_record->host_id, se, sv );
        }
    }
    else if ( skew_difference > 0.99 )
    {
        // Seem to be clock jump, reset monitor
        if ( fabs( average_skew ) > SKEW_THRESHOLD && delta_t( age, time ) > AGE_THRESHOLD )
        {
            const char * mt = GetMeasurementTime();
            double dv = fabs( diff_value ) / 1000;
            double dt = fabs( diff_time ) / 1E6;
            host_record->clock_skew_extent += SkewExtent( -1 );
            host_record->clock_skew_indicator = SKEW_SUPPRESS_TIME;
            fprintf( clock_skew_fp, "%s %d-%d Clock jump: %5g(ms) in %5g(sec)\n", mt, my_host_id, host_record->host_id, dv, dt );
        }
        average_skew = 9E9;
        first_value = mv->value;
        first_time = time;
    }
    last_value = mv->value;
    last_time = time;
}

// *********** End of SkewMonitor member functions ***********


void OpenClockSkewFile( void )
{
    const char * file = "tmon_clock_skew";
    if ( ( clock_skew_fp = fopen( file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", file );
        exit(1);
    }
}


// ************** BinCount member functions **************

BinCount::BinCount( void )
{
    Reset();
}


void BinCount::Reset( void )
{
    for ( int i = 0; i < BIN_NUM; i ++ )
        bin_count[i] = 0;
    total_count = 0;
    percent_below = percent_above = 0;
}


void BinCount::Update( int bin_index )
{
    total_count ++;
    bin_count[bin_index] ++;
    if ( total_count >= 2000 )
        Aging( 0.8 );
}


void BinCount::Aging( float factor )
{
    if ( total_count < 100 )
        return;
    total_count = 0;
    for ( int i = 0; i < BIN_NUM; i ++ )
    {
        bin_count[i] = (int)( (float)bin_count[i] * factor + 0.5 );
        total_count += bin_count[i];
    }
}


void BinCount::Output( FILE *fp )
{
    for ( int i = 0; i < BIN_NUM; i++ )
    {
        int fraction = ( total_count == 0 ? 0 : bin_count[i] * 1000 / total_count );
        fprintf( fp, "%d\t", fraction );
    }
    fprintf( fp, "%d\n", total_count );
}


void BinCount::CheckThreshold( int threshold_index )
{
    int i;
    int count;
    count = 0;
    for ( i = 0; i <= threshold_index; i ++ )
        count += bin_count[i];
    percent_below = count * 100.0 / total_count;
    count = 0;
    for ( i = threshold_index + 1; i < BIN_NUM; i ++ )
        count += bin_count[i];
    percent_above = count * 100.0 / total_count;
}


#define ASCEND      1
#define DESCEND     2

bool BinCount::LocateValley( int& bin_index, int& pattern )
{
    if ( total_count < 100 )    // This threshold is different for TMon (set to 100) and Pong (set to 30)
        return false;

    double bin_percent[BIN_NUM];
    int i;
    for ( i = 0; i < BIN_NUM; i ++ )
        bin_percent[i] = bin_count[i] * 100.0 / total_count;

    double bin_percent_0 = bin_percent[0];
    if ( bin_percent_0 > 90 )
    {
        bin_index = 1;
        while ( bin_index <= 4 && bin_percent[bin_index] > 3 )
            bin_index ++;
        bin_index = bin_index - 1;
        pattern = PATTERN_A;
        return true;
    }

    // ************* Find dominant valley (Pattern B) *************
    
    // Find the last bin that has percentage >= 1%
    int last_bin = BIN_NUM - 2;
    while ( last_bin > 1 && bin_percent[last_bin] < 1 )
        last_bin --;
    int first_bin = 1;
    // Find the first bin that has percentage >= 1%
    while ( first_bin < last_bin && bin_percent[first_bin] < 1 )
        first_bin ++;

    // Locate all valley points
    int state = ASCEND;
    int valley_point[BIN_NUM / 2];
    int vp_count = 0;
    bin_index = first_bin + 1;
    while ( bin_index <= last_bin )
    {
        if ( bin_percent[bin_index] < bin_percent[bin_index - 1] )
        {
            if ( state == ASCEND )
                state = DESCEND;
        }
        else if ( bin_percent[bin_index] > bin_percent[bin_index - 1] )
        {
            if ( state == DESCEND )
            {
                valley_point[vp_count ++] = bin_index - 1;
                state = ASCEND;
            }
        }
        bin_index ++;
    }
    if ( vp_count > 0 )
    {
        // Search for the dominant valley point
        int max_valley_length = -1;
        int dominant_valley_point = -1;
        for ( i = 0; i < vp_count; i ++ )
        {
            int valley_index = valley_point[i];
            int max_index = valley_index;
            bin_index = valley_index - 1;
            while ( bin_index > first_bin )
            {
                if ( bin_percent[bin_index] > bin_percent[max_index] )
                    max_index = bin_index;
                bin_index --;
            }
            int valley_length = valley_index - max_index;
            double max_value = bin_percent[max_index];
            
            max_index = valley_index;
            bin_index = valley_index + 1;
            while ( bin_index <= last_bin )
            {
                if ( bin_percent[bin_index] > bin_percent[max_index] )
                    max_index = bin_index;
                bin_index ++;
            }
            if ( max_index - valley_index < valley_length )
                valley_length = max_index - valley_index;
            if ( max_value > bin_percent[max_index] )
                // Get the smaller one of the two maximum values
                max_value = bin_percent[max_index];

            if ( max_value * 0.7 > bin_percent[valley_index] && max_valley_length < valley_length )
            {
                max_valley_length = valley_length;
                dominant_valley_point = valley_index;
            }
        }
        if ( dominant_valley_point != -1 )
        {
            bin_index = dominant_valley_point;
            pattern = PATTERN_B;
            return true;
        }
    }
    // ********** End of Find dominant valley (Pattern B) ********** 

    double bin_threshold;

    if ( bin_percent_0 > 15 && bin_percent[BIN_NUM - 1] < 3 )
    {
        bin_index = 1;
        bin_threshold = bin_percent_0 * 0.1;
        if ( bin_threshold > 3 )
            bin_threshold = 3;
        while ( bin_index <= last_bin && bin_percent[bin_index] > bin_threshold )
            bin_index ++;
        if ( bin_index > 1 )
        {
            bin_threshold = bin_percent[bin_index - 1] * 0.7;
            while ( bin_index <= last_bin && bin_percent[bin_index] > bin_threshold )
                bin_index ++;
        }
        bin_index = bin_index - 1;
        pattern = PATTERN_C;
        return true;
    }
    
    int max_index = 1;
    bin_index = 2;
    while ( bin_index <= last_bin )
    {
        if ( bin_percent[bin_index] > bin_percent[max_index] )
            max_index = bin_index;
        bin_index ++;
    }
    bin_threshold = bin_percent[max_index] * 0.2;
    if ( bin_threshold < 1 )
        bin_threshold = 1;
    bin_index = max_index + 1;
    while ( bin_index <= last_bin && bin_percent[bin_index] > bin_threshold )
        bin_index ++;
    bin_index = bin_index - 1;
    pattern = PATTERN_D;
    return true;
}

// *********** End of BinCount member functions ***********


// ************** ProbeThreshold member functions **************

void ProbeThreshold::Initialize( double small, int count_max )
{
    this->small = small;
    this->big = small * 1.5;
    this->count_max = count_max;
    bin_index = pattern = -1;
    percent_below = percent_between = percent_above = 0;
    count = count_below = count_between = count_above = 0;
    average_below = average_between = average_above = 0;
    deviation_below = deviation_between = deviation_above = 0;
    max_below = min_above = VALUE_UNDEFINED;

    avg_below = avg_betwn = avg_above = 0;
    dev_below = dev_betwn = dev_above = 0;

    if ( bin_count.total_count != 0 )
        bin_count.Reset();
}


// Exponential 30 Bin
const double BIN_UPPER[] = { 150, 195, 254, 330, 428, 557, 724, 941, 1224, 1591, 2068, 2688, 3495, 4543, 5906, 7678, 9981, 12976, 16868, 21929, 28507, 37060, 48178, 62631, 81420, 105846, 137600, 178880, 232544, 302307 };


// Look up a corresponding index of bin for the input queuing delay value

int ProbeThreshold::GetBinIndex( double value )
{
    int i;
    for ( i = 0; i < BIN_NUM - 1; i ++ )
        if ( value < BIN_UPPER[i] )
            break;
    return i;
}


void ProbeThreshold::Update( double value )
{
    if ( value < small )
    {
        // Below
        count_below ++;
        average_below += value;
        deviation_below += fabs( value - average_below / (double)count_below );
        if ( max_below == VALUE_UNDEFINED || value > max_below )
            max_below = value;
    }
    else if ( value > big )
    {
        // Above
        count_above ++;
        average_above += value;
        deviation_above += fabs( value - average_above / (double)count_above );
        if ( min_above == VALUE_UNDEFINED || value < min_above )
            min_above = value;
    }
    else
    {
        // Between
        count_between ++;
        average_between += value;
        deviation_between += fabs( value - average_between / (double)count_between );
    }

    // Update per bin statistics
    bin_count.Update( GetBinIndex( value ) );

    if ( ++ count >= count_max && bin_count.total_count >= TH_MIN_TOTAL_SAMPLE )
    {
        //assert( count_below + count_between + count_above == count );
        percent_below = (double)count_below / (double)count;
        percent_above = (double)count_above / (double)count;
        percent_between = (double)count_between / (double)count;
        if ( count_below > 0 )
        {
            average_below /= (double)count_below;
            deviation_below /= (double)count_below;
        }
        else
            average_below = small;
        if ( count_above > 0 )
        {
            average_above /= (double)count_above;
            deviation_above /= (double)count_above;
        }
        else
            average_above = big;
        if ( count_between > 0 )
        {
            average_between /= (double)count_between;
            deviation_between /= (double)count_between;
        }
        else
            average_between = ( small + big ) * 0.5;


        //***********************************************************************
        // Adjust 'small' and 'big' in the following way (Mar. 1, 2007):
        // 1) Initial threshold:
        //    small = 400; big = 600;
        // 2) Use LocateValley Algorithm to get bin_index:
        //    small = BIN_UPPER[bin_index], big = BIN_UPPER[bin_index+1]
        //    bin_index = 0..28
        //***********************************************************************

        if ( bin_count.LocateValley( bin_index, pattern ) )
        {
            bin_count.CheckThreshold( bin_index );
            small = BIN_UPPER[bin_index];
            big = BIN_UPPER[bin_index+1];
        }

        avg_below = average_below;
        avg_betwn = average_between;
        avg_above = average_above;
        dev_below = deviation_below;
        dev_betwn = deviation_between;
        dev_above = deviation_above;

        count = count_below = count_between = count_above = 0;
        average_below = average_between = average_above = 0;
        deviation_below = deviation_between = deviation_above = 0;
        max_below = min_above = VALUE_UNDEFINED;
    }
}

// *********** End of ProbeThreshold member functions ***********
