#ifndef TMON_H
#define TMON_H

#include "Pong_Common.h"
#include "Message.h"
#include <vector>
#include <map>
#include <list>

using std::vector;
using std::map;
using std::list;

#define HAVE_PCAP

#define PROBE_PACKET_SIZE       sizeof( ProbeUdpPayload )   // = 28 (plus ip and udp header => 56 bytes)
#define VALUE_UNDEFINED         -1
#define VALUE_UNDEFINED_2       -8.88

const int MAX_SEND_BUF_SIZE             = 100;
const int ORIGIN_TTL                    = 255;
const int TRACEROUTE_RETRY              = 3;
const char * const FAKE_TARGET          = "www.google.com";
const int MAX_LINE_WIDTH                = 300;
//const int DEFAULT_PROBE_PERIOD          = 3;
const int DEFAULT_PROBE_PERIOD          = 300;      // 300 seconds
const int INVALID_PROBE                 = -1;
const int MIN_HOP_NUM                   = 6;        // Minumum number of hops required for a path
const int MAX_PENDING_PROBE             = 30;
const double DEFAULT_PROBE_PERIOD_FP    = 0.2;      // 0.2 second
const int FRP_FLAG                      = 0x4000;   // A flag bit in host_id used to differetiate fast-rate probe and low-rate probe
const float TH_RC_COUNT                 = 1.5;      // Threshold of route change count above which topology updates will be suppressed
const float MAX_RC_COUNT                = 4;        // Maximum value of route change count
const int TIME_E2E_CHECK                = 30;       // Check end-to-end congestion statistics every 30 seconds
const float TH_BIN_ABOVE                = 0.95;     // Maximum sample percentage for bins above queing delay threshold
const float TH_E2E_CI                   = 0.01;     // Threshold of end-to-end congestion intensity above which MSG_E2EC will be sent; threshold of congestion level 1
const float TH_E2E_CI_LEVEL2            = 0.03;     // Threshold of congestion level 2
const float TH_E2E_CI_LEVEL3            = 0.1;      // Threshold of congestion level 3

// Linux networking compatability macros.  For some unfathomable
// reason, Linux systems seem to have named many of their networking
// constants differently than those used by virtually every other
// sockets API implementation.  We try to bring it in line with
// the more widely-used sockets API standards here.
#ifdef linux

// UDP
#define    uh_sport         source
#define    uh_dport         dest
#define    uh_ulen          len
#define    uh_sum           check

#endif /* linux */


// Action codes. ICMPv4 and ICMPv6 have different values for their type
// and code fields. The TMonActionType abstracts these differences.
typedef enum {
    TMonActionReserved = 0,        // reserved code
    TMonActionValid = 1,           // store valid measurement (e.g. ICMP time exceeded)
    TMonActionValidLasthop = 2,    // store valid measurement, this is last hop (e.g. ICMP port unreachable)
    TMonActionFiltered = 3,        // packets filtered, give up (e.g. ICMP prohibited)
    TMonActionTimeout = 4,         // Timeout
    TMonActionAbort = 255          // huh? we haven't a clue
} TMonActionType;


typedef enum {
    probe_f_out = 1,
    probe_f_in = 2,
    probe_b_out = 3,
    probe_b_in = 4,
    probe_i_out = 5,
    probe_i_in = 6,
    probe_s_out = 7,
    probe_s_in = 8,
    probe_icmp_error = 9
} ProbeType;


extern const char * ptype_code[];


// Type codes of ProbeUdpPayload

#define PL_UDP_F    0   // End-to-end udp probe, forward (i.e., f probe)
#define PL_UDP_B    1   // End-to-end udp probe, backward (i.e., b probe)
#define PL_UDP_I    2   // Informational udp packet, backward
#define PL_ROUTER   3   // Router probe (i.e., s probe)

#define PAD_NUM     8

struct ProbeUdpPayload
{
    char type_code;
    char hop_count;
    short host_id;
    long probe_id;          // Use to identify probes of different rounds, usually timestamp
    timeval time_value;     // For info packet, it's send time of b probe
                            // For b probe, it's receive time of f probe
    char signature[4];
    
    char padding[PAD_NUM];  // padding up to 28 bytes (plus ip and udp header => 56 bytes)

    ProbeUdpPayload( void );
    void SetRouterProbe( void );
    void SetUdpFprobe( int host_id, long probe_id );
    void SetUdpBprobe( int host_id, long probe_id, int hop_count, timeval f_recv_time );
    void SetUdpIprobe( int host_id, long probe_id, timeval b_send_time );
    bool IsValid( void );
};


struct ProbeResult
{
    timeval tv;                         // timestamp recorded
    union {
    struct in_addr icmpSourceAddress;   // source address of ICMP packet
    struct in_addr remote_address;      // source address of UDP packet
    };
    int icmpSourceAddressLength;        // length of source address
    TMonActionType icmpResult;            // icmp result
    unsigned int packet_size;           // bytes in the ip packet
    int host_id;
    long probe_id;
    int hop_id;
    int hop_count;
    timeval time_value;
    ProbeType probe_type;
    int ttl;

    ProbeResult( void );
    void ShowRecord( void );
};


struct HostInfo
{
    int host_id;
    const char * host_name;
    long probe_id;
    timeval send_time_f,
            recv_time_f,
            send_time_b,
            recv_time_b;    // Raw data for RTT measurement
    struct in_addr ip;      // IP address of the destination host
    long last_contact;      // Probe id of the last contact with the destination
    int hop_count;          // Hop count of the forward path
    vector<struct in_addr> hop_list;        // IP address list of each hop
    vector<struct in_addr> last_hop_list;   // Last hop_list
    time_t hop_list_ver;    // Version of hop_list. When reporting topology updates to control node, the version is annotated.
    time_t time_last_topo_update;   // Time when the last topology update is sent
    time_t acked_hl_ver;    // Version of hop_list that has been acked by control node
    float route_change_count;   // Counter of route changes
    bool tracing;           // Whether tracing route is going on. (If it is, write permssion of 'hop_list' is granted.)
    pthread_t tid_trace;    // Thread id for traceroute process
    double ewma_rtt;        // EWMA value of round-trip time in usecond

    HostInfo( int host_id, const char * host_name );
    void ResetHopInfo( void );
    void ResetRawTime( void );
    void UpdateContact( void );
    void UpdateRttEwma( double rtt );
    double GetRtt( void );
    void PrintHops( FILE * fp, const char * sep );
    bool TopologyChanged( void );
};


struct MinimumValue
{
    double value, candidate_value;
    int age_max;
    int age;

    void Initialize( int age_max );
    int Update( double value );
};


#define SKEW_SUPPRESS_TIME  5

struct ProbeRecord;

struct SkewMonitor
{
    MinimumValue * mv;
    double first_value, last_value;
    timeval first_time, last_time;
    double average_skew;
    timeval age;
    ProbeRecord * host_record;

    void Initialize( MinimumValue * mv, ProbeRecord * host_record );
    void CheckSkew( const timeval &time );
    void ResetAge( const timeval &current_time );
};


#define BIN_NUM             30

#define PATTERN_A           1
#define PATTERN_B           2
#define PATTERN_C           3
#define PATTERN_D           4

struct BinCount
{
    int bin_count[BIN_NUM];
    int total_count;
    float percent_below, percent_above;

    BinCount( void );
    void Reset( void );
    void Update( int bin_index );
    void Aging( float factor );
    void Output( FILE *fp );
    bool LocateValley( int& bin_index, int& pattern );
    void CheckThreshold( int threshold_index );
};


#define TH_INITIAL_SMALL    9999999
#define TH_FB_CNT_MAX       100     // Threshold of f/b probes is updated every 100 probes
#define TH_MIN_TOTAL_SAMPLE 300     // Threshold of f/b probes are updated only when more than 300 samples are collected

struct ProbeThreshold
{
    double small;   // the threshold below which a value is considerred to be small
    double big;     // the threshold above which a value is considerred to be big
    int count_max;  // statistics is performed on every count_max samples
    int bin_index, pattern;

    // ***************************************************
    // Statistic variables
    int count, count_below, count_between, count_above;
    double percent_below, percent_between, percent_above;
    double average_below, average_between, average_above;
    double deviation_below, deviation_between, deviation_above;
    double max_below, min_above;

    double avg_below, avg_betwn, avg_above;
    double dev_below, dev_betwn, dev_above;
    
    BinCount bin_count;
    // ***************************************************

    void Initialize( double small, int count_max );
    void Update( double value );
    static int GetBinIndex( double value );
};


// Data of a pending probe for fast-rate probing

struct PendingProbe
{
    long probe_id;
    timeval send_time_f,
            recv_time_f,
            send_time_b,
            recv_time_b;    // Raw data for RTT measurement

    PendingProbe( long probe_id, timeval& send_time_f );
};


enum UpdateType
{
    recv_time_fb = 1,
    send_time_b  = 2
};


// Record of each target host for fast-rate probing

struct ProbeRecord
{
    int host_id;
    MinimumValue min_f, min_b;  // Mininum delays measured by f and b probes.
    MinimumValue min_r;         // Minimum round trip delay
    ProbeThreshold th_f, th_b;  // Thresholds of queuing delay measured by f and b probes
    ProbeThreshold th_r;        // Thresholds of round trip queuing delay
    SkewMonitor skew_monitor_f, skew_monitor_b;
    int clock_skew_extent;
    int clock_skew_indicator;
    int min_f_update_count;     // Count of times that min_f is updated
    int min_b_update_count;     // Count of times that min_b is updated
    int min_r_update_count;     // Count of times that min_r is updated
    float route_change_count;
    list<PendingProbe> pending_probes;
    long last_contact;          // FRP probe id of the last contact with the destination
    int hop_count;
    float clevel_edsum_f;       // Exponentially decayed sum of congestion levels over all congestion events (forward path)
    float clevel_edsum_b;       // Exponentially decayed sum of congestion levels over all congestion events (backward path)
    float e2e_ci_f, e2e_ci_b;   // End-to-end congestion intensities of forward path and backward path
    float rtt_ci;               // Congestion intensity of round trip path
    time_t last_check_time;
    int e2e_cc_f, e2e_cc_b;     // End-to-end congestion counts of forward path and backward path
    int rtt_cc;                 // Congestion count of round trip path
    int sample_count;           // Total number of probing samples

    ProbeRecord( void );
    void Reset( void );
    void Enqueue( PendingProbe& new_record );
    list<PendingProbe>::iterator FindRecord( long probe_id );
    void UpdateRecord( long probe_id, UpdateType type, ProbeResult &pr );
};


class TMon
{
protected:
    long probe_id;      // Probe id is used to differentiate different rounds of probes
    long frp_probe_id;  // Similar to probe_id, but is used for fast-rate probing

    map<int, HostInfo *> dest_host; // Host list for low-rate probing
    pthread_mutex_t hl_mutex;       // Mutex for 'dest_host'

    vector<ProbeRecord> dest_host_frp;  // Host list for fast-rate probing
    int num_host_frp;
    pthread_mutex_t hlfrp_mutex;    // Mutex for 'dest_host_frp'

    FILE *log_fp;
    pthread_mutex_t log_mutex;      // Mutex for 'log_fp'

    struct in_addr originAddress;   // Local ip
    char *originName;               // Local hostname
    struct in_addr targetAddress;
    char *targetName;

    struct sockaddr_in target_udp_addr;
    struct sockaddr_in origin_udp_addr;

    timeval syscallTime;    // Estimated overhead for gettimeofday()

    int sock_udp;   // Socket for e2e (f/b) probes
    int socketOut;  // Output socket (RAW) for s probe
    int socketIn;   // Input socket (ICMP) for s probe

    pthread_t tid_sink, tid_frp;

    timeval time_baseline;
    int round_id;

    TMonActionType GetAction( int icmpType, int icmpCode );
    char * GetName( const void * a );
    char * GetPrintableAddress( const void * a );

    int ProcessIcmpPacket( const char *ipPacket, ProbeResult &pr, bool silent = false );
    int ProcessUdpPacket( const char *ipPacket, ProbeResult &pr );
    int ProcessUdpPayload( const char *udp_payload, ProbeResult &pr );

    int SetOriginName( const char *o );
    int SetTargetName( const char *t );

    int GetSocketIn();
    int GetSocketOut();

    void SetUdpSocket( void );

    int SendUdpFprobe( struct in_addr target_ip, int host_id, long probe_id );
    int SendUdpBprobe( struct in_addr target_ip, int host_id, long probe_id, int hop_count, timeval f_recv_time );
    int SendUdpIprobe( struct in_addr target_ip, int host_id, long probe_id, timeval b_send_time );
    int SendRouterProbe( struct in_addr target_ip, int host_id, long probe_id, int hop_id );
    int SendProbePacket( struct in_addr origin_ip, struct in_addr target_ip, int host_id, int hop_id, ProbeUdpPayload &udpPayload );

    int RecvUdpProbe( ProbeResult &pr );
    int RecvRouterProbe( ProbeResult &pr );

    void TraceRouteStart( HostInfo *host );

    void SetupPacketSink();

    void OpenLogFile( void );

    double GetProbeTime( long probe_id, double probe_period, double time_offset );

    void BuildHostList( void );
    void ResolveIpParallel( HostInfo * host );
    void RemoveHost( int host_id );

    vector<ProbeRecord>::iterator FindFrpHost( int host_id );
    void RandomFrpHost( void ); // For test purpose

    void SendTopologyUpdate( HostInfo * host, bool rc_stop );

    void CreatePathBitmap( PathBitmap * bitmap );

#ifdef HAVE_PCAP

    pthread_t tid_pcap;

    void SetupPCAP();

#endif // HAVE_PCAP

public:

    timeval current_time;
    list<int> pending_host_list;  // Hosts whose hostnames are being resolving
    time_t clock_offset;    // Clock offset between this node and the control node

    TMon( void );

    void InitializeTMon( void );
    void Initialize_Common( void );
    
    void Finalize( void );

    void SendProbes( void );
    void SendFrProbes( void );
    void ProcessProbes( ProbeResult &pr );

    void ReceiveMessages( void );

    int ParsePacket( char *ipPacket, ProbeResult &pr );
    
    void TraceRoute( HostInfo *host );

    void ResolveIP( HostInfo * host_info );

    int GetHostNum( void );

    void OutputFrpResult( ProbeRecord& host_record, PendingProbe& record );

    void ProcessTopoAck( const Packet_MSG_TOPO_ACK * ack );

    void SendE2ecReport( ProbeRecord& host_record );
    void SendClockSkewReport( ProbeRecord& host_record );
    void SendPongFailure( char role, short peer, char reason );

    void CheckFrpHost( const TMonList & tmon_list );
    bool AddFrpHost( int host_id );
    bool RemoveFrpHost( int host_id );
    void ShowFrpHost( void );   // For test purpose

    void SendTMonList( short dID_add, short dID_rmv );

    void SendKeepAlive( void );

    const char * GetHostName( short host_id ) const;

#ifdef HAVE_PCAP

    pcap_t *pc; // Pcap structure
    unsigned int packet_offset;

#endif // HAVE_PCAP
};

extern bool done;
extern TMon *tmon;
extern pthread_t tid_tmon;
extern char *OriginHost;        // Hostname of local ip
extern char *TargetHostList;    // Filename of target host list
extern char *Interface;         // Physical name of local interface
extern double probe_period;     // Probing period of low-rate RTT measurement; traceroute period is also probe_period
extern int my_host_id;
extern time_t checkpoint_time;
extern int state;
extern int peer_id;

u_short InCksum( u_short *addr, int len );

void * sink_process( void * arg );

const char * GetMeasurementTime( void );

void TimeSyscall( timeval &diff );

void Sig_Term( int sig_num );

void SetSockAddr( struct sockaddr_in &sock_address, struct in_addr ip_address );

void * traceroute_process( void * arg );

void * resolve_ip_process( void * arg );

void SetupTMonProcess( void );

void * tmon_process( void * arg );

void * frp_process( void * arg );

void ReadHostID( void );

void OpenClockSkewFile( void );

void ProcessUdpMessage( const char * packet );

void RunSender( const char * receiver_id, bool silent = false );
void RunReceiver( const char * target_host, bool silent = false );
void EndPong( bool silent = false );

bool IsPongRunning( void );

#ifdef HAVE_PCAP

// Pcap thread process
void * pcap_process( void * arg );

// Pcap callback functions
void pcap_callback( u_char *puc, const struct pcap_pkthdr *ph, const u_char *pd );

#endif // HAVE_PCAP

#endif
