#include "Pong.h"
#include "Pong_Report.h"

bool IgnoreConfiguration = false;
int DebugLevel = 3;
int ProbeToDebug = -1;
int SavedDebugLevel = -99;
char *OriginHost = NULL;
char *TargetHost = NULL;
Pong *pong = NULL;
char *Interface = NULL;
unsigned int Mtu = 1500;
const char * ptype_code[] = { "n/a", "fb_out", "fb_in", "sd_out", "sd_in", "icmp_msg", "hop_in", "hop_out" };
const char * n_a = "n/a";
const char * mmt_data_dir = "data/";
bool enable_interactive_command = true;
FILE * clock_skew_fp;
bool SIGTERM_received = false;
bool suppress_clock_skew_info = false;
bool cancel_measurement = false;
int sender_control_port = TCPSND_PORT;
int clock_skew_extent = 0;
int clock_skew_indicator = 0;
int min_hop = MIN_HOP;
int min_probe_num = MIN_PROBE_NUM;
int receiver_id = VALUE_UNDEFINED;
int my_host_id = VALUE_UNDEFINED;
unsigned long msg_signature = MSG_SIGNATURE_PONG;
bool pong_report = true;    // Whether enable reporting mechanism

Pong::Pong( bool is_sender )
{
    this->is_sender = is_sender;
    phase = PP_INIT;
    GetCurrentTime( current_time );
    PrintPongStartTime();
    Initialize_Common_Pre();

    HeartBeat( "booting" );
    signal( SIGTERM, Sig_Term );
    srand( my_host_id + time( NULL ) );
    time_baseline.tv_sec = 0;

    done = false;
    quit_reason = NULL;
    TimeSyscall( syscallTime );
    pthread_mutex_init( &ctr_mutex, NULL );
    pthread_mutex_init( &hrb_mutex, NULL );
    pthread_mutex_init( &pih_mutex, NULL );
    if ( is_sender )
    {
        pthread_mutex_init( &frb_mutex, NULL );
        pthread_mutex_init( &tpt_mutex, NULL );
    }
    time_offset_adjusted = false;
    pairing_updated = false;
    technique_updated = false;
    PUA_round_id = 0;
    TUA_round_id = 0;
    hop_num_2 = 0;
    if ( is_sender )
    {
        auto_tpt = true;
        link_cc_update_count = 0;
        min_f.Initialize( 1000 );
        min_b.Initialize( 1000 );
        skew_monitor_f.Initialize( &min_f );
        skew_monitor_b.Initialize( &min_b );
        OpenClockSkewFile();
        last_Df = VALUE_UNDEFINED;
        last_Db = VALUE_UNDEFINED;
        th_f.Initialize( TH_INITIAL_SMALL, TH_FB_CNT_MAX );
        th_b.Initialize( TH_INITIAL_SMALL, TH_FB_CNT_MAX );
        dominant_bottleneck = 0;
    }
}


void Pong::Finalize( void )
{
#ifdef HAVE_PCAP
    pcap_breakloop( pc );
    pcap_close( pc );
    //pthread_join( tid_pcap, NULL );
#endif // HAVE_PCAP
    //pthread_join( tid_sink, NULL );
    close( sock_udp );
    close( socketOut );
    close( socketIn );
    close( sock_tcp );
    close( sock_monitor_listener );
    pthread_mutex_destroy( &ctr_mutex );
    pthread_mutex_destroy( &hrb_mutex );
    pthread_mutex_destroy( &pih_mutex );
    ReportFinalize();
    if ( is_sender )
    {
        pthread_mutex_destroy( &frb_mutex );
        pthread_mutex_destroy( &tpt_mutex );
        fclose( clock_skew_fp );
    }
    const char * status = ( cancel_measurement ? "cancelled" : "quitted" );
    HeartBeat( status, quit_reason );
    IF_DEBUG( 1, fprintf( stderr, "Quitted.\n" ) );
}


void Pong::PrintPongStartTime( void )
{
    char time_string[50];
    time_t gmt_time = current_time.tv_sec;
    strcpy( time_string, asctime( gmtime( &gmt_time ) ) );
    time_string[strlen(time_string) - 1] = 0;
    const char * pong_type = ( is_sender ? "sender" : "receiver" );
    IF_DEBUG( 1, fprintf( stderr, "\nPong %s starts at %s UTC.\n", pong_type, time_string ) );
}


//********************************************
// Sig_Term( sig_num )
// -------------------------------------------
// The signal handler for SIGTERM
//********************************************

void Sig_Term( int sig_num )
{
    if ( pong && pong->phase >= PP_PROBE )
    {
        IF_DEBUG( 1, fprintf( stderr, "%s SIGTERM received\n", GetMeasurementTime() ) );
        SIGTERM_received = true;
    }
    else
    {
        if ( pong )
            pong->HeartBeat( "quitted" );
        IF_DEBUG( 1, fprintf( stderr, "%s SIGTERM received, call exit().\n", GetMeasurementTime() ) );
        exit(1);
    }
}


int Pong::GetSocketOut()
{
    int rc;
    
    // NOTE: Using IPPROTO_UDP, IPPROTO_ICMP or IPPROTO_RAW works the same here
    // We send both UDP packets and ICMP packets through this socket
    socketOut = socket( AF_INET, SOCK_RAW, IPPROTO_RAW );
    if ( socketOut < 0 )
    {
        perror( "socket" );
        return socketOut;
    }
    
    int hdrinclOption = 1;
    rc = setsockopt( socketOut, IPPROTO_IP, IP_HDRINCL, &hdrinclOption, sizeof( hdrinclOption ) );
    if ( rc < 0 )
    {
        perror( "setsockopt( IP_HDRINCL )" );
        return rc;
    }

    // Bind remote side of socket
    rc = connect( socketOut, (struct sockaddr *)&target_udp_addr, sizeof( struct sockaddr_in ) );
    if ( rc < 0 )
    {
        perror( "connect" );
        return rc;
    }

    return socketOut;
}


void Pong::SetUdpSocket( void )
{
    /* 
    Data stream: UDP socket 
    */
    if ( ( sock_udp = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_DGRAM, 0 )" );
        exit(1);
    }

#ifdef linux
    int bsdcompatOption;
    bsdcompatOption = 1;
    if ( setsockopt( sock_udp, SOL_SOCKET, SO_BSDCOMPAT, &bsdcompatOption, sizeof( bsdcompatOption ) ) < 0 )
    {
        perror( "setsockopt( SO_BSDCOMPAT )" );
        exit(1);
    }
#endif /* linux */

    memset( (void *) &origin_udp_addr, 0, sizeof( struct sockaddr_in ) );
#ifdef HAVE_SOCKADDR_SA_LEN
    origin_udp_addr.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
    origin_udp_addr.sin_family = AF_INET;
    origin_udp_addr.sin_port = htons( UDPRCV_PORT );
    memcpy( &(origin_udp_addr.sin_addr), &originAddress, sizeof( struct in_addr ) );
    
    if ( bind( sock_udp, (struct sockaddr *)&origin_udp_addr, sizeof( origin_udp_addr ) ) < 0 )
    {
        perror( "bind( sock_udp )" );
        exit(1);
    }
}


char* Pong::GeneratePayload( int size )
{
    char *buf;
    int i;

    if ( size <= 0 )
        return NULL;

    buf = new char[size];
    if ( buf == NULL )
        return buf;

    for (i = 0; i < size; i++)
    {
        buf[i] = random() & 0xff;
    }
    return buf;
}


//
// TimeSyscall
//
// Input:  None
//
// Output:  timeval to hold gettimeofday overhead
//
// Determine the gettimeofday() syscall overhead. Probably this is going
// to be negligible compared to the data we're getting back from the
// network.
//

void TimeSyscall( timeval &diff )
{
    timeval t1, t2;

    gettimeofday( &t1, NULL );
    gettimeofday( &t2, NULL );
    
    diff.tv_sec = t2.tv_sec - t1.tv_sec;
    diff.tv_usec = t2.tv_usec - t1.tv_usec;
    if ( diff.tv_usec < 0 )
    {
        diff.tv_usec += 1000000;
        diff.tv_sec--;
    }
}


//
// GetAction
//
// Input:  a test record
//
// Output:  action code
//
// Figure out the meaning of a particular combination of ICMPv4 type
// and code values.
//

PongActionType Pong::GetAction( int icmpType, int icmpCode ) 
{
    if ( icmpType == ICMP_TIMXCEED )
    {
        return PongActionValid;
    }
    else if ( ( icmpType == ICMP_UNREACH ) && ( icmpCode == ICMP_UNREACH_PORT ) )
    {
        return PongActionValidLasthop;
    }
    else if ( ( icmpType == ICMP_UNREACH ) && ( icmpCode == ICMP_UNREACH_FILTER_PROHIB ) )
    {
        return PongActionFiltered;
    }
    else
    {
        return PongActionAbort;
    }
}


//
// SetOriginName
//
// Input:  origin hostname
//
// Output:  success code (negative if error)
//
// Attempt to set the origin of probe packets emanating from this host,
// for tests where it is applicable, using protocol family-dependent
// resolution as necessary.  If no origin hostname is given, 
// then use the target address to compute the correct outgoing interface.
// originName and originAddress are set.
//

int Pong::SetOriginName( char *o )
{
    // If nothing was passed in, then bind a dummy socket to try
    // to figure out where packets exit this host.
    if (o == NULL)
    {
        int dummySock;
        struct sockaddr_in dummyAddr, localAddr;
#ifdef HAVE_SOCKLEN_T
        socklen_t localAddrLength;
#else /* HAVE_SOCKLEN_T */
#ifdef NEED_GETSOCKNAME_HACK
        int localAddrLength;
#else /* NEED_GETSOCKNAME_HACK */
        size_t localAddrLength;
#endif /* NEED_GETSOCKNAME_HACK */
#endif /* HAVE_SOCKLEN_T */

        // Create socket, then connect to it, and then read out
        // the local socket address with a getsockname call.
        dummySock = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
        if ( dummySock < 0 )
        {
            perror( "socket" );
            return -1;
        }

        memset( (void *)&dummyAddr, 0, sizeof( struct sockaddr_in ) );
#ifdef HAVE_SOCKADDR_SA_LEN
        dummyAddr.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
        dummyAddr.sin_family = AF_INET;
        dummyAddr.sin_port = htons( 32768 ); // port number irrelevant
        memcpy( &dummyAddr.sin_addr, &targetAddress, sizeof( struct in_addr ) );

        if ( connect( dummySock, (sockaddr *)&dummyAddr, sizeof( struct sockaddr_in ) ) < 0 )
        {
            perror( "connect" );
            return -1;
        }

        localAddrLength = sizeof( sockaddr_in );
        memset( (void *)&localAddr, 0, sizeof( struct sockaddr_in ) );
        if ( getsockname( dummySock, (struct sockaddr *)&localAddr, &localAddrLength ) < 0)
        {
            perror( "getsockname" );
            return -1;
        }

        // Got the local address, now do a reverse DNS looup
        memcpy( &originAddress, &localAddr.sin_addr, sizeof( struct in_addr ) );
        originName = strdup( GetName( &originAddress ) );
        if ( originName == NULL )
        {
            fprintf( stderr, "Couldn't allocate memory for origin hostname.\n" );
            return -1;
        }

        close( dummySock );

        IF_DEBUG( 3, fprintf( stderr, "originName = %s\n", originName ) );
        IF_DEBUG( 3, fprintf( stderr, "originAddress = %s\n", inet_ntoa( originAddress ) ) );
    }

    // User gave a source name/address.  Attempt to resolve, if possible.
    else {
        struct hostent *host;    // resolver host entry
        host = gethostbyname( o );

        // Resolver failed
        if ( host == NULL )
        {

#ifdef HAVE_HERROR
            herror( o );
#else
            fprintf( stderr, "%s: Host not found\n", o );
#endif /* HAVE_HERROR */

            memset( (void *)&originAddress, 0, sizeof( struct in_addr ) );
            originName = strdup( o );

            if ( originName == NULL )
            {
                fprintf( stderr, "Couldn't allocate space for origin hostname.\n" );
                return -1;
            }
            return -1;
        }

        IF_DEBUG( 3, fprintf( stderr, "originName = %s\n", host->h_name ) );
        //IF_DEBUG( 3, fprintf( stderr, "h_length = %d\n", host->h_length) );
        IF_DEBUG( 3, fprintf( stderr, "originAddress = %s\n", inet_ntoa( *((struct in_addr *)(host->h_addr_list[0])) ) ) );
             
        // Get IP address
        memcpy( &originAddress, host->h_addr_list[0], host->h_length );

        // Make a copy of the canonical hostname
        originName = strdup( host->h_name );
        if ( originName == NULL )
        {
            fprintf( stderr, "Couldn't allocate memory for origin hostname.\n" );
            return -1;
        }
    }

    return 0;
}


//
// SetTargetName
//
// Input:  target hostname (target)
//
// Output:  success code (negative if error)
//
// Set target name and do protocol-dependent resolving to get a 
// network address.  In case of an error, we're responsible for
// printing some error message.
//

int Pong::SetTargetName( char *t )
{
    struct hostent *host;    // resolver host entry

    // Attempt to resolve, if possible
    host = gethostbyname( t );

    // Resolver failed
    if ( host == NULL )
    {

        // Some systems don't have herror (non-BSD?), so for those,
        // we'll cobble together an error message.
#ifdef HAVE_HERROR
        herror( t );
#else
        fprintf( stderr, "%s: Host not found\n", t );
#endif /* HAVE_HERROR */

        memset( (void *)&targetAddress, 0, sizeof( struct in_addr ) );
        targetName = strdup( t );

        if ( targetName == NULL )
        {
            fprintf( stderr, "Couldn't allocate memory for target hostname.\n" );
            return -1;
        }
        return -1;
    }

    IF_DEBUG( 3, fprintf( stderr, "targetName = %s\n", host->h_name ) );
    //IF_DEBUG( 3, fprintf( stderr, "h_length = %d\n", host->h_length ) );
    IF_DEBUG( 3, fprintf( stderr, "targetAddress = %s\n", inet_ntoa( *((struct in_addr *)(host->h_addr_list[0])) ) ) );
         
    // Get IP address
    memcpy( &targetAddress, host->h_addr_list[0], host->h_length );

    memset( (void *)&target_udp_addr, 0, sizeof( struct sockaddr_in ) );
    // Note:  Only BSD4.3Reno and later have sin_len in struct
    // sockaddr_in, so we need to test for it.
#ifdef HAVE_SOCKADDR_SA_LEN
    target_udp_addr.sin_len = sizeof( struct sockaddr_in );
#endif /* HAVE_SOCKADDR_SA_LEN */
    target_udp_addr.sin_family = AF_INET;
    target_udp_addr.sin_port = htons( UDPRCV_PORT );

    memcpy( &target_udp_addr.sin_addr, host->h_addr_list[0], host->h_length );

    // Make a copy of the canonical hostname
    targetName = strdup( host->h_name );
    if ( targetName == NULL )
    {
        fprintf( stderr, "Couldn't allocate memory for target hostname.\n" );
        return -1;
    }

    return 0;
}


//
// GetSocketIn
//
// Input:  None
//
// Output:  In return value, returns socket number.
//
// Get input socket of an appropriate type.
//

int Pong::GetSocketIn()
{
#ifdef _PLANETLAB_H_
    socketIn = socket( PF_INET, SOCK_RAW, IPPROTO_ICMP_UDP );
#else
    socketIn = socket( PF_INET, SOCK_RAW, IPPROTO_ICMP );
#endif /* _PLANETLAB_H_ */

    if ( socketIn < 0 )
    {
        perror( "socket" );
        return socketIn;
    }

#ifdef _PLANETLAB_H_
    struct sockaddr_in sin;
    memset( &sin, 0, sizeof( sin ) );
    sin.sin_port = htons( UDPRCV_PORT );
  
    bind( socketIn, (struct sockaddr *)&sin, sizeof( sin ) );
#endif /* _PLANETLAB_H_ */

    return socketIn;
}


//
// GetName
//
// Input:  Pointer to address structure
//
// Output:  Pointer to ASCII representation of name (in return
// value)
//

char * Pong::GetName( const void *a )
{
    struct hostent *host;
    host = gethostbyaddr( a, sizeof( struct in_addr ), AF_INET );

    if ( host )
    {
        return ( host->h_name );
    }
    else
    {
        return ( GetPrintableAddress( a ) );
    }
}


//
// GetPrintableAddress
//
// Input:  Pointer to address structure
//
// Output:  Pointer to ASCII representation of address (in return
// value)
//

char * Pong::GetPrintableAddress( const void * a )
{
    return ( inet_ntoa( *((struct in_addr *)a) ) );
}


void Pong::SendHalfRecord( HalfRecord &hr )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_HFR );
    const int data_size = sizeof(HalfRecord);
    if ( write( sock_tcp, &hr, data_size ) != data_size )
    {
        fprintf( stderr, "%s Send HalfRecord failed\n", GetMeasurementTime() );
        exit(1);
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


void Pong::RecvHalfRecord( HalfRecord &hr )
{
    const int data_size = sizeof(HalfRecord);
    if ( recv( sock_tcp, &hr, data_size, MSG_WAITALL ) != data_size )
    {
        fprintf( stderr, "%s Receive HalfRecord failed\n", GetMeasurementTime() );
        exit(1);
    }
}


void Pong::SendTimeBaseline( void )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_TBL );
    if ( write( sock_tcp, &time_baseline, sizeof(timeval) ) != sizeof(timeval) )
    {
        fprintf( stderr, "%s Send time baseline failed\n", GetMeasurementTime() );
        exit(1);
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


void Pong::RecvTimeBaseline( void )
{
    if ( recv( sock_tcp, &time_baseline_receiver, sizeof(timeval), MSG_WAITALL ) != sizeof(timeval) )
    {
        fprintf( stderr, "%s Receive time baseline failed\n", GetMeasurementTime() );
        exit(1);
    }
}


void Pong::SendDoubleValue( double value )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_SDV );
    if ( write( sock_tcp, &value, sizeof(double) ) != sizeof(double) )
    {
        fprintf( stderr, "%s Send double value failed\n", GetMeasurementTime() );
        exit(1);
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


double Pong::RecvDoubleValue( void )
{
    assert( RecvCtrCode() == CC_SDV );
    double value;
    if ( recv( sock_tcp, &value, sizeof(double), MSG_WAITALL ) != sizeof(double) )
    {
        fprintf( stderr, "%s Receive double value failed\n", GetMeasurementTime() );
        exit(1);
    }
    return value;
}


// Send receiver's hostname to the sender side

void Pong::SendReceiverName( void )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_RHN );

    int data_size = strlen( originName ) + 1;
    SendIntValue( data_size );

    if ( write( sock_tcp, originName, data_size ) != data_size )
    {
        fprintf( stderr, "Send receiver hostname failed\n" );
        exit(1);
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive receiver's hostname from the receiver side

void Pong::RecvReceiverName( void )
{
    assert( RecvCtrCode() == CC_RHN );

    int data_size = RecvIntValue();

    targetName = new char[data_size];
    if ( recv( sock_tcp, targetName, data_size, MSG_WAITALL ) != data_size )
    {
        fprintf( stderr, "Receive receiver hostname failed\n" );
        exit(1);
    }
}


// Send TDF(failure of targeted d probe) message to the sender side

void Pong::SendTDF( int probe_id )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_TDF );
    SendIntValue( probe_id );

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Send ATO(Adjust Time Offset) message to the receiver side
// Unit of <time_offset_increment>: second

void Pong::SendATO( int probe_id, double time_offset_increment )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_ATO );
    
    const int data_size = sizeof(char) + sizeof(double);
    char buffer[data_size];
    memcpy( buffer, &probe_id, sizeof(char) );
    memcpy( buffer + sizeof(char), &time_offset_increment, sizeof(double) );
    if ( write( sock_tcp, buffer, data_size ) != data_size )
    {
        fprintf( stderr, "%s Send ATO message failed\n", GetMeasurementTime() );
        exit(1);
    }
    
    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive ATO(Adjust Time Offset) message from the sender side
// Unit of <time_offset_increment>: second

void Pong::RecvATO( int &probe_id, double &time_offset_increment )
{
    const int data_size = sizeof(char) + sizeof(double);
    char buffer[data_size];
    if ( recv( sock_tcp, buffer, data_size, MSG_WAITALL ) != data_size )
    {
        fprintf( stderr, "%s Receive ATO failed\n", GetMeasurementTime() );
        exit(1);
    }
    probe_id = 0;
    memcpy( &probe_id, buffer, sizeof(char) );
    memcpy( &time_offset_increment, buffer + sizeof(char), sizeof(double) );
}


// Send ATA(ATO acknowledge) message from sender to receiver

void Pong::SendATA( int probe_id, int round_id )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_ATA );
    
    const int data_size = sizeof(char) + sizeof(int);
    char buffer[data_size];
    memcpy( buffer, &probe_id, sizeof(char) );
    memcpy( buffer + sizeof(char), &round_id, sizeof(int) );
    if ( write( sock_tcp, buffer, data_size ) != data_size )
    {
        fprintf( stderr, "%s Send ATA message failed\n", GetMeasurementTime() );
        exit(1);
    }
    
    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive ATA(ATO acknowledge) message from sender

void Pong::RecvATA( int &probe_id, int &round_id )
{
    const int data_size = sizeof(char) + sizeof(int);
    char buffer[data_size];
    if ( recv( sock_tcp, buffer, data_size, MSG_WAITALL ) != data_size )
    {
        fprintf( stderr, "%s Receive ATA failed\n", GetMeasurementTime() );
        exit(1);
    }
    probe_id = 0;
    memcpy( &probe_id, buffer, sizeof(char) );
    memcpy( &round_id, buffer + sizeof(char), sizeof(int) );
}


void Pong::AdjustTimeOffset( int probe_id, double time_offset_increment )
{
    // fprintf( stdout, "pih_mutex 6\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    ProbeInfo * pi = FindProbeInfo( probe_id );
    if ( pi == NULL )
    {
        fprintf( stderr, "Invalid probe_id %d\n", probe_id );
        assert(0);
    }
    pi->time_offset += time_offset_increment;
    time_offset_adjusted = true;
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
}


bool Pong::TptPending( void )
{
    return pairing_update_pending || technique_update_pending;
}


// Get the hop_d range of the complementary d probe candidates

void Pong::GetPairingRange( ProbeInfo * pi, int &lower_bound, int &upper_bound )
{
    lower_bound = hop_num + 1 - pi->ttl;
    upper_bound = hop_num_2  + 1 - pi->ttl;
    if ( lower_bound > upper_bound )
    {
        int temp = lower_bound;
        lower_bound = upper_bound;
        upper_bound = temp;
    }
    lower_bound -= PAIRING_DEV;
    upper_bound += PAIRING_DEV;
    if ( lower_bound < 1 )
        lower_bound = 1;
    if ( upper_bound > hop_num_2 )
        upper_bound = hop_num_2;
}


// Tune pairing of s and d probes automatically

bool Pong::TuningPairings( void )
{
    assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
    // fprintf( stdout, "pih_mutex 7\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    double slot_hop_d[64];
    int ttl_d[63];
    
    bool result = false;

    if ( TptPending() )
    {
        // const char * pending_type = ( pairing_update_pending ? "pairing" : "technique" );
        // IF_DEBUG( 4, fprintf( stderr, "Previous %s updates is still pending, postpone automatic tuning of pairing\n", pending_type ) );
        goto Exit_TuningPairing;
    }

    // Initialize slot_hop_d(based on whether this hop_d respond TTL limited probes)
    for ( int i = 0; i < hop_num_2; i ++ )
    {
        HopInfo * hi = hop_info_2 + i;
        if ( hi->ip.s_addr != 0 )
            slot_hop_d[hi->ttl] = 0;        // slot available
        else
            slot_hop_d[hi->ttl] = 999999;   // slot not available
    }
    
    // Initialize slot_hop_d(based on current pairings) and ttl_d
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        ProbeHistory * ph = FindProbeHistory( pi->id );
        assert( ph != NULL );
        ttl_d[i] = ph->ttl_d;
        if ( pi->technique == TECH_4P || pi->technique == TECH_FSD )
        {
            if ( ph->ttl_d != TTL_TARGETED_D && ph->ttl_d != VALUE_UNDEFINED )
                slot_hop_d[ph->ttl_d] = 999999; // slot not available
        }
        else if ( ph->ttl_d != VALUE_UNDEFINED )
        {
            if ( ph->QoM_OldAge() )
            {
                ttl_d[i] = VALUE_UNDEFINED;     // slot available, release it from current probe
                // Snapshot min_d and th_d
                if ( ph->ttl_d == TTL_TARGETED_D )
                {
                    TargetedDprobeInfo * ti = FindTdpInfo( pi->id );
                    assert( ti != NULL );
                    ti->min_d = ph->min_d.value;
                    ti->th_d_small = ph->th_d.small;
                    ti->th_d_bin = ph->th_d.bin_count;
                }
                else
                {
                    HopInfo * hi = FindHopInfo( ph->ttl_d, true );
                    assert( hi != NULL );
                    hi->min_d = ph->min_d.value;
                    hi->th_d_small = ph->th_d.small;
                    hi->th_d_bin = ph->th_d.bin_count;
                }
                UpdateQoMLog( pi, ph );
                ph->UpdatePairingHistory();
            }
            else
            {
                if ( ph->ttl_d != TTL_TARGETED_D )
                {
                    PairingHistory * pah = ph->FindPairingHistory( ph->ttl_d );
                    if ( pah != NULL )
                        slot_hop_d[ph->ttl_d] = pah->Priority();    // slot might be available
                }
            }
        }
    }

    // Figure out new pairings
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        ProbeHistory * ph = FindProbeHistory( pi->id );
        assert( ph != NULL );
        if ( ttl_d[i] == VALUE_UNDEFINED )
        {
            int num = ph->pairing_history.size();
            int j;
            for ( j = 0; j < num; j ++ )
            {
                int candidate_ttl_d = ph->pairing_history[j].ttl_d;

                if ( candidate_ttl_d == TTL_TARGETED_D )
                {
                    // targeted d probe, always safe to set
                    ttl_d[i] = candidate_ttl_d;
                    break;
                }
                
                // Check whether this probe has priority to pair with this candidate_ttl_d
                if ( ph->ttl_d != candidate_ttl_d && ph->pairing_history[j].Priority() > slot_hop_d[candidate_ttl_d] )
                {
                    for ( int k = 0; k < probe_num; k ++ )
                    {
                        if ( ttl_d[k] == candidate_ttl_d )
                        {
                            ttl_d[k] = VALUE_UNDEFINED;
                            break;
                        }
                    }
                    ttl_d[i] = candidate_ttl_d;
                    slot_hop_d[candidate_ttl_d] = ph->pairing_history[j].Priority();
                    break;
                }
            }
        }
    }

    assert( pairing_updates.empty() );

    // Collect pairing updates
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        ProbeHistory * ph = FindProbeHistory( pi->id );
        if ( ph->ttl_d != ttl_d[i] )
        {
            PairingUpdate pu;
            pu.probe_id = pi->id;
            pu.ttl_d = ttl_d[i];
            assert( pu.ttl_d > 0 && pu.ttl_d <= 64 || pu.ttl_d == VALUE_UNDEFINED );
            pu.time_offset_d = VALUE_UNDEFINED;
            pairing_updates.push_back( pu );
        }
    }

    // Send pairing updates to receiver
    if ( ! pairing_updates.empty() )
    {
        IF_DEBUG( 4, fprintf( stderr, "%s Send automatic pairing updates (%d updates).\n", GetMeasurementTime(), pairing_updates.size() ) );
        SendPairingUpdates();
        pairing_update_pending = true;
        result = true;
    }

Exit_TuningPairing:
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
    return result;
}


// Change pairings of s and d probes
// Return true if changes are made, otherwise return false

bool Pong::ChangePairings( int hop_s_start, int hop_s_end, int hop_d_start, bool verbose )
{
    assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
    // fprintf( stdout, "pih_mutex 8\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    bool result = false;
    bool invalid = false;
    int ttl_d[63], technique[64];

    if ( TptPending() )
    {
        if ( verbose )
        {
            const char * pending_type = ( pairing_update_pending ? "pairing" : "technique" );
            fprintf( stdout, "Error: previous %s updates is still pending\n", pending_type );
        }
        goto Exit_ChangePairings;
    }
    assert( pairing_updates.empty() );

    if ( hop_s_start <= 0 || hop_s_start > hop_num )
    {
        if ( verbose )
            fprintf( stdout, "Error: invalid <hop_s_start>: %d; use 'p' or 'pb' command to check valid hop range\n", hop_s_start );
        invalid = true;
    }
    if ( hop_s_end <= 0 || hop_s_end > hop_num )
    {
        if ( verbose )
            fprintf( stdout, "Error: invalid <hop_s_end>: %d; use 'p' or 'pb' command to check valid hop range\n", hop_s_end );
        invalid = true;
    }
    if ( hop_s_start > hop_s_end )
    {
        if ( verbose )
            fprintf( stdout, "Error: <hop_s_start> should be no larger than <hop_s_end>\n" );
        invalid = true;
    }
    if ( hop_d_start <= 0 || ( hop_d_start > hop_num_2 && hop_d_start != TTL_TARGETED_D ) )
    {
        if ( verbose )
            fprintf( stdout, "Error: invalid <hop_d_start>: %d; use 'p' or 'pb' command to check valid hop range\n", hop_d_start );
        invalid = true;
    }
    if ( hop_d_start - ( hop_s_end - hop_s_start ) <= 0 )
    {
        if ( verbose )
            fprintf( stdout, "Error: <hop_d_start> is too small such that it makes <hop_d_end> below 1; use 'p' or 'pb' command to check valid hop range\n" );
        invalid = true;
    }
    if ( invalid )
        goto Exit_ChangePairings;

    // Initialize temporary array 'technique'
    for ( int i = 0; i <= hop_num_2; i ++ )
        technique[i] = VALUE_UNDEFINED;
    
    // Snapshot current configuration
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        ProbeHistory * ph = FindProbeHistory( pi->id );
        assert( ph != NULL );
        assert( ph->ttl_d > 0 && ph->ttl_d <= hop_num_2 || ph->ttl_d == VALUE_UNDEFINED || ph->ttl_d == TTL_TARGETED_D );
        
        ttl_d[i] = ph->ttl_d;

        if ( ph->ttl_d != VALUE_UNDEFINED )
        {
            if ( ph->ttl_d != TTL_TARGETED_D )
            {
                // Set values for array 'technique'
                if ( hop_d_start != TTL_TARGETED_D && pi->ttl >= hop_s_start && pi->ttl <= hop_s_end )
                {
                    int new_ttl_d = hop_d_start - ( pi->ttl - hop_s_start );
                    assert( new_ttl_d > 0 && new_ttl_d <= hop_num_2 );
                    HopInfo * hi = FindHopInfo( new_ttl_d, true );
                    assert( hi != NULL );
                    if ( hi->ip.s_addr != 0 )
                        technique[ph->ttl_d] = VALUE_UNDEFINED;
                    else
                        technique[ph->ttl_d] = pi->technique;
                }
                else
                    technique[ph->ttl_d] = pi->technique;
            
                // Snapshot min_d and th_d
                HopInfo * hi = FindHopInfo( ph->ttl_d, true );
                assert( hi != NULL );
                hi->min_d = ph->min_d.value;
                hi->th_d_small = ph->th_d.small;
                hi->th_d_bin = ph->th_d.bin_count;
            }
            else
            {
                // Snapshot min_d and th_d
                TargetedDprobeInfo * ti = FindTdpInfo( pi->id );
                assert( ti != NULL );
                ti->min_d = ph->min_d.value;
                ti->th_d_small = ph->th_d.small;
                ti->th_d_bin = ph->th_d.bin_count;
            }
        }
    }

    invalid = false;

    // Figure out new pairings
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;

        if ( pi->ttl >= hop_s_start && pi->ttl <= hop_s_end )
        {
            if ( hop_d_start == TTL_TARGETED_D )
            {
                // Set to targeted d probe
                ttl_d[i] = TTL_TARGETED_D;
            }
            else
            {
                int new_ttl_d = hop_d_start - ( pi->ttl - hop_s_start );
                HopInfo * hi = FindHopInfo( new_ttl_d, true );
                if ( hi->ip.s_addr != 0 )
                {
                    if ( technique[new_ttl_d] == TECH_4P || technique[new_ttl_d] == TECH_FSD )
                    {
                        invalid = true;
                        if ( verbose )
                            fprintf( stdout, "Conflicts: Cannot pair hop_s %d with hop_d %d, %s probing is enabled on this hop_d\n", pi->ttl, new_ttl_d, tech_code[technique[new_ttl_d]] );
                    }
                    else
                    {
                        for ( int j = 0; j < probe_num; j ++ )
                        {
                            if ( ttl_d[j] == new_ttl_d )
                            {
                                ttl_d[j] = VALUE_UNDEFINED;
                                break;
                            }
                        }
                        ttl_d[i] = new_ttl_d;
                    }
                }
                else
                {
                    invalid = true;
                    if ( verbose )
                        fprintf( stdout, "Conflicts: Cannot pair hop_s %d with hop_d %d, this hop_d does not respond TTL limited probes\n", pi->ttl, new_ttl_d );
                }
            }
        }
    }
    
    // Submit pairing updates
    if ( ! invalid )
    {
        // Compare changes of hop_d and collect pairing updates
        if ( verbose )
        {
            fprintf( stdout, "******** Pairing updates ********\n" );
            fprintf( stdout, "id\thop_s\t  hop_d  \t  min_d\t  th_d\n" );
        }
        for ( int i = 0; i < probe_num; i ++ )
        {
            ProbeInfo * pi = probe_info + i;
            ProbeHistory * ph = FindProbeHistory( pi->id );

            if ( ph->ttl_d != ttl_d[i] )
            {
                PairingUpdate pu;
                pu.probe_id = pi->id;
                pu.ttl_d = ttl_d[i];
                pu.time_offset_d = VALUE_UNDEFINED;
                pairing_updates.push_back( pu );

                if ( verbose )
                {
                    fprintf( stdout, "%2d\t%5d\t%2d --> %2d", pi->id, pi->ttl, ph->ttl_d, ttl_d[i] );
                    if ( ttl_d[i] != VALUE_UNDEFINED )
                    {
                        double min_d, th_d_small;
                        if ( ttl_d[i] != TTL_TARGETED_D )
                        {
                            HopInfo * hi = FindHopInfo( ttl_d[i], true );
                            assert( hi != NULL );
                            min_d = ( hi->min_d != 9E9 ? hi->min_d / 1000 : 8888.88 );
                            th_d_small = ( hi->th_d_small != VALUE_UNDEFINED ? hi->th_d_small / 1000 : 888.88 );
                        }
                        else
                        {
                            TargetedDprobeInfo * ti = FindTdpInfo( pi->id );
                            assert( ti != NULL );
                            min_d = ( ti->min_d != 9E9 ? ti->min_d / 1000 : 8888.88 );
                            th_d_small = ( ti->th_d_small != VALUE_UNDEFINED ? ti->th_d_small / 1000 : 888.88 );
                        }
                        fprintf( stdout, "\t%7.2lf\t%6.2lf", min_d, th_d_small );
                    }
                    fprintf( stdout, "\n" );
                }
            }
        }
        if ( verbose )
            fprintf( stdout, "*********************************\n" );
    }

    // Send pairing updates to receiver
    if ( ! pairing_updates.empty() )
    {
        if ( verbose )
            fprintf( stdout, "Send pairing updates to receiver.\n" );
        SendPairingUpdates();
        pairing_update_pending = true;
        result = true;
    }

Exit_ChangePairings:
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
    if ( verbose )
        fflush( stdout );
    return result;
}


void Pong::UpdatePairings( void )
{
    int update_num = pairing_updates.size();
    if ( is_sender )
    {
        // fprintf( stdout, "pih_mutex 9\n" );
        assert( pthread_mutex_lock( &pih_mutex ) == 0 );
        for ( int i = 0; i < update_num; i ++ )
        {
            // Karl
            if ( pairing_updates[i].ttl_d > 64 )
            {
                fprintf( stderr, "pairing_updates[%d].ttl_d=%d\n", i, pairing_updates[i].ttl_d );
                exit(1);
            }
            ProbeHistory * ph = FindProbeHistory( pairing_updates[i].probe_id );
            assert( ph != NULL );
            int previous_ttl_d = ph->ttl_d;
            ph->ttl_d = pairing_updates[i].ttl_d;
            IF_DEBUG( 4, fprintf( stderr, "p%d(%d->%d) ", ph->probe_id, previous_ttl_d, ph->ttl_d ) );
            if ( ph->ttl_d != VALUE_UNDEFINED )
            {
                // Karl
                assert( ph->ttl_d != 0 );

                double min_d, th_d_small;
                if ( ph->ttl_d == TTL_TARGETED_D )
                {
                    TargetedDprobeInfo * ti = FindTdpInfo( pairing_updates[i].probe_id );
                    if ( ti == NULL )
                    {
                        fprintf( stderr, "%s Error: FindTdpInfo() retures NULL, probe_id = %d, probe_num = %d\n", GetMeasurementTime(), pairing_updates[i].probe_id, probe_num );
                        exit(1);
                    }
                    //assert( ti != NULL );
                    min_d = ti->min_d;
                    th_d_small = ti->th_d_small;
                    ph->th_d.bin_count = ti->th_d_bin;
                }
                else
                {
                    HopInfo * hi = FindHopInfo( ph->ttl_d, true );

                    // Karl
                    if ( hi == NULL )
                    {
                        fprintf( stderr, "ph->ttl_d=%d, id=%d, id2=%d \n", ph->ttl_d, ph->probe_id, pairing_updates[i].probe_id );
                        for ( int ii = 0; ii < probe_num; ii ++ )
                        {
                            fprintf( stderr, "ii=%d, probe_id=%d, ttl_d=%d\n", ii, phistory[ii].probe_id, phistory[ii].ttl_d );
                        }
                        exit(1);
                    }
                    assert( hi != NULL );
                    min_d = hi->min_d;
                    th_d_small = hi->th_d_small;
                    ph->th_d.bin_count = hi->th_d_bin;
                }
                ph->min_d.Initialize( MIN_SD_AGE_MAX );
                ph->min_d.Update( min_d );
                ph->last_Dd = VALUE_UNDEFINED;
                if ( th_d_small == VALUE_UNDEFINED )
                    th_d_small = TH_INITIAL_SMALL;
                ph->th_d.Initialize( th_d_small, TH_SD_CNT_MAX );
                ph->ResetQoMs();    // Reset QoMs of JP, 4P, fsd (except fsb)
            }
        }
        assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    }
    else
    {
        // fprintf( stdout, "pih_mutex 10\n" );
        assert( pthread_mutex_lock( &pih_mutex ) == 0 );
        for ( int i = 0; i < update_num; i ++ )
        {
            ProbeInfo * pi = FindProbeInfo( pairing_updates[i].probe_id );
            assert( pi != NULL );
            int previous_ttl = pi->ttl;
            pi->ttl = pairing_updates[i].ttl_d;
            IF_DEBUG( 4, fprintf( stderr, "p%d(%d->%d) ", pi->id, previous_ttl, pi->ttl ) );
        }
        assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    }
    IF_DEBUG( 4, fprintf( stderr, "... " ) );
}


// Change probing techniques
// Return true if changes are made, otherwise return false

bool Pong::ChangeTechniques( int hop_s_start, int hop_s_end, int technique, bool verbose )
{
    assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
    // fprintf( stdout, "pih_mutex 11\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    bool result = false;
    bool invalid = false;
    int new_technique[63];

    if ( TptPending() || ! technique_updates.empty() )
    {
        if ( verbose )
        {
            const char * pending_type = ( pairing_update_pending ? "pairing" : "technique" );
            fprintf( stdout, "Error: previous %s updates is still pending\n", pending_type );
        }
        goto Exit_ChangeTechniques;
    }
    
    if ( hop_s_start <= 0 || hop_s_start > hop_num )
    {
        if ( verbose )
            fprintf( stdout, "Error: invalid <hop_s_start>: %d; use 'p' command to check valid hop range\n", hop_s_start );
        invalid = true;
    }
    if ( hop_s_end <= 0 || hop_s_end > hop_num )
    {
        if ( verbose )
            fprintf( stdout, "Error: invalid <hop_s_end>: %d; use 'p' command to check valid hop range\n", hop_s_end );
        invalid = true;
    }
    if ( hop_s_start > hop_s_end )
    {
        if ( verbose )
            fprintf( stdout, "Error: <hop_s_start> should be no larger than <hop_s_end>\n" );
        invalid = true;
    }
    if ( invalid )
        goto Exit_ChangeTechniques;

    // Set new techniques
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;

        new_technique[i] = pi->technique;
        if ( pi->ttl >= hop_s_start && pi->ttl <= hop_s_end )
        {
            if ( technique == TECH_4P || technique == TECH_FSD )
            {
                ProbeHistory * ph = FindProbeHistory( pi->id );
                assert( ph != NULL );
                if ( ph->ttl_d == VALUE_UNDEFINED )
                {
                    if ( verbose )
                        fprintf( stdout, "Conflicts: Cannot change probing technique to '%s' for hop_s %d, hop_d has not been set yet\n", tech_code[technique], pi->ttl );
                }
                else
                    new_technique[i] = technique;
            }
            else
                new_technique[i] = technique;
        }
    }

    // Submit technique updates
    // Compare changes of technique and collect technique updates
    if ( verbose )
    {
        fprintf( stdout, "******** Technique updates ********\n" );
        fprintf( stdout, "id\thop_s\thop_d\ttechnique\n" );
    }
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        ProbeHistory * ph = FindProbeHistory( pi->id );

        if ( pi->technique != new_technique[i] )
        {
            TechniqueUpdate tu;
            tu.probe_id = pi->id;
            tu.technique = new_technique[i];
            technique_updates.push_back( tu );

            if ( verbose )
                fprintf( stdout, "%2d\t%5d\t%5d\t%3s --> %3s\n", pi->id, pi->ttl, ph->ttl_d, tech_code[pi->technique], tech_code[new_technique[i]] );
        }
    }
    if ( verbose )
        fprintf( stdout, "***********************************\n" );

    // Send technique updates to receiver
    if ( ! technique_updates.empty() )
    {
        if ( verbose )
            fprintf( stdout, "Send technique updates (%d updates).\n", technique_updates.size() );
        SendTechniqueUpdates();
        technique_update_pending = true;
        result = true;
    }

Exit_ChangeTechniques:
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
    if ( verbose )
        fflush( stdout );
    return result;
}


void Pong::UpdateTechniques( void )
{
    int update_num = technique_updates.size();
    // fprintf( stdout, "pih_mutex 12\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < update_num; i ++ )
    {
        ProbeInfo * pi = FindProbeInfo( technique_updates[i].probe_id );
        assert( pi != NULL );
        int previous_technique = pi->technique;
        pi->technique = technique_updates[i].technique;
        IF_DEBUG( 4, fprintf( stderr, "p%d(%s->%s) ", pi->id, tech_code[previous_technique], tech_code[pi->technique] ) );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    IF_DEBUG( 4, fprintf( stderr, "... " ) );
}


void Pong::ProcessRecord( ProbeInfo *pi, ProbeHistory *ph, FullRecord &fr )
{
    UpdateThresholds( ph, fr );
    if ( ! clock_skew_indicator )
    {
        UpdateCongestionStatus( pi, ph, fr );
        UpdateQoM( pi, ph, fr );
    }
}


void Pong::UpdateCongestionStatus( ProbeInfo *pi, ProbeHistory *ph, FullRecord &fr )
{
    double QoM;
    double eDs, eDd, eDf, eDb;
    eDs = eDd = eDf = eDb = 0;

    // Update end-to-end congestion count
    sample_count ++;
    sample_total ++;
    if ( fr.Df > th_f.small )
        sum_e2e_cc ++;
    sum_round_id += fr.round_id;

    switch ( pi->technique )
    {
    case TECH_4P:
        {
            double Lfs1 = fr.Df - fr.Dd;
            double Lfs2 = fr.Ds - fr.Db;
    
            double Lfd1 = fr.Df - fr.Ds;
            double Lfd2 = fr.Dd - fr.Db;
    
            eDs = 0.4 * fr.es;
            eDd = 0.4 * fr.ed;
            eDf = fr.ef;
            eDb = fr.eb;
    
            double e1, e2, I1, I2, upper_bound;
    
            // Estimate fs
            e1 = eDf + eDd;
            if ( e1 < 1 )
                e1 = 1;
            e2 = eDs + eDb;
            if ( e2 < 1 )
                e2 = 1;
            //I1 = fr.Dd + eDd;
            I1 = fr.Dd;
            //I2 = fr.Db + eDb;
            I2 = fr.Db;
            fr.Lfs = ( e2 * Lfs1 + e1 * Lfs2 ) / ( e1 + e2 );
            fr.Ufs = fr.Lfs + MIN( I1, I2 );

            //upper_bound = MIN( fr.Ds + eDs, fr.Df + eDf );
            upper_bound = MIN( fr.Ds, fr.Df );
            if ( fr.Ufs > upper_bound )
            {
                fr.Ufs = upper_bound;
                if ( fr.Lfs > upper_bound )
                    fr.Lfs = upper_bound;
            }
    
            // Estimate fd
            e1 = eDf + eDs;
            if ( e1 < 1 )
                e1 = 1;
            e2 = eDd + eDb;
            if ( e2 < 1 )
                e2 = 1;
            //I1 = fr.Ds + eDs;
            I1 = fr.Ds;
            //I2 = fr.Db + eDb;
            I2 = fr.Db;
            fr.Lfd = ( e2 * Lfd1 + e1 * Lfd2 ) / ( e1 + e2 );
            fr.Ufd = fr.Lfd + MIN( I1, I2 );

            //upper_bound = MIN( fr.Dd + eDd, fr.Df + eDf );
            upper_bound = MIN( fr.Dd, fr.Df );
            if ( fr.Ufd > upper_bound )
            {
                fr.Ufd = upper_bound;
                if ( fr.Lfd > upper_bound )
                    fr.Lfd = upper_bound;
            }

            QoM = fr.Q_4P;
        }
        break;
    
    case TECH_2P:
        {
            eDs = 0.4 * fr.es;
            eDf = fr.ef;
            fr.Lfs = 0;
            //fr.Ufs = MIN( fr.Ds + eDs, fr.Df + eDf );
            fr.Ufs = MIN( fr.Ds, fr.Df );
            //fr.Lfd = fr.Df - fr.Ds - eDf - eDs;
            fr.Lfd = fr.Df - fr.Ds;
            //fr.Ufd = fr.Df + eDf;
            fr.Ufd = fr.Df;

            QoM = 1;
        }
        break;
    
    case TECH_FSD:
        {
            eDs = 0.4 * fr.es;
            eDd = 0.4 * fr.ed;
            eDf = fr.ef;

            //fr.Lfs = fr.Df - fr.Dd - eDf - eDd;
            fr.Lfs = fr.Df - fr.Dd;
            //fr.Ufs = MIN( fr.Ds + eDs, fr.Df + eDf );
            fr.Ufs = MIN( fr.Ds, fr.Df );
            //fr.Lfd = fr.Df - fr.Ds - eDf - eDs;
            fr.Lfd = fr.Df - fr.Ds;
            //fr.Ufd = MIN( fr.Dd + eDd, fr.Df + eDf );
            fr.Ufd = MIN( fr.Dd, fr.Df );

            QoM = fr.Q_fsd;
        }
        break;
    
    case TECH_FSB:
        {
            eDs = 0.4 * fr.es;
            eDf = fr.ef;
            eDb = fr.eb;

            //fr.Lfs = fr.Ds - fr.Db - eDs - eDb;
            fr.Lfs = fr.Ds - fr.Db;
            //fr.Ufs = MIN( fr.Ds + eDs, fr.Df + eDf );
            fr.Ufs = MIN( fr.Ds, fr.Df );
            //fr.Lfd = fr.Df - fr.Ds - eDf - eDs;
            fr.Lfd = fr.Df - fr.Ds;
            /*
            if ( fr.Ds - eDs > fr.Db + eDb )
                fr.Ufd = fr.Df - fr.Ds + fr.Db + eDf + eDs + eDb;
            else
                fr.Ufd = fr.Df + eDf;
            */
            if ( fr.Ds > fr.Db )
                fr.Ufd = fr.Df - fr.Ds + fr.Db;
            else
                fr.Ufd = fr.Df;

            QoM = fr.Q_fsb;
        }
        break;

    default:
        fprintf( stderr, "Invalid technique code: %d\n", pi->technique );
        assert(0);
    }

    if ( fr.Lfs < 0 ) fr.Lfs = 0;
    if ( fr.Lfd < 0 ) fr.Lfd = 0;
    if ( fr.Lfs > fr.Ufs ) fr.Lfs = fr.Ufs;
    if ( fr.Lfd > fr.Ufd ) fr.Lfd = fr.Ufd;

    double threshold = MIN( ph->th_s.small, th_f.small );
    GetCongestionExtent( fr.Efs, fr.Cfs, fr.Lfs, fr.Ufs, threshold );
    if ( pi->technique == TECH_4P || pi->technique == TECH_FSD )
        threshold = MIN( ph->th_d.small, th_f.small );
    else
        threshold = th_f.small;
    GetCongestionExtent( fr.Efd, fr.Cfd, fr.Lfd, fr.Ufd, threshold );

    if( QoM != VALUE_UNDEFINED )
    {
        fr.Cfs *= QoM;
        fr.Cfd *= QoM;
    }
    else
        return;

    if ( DebugOn( pi->id ) )
        IF_DEBUG( 6, fprintf( stderr, "- Technique=%s, Efs=%.2lf, Cfs=%.2lf, Efd=%.2lf, Cfd=%.2lf (Dfs=[%.2lf, %.2lf], Dfd=[%.2lf, %.2lf], eDs=%.2lf, eDd=%.2lf, eDf=%.2lf, eDb=%.2lf)\n",
            tech_code[pi->technique], fr.Efs, fr.Cfs, fr.Efd, fr.Cfd, fr.Lfs / 1000, fr.Ufs / 1000, fr.Lfd / 1000, fr.Ufd / 1000, eDs / 1000, eDd / 1000, eDf / 1000, eDb / 1000 ) );

    last_pr[pi->id].fr = fr;
    last_pr[pi->id].technique = pi->technique;

    if ( pi->id == 0 || pi->id == probe_num - 1 )
        UpdateLCD( last_pr[pi->id] );
    if ( pi->id > 0 )
    {
        if ( last_pr[pi->id - 1].fr.round_id == last_pr[pi->id].fr.round_id )
        {
            if ( pi->ttl > last_pr[pi->id - 1].ttl )
                UpdateLCD( last_pr[pi->id - 1], last_pr[pi->id] );
            else
                UpdateLCD( last_pr[pi->id], last_pr[pi->id - 1] );
        }
        /*
        else
            fprintf( stderr, "UpdateLCD( %d, %d ), round_id: %d, %d\n", pi->id - 1, pi->id, last_pr[pi->id - 1].fr.round_id, last_pr[pi->id].fr.round_id );
        */
    }
    if ( pi->id < probe_num - 1 )
    {
        if ( last_pr[pi->id + 1].fr.round_id == last_pr[pi->id].fr.round_id )
        {
            if ( pi->ttl > last_pr[pi->id + 1].ttl )
                UpdateLCD( last_pr[pi->id + 1], last_pr[pi->id] );
            else
                UpdateLCD( last_pr[pi->id], last_pr[pi->id + 1] );
        }
        /*
        else
            fprintf( stderr, "UpdateLCD( %d, %d ), round_id: %d, %d\n", pi->id + 1, pi->id, last_pr[pi->id + 1].fr.round_id, last_pr[pi->id].fr.round_id );
        */
    }

    UpdateICS( last_pr[pi->id] );
}


// Update long-term congestion distribution

void Pong::UpdateLCD( const ProbeRecord &pr1, const ProbeRecord &pr2 )
{
    assert( pr1.ttl < pr2.ttl && pr1.ttl > 0 && pr2.ttl <= hop_num );
    if ( Congested( pr1.fr.Efd ) && Congested( pr2.fr.Efs ) )
    {
        if ( ! Congested( pr1.fr.Efs ) || ! Congested( pr2.fr.Efd ) )
        {
            double delta = pr1.fr.Cfd + pr2.fr.Cfs;
            if ( ! Congested( pr1.fr.Efs ) && ! Congested( pr2.fr.Efd ) )
            {
                delta = 0.4 + 0.6 * ( delta + pr1.fr.Cfs + pr2.fr.Cfd ) / 4;
                double increment = 2 * delta;
                link_cc_sum_all += increment;
                for ( int i = pr1.ttl; i < pr2.ttl; i ++ )
                    link_cc[i] += increment;
            }
            else
            {
                if ( ! Congested( pr1.fr.Efs ) )
                {
                    if ( pr2.technique == TECH_2P )
                        // Give small delta if the second node uses 2P probing and says both half paths are congested
                        delta = 0.1 + 0.9 * ( delta + pr1.fr.Cfs ) / 3;
                    else
                        delta = 0.4 + 0.6 * ( delta + pr1.fr.Cfs ) / 3;
                }
                else
                    delta = 0.4 + 0.6 * ( delta + pr2.fr.Cfd ) / 3;
                double increment = 0.7 * delta;
                link_cc_sum_all += increment;
                for ( int i = pr1.ttl; i < pr2.ttl; i ++ )
                    link_cc[i] += increment;
            }
        }
    }
}


void Pong::UpdateLCD( const ProbeRecord &pr )
{
    assert( pr.ttl > 0 && pr.ttl <= hop_num );
    if ( pr.fr.probe_id == 0 && Congested( pr.fr.Efs ) )
    {
        double increment = 0.4 + 0.6 * pr.fr.Cfs;
        link_cc_sum_all += increment;
        for ( int i = 0; i < pr.ttl; i ++ )
            link_cc[i] += increment;
    }
    else if ( pr.fr.probe_id == probe_num - 1 && Congested( pr.fr.Efd ) )
    {
        double increment;
        if ( pr.technique == TECH_2P && Congested( pr.fr.Efs ) )
            // Give small increment if the node uses 2P probing and says both half paths are congested
            increment = 0.1 + 0.9 * pr.fr.Cfd;
        else
            increment = 0.4 + 0.6 * pr.fr.Cfd;
        link_cc_sum_all += increment;
        for ( int i = pr.ttl; i <= hop_num; i ++ )
            link_cc[i] += increment;
    }
}


// UPdate instantaneous congestion status

void Pong::UpdateICS( const ProbeRecord &pr )
{
    double LCD_positive_rate[64];
    int positive_count;
    assert( pr.ttl >= 0 && pr.ttl <= hop_num );
    if ( Congested( pr.fr.Efs ) )
    {
        positive_count = 0;
        for ( int i = 0; i < pr.ttl; i ++ )
        {
            LCD_positive_rate[i] = LCD_PositiveRate( i );
            if ( LCD_positive_rate[i] > 0 && ! redundant_link[i] )
                positive_count ++;
        }
        if ( positive_count > 0 )
        {
            // Single congestion point has a confidence much higher than that of multiple congestion points
            double Cf = ( positive_count == 1 ? 1.0 : 0.25 );
            Cf *= 0.3 + 0.7 * pr.fr.Cfs;
            for ( int i = 0; i < pr.ttl; i ++ )
            {
                if ( LCD_positive_rate[i] > 0 )
                {
                    double increment = Cf * LCD_positive_rate[i];
                    sum_link_ci[i] += increment;
                    if ( ! redundant_link[i] )
                        link_ci_sum_all += increment;
                }
            }
        }
    }
    if ( Congested( pr.fr.Efd ) )
    {
        positive_count = 0;
        for ( int i = pr.ttl; i <= hop_num; i ++ )
        {
            LCD_positive_rate[i] = LCD_PositiveRate( i );
            if ( LCD_positive_rate[i] > 0 && ! redundant_link[i] )
                positive_count ++;
        }
        if ( positive_count > 0 )
        {
            // Single congestion point has a confidence much higher than that of multiple congestion points
            double Cf = ( positive_count == 1 ? 1.0 : 0.25 );
            Cf *= 0.3 + 0.7 * pr.fr.Cfd;
            for ( int i = pr.ttl; i <= hop_num; i ++ )
            {
                if ( LCD_positive_rate[i] > 0 )
                {
                    double increment = Cf * LCD_positive_rate[i];
                    sum_link_ci[i] += increment;
                    if ( ! redundant_link[i] )
                        link_ci_sum_all += increment;
                }
            }
        }
    }
}


#define POSITIVE_FACTOR 0.5
#define BLUR_FACTOR     3
#define TH_DOMI_LINK_CC 20

double Pong::LCD_PositiveRate( int link_id )
{
    double max_link_cc;
    if ( link_id != dominant_bottleneck && ( max_link_cc = LinkCC_30sec( dominant_bottleneck ) ) > TH_DOMI_LINK_CC &&
         LinkCC_30sec( link_id ) * BLUR_FACTOR < max_link_cc )
        return 0;

    if ( link_cc[link_id] > 3.0 * POSITIVE_FACTOR || link_cc_10sec[5][link_id] > 3.0 * POSITIVE_FACTOR )
        return 1.0;
/*
    if ( LCD_30sec[link_id] > 5.0 * POSITIVE_FACTOR )
        return 0.9;
    if ( LCD_1min[link_id] > 8.0 * POSITIVE_FACTOR )
        return 0.7;
    if ( LCD_1min[link_id] > 5.0 * POSITIVE_FACTOR )
        return 0.6;
    if ( link_cc_1min[9][link_id] > 5.0 * POSITIVE_FACTOR )
        return 0.5;
    if ( LCD_3min[link_id] > 10.0 * POSITIVE_FACTOR )
        return 0.3;
    if ( LCD_3min[link_id] > 6.0 * POSITIVE_FACTOR )
        return 0.2;
*/

    double link_cc_subtotal = link_cc_10sec[5][link_id] + link_cc_10sec[4][link_id];
    if ( link_cc_subtotal > 4.0 * POSITIVE_FACTOR )
        return 0.9;
    link_cc_subtotal += link_cc_10sec[3][link_id];
    if ( link_cc_subtotal > 5.0 * POSITIVE_FACTOR )
        return 0.8;
    link_cc_subtotal += link_cc_10sec[2][link_id];
    if ( link_cc_subtotal > 5.5 * POSITIVE_FACTOR )
        return 0.65;
    link_cc_subtotal += link_cc_10sec[1][link_id];
    if ( link_cc_subtotal > 6.0 * POSITIVE_FACTOR )
        return 0.5;
    link_cc_subtotal += link_cc_10sec[0][link_id];
    if ( link_cc_subtotal > 6.5 * POSITIVE_FACTOR )
        return 0.35;
    
    return 0;
}


double Pong::LinkCC_30sec( int link_id )
{
    return link_cc[link_id] + link_cc_10sec[5][link_id] + link_cc_10sec[4][link_id] + link_cc_10sec[3][link_id];
}


void Pong::UpdateDominantBottleneck( void )
{
    double max_link_cc_30sec = 0;
    int link_id = 0;
    for ( int i = 0; i <= hop_num; i ++ )
    {
        double link_cc_30sec = LinkCC_30sec( i );
        if ( link_cc_30sec > max_link_cc_30sec )
        {
            max_link_cc_30sec = link_cc_30sec;
            link_id = i;
        }
    }
    dominant_bottleneck = link_id;
}


void GetCongestionExtent( double &extent, double &confidence, double L, double U, double threshold )
{
    if ( L >= threshold )
    {
        extent = 1;
        confidence = 1;
    }
    else if ( U <= threshold )
    {
        extent = 0;
        confidence = 1;
    }
    else
    {
        // Convert queuing delay values roughly to logarithmic counterparts
        double U_idx = ProbeThreshold::GetBinIndex( U );
        double L_idx = ProbeThreshold::GetBinIndex( L );
        double T_idx = ProbeThreshold::GetBinIndex( threshold );

        if ( U_idx == L_idx )
        {
            extent = 1;
            confidence = 1;
        }
        else
        {
            extent = ( U_idx - T_idx ) / ( U_idx - L_idx );
            confidence = 0.3;
            double I_threshold = I_THRESHOLD_FACTOR * T_idx;
            if ( U_idx - L_idx > I_threshold )
                confidence *= I_threshold / ( U_idx - L_idx );
        }
    }
}


bool Congested( double extent )
{
    return ( extent > 0.5 );
}


void Pong::UpdateThresholds( ProbeHistory *ph, FullRecord &fr )
{
    if ( ! clock_skew_indicator )
    {
        if ( fr.Isset_f() )
            th_f.Update( fr.Df );
        if ( fr.Isset_b() )
            th_b.Update( fr.Db );
    }
    if ( fr.Isset_s() )
        ph->th_s.Update( fr.Ds );
    if ( fr.Isset_d() )
        ph->th_d.Update( fr.Dd );
}


void Pong::UpdateQoM( ProbeInfo *pi, ProbeHistory *ph, FullRecord &fr )
{
    double C_4P, C_fsd, C_fsb;   // Confidence
    fr.Q_JP = fr.Q_4P = fr.Q_fsd = fr.Q_fsb = VALUE_UNDEFINED;
    
    if ( fr.Isset_f() && fr.Isset_s() )
    {
        if ( fr.Isset_b() )
        {
            fr.Q_fsb = fr.QoM_fsb();
            if ( fr.Df > th_f.small || fr.Ds > ph->th_s.small || fr.Db > th_b.small )
                C_fsb = 1;
            else
                C_fsb = 0.1;
            if ( fr.Isset_d() )
            {
                fr.Q_4P = fr.QoM_4P();
                if ( fr.Df > th_f.small || fr.Db > th_b.small || fr.Ds > ph->th_s.small || fr.Dd > ph->th_d.small )
                    C_4P = 1;
                else
                    C_4P = 0.1;
                fr.Q_JP = fr.QoM_JP();
            }
        }
        if ( fr.Isset_d() )
        {
            fr.Q_fsd = fr.QoM_fsd();
            if ( fr.Df > th_f.small || fr.Ds > ph->th_s.small || fr.Dd > ph->th_d.small )
                C_fsd = 1;
            else
                C_fsd = 0.1;
        }
    }

    if ( DebugOn( ph->probe_id ) )
        IF_DEBUG( 6, fprintf( stderr, "- QoM: " ) );
    if ( fr.Q_JP != VALUE_UNDEFINED )
    {
        ph->QoM_JP.Update( fr.Q_JP, 1 );
        if ( DebugOn( ph->probe_id ) )
            IF_DEBUG( 6, fprintf( stderr, "Q_JP=%.2lf(%.2lf)", fr.Q_JP, ph->QoM_JP.value ) );
    }
    if ( fr.Q_4P != VALUE_UNDEFINED )
    {
        ph->QoM_4P.Update( fr.Q_4P, C_4P );
        if ( DebugOn( ph->probe_id ) )
            IF_DEBUG( 6, fprintf( stderr, ", Q_4P=%.2lf(%.2lf)", fr.Q_4P, ph->QoM_4P.value ) );
    }
    if ( fr.Q_fsd != VALUE_UNDEFINED )
    {
        ph->QoM_fsd.Update( fr.Q_fsd, C_fsd );
        if ( DebugOn( ph->probe_id ) )
            IF_DEBUG( 6, fprintf( stderr, ", Q_fsd=%.2lf(%.2lf)", fr.Q_fsd, ph->QoM_fsd.value ) );
    }
    if ( fr.Q_fsb != VALUE_UNDEFINED )
    {
        ph->QoM_fsb.Update( fr.Q_fsb, C_fsb );
        if ( DebugOn( ph->probe_id ) )
            IF_DEBUG( 6, fprintf( stderr, ", Q_fsb=%.2lf(%.2lf)", fr.Q_fsb, ph->QoM_fsb.value ) );
    }
    if ( DebugOn( ph->probe_id ) )
        IF_DEBUG( 6, fprintf( stderr, "\n" ) );

    if ( auto_tpt )
    {
        int previous_technique = pi->technique;
        int technique = previous_technique;

        if ( technique == TECH_4P  && ph->QoM_4P.value_l  < TH_DEMOTE_4P  && ph->QoM_4P.age  > TH_QoM_AGE_MIN ||
             technique == TECH_FSD && ph->QoM_fsd.value_l < TH_DEMOTE_FSD && ph->QoM_fsd.age > TH_QoM_AGE_MIN ||
             technique == TECH_FSB && ph->QoM_fsb.value_l < TH_DEMOTE_FSB && ph->QoM_fsb.age > TH_QoM_AGE_MIN )
            // Demote to 2P
            technique = TECH_2P;
 
        if ( ph->QoM_4P.value > TH_PROMOTE_4P && ph->QoM_4P.age > TH_QoM_AGE_MIN && technique != TECH_4P )
            // Promote to 4P
            technique = TECH_4P;
        else if ( ph->QoM_fsd.value > TH_PROMOTE_FSD && ph->QoM_fsd.age > TH_QoM_AGE_MIN && technique != TECH_4P && technique != TECH_FSD )
            // Promote to fsd
            technique = TECH_FSD;
        else if ( ph->QoM_fsb.value > TH_PROMOTE_FSB && ph->QoM_fsb.age > TH_QoM_AGE_MIN && technique != TECH_4P && technique != TECH_FSD && technique != TECH_FSB )
            // Promote to fsb
            technique = TECH_FSB;
    
        if ( technique != previous_technique )
            EnqueueTechniqueUpdate( pi->id, technique );
    }
}


bool Pong::EnqueueTechniqueUpdate( int probe_id, int technique )
{
    assert( pthread_mutex_lock( &tpt_mutex ) == 0 );

    bool result = false;
    if ( ! TptPending() )
    {
        // Check whether this probe_id is already queued
        int update_num = technique_updates.size();
        for ( int i = 0; i < update_num; i ++ )
            if ( technique_updates[i].probe_id == probe_id )
                goto Exit_EnqueueTechniqueUpdate;
        
        TechniqueUpdate tu;
        tu.probe_id = probe_id;
        tu.technique = technique;
        technique_updates.push_back( tu );
        result = true;
    }

Exit_EnqueueTechniqueUpdate:
    assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
    return result;
}


void Pong::UpdateFullRecord( HalfRecord &hr, bool is_sender_record )
{
    assert( pthread_mutex_lock( &frb_mutex ) == 0 );

    int probe_id = hr.probe_id;
    ProbeInfo * pi = NULL;
    ProbeHistory * ph = NULL;
    FullRecordBuffer * frb = FindFrBuffer( probe_id );
    if ( frb == NULL )
    {
        fprintf( stderr, "Invalid probe_id %d, round_id=%d, status=%d\n", probe_id, hr.round_id, hr.status );
        assert(0);
    }
    list<FullRecord>::iterator fr = frb->GetRecord( hr.round_id );
    if ( fr == frb->buffer.end() )
    {
        if ( DebugOn( probe_id ) )
            IF_DEBUG( 4, fprintf( stderr, "Invalid round_id %d of FullRecord: probe_id=%d, status=%X\n", hr.round_id, probe_id, hr.status ) );
        goto Exit_UpdateFullRecord;
    }
    if ( is_sender_record )
    {
        fr->send_time_s = hr.send_time_sd;
        fr->recv_time_s = hr.recv_time_sd;
        fr->send_time_f = hr.send_time_fb;
        fr->recv_time_b = hr.recv_time_bf;
        fr->send_time_f_scheduled = hr.send_time_fb_scheduled;
        fr->status |= hr.status & HR_ALL_TIME;
        fr->status |= SENDER_DONE;
    }
    else
    {
        fr->send_time_d = hr.send_time_sd;
        fr->recv_time_d = hr.recv_time_sd;
        fr->send_time_b = hr.send_time_fb;
        fr->recv_time_f = hr.recv_time_bf;
        fr->send_time_b_scheduled = hr.send_time_fb_scheduled;
        fr->status |= ( hr.status & HR_ALL_TIME ) << 4;
        fr->status |= RECEIVER_DONE;
        ph = FindProbeHistory( probe_id );
        if ( ph == NULL )
        {
            fprintf( stderr, "Invalid probe_id %d\n", probe_id );
            assert(0);
        }
        ph->time_offset_d = hr.time_offset_d;
        ph->ttl_d = hr.ttl_d;
    }
    if ( fr->RecordComplete() )
    {
        if ( ph == NULL )
            ph = FindProbeHistory( probe_id );
        if ( ph == NULL )
        {
            fprintf( stderr, "Invalid probe_id %d\n", probe_id );
            assert(0);
        }
        pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        if ( fr->Isset_s() )
        {
            fr->s = delta_t( fr->send_time_s, fr->recv_time_s );
            ph->min_s.Update( fr->s );
            fr->Ds = fr->s - ph->min_s.value;
            fr->es = ( ph->last_Ds != VALUE_UNDEFINED ? fabs( fr->Ds - ph->last_Ds ) : 0 );
            ph->last_Ds = fr->Ds;
        }
        if ( fr->Isset_d() )
        {
            fr->d = delta_t( fr->send_time_d, fr->recv_time_d );
            ph->min_d.Update( fr->d );
            fr->Dd = fr->d - ph->min_d.value;
            fr->ed = ( ph->last_Dd != VALUE_UNDEFINED ? fabs( fr->Dd - ph->last_Dd ) : 0 );
            ph->last_Dd = fr->Dd;
        }
        if ( fr->Isset_f() )
        {
            fr->f = delta_t( fr->send_time_f, fr->recv_time_f );
            if ( min_f.Update( fr->f ) == 1 )
                skew_monitor_f.CheckSkew( fr->recv_time_f );
            fr->Df = fr->f - min_f.value;
            fr->ef = ( last_Df != VALUE_UNDEFINED ? fabs( fr->Df - last_Df ) : 0 );
            last_Df = fr->Df;
        }
        if ( fr->Isset_b() )
        {
            fr->b = delta_t( fr->send_time_b, fr->recv_time_b );
            if ( min_b.Update( fr->b ) == 1 )
                skew_monitor_b.CheckSkew( fr->recv_time_b );
            fr->Db = fr->b - min_b.value;
            fr->eb = ( last_Db != VALUE_UNDEFINED ? fabs( fr->Db - last_Db ) : 0 );
            last_Db = fr->Db;
        }
        double time_difference_fd = 9E9, time_difference_sb = 9E9;
        if ( fr->Isset_f() && fr->Isset_d() )
        {
            // time_difference_fd = ( send_time_f_scheduled + min_f ) - ( send_time_b_scheduled + min_d )
            time_difference_fd = delta_t( fr->send_time_b_scheduled, fr->send_time_f_scheduled ) + min_f.value - ph->min_d.value;
        }        
        if ( fr->Isset_s() && fr->Isset_b() )
        {
            // time_difference_sb = ( send_time_f_scheduled + min_s ) - ( send_time_b_scheduled + min_b )
            time_difference_sb = delta_t( fr->send_time_b_scheduled, fr->send_time_f_scheduled ) + ph->min_s.value - min_b.value;
        }
        double time_offset_increment;
        if ( time_difference_fd != 9E9 )
            time_offset_increment = time_difference_fd / 1E6;
        else if ( time_difference_sb != 9E9 )
            time_offset_increment = time_difference_sb / 1E6;
        else
        {
            fprintf( stderr, "Invalid FullRecord\n" );
            assert(0);
        }

        if ( DebugOn( probe_id ) )
        {
            IF_DEBUG( 6, fprintf( stderr, "FullRecord: probe_id=<%d>", probe_id ) );
            if ( time_difference_fd != 9E9 )
                IF_DEBUG( 6, fprintf( stderr, ", dsrt_fd=%.3lf", time_difference_fd / 1000 ) );
            if ( time_difference_sb != 9E9 )
                IF_DEBUG( 6, fprintf( stderr, ", dsrt_sb=%.3lf", time_difference_sb / 1000 ) );
            // Comment: dsrt - difference of scheduled receiving time
            if ( fr->Isset_f() )
                IF_DEBUG( 6, fprintf( stderr, ", dt_f=%.3lf", delta_t( fr->send_time_f_scheduled, fr->send_time_f ) / 1000 ) );
            if ( fr->Isset_b() )
                IF_DEBUG( 6, fprintf( stderr, ", dt_b=%.3lf", delta_t( fr->send_time_b_scheduled, fr->send_time_b ) / 1000 ) );
            // Comment: dt - difference of sending time between scheduled one and actual one

            IF_DEBUG( 6, fprintf( stderr, "\n- " ) );
            IF_DEBUG( 6, fr->Print() );
        }

        if ( fabs( time_offset_increment ) > 0.001 )
        {
            if ( ! ph->ATO_pending && fr->round_id >= ph->ATA_round_id )
            {
                if ( DebugOn( probe_id ) )
                    IF_DEBUG( 5, fprintf( stderr, "Send ATO message, probe_id=%d, time_offset_increment=%.3lf\n", probe_id, time_offset_increment * 1000 ) );
                SendATO( probe_id, time_offset_increment );
                ph->ATO_pending = true;
            }
        }
        else
        {
            ProcessRecord( pi, ph, *fr );
        }

        // Remove this FullRecord
        frb->buffer.erase( fr );
        // Check for FullRecords that are older than this one
        frb->CheckIncompleteRecords( hr.round_id );
    }

Exit_UpdateFullRecord:
    assert( pthread_mutex_unlock( &frb_mutex ) == 0 );
}


bool Pong::UpdateHopIP( int ttl, int ip )
{
    vector<int> & hips = hop_ips[ttl - 1];
    int num = hips.size();
    int i;
    for ( i = 0; i < num; i ++ )
    {
        if ( hips[i] == ip )
            break;
    }
    if ( i == num )
    {
        hop_ips[ttl - 1].push_back( ip );
        return true;
    }
    return false;
}


void Pong::UpdateRouteAlteration( ProbeResult &pr )
{
    if ( ! is_sender || pr.probe_type != probe_sd_in )
        return;

    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    ProbeInfo *pi = FindProbeInfo( pr.probe_id );
    if ( pi == NULL )
    {
        fprintf( stderr, "Invalid probe_id %d, round_id=%d, probe_type=%s\n", pr.probe_id, pr.round_id, ptype_code[pr.probe_type] );
        assert(0);
    }
    assert( pi->ttl >= 1 );
    if ( pr.icmpSourceAddress.s_addr != hop_info[pi->ttl - 1].ip.s_addr )
    {
        route_alteration_count ++;
        if ( ! route_altered[pi->ttl - 1] )
        {
            route_altered[pi->ttl - 1] = true;
            altered_hop_count ++;
        }
        hop_info[pi->ttl - 1].ip = pr.icmpSourceAddress;
        UpdateHopIP( pi->ttl, pr.icmpSourceAddress.s_addr );
        route_converge_count = 0;
    }

    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
}


void Pong::UpdateHalfRecord( ProbeResult &pr )
{
    assert( pthread_mutex_lock( &hrb_mutex ) == 0 );
    // fprintf( stdout, "pih_mutex 14\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    ProbeInfo * pi;
    list<HalfRecord>::iterator hr;
    HalfRecordBuffer * hrb = FindHrBuffer( pr.probe_id );
    if ( hrb == NULL )
    {
        fprintf( stderr, "%s Error: Invalid probe_id %d, round_id=%d, probe_type=%s\n", GetMeasurementTime(), pr.probe_id, pr.round_id, ptype_code[pr.probe_type] );
        //if ( pr.probe_id == 99999 && pr.round_id == 99999 )
        goto Exit_UpdateHalfRecord;
    }
    hr = hrb->GetRecord( pr.round_id );
    if ( hr == hrb->buffer.end() )
    {
        if ( DebugOn( pr.probe_id ) )
        {
            IF_DEBUG( 4, fprintf( stderr, "%s Expired or invalid round_id %d of HalfRecord: probe_id=%d, probe_type=%s, rid=%d", GetMeasurementTime(), pr.round_id, pr.probe_id, ptype_code[pr.probe_type], round_id ) );
            if ( pr.probe_type == probe_sd_in )
                IF_DEBUG( 4, fprintf( stderr, ", icmpSourceAddress=%s", inet_ntoa( pr.icmpSourceAddress ) ) );
            IF_DEBUG( 4, fprintf( stderr, "\n" ) );
        }
        goto Exit_UpdateHalfRecord;
    }

    pi = FindProbeInfo( pr.probe_id );
    if ( pi == NULL )
    {
        fprintf( stderr, "Invalid probe_id %d, round_id=%d, probe_type=%s\n", pr.probe_id, pr.round_id, ptype_code[pr.probe_type] );
        assert(0);
    }
    
    switch ( pr.probe_type )
    {
    case probe_fb_out:
        hr->SetSendTime_fb( pr.tv );
        hr->send_time_fb_scheduled = pr.timestamp;
        break;
    case probe_fb_in:
        hr->SetRecvTime_bf( pr.tv );
        hrb->CheckProbeLoss( pr );
        break;
    case probe_sd_out:
        hr->SetSendTime_sd( pr.tv );
        break;
    case probe_sd_in:
        if ( pi->ttl >= 1 )
        {
            if ( pi->ttl != TTL_TARGETED_D && ! pairing_update_pending && pr.round_id >= PUA_round_id && pr.icmpSourceAddress.s_addr != hop_info[pi->ttl - 1].ip.s_addr )
            {
                route_alteration_count ++;
                if ( ! route_altered[pi->ttl - 1] )
                {
                    route_altered[pi->ttl - 1] = true;
                    altered_hop_count ++;
                }
                if ( DebugOn( pr.probe_id ) )
                {
                    IF_DEBUG( 4, fprintf( stderr, "%s WARNING: Route alteration detected, hop=%d, current_ip=%s, ", GetMeasurementTime(), pi->ttl, inet_ntoa( pr.icmpSourceAddress ) ) );
                    IF_DEBUG( 4, fprintf( stderr, "previous_ip=%s, probe_id=%d, route_alteration_count=%d, altered_hop_count=%d\n", inet_ntoa( hop_info[pi->ttl - 1].ip ), pr.probe_id, (int)route_alteration_count, altered_hop_count ) );
                }
            }
            hr->SetRecvTime_sd( pr.tv );
            hrb->CheckProbeLoss( pr );
        }
        break;
    default:;
    }
    
    if ( hr->RecordComplete( pi->ttl >= 1 ) )
    {
        if ( DebugOn( pr.probe_id ) )
        {
            IF_DEBUG( 7, fprintf( stderr, "HalfRecord: probe_id=%d, ", pr.probe_id ) );
            IF_DEBUG( 7, hr->Print() );
        }
        if ( is_sender )
            // Update FullRecord with this HalfRecord
            UpdateFullRecord( *hr, true );
        else
        {
            if ( ! is_sender )
            {
                // Set time_offset_d and ttl_d fields
                hr->time_offset_d = pi->time_offset;
                hr->ttl_d = pi->ttl;
            }
            // Send this HalfRecord to the sender side
            SendHalfRecord( *hr );
        }
        // Remove this HalfRecord
        hrb->buffer.erase( hr );
        // Check for HalfRecords that are older than this one
        hrb->CheckIncompleteRecords( pr.round_id );
    }

    if ( pi->ttl == TTL_TARGETED_D && hrb->buffer.size() > 10 )
    {
        if ( pi->tdp_state == TDP_ICMP )
        {
            pi->tdp_state = TDP_UDP;
            hrb->buffer.clear();
            IF_DEBUG( 3, fprintf( stderr, "%s Targeted ICMP d probe fails, try UDP instead, probe_id = %d\n", GetMeasurementTime(), pr.probe_id ) );
        }
        else if ( pi->tdp_state == TDP_UDP )
        {
            pi->tdp_state = TDP_NONE;
            hrb->buffer.clear();
            IF_DEBUG( 3, fprintf( stderr, "%s Targeted UDP d probe fails, sending TDF, probe_id = %d\n", GetMeasurementTime(), pr.probe_id ) );
            SendTDF( pr.probe_id );
        }
    }

Exit_UpdateHalfRecord:    
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    assert( pthread_mutex_unlock( &hrb_mutex ) == 0 );
}

double TimeInMsec( timeval &tv )
{
    return tv.tv_sec * 1000.0 + tv.tv_usec / 1000.0;
}

double delta_t( const timeval &t1, const timeval &t2 )
{
    double dt;
    dt = (double)( t2.tv_sec - t1.tv_sec) * 1000000.0
       + (double)( t2.tv_usec - t1.tv_usec);
    return dt;
}


timeval Timeval( double usec )
{
    timeval tv;
    tv.tv_sec = (int)( usec / 1E6 );
    tv.tv_usec = (int)( usec - tv.tv_sec * 1E6 );
    return tv;
}


double TimeInUsec( timeval &tv )
{
    return tv.tv_sec * 1000000.0 + tv.tv_usec;
}

void GetCurrentTime( timeval &time )
{
    gettimeofday( &time, (struct timezone *)NULL );
}

//
// InCksum
//
// Input:  addr, len (buffer to checksum)
//
// Output:  IP checksum for this buffer in return value
//
// Compute IP checksum for a buffer.  It's put here because it's fairly
// general-purpose, and both the IPv4 and IPv6 tests could potentially
// make use of it.  RFC 1071 has implementation notes, and we use
// the sample implementation (essentially unmodified) therein.
//
u_short InCksum( u_short *addr, int len )
{
    register int sum = 0;
    u_short checksum;
    
    while ( len > 1 )
    {
        sum += *addr++;
        len -= 2;
    }
    
    // Add left-over byte, if any
    if ( len > 0 )
    {
        sum += *(u_char *)addr;
    }
    
    // fold 32-bit sum to 16 bits
    while ( sum >> 16 )
    {
        sum = (sum & 0xffff) + (sum >> 16);
    }
    
    checksum = ~sum;
    return checksum;
}


/*
Send a message through the control channel
*/

void Pong::SendCtrCode( long ctr_code )
{
    char ctr_buff[sizeof(long)];
    long ctr_code_n = htonl( ctr_code );
    memcpy( (void *)ctr_buff, &ctr_code_n, sizeof(long) );

    if ( write( sock_tcp, ctr_buff, sizeof(long) ) != sizeof(long) )
    {
        fprintf( stderr, "%s Send control message failed\n", GetMeasurementTime() );
        exit(1);
    }
}


/*
Receive an empty message from the control channel
*/

long Pong::RecvCtrCode( void )
{
    long ctr_code;
    int rc = recv( sock_tcp, &ctr_code, sizeof(long), MSG_WAITALL );
    if ( rc == 0 )
    {
        IF_DEBUG( 1, fprintf( stderr, "%s Control channel closed, interpret this as a quit command\n", GetMeasurementTime() ) );
        return CC_QUIT;
    }
    if ( rc != sizeof(long) )
    {
        fprintf( stderr, "%s Receive control message failed\n", GetMeasurementTime() );
        exit(1);
    }
    return ntohl( ctr_code );
}


// Receive probes and other messages (control message, interactive commands, etc.)

void Pong::ReceiveMessages( void )
{

    int max_fd = MAX( socketIn, sock_udp );
    max_fd = MAX( max_fd, sock_tcp );
    max_fd = MAX( max_fd, sock_monitor_listener );
    max_fd = MAX( max_fd, sock_report );

    while ( ! done )
    {
        fd_set read_fds;

        FD_ZERO( &read_fds );
        FD_SET( socketIn, &read_fds );
        FD_SET( sock_udp, &read_fds );
        FD_SET( sock_monitor_listener, &read_fds );
        FD_SET( sock_report, &read_fds );
        if ( phase >= PP_PROBE )
        {
            FD_SET( sock_tcp, &read_fds );
            if ( enable_interactive_command || sock_monitor >= 0 )
                FD_SET( STDIN, &read_fds );
        }

        int rc;
        rc = select( max_fd + 1, &read_fds, NULL, NULL, NULL );
        if ( rc < 0 )
        {
            perror( "select" );
            exit(1);
        }

        ProbeResult pr;
        
        GetCurrentTime( pr.tv );
        
        if ( FD_ISSET( socketIn, &read_fds ) )
        {
            //IF_DEBUG( 9, fprintf( stderr, "%c probe received\n", pn_sd ) );
            RecvRouterProbe( pr );
        }
        if ( FD_ISSET( sock_udp, &read_fds ) )
        {
            //IF_DEBUG( 9, fprintf( stderr, "%c probe received\n", pn_fb_in ) );
            RecvUdpProbe( pr );
        }
        if ( FD_ISSET( sock_tcp, &read_fds ) )
        {
            ProcessControlMessage();
        }
        if ( FD_ISSET( STDIN, &read_fds ) )
        {
            ProcessInteractiveCommand();
        }
        if ( FD_ISSET( sock_monitor_listener, &read_fds ) )
        {
            AcceptMonitorClient();
        }
        if ( FD_ISSET( sock_report, &read_fds ) )
        {
            RecvUdpMessages();
        }
    }
}


void Pong::OutputStartTime( void )
{
    char time_string[50];
    time_t gmt_time = time_baseline.tv_sec;
    strcpy( time_string, asctime( gmtime( &gmt_time ) ) );
    time_string[strlen(time_string) - 1] = 0;
    IF_DEBUG( 1, fprintf( stderr, "Measurement started at %s UTC.\n", time_string ) );

    if ( is_sender )
    {
        char path[200];
        FILE * fp;

        strcpy( path, mmt_data_dir );
        strcat( path, "start_time" );
        if ( ( fp = fopen( path, "w" ) ) == NULL )
            return;
        fprintf( fp, "%s UTC\n%ld\n", time_string, gmt_time );
        // Output path_id
        fprintf( fp, "%d-%d\n", host_id, peer_host_id );
        fclose( fp );
    }
}


void Pong::OutputDataID( void )
{
    char path[200];
    FILE * fp;
    strcpy( path, mmt_data_dir );
    strcat( path, "data_id" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", path );
        exit(1);
    }
    time_t data_id = ( is_sender ? time_baseline_receiver.tv_sec : time_baseline.tv_sec );
    fprintf( fp, "%ld\n", data_id );
    // Output path_id
    int sID = ( is_sender ? host_id : peer_host_id );
    int dID = ( is_sender ? peer_host_id : host_id );
    fprintf( fp, "%d-%d\n", sID, dID );
    fclose( fp );
}


// Check whether the path is too short. If it is cancel the measurement.

bool Pong::CheckHopNum( void )
{
    bool error = false;
    if ( probe_num < min_probe_num )
    {
        IF_DEBUG( 1, fprintf( stderr, "Too few measurable points, probe_num = %d\n", probe_num ) );
        error = true;
    }
    else if ( hop_num < min_hop || hop_num > MAX_HOP )
    {
        const char * result = ( hop_num < min_hop ? "short" : "long" );
        IF_DEBUG( 1, fprintf( stderr, "Forward path is too %s, hop_num = %d\n", result, hop_num ) );
        error = true;
    }
    else if ( hop_num_2 > MAX_HOP )
    {
        IF_DEBUG( 1, fprintf( stderr, "Backward path is too long, hop_num = %d\n", hop_num_2 ) );
        error = true;
    }

    if ( error )
    {
        quit_reason = "invalid_hop_num";
        Cancel();
        return false;
    }
    return true;
}


void Pong::SendProbes( void )
{

    if ( is_sender )
    {
        IF_DEBUG( 3, fprintf( stderr, "Synchronze with receiver ... " ) );
    }
    else
    {
        IF_DEBUG( 3, fprintf( stderr, "Synchronze with sender ... " ) );
    }
    Synchronize();
    
    GetCurrentTime( time_baseline );
    last_tuning_paring_time = time_baseline;
    last_tuning_paring_time.tv_sec += TH_QoM_AGE_MIN;
    last_link_cc_update_time = time_baseline;
    heart_beat_update_time = time_baseline.tv_sec;
    last_update_time = time_baseline.tv_sec;

    skew_monitor_f.ResetAge( time_baseline );
    skew_monitor_b.ResetAge( time_baseline );

    IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
    if ( ! is_sender )
    {
        SendTimeBaseline();
        OutputDataID();
    }

    GetCurrentTime( current_time );
    
    if ( is_sender )
        CheckHopNum();  // Check whether the path is too short. If it is cancel the measurement.

    if ( ! done )
    {
        HeartBeat( "running" );
        OutputStartTime();
    }

    round_id = 1;   // Must start from 1

    phase = PP_PROBE;
    phase_done = false;

    while ( ! done && ! phase_done )
    {
        for ( int i = 0; i < probe_num; i ++ )
        {
            double probe_time = GetProbeTime( probe_info[i].time_offset );
            GetCurrentTime( current_time );
            double sleep_usec = probe_time - TimeInUsec( current_time ) - TimeInUsec( syscallTime );
            // Sleep waiting
            if ( sleep_usec > 0 )
            {
                usleep( (useconds_t)sleep_usec );
            }
            /*
            // Busy waiting
            while ( TimeInUsec( current_time ) + TimeInUsec( syscallTime ) < probe_time )
            {
                GetCurrentTime( current_time );
            }
            */
            SendUdpProbe( round_id, probe_info[i].id, probe_time );
            if ( probe_info[i].ttl >= 1 )
                SendRouterProbe( round_id, probe_info[i].id, probe_info[i].ttl );
        }
        if ( ! is_sender && pairing_updated )
        {
            assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
            assert( ! pairing_updates.empty() );
            IF_DEBUG( 4, fprintf( stderr, "%s Update pairings ... ", GetMeasurementTime() ) );
            UpdatePairings();
            pairing_updates.clear();
            IF_DEBUG( 4, fprintf( stderr, "done.\n" ) );
            assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
            assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
            SendCtrCode( CC_PUA );
            PUA_round_id = round_id + 1;
            SendIntValue( PUA_round_id );
            assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
            pairing_updated = false;
        }
        if ( ! is_sender && technique_updated )
        {
            assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
            assert( ! technique_updates.empty() );
            IF_DEBUG( 4, fprintf( stderr, "%s Update techniques ... ", GetMeasurementTime() ) );
            UpdateTechniques();
            technique_updates.clear();
            IF_DEBUG( 4, fprintf( stderr, "done.\n" ) );
            assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
            assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
            SendCtrCode( CC_TUA );
            TUA_round_id = round_id + 1;
            SendIntValue( TUA_round_id );
            assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
            technique_updated = false;
        }

        GetCurrentTime( current_time );
        if ( current_time.tv_sec - heart_beat_update_time >= 30 )
        {
            HeartBeat( "running" );
            heart_beat_update_time = current_time.tv_sec;
        }

        // Check pending report messages every 1 second
        CheckPendingMessages( current_time.tv_sec );

        // ReportTest();

        if ( is_sender )
        {
            UpdateDominantBottleneck();

            assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
            if ( ! TptPending() )
            {
                if ( ! technique_updates.empty() )
                {
                    SendTechniqueUpdates();
                    technique_update_pending = true;
                    IF_DEBUG( 4, fprintf( stderr, "%s Send automatic technique updates (%d updates).\n", GetMeasurementTime(), technique_updates.size() ) );
                }
            }
            assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );

            if ( auto_tpt )
            {
                GetCurrentTime( current_time );
                if ( current_time.tv_sec - last_tuning_paring_time.tv_sec >= TP_PERIOD )
                {
                    if ( TuningPairings() )
                        last_tuning_paring_time = current_time;
                }
            }

            // Aging clock_skew_indicator every 1 second
            if ( clock_skew_indicator > 0 )
                clock_skew_indicator --;
            
            GetCurrentTime( current_time );
            if ( current_time.tv_sec - last_link_cc_update_time.tv_sec >= 10 )
            {
                // Update every 10sec
                last_link_cc_update_time = current_time; 
                link_cc_update_count ++;
                OutputCurrentPairing();
                UpdateLinkCC_10sec();
                if ( link_cc_update_count % 3 == 0 )
                {
                    // Update every 30sec
                    if ( sample_count == 0 ) sample_count = 1;
                    UpdateLinkCCmax();
                    OutputLinkCC();
                    UpdateAverageLinkCI();
                    OutputLinkCI();
                    OutputQoMLog();
                    OutputErrorMetrics();
                    if ( pong_report && max_link_ci > TH_LINK_CI )
                        ReportLinkCI(); // NOTE: Must be put after OutputLinkCI(), OutputQoMLog(), etc.
                    sample_count = 0;
                    if ( link_cc_update_count % 6 == 0 )
                    {
                        // Update every 1min
                        UpdateLinkCC_1min();
                        OutputSummary();
                        OutputMTLog();
                        OutputQDBin();
                        CheckTDPA();
                        if ( link_cc_update_count % 60 == 0 )
                        {
                            // Update every 10min
                            UpdateLinkCC_10min();
                            if ( link_cc_update_count % 360 == 0 )
                            {
                                // Update every 1hour
                                UpdateLinkCC_1hour();
                                if ( link_cc_update_count % 1080 == 0 )
                                {
                                    // Update every 3hour
                                    UpdateLinkCC_3hour();
                                }
                            }
                        }
                    }
                }
            }
        }
        if ( current_time.tv_sec - last_update_time >= 60 )
        {
            // Common stuff for sender and receiver side, updated every 1min
            last_update_time = current_time.tv_sec;
            OutputRouteAlteration();
            OutputProbeLoss();
        }
        if ( time_offset_adjusted )
        {
            time_offset_adjusted = false;
            SortProbes();
        }
        round_id ++;
        if ( route_alteration_count > 0 )
        {
            if ( is_sender && ( altered_hop_count > 3 || route_alteration_count > 10 ) )
            {
                fprintf( stderr, "%s Significant route alteration encountered, altered_hop_count=%d, stop measurement.\n", GetMeasurementTime(), altered_hop_count );
                quit_reason = "route_alteration";
                if ( round_id < 300 )
                    // Path has not been stable for more than 5 minutes, cancel the measurement
                    Cancel();
                else
                    StartTraceRA();
            }
            route_alteration_count -= 0.1;
        }

        if ( SIGTERM_received )
        {
            quit_reason = "sigterm";
            Quit();
        }
    }

    if ( ! done && phase == PP_TRA )
        TraceRouterAlteration();
}


void Pong::TraceRouterAlteration( void )
{
    GetCurrentTime( current_time );
    TRA_start_time = current_time.tv_sec;

    HeartBeat( "trace_ra" );
    route_converge_count = 0;
    phase_done = false;

    while ( ! done && ! phase_done )
    {
        for ( int i = 0; i < probe_num; i ++ )
        {
            double probe_time = GetProbeTime( probe_info[i].time_offset );
            GetCurrentTime( current_time );
            double sleep_usec = probe_time - TimeInUsec( current_time ) - TimeInUsec( syscallTime );
            // Sleep waiting
            if ( sleep_usec > 0 )
            {
                usleep( (useconds_t)sleep_usec );
            }
            if ( is_sender && probe_info[i].ttl >= 1 )
                SendRouterProbe( round_id, probe_info[i].id, probe_info[i].ttl );
        }
        round_id ++;

        GetCurrentTime( current_time );
        if ( current_time.tv_sec - heart_beat_update_time >= 30 )
        {
            HeartBeat( "trace_ra" );
            heart_beat_update_time = current_time.tv_sec;
            OutputRouteAlteration();
            ResetRaCount();

            if ( is_sender )
            {
                TRA_stop_time = current_time.tv_sec;
                DumpRouteAlterationInfo( false );
            }
        }
        
        if ( is_sender )
        {
            if ( ++ route_converge_count >= TH_TRA_CONVERGE_TIME )
            {
                fprintf( stderr, "%s Route alteration stopped.\n", GetMeasurementTime() );
                TRA_stop_time = current_time.tv_sec - TH_TRA_CONVERGE_TIME;
                DumpRouteAlterationInfo( true );
                Quit();
            }
        }

        if ( SIGTERM_received )
        {
            quit_reason = "sigterm";
            Quit();
        }
    }
}


bool Pong::DumpRouteAlterationInfo( bool converged )
{
    int i;

    // List the last used IP as the first element for each hop
    for ( i = 0; i < hop_num; i ++ )
    {
        int num = hop_ips[i].size();
        int last_ip = hop_info[i].ip.s_addr;
        int j;
        for ( j = 0; j < num; j ++ )
        {
            if ( hop_ips[i][j] == last_ip )
                break;
        }
        if ( j < num && j != 0 )
        {
            hop_ips[i][j] = hop_ips[i][0];
            hop_ips[i][0] = last_ip;
        }
    }

    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "TRA_info" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;

    const char * result = ( converged ? "converged" : "not_converged" );
    fprintf( fp, "%.1lf\t%.1lf\t%s\n", (double)( this->TRA_start_time - time_baseline.tv_sec ) / 60, (double)( this->TRA_stop_time - time_baseline.tv_sec ) / 60, result );
    for ( i = 0; i < hop_num; i ++ )
    {
        fprintf( fp, "%d", i + 1 );
        int num = hop_ips[i].size();
        for ( int j = 0; j < num; j ++ )
        {
            struct in_addr ip;
            ip.s_addr = hop_ips[i][j];
            fprintf( fp, "\t%s", inet_ntoa( ip ) );
        }
        fprintf( fp, "\n" );
    }

    fclose( fp );
    return true;
}


// Restore round_id from the last 10 bits

int Pong::RestoreRoundId( int truncated_round_id )
{
    truncated_round_id = truncated_round_id & 0x3FF;
    int round_id_offset = round_id & 0x3FF;
    int round_id_base = round_id & ~0x3FF;

    if ( truncated_round_id < round_id_offset )
    {
        if ( round_id_offset - truncated_round_id < 0x200 )
            return round_id_base + truncated_round_id;
        else
            return round_id_base + 0x400 + truncated_round_id;
    }
    else
    {
        if ( truncated_round_id - round_id_offset < 0x200 )
            return round_id_base + truncated_round_id;
        else
            return round_id_base - 0x400 + truncated_round_id;
    }
}


// Synchronize sender and receiver

void Pong::Synchronize( void )
{
    if ( is_sender )
    {
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_SYNC );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
        assert( RecvCtrCode() == CC_SYNC );
    }
    else
    {
        assert( RecvCtrCode() == CC_SYNC );
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_SYNC );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
        usleep( (int)( e2e_rtt / 2 ) );
    }
}


void Pong::ExchangeHopnum( void )
{
    SendHopnum();
    RecvHopnum();
}


void Pong::SendHopnum( void )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_EHN );
    SendIntValue( hop_num_2 );

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


void Pong::RecvHopnum( void )
{
    assert( RecvCtrCode() == CC_EHN );
    hop_num = RecvIntValue();
}


// Read version of Pong from file

void Pong::ReadVersion( void )
{
    const char * file = "version";
    FILE *fp;
    if ( ( fp = fopen( file, "r" ) ) == NULL )
    {
        fprintf( stderr, "Fail to read file '%s'\n", file );
        exit(1);
    }
    if ( fscanf( fp, "%d", &version ) != 1 )
    {
        fprintf( stderr, "Corrupt file '%s'\n", file );
        exit(1);
    }
    fclose( fp );
    //IF_DEBUG( 3, fprintf( stderr, "Version = %d\n", version ) );
}


// Compare version between sender and receiver

void Pong::CheckVersion( void )
{
    int peer_version;
    if ( is_sender )
    {
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_VER );
        SendIntValue( version );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
        assert( RecvCtrCode() == CC_VER );
        peer_version = RecvIntValue();
    }
    else
    {
        assert( RecvCtrCode() == CC_VER );
        peer_version = RecvIntValue();
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_VER );
        SendIntValue( version );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    }
    if ( peer_version != version )
    {
        if ( version > peer_version )
        {
            fprintf( stderr, "Pong binary on remote machine is outdated\n" );
            sleep( 10 );    // To ensure that the old version side exit first
            HeartBeat( "peer_outdated" );
        }
        else
        {
            fprintf( stderr, "Pong binary on this machine is outdated\n" );
            HeartBeat( "outdated" );
        }
        exit(1);
    }
}


void Pong::ReadHostID( void )
{
    const char * file = "host_id";
    FILE * fp;
    if ( ( fp = fopen( file, "r" ) ) == NULL )
    {
        fprintf( stderr, "Fail to read file '%s'\n", file );
        exit(1);
    }
    if ( fscanf( fp, "%d", &host_id ) != 1 )
    {
        fprintf( stderr, "Corrupt file '%s'\n", file );
        exit(1);
    }
    fclose( fp );
    //IF_DEBUG( 3, fprintf( stderr, "host_id = %d\n", host_id ) );
    my_host_id = host_id;
}


void Pong::ExchangeHostID( void )
{
    if ( is_sender )
    {
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_HID );
        SendIntValue( host_id );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
        assert( RecvCtrCode() == CC_HID );
        peer_host_id = RecvIntValue();
    }
    else
    {
        assert( RecvCtrCode() == CC_HID );
        peer_host_id = RecvIntValue();
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_HID );
        SendIntValue( host_id );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    }
}


void Pong::InitializeProbeInfo( void )
{
    int ctr_code;
    if ( is_sender )
    {
        IF_DEBUG( 3, fprintf( stderr, "Loading/initialize probe configuration ... " ) );
        if ( IgnoreConfiguration || ! LoadProbeConfiguration() )
        {
            if ( IgnoreConfiguration )
            {
                IF_DEBUG( 3, fprintf( stderr, "\nIgnore probe configuration history" ) );
            }
            else
            {
                IF_DEBUG( 3, fprintf( stderr, "\nNo probe configuration history found" ) );
            }
            IF_DEBUG( 3, fprintf( stderr, ", initialize to default configuration.\n" ) );
            ctr_code = CC_INI;
        }
        else
        {
            IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
            ctr_code = CC_HIS;
        }
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( ctr_code );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    }
    else
    {
        ctr_code = RecvCtrCode();
        assert( ctr_code == CC_INI || ctr_code == CC_HIS );
    }

    if ( ctr_code == CC_INI )
    {
        IF_DEBUG( 3, fprintf( stderr, "Measuring hop number ... " ) );
        Synchronize();
        MeasureHopnum();
        ExchangeHopnum();
        IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
    
        IF_DEBUG( 3, fprintf( stderr, "Exploring information of hops ... " ) );
        ExploreHopInfo();
        if ( is_sender )
        {
            assert( RecvCtrCode() == CC_HOI );
            IF_DEBUG( 3, fprintf( stderr, "recv" ) );
            RecvHopInformation( true );
        }
        else
        {
            IF_DEBUG( 3, fprintf( stderr, "send" ) );
            SendHopInformation( false );
        }
        IF_DEBUG( 3, fprintf( stderr, " ... done.\n" ) );

        if ( is_sender )
        {
            IF_DEBUG( 3, fprintf( stderr, "Initializing probe configuration ... " ) );
            InitializeProbeConfiguration();
            IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
        }
        else
        {
            IF_DEBUG( 3, fprintf( stderr, "Waiting for and loading configuration from sender -->\n" ) );
        }
    }
    else
    {
        if ( is_sender )
        {
            IF_DEBUG( 3, fprintf( stderr, "Sending hop information ... " ) );
            SendHopInformation( true );
            IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
        }
        else
        {
            IF_DEBUG( 3, fprintf( stderr, "Waiting for and loading configuration from sender -->\n" ) );
            assert( RecvCtrCode() == CC_HOI );
            IF_DEBUG( 3, fprintf( stderr, "--> Receiving hop information ... " ) );
            RecvHopInformation( false );
            IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
        }
    }
    
    if ( is_sender )
    {
        IF_DEBUG( 3, fprintf( stderr, "Sending probe configuration ... " ) );
        SendProbeConfiguration();
        IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
        InitializePairingHistory();
        InitializeRecords();
        InitializeLinkInfo();
        InitializeStatistics();
        OutputHopInformation();
        OutputGlobalInfo();
        OutputProbeConfiguration();
    }
    else
    {
        assert( RecvCtrCode() == CC_CFG );
        IF_DEBUG( 3, fprintf( stderr, "--> Receiving probe configuration ... " ) );
        RecvProbeConfiguration();
        IF_DEBUG( 3, fprintf( stderr, "done.\n" ) );
    }

    InitializeRouteAlteration();

    // Remove data files applicable for both sender and receiver
    RemoveDataFile( "probe_loss" );
    RemoveDataFile( "data_id" );

    SortProbes();
}


void Pong::SortProbes( void )
{
    for ( int i = 1; i < probe_num; i ++ )
    {
        for ( int j = i - 1; j >= 0; j -- )
        {
            if ( probe_info[j+1].time_offset < probe_info[j].time_offset )
            {
                ProbeInfo temp = probe_info[j];
                probe_info[j] = probe_info[j+1];
                probe_info[j+1] = temp;
            }
            else break;
        }
    }
}


void Pong::InitializeRecords( void )
{
    last_pr = new ProbeRecord[probe_num];
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = FindProbeInfo( i );
        assert( pi != NULL );
        last_pr[i].Initialize( pi );
    }
    
    link_cc = new double[hop_num + 1];
    ResetLinkCC( link_cc );
    link_cc_max = new double[hop_num + 1];
    ResetLinkCC( link_cc_max );
    LCD_30sec = new double[hop_num + 1];
    ResetLinkCC( LCD_30sec );
    LCD_1min = new double[hop_num + 1];
    ResetLinkCC( LCD_1min );
    LCD_3min = new double[hop_num + 1];
    ResetLinkCC( LCD_3min );
    LCD_10min = new double[hop_num + 1];
    ResetLinkCC( LCD_10min );
    LCD_30min = new double[hop_num + 1];
    ResetLinkCC( LCD_30min );
    LCD_1hour = new double[hop_num + 1];
    ResetLinkCC( LCD_1hour );
    LCD_3hour = new double[hop_num + 1];
    ResetLinkCC( LCD_3hour );
    LCD_6hour = new double[hop_num + 1];
    ResetLinkCC( LCD_6hour );
    LCD_12hour = new double[hop_num + 1];
    ResetLinkCC( LCD_12hour );
    LCD_24hour = new double[hop_num + 1];
    ResetLinkCC( LCD_24hour );

    for ( int i = 0; i < 6; i ++ )
    {
        link_cc_10sec[i] = new double[hop_num + 1];
        ResetLinkCC( link_cc_10sec[i] );
    }
    for ( int i = 0; i < 10; i ++ )
    {
        link_cc_1min[i] = new double[hop_num + 1];
        ResetLinkCC( link_cc_1min[i] );
    }
    for ( int i = 0; i < 6; i ++ )
    {
        link_cc_10min[i] = new double[hop_num + 1];
        ResetLinkCC( link_cc_10min[i] );
    }
    for ( int i = 0; i < 6; i ++ )
    {
        link_cc_1hour[i] = new double[hop_num + 1];
        ResetLinkCC( link_cc_1hour[i] );
    }
    for ( int i = 0; i < 8; i ++ )
    {
        link_cc_3hour[i] = new double[hop_num + 1];
        ResetLinkCC( link_cc_3hour[i] );
    }

    sum_link_ci = new double[hop_num + 1];
    ResetLinkCC( sum_link_ci );
    average_link_ci = new double[hop_num + 1];
    average_link_ci_max = new double[hop_num + 1];
    ResetLinkCC( average_link_ci_max );
    sum_e2e_cc = average_e2e_cc_max = 0;
    sample_count = 0;
    sum_round_id = 0;

    RemoveLinkCCdata( "10sec" );
    RemoveLinkCCdata( "30sec" );
    RemoveLinkCCdata( "1min" );
    RemoveLinkCCdata( "3min" );
    RemoveLinkCCdata( "10min" );
    RemoveLinkCCdata( "30min" );
    RemoveLinkCCdata( "1hour" );
    RemoveLinkCCdata( "3hour" );
    RemoveLinkCCdata( "6hour" );
    RemoveLinkCCdata( "12hour" );
    RemoveLinkCCdata( "24hour" );
    RemoveDataFile( "link_cc" );
    RemoveDataFile( "link_ci" );
    RemoveDataFile( "current_time" );
    RemoveDataFile( "QoM_log" );
    RemoveDataFile( "error_metrics" );
    RemoveDataFile( "MT_log" );
    RemoveDataFile( "pong_summary" );
    RemoveDataFile( "tdp_available" );
    RemoveDataFile( "probe_config" );
}


// Initialize Router Alteration Counter
void Pong::InitializeRouteAlteration( void )
{
    route_altered = new bool[hop_num];
    ResetRaCount();

    RemoveDataFile( "route_alteration" );
    if ( is_sender )
        RemoveDataFile( "TRA_info" );
}


void Pong::InitializeLinkInfo( void )
{
    redundant_link = new bool[hop_num + 1];
    redundant_link[0] = false;
    for ( int i = 0; i < hop_num; i ++ )
    {
        assert( hop_info[i].ttl == i + 1 );
        redundant_link[i + 1] = ( hop_info[i].ip.s_addr == 0 ? true : false );
    }
}


// Initialize statistic parameters
void Pong::InitializeStatistics( void )
{
    link_cc_sum_all = 0;
    link_ci_sum_all = 0;
    sample_total = 0;

    QoM_log = new QoM_Log[probe_num];
    ResetQoMLog();

    tdp_available = new bool[probe_num];
    for ( int i = 0; i < probe_num; i ++ )
        tdp_available[i] = true;
    tdpa_updated = true;    // Force it to output for the first CheckTDPA() call
}


void Pong::ResetQoMLog( void )
{
    for ( int i = 0; i < probe_num; i ++ )
        QoM_log[i].isset = false;
}


void Pong::UpdateQoMLog( ProbeInfo * pi, ProbeHistory * ph )
{
    QoM_Log *ql =  QoM_log + pi->id;
    ql->isset = true;
    ql->ttl_d = ph->ttl_d;
    ql->technique = pi->technique;
    ql->QoM_JP = ph->QoM_JP.value_l;
    ql->QoM_4P = ph->QoM_4P.value_l;
    ql->QoM_fsd = ph->QoM_fsd.value_l;
    ql->QoM_fsb = ph->QoM_fsb.value_l;
    CopyThresholdInfo( ql->th_s, ph->th_s );
    CopyThresholdInfo( ql->th_d, ph->th_d );
    CopyThresholdInfo( ql->th_f, th_f );
    CopyThresholdInfo( ql->th_b, th_b );
}


void Pong::CopyThresholdInfo( ThresholdInfo_Lite& dest, const ProbeThreshold& source )
{
    dest.bin_index = source.bin_index;
    dest.pattern = source.pattern;
    dest.percent_below = source.bin_count.percent_below;
    dest.percent_above = source.bin_count.percent_above;
}


void Pong::FinishQoMLog( void )
{
    // fprintf( stdout, "pih_mutex FinishQoMLog\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( ! QoM_log[i].isset )
        {
            ProbeInfo * pi = FindProbeInfo( i );
            assert( pi != NULL );
            ProbeHistory * ph = FindProbeHistory( i );
            assert( ph != NULL );
            UpdateQoMLog( pi, ph );
        }
    }

    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
}


bool Pong::OutputQoMLog( void )
{
    FinishQoMLog();

    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "QoM_log" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;
    fprintf( fp, "%.1lf\t", (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        QoM_Log *ql =  QoM_log + i;
        fprintf( fp, "%d\t%d\t%.2lf\t%.2lf\t%.2lf\t%.2lf\t", ql->ttl_d, ql->technique, ql->QoM_JP, ql->QoM_4P, ql->QoM_fsd, ql->QoM_fsb );
    }
    fprintf( fp, "\n" );
    fclose( fp );

    ResetQoMLog();
    return true;
}


bool Pong::OutputErrorMetrics( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "error_metrics" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;
    fprintf( fp, "%.1lf\t", (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60 );
    fprintf( fp, "%d\t%d\t%.1lf\t", clock_skew_extent, altered_hop_count, route_alteration_count );
    fprintf( fp, "\n" );
    fclose( fp );

    clock_skew_extent = 0;
    return true;
}


void Pong::CheckTDPA( void )
{
    if ( tdpa_updated )
    {
        tdpa_updated = false;

        // Output TDPA
        char path[200];
        FILE * fp;

        strcpy( path, mmt_data_dir );
        strcat( path, "tdp_available" );
        if ( ( fp = fopen( path, "w" ) ) == NULL )
            return;
        for ( int i = 0; i < probe_num; i ++ )
            fprintf( fp, "%d\t", tdp_available[i] );
        fprintf( fp, "\n" );
        fclose( fp );
    }
}


bool Pong::OutputSummary( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "pong_summary" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    fprintf( fp, "link_cc_sum\tlink_ci_sum\tsample_total\n" );
    fprintf( fp, "%.1lf\t%.1lf\t%d\n", link_cc_sum_all, link_ci_sum_all, sample_total );
    fclose( fp );

    return true;
}


bool Pong::OutputProbeLoss( void )
{
    char path[200];
    FILE * fp;

    if ( is_sender )
        strcpy( path, mmt_data_dir );
    else
        strcpy( path, "" );
    strcat( path, "probe_loss" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;
    fprintf( fp, "%.1lf\t", (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60 );
    
    assert( pthread_mutex_lock( &hrb_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        HalfRecordBuffer * hrb = hrbuffer + i;
        fprintf( fp, "%d\t%d\t%d\t%d\t%d\t", hrb->probe_id, hrb->lost_bf, hrb->expected_total_bf, hrb->lost_sd, hrb->expected_total_sd );
        hrb->ResetLossCount();
    }
    assert( pthread_mutex_unlock( &hrb_mutex ) == 0 );

    fprintf( fp, "\n" );
    fclose( fp );

    return true;
}


bool Pong::OutputRouteAlteration( void )
{
    char path[200];
    FILE * fp;

    if ( is_sender )
        strcpy( path, mmt_data_dir );
    else
        strcpy( path, "" );
    strcat( path, "route_alteration" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;
    if ( is_sender )
        fprintf( fp, "%.1lf\t", (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60 );
    else
        fprintf( fp, "%ld\t", current_time.tv_sec );
    fprintf( fp, "%d\t%.1lf\t", altered_hop_count, route_alteration_count );
    for ( int i = 0; i < hop_num; i ++ )
        fprintf( fp, "%d\t", route_altered[i] );
    fprintf( fp, "\n" );
    fclose( fp );

    return true;
}


void Pong::RemoveLinkCCdata( const char * filename )
{
    char path[200];
    strcpy( path, mmt_data_dir );
    strcat( path, filename );
    remove( path );
    strcat( path, ".max" );
    remove( path );
}


const char * GetMeasurementTime( void )
{
    static char mt[20];
    int measurement_time = ( pong->time_baseline.tv_sec == 0 ? 0: pong->current_time.tv_sec - pong->time_baseline.tv_sec );
    sprintf( mt, "%d:%02d:%02d", measurement_time / 3600, measurement_time % 3600 / 60, measurement_time % 60 );
    return mt;
}


bool Pong::OutputCurrentPairing( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "pairing" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;

    OutputPairing( fp, "Current Hop Pairings" );
    fclose( fp );

    return true;
}


void Pong::UpdateLinkCCmax( void )
{
    for ( int i = 0; i <= hop_num; i ++ )
    {
        if ( LCD_1min[i] > link_cc_max[i] )
            link_cc_max[i] = LCD_1min[i];
    }
    double average_e2e_cc = sum_e2e_cc / sample_count;
    if ( average_e2e_cc > average_e2e_cc_max )
        average_e2e_cc_max = average_e2e_cc;
}


void Pong::UpdateLinkCC_10sec( void )
{
    double * temp = link_cc_10sec[0];
    for ( int i = 1; i < 6; i ++ )
        link_cc_10sec[i-1] = link_cc_10sec[i];
    link_cc_10sec[5] = link_cc;
    link_cc = temp;
    ResetLinkCC( link_cc );
    OutputLCD( "10sec", link_cc_10sec[5] );
    
    CopyLinkCC( LCD_30sec, link_cc_10sec[3] );
    for ( int i = 4; i < 6; i ++ )
        AddLinkCC( LCD_30sec, link_cc_10sec[i] );
    OutputLCD( "30sec", LCD_30sec );
    CopyLinkCC( LCD_1min, link_cc_10sec[0] );
    for ( int i = 1; i < 6; i ++ )
        AddLinkCC( LCD_1min, link_cc_10sec[i] );
    OutputLCD( "1min", LCD_1min );
}


void Pong::UpdateLinkCC_1min( void )
{
    double * temp = link_cc_1min[0];
    for ( int i = 1; i < 10; i ++ )
        link_cc_1min[i-1] = link_cc_1min[i];
    link_cc_1min[9] = temp;
    CopyLinkCC( link_cc_1min[9], LCD_1min );

    CopyLinkCC( LCD_3min, link_cc_1min[7] );
    for ( int i = 8; i < 10; i ++ )
        AddLinkCC( LCD_3min, link_cc_1min[i] );
    OutputLCD( "3min", LCD_3min );
    CopyLinkCC( LCD_10min, link_cc_1min[0] );
    for ( int i = 1; i < 10; i ++ )
        AddLinkCC( LCD_10min, link_cc_1min[i] );
    OutputLCD( "10min", LCD_10min );
}


void Pong::UpdateLinkCC_10min( void )
{
    double * temp = link_cc_10min[0];
    for ( int i = 1; i < 6; i ++ )
        link_cc_10min[i-1] = link_cc_10min[i];
    link_cc_10min[5] = temp;
    CopyLinkCC( link_cc_10min[5], LCD_10min );

    CopyLinkCC( LCD_30min, link_cc_10min[3] );
    for ( int i = 4; i < 6; i ++ )
        AddLinkCC( LCD_30min, link_cc_10min[i] );
    OutputLCD( "30min", LCD_30min );
    CopyLinkCC( LCD_1hour, link_cc_10min[0] );
    for ( int i = 1; i < 6; i ++ )
        AddLinkCC( LCD_1hour, link_cc_10min[i] );
    OutputLCD( "1hour", LCD_1hour );
}


void Pong::UpdateLinkCC_1hour( void )
{
    double * temp = link_cc_1hour[0];
    for ( int i = 1; i < 6; i ++ )
        link_cc_1hour[i-1] = link_cc_1hour[i];
    link_cc_1hour[5] = temp;
    CopyLinkCC( link_cc_1hour[5], LCD_1hour );

    CopyLinkCC( LCD_3hour, link_cc_1hour[3] );
    for ( int i = 4; i < 6; i ++ )
        AddLinkCC( LCD_3hour, link_cc_1hour[i] );
    OutputLCD( "3hour", LCD_3hour );
    CopyLinkCC( LCD_6hour, link_cc_1hour[0] );
    for ( int i = 1; i < 6; i ++ )
        AddLinkCC( LCD_6hour, link_cc_1hour[i] );
    OutputLCD( "6hour", LCD_6hour );
}


void Pong::UpdateLinkCC_3hour( void )
{
    double * temp = link_cc_3hour[0];
    for ( int i = 1; i < 8; i ++ )
        link_cc_3hour[i-1] = link_cc_3hour[i];
    link_cc_3hour[7] = temp;
    CopyLinkCC( link_cc_3hour[7], LCD_3hour );

    CopyLinkCC( LCD_12hour, link_cc_3hour[4] );
    for ( int i = 5; i < 8; i ++ )
        AddLinkCC( LCD_12hour, link_cc_3hour[i] );
    OutputLCD( "12hour", LCD_12hour );
    CopyLinkCC( LCD_24hour, link_cc_3hour[0] );
    for ( int i = 1; i < 8; i ++ )
        AddLinkCC( LCD_24hour, link_cc_3hour[i] );
    OutputLCD( "24hour", LCD_24hour );
}


void Pong::ResetLinkCC( double * link_cc )
{
    for ( int i = 0; i <= hop_num; i ++ )
        link_cc[i] = 0;
}


void Pong::CopyLinkCC( double * to, const double * from )
{
    for ( int i = 0; i <= hop_num; i ++ )
        to[i] = from[i];
}


void Pong::AddLinkCC( double * to, const double * from )
{
    for ( int i = 0; i <= hop_num; i ++ )
        to[i] += from[i];
}


bool Pong::OutputLCD( const char * filename, double * link_cc )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, filename );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    double max_cc = 0;
    for ( int i = 0; i <= hop_num; i ++ )
    {
        fprintf( fp, "%.1lf\t%.1lf\n", i + 0.5, link_cc[i] );
        if ( link_cc[i] > max_cc )
            max_cc = link_cc[i];
    }
    fclose( fp );

    strcat( path, ".max" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    fprintf( fp, "%d\t%.1lf\t%ld\n", hop_num + 1, max_cc, current_time.tv_sec - time_baseline.tv_sec );
    fclose( fp );

    return true;
}


void Pong::RemoveDataFile( const char * filename )
{
    char path[200];
    if ( is_sender )
        strcpy( path, mmt_data_dir );
    else
        strcpy( path, "" );
    strcat( path, filename );
    remove( path );
}


bool Pong::OutputLinkCC( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "link_cc" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;
    fprintf( fp, "%.1lf\t", (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60 );
    for ( int i = 0; i <= hop_num; i ++ )
        fprintf( fp, "%.1lf\t", LCD_1min[i] );
    fprintf( fp, "%.3lf\t%d\t%.2lf", sum_e2e_cc / sample_count, sample_count, (double)sum_round_id / sample_count / 60 );
    fprintf( fp, "\n" );
    fclose( fp );

    sum_e2e_cc = 0;
    sum_round_id = 0;

    strcat( path, ".max" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    for ( int i = 0; i <= hop_num; i ++ )
        fprintf( fp, "%.1lf\t", link_cc_max[i] );
    fprintf( fp, "%.3lf", average_e2e_cc_max );
    fprintf( fp, "\n" );
    fclose( fp );

    strcpy( path, mmt_data_dir );
    strcat( path, "current_time" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    fprintf( fp, "%ld", current_time.tv_sec - time_baseline.tv_sec );
    fclose( fp );

    return true;
}


bool Pong::OutputLinkCI( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "link_ci" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;
    fprintf( fp, "%.1lf\t", (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60 );
    for ( int i = 0; i <= hop_num; i ++ )
        fprintf( fp, "%.3lf\t", average_link_ci[i] );
    fprintf( fp, "\n" );
    fclose( fp );

    ResetLinkCC( sum_link_ci );

    strcat( path, ".max" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    for ( int i = 0; i <= hop_num; i ++ )
        fprintf( fp, "%.3lf\t", average_link_ci_max[i] );
    fprintf( fp, "\n" );
    fclose( fp );

    return true;
}


void Pong::ReportLinkCI( void )
{
    for ( int i = 0; i <= hop_num; i ++ )
    {
        if ( average_link_ci[i] > TH_LINK_CI )
        {
            char ttl = hop_info[i].ttl;
            in_addr ip1 = ( i > 0 ? hop_info[i - 1].ip : originAddress );
            in_addr ip2 = ( i < hop_num ? hop_info[i].ip : targetAddress );
            if ( ip1.s_addr != 0 && ip2.s_addr != 0 )
            {
                LinkNodeInfo lni1, lni2;
                float lms = 0;
                lms += GetLinkNodeInfo( ttl - 1, lni1 );
                lms += GetLinkNodeInfo( ttl, lni2 );
                SendLinkCI( ip1, ip2, average_link_ci[i], probe_num, ttl, hop_num, sample_count, lms, lni1, lni2 );
            }
        }
    }
}


float Pong::GetLinkNodeInfo( int ttl, LinkNodeInfo& lni )
{
    float lms_component = 0;
    bzero( &lni, sizeof( LinkNodeInfo ) );

    ProbeInfo * pi = FindProbeInfoByTTL( ttl );
    if ( pi )
    {
        QoM_Log *ql =  QoM_log + pi->id;
        lni.isset = true;
        lni.technique = ql->technique;
        lni.th_s = ql->th_s;
        lni.th_d = ql->th_d;
        lni.th_f = ql->th_f;
        lni.th_b = ql->th_b;
        switch ( lni.technique )
        {
        case TECH_4P:
            lni.QoM = ql->QoM_4P;
            lms_component = 3 * lni.QoM;
            break;

        case TECH_FSD:
            lni.QoM = ql->QoM_fsd;
            lms_component = 3 * lni.QoM;
            break;

        case TECH_FSB:
            lni.QoM = ql->QoM_fsb;
            lms_component = 2 * lni.QoM;
            break;
        }
    }

    return lms_component;
}


void Pong::SendLinkCI( in_addr ip1, in_addr ip2, float intensity, char probe_num, char hop_num, char total_hop_num, short sample_count, float lms, const LinkNodeInfo& lni1, const LinkNodeInfo& lni2 )
{
    char * message = CreateMessage( MSG_LINKC, GetNextMessageID(), sizeof( Payload_MSG_LINKC ) );
    if ( message )
    {
        Payload_MSG_LINKC * payload = (Payload_MSG_LINKC *)( message + sizeof( MsgHeader ) );
        payload->ip1 = ip1;
        payload->ip2 = ip2;
        payload->intensity = intensity;
        payload->probe_num = probe_num;
        payload->hop_num = hop_num;
        payload->total_hop_num = total_hop_num;
        payload->sample_count = sample_count;
        payload->lms = lms;
        payload->lni1 = lni1;
        payload->lni2 = lni2;
        SendMessage( message );
    }
}


void Pong::UpdateAverageLinkCI( void )
{
    max_link_ci = 0;
    for ( int i = 0; i <= hop_num; i ++ )
    {
        average_link_ci[i] = sum_link_ci[i] / sample_count;
        if ( average_link_ci[i] > average_link_ci_max[i] )
            average_link_ci_max[i] = average_link_ci[i];
        if ( average_link_ci[i] > max_link_ci )
            max_link_ci = average_link_ci[i];
    }
}


void OpenClockSkewFile( void )
{
    char path[200];

    strcpy( path, mmt_data_dir );
    strcat( path, "clock_skew" );
    if ( ( clock_skew_fp = fopen( path, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", path );
        exit(1);
    }
}


void Pong::HeartBeat( const char * status, const char * comment )
{
    const char * file = "heartbeat";
    FILE *fp;
    if ( ( fp = fopen( file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", file );
        exit(1);
    }
    const char * pong_type = ( is_sender ? "sender" : "receiver" );
    if ( ! comment ) comment = "n/a";
    fprintf( fp, "%s\t%ld\t%s\t%s\n", status, current_time.tv_sec, pong_type, comment );
    fclose( fp );
}


void CheckProcess( void )
{
    const char * file = "pong_pid";
    FILE * fp;
    if ( ( fp = fopen( file, "r" ) ) != NULL )
    {
        pid_t pid;
        fscanf( fp, "%d", &pid );
        fclose( fp );
        if ( kill( pid, 0 ) == 0 )
        {
            fprintf( stderr, "Pong is already running\n" );
            exit(1);
        }
    }
}


void PrintPID( void )
{
    const char * file = "pong_pid";
    FILE *fp;
    if ( ( fp = fopen( file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Fail to write file '%s'\n", file );
        exit(1);
    }

    pid_t pid = getpid();
    fprintf( fp, "%d", pid );
    fclose( fp );
}


// Read the next parameter in the configuration file

char ReadParameter( FILE * fp, const char * parameter_name, const char * format_string, void * parameter_ptr )
{
    char buffer[MAX_CONFIG_LINE_WIDTH + 1];
    char match = 0;
    while ( ! match && fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
    {
        if ( buffer[0] != '#' ) // Not a comment line
        {
            int str_len = strlen( parameter_name );
            if ( strncmp( buffer, parameter_name, str_len ) == 0 )
            {
                match = 1;
                sscanf( buffer + str_len + 1, format_string, parameter_ptr );
            }
        }
    }
    if ( ! match )
        return -1;
    return 0;
}


// Load probe configuration from file

bool Pong::LoadProbeConfiguration( void )
{
    FILE * fp;
    if( ( fp = fopen( probe_config_file, "r" ) ) == NULL )
    {
        //fprintf( stderr, "Cannot open configuration file '%s'\n", probe_config_file );
        //exit(1);
        return false;
    }

    if ( ReadParameter( fp, "probe_period", "%lf", &probe_period ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'probe_period')\n" );
        exit(1);
    }
    if ( ReadParameter( fp, "probe_num", "%d", &probe_num ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'probe_num')\n" );
        exit(1);
    }

    probe_info = new ProbeInfo[probe_num];
    hrbuffer = new HalfRecordBuffer[probe_num];
    frbuffer = new FullRecordBuffer[probe_num];
    phistory = new ProbeHistory[probe_num];
    tdp_info = new TargetedDprobeInfo[probe_num];

    char buffer[MAX_CONFIG_LINE_WIDTH + 1];
    fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp );

    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( ! fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
        {
            fprintf( stderr, "Invalid configuration\n" );
            exit(1);
        }
        int ttl_d;
        double time_offset_d, min_s, min_d;
        sscanf( buffer, "%d %d %d %lf %lf %d %lf %lf", &(probe_info[i].id), &(probe_info[i].ttl), &ttl_d, &(probe_info[i].time_offset), &time_offset_d, &(probe_info[i].technique), &min_s, &min_d );
        if ( probe_info[i].technique < TECH_NA || probe_info[i].technique > TECH_4P )
        {
            fprintf( stderr, "Invalid technique code %d\n", probe_info[i].technique );
            exit(1);
        }
        hrbuffer[i].Initialize( probe_info[i].id );
        frbuffer[i].Initialize( probe_info[i].id );
        phistory[i].Initialize( probe_info[i].id );
        phistory[i].ttl_d = ttl_d;
        phistory[i].time_offset_d = time_offset_d;
        phistory[i].min_s.Update( min_s * 1000 );
        phistory[i].min_d.Update( min_d * 1000 );
        tdp_info[i].probe_id = probe_info[i].id;
    }

    HopInfo *hi;
    char ip[20], host_name[300];

    if ( ReadParameter( fp, "hop_num", "%d", &hop_num ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'hop_num')\n" );
        exit(1);
    }
    
    hop_info = new HopInfo[hop_num];

    fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp );
    
    for ( int i=0; i < hop_num; i ++ )
    {
        if ( ! fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
        {
            fprintf( stderr, "Invalid configuration\n" );
            exit(1);
        }
        hi = hop_info + i;
        sscanf( buffer, "%d %s %s", &(hi->ttl), ip, host_name );
        inet_aton( ip, &(hi->ip) );
        hi->host_name = strdup( host_name );
    }

    if ( ReadParameter( fp, "hop_num_2", "%d", &hop_num_2 ) != 0 )
    {
        fprintf( stderr, "Invalid configuration (read 'hop_num_2')\n" );
        exit(1);
    }

    hop_info_2 = new HopInfo[hop_num_2];

    fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp );

    for ( int i=0; i < hop_num_2; i ++ )
    {
        if ( ! fgets( buffer, MAX_CONFIG_LINE_WIDTH, fp ) )
        {
            fprintf( stderr, "Invalid configuration\n" );
            exit(1);
        }
        hi = hop_info_2 + i;
        sscanf( buffer, "%d %s %s", &(hi->ttl), ip, host_name );
        inet_aton( ip, &(hi->ip) );
        hi->host_name = strdup( host_name );
    }

    fclose( fp );
    return true;
}


// Save probe configuration to file

void Pong::SaveProbeConfiguration( void )
{
    FILE * fp;
    if( ( fp = fopen( probe_config_file, "w" ) ) == NULL )
    {
        fprintf( stderr, "Cannot open configuration file '%s' for writing\n", probe_config_file );
        return;
    }

    WriteProbeConfiguration( fp );

    fclose( fp );
    
    IF_DEBUG( 1, fprintf( stdout, "Probe configuration saved.\n" ) );
    fflush( stdout );
}


void Pong::WriteProbeConfiguration( FILE * fp )
{
    fprintf( fp, "# Probe configuration file for destination %s(%s)\n", targetName, inet_ntoa( targetAddress ) );
    fprintf( fp, "\n# Probe pairing, scheduling and techniques\n" );
    fprintf( fp, "probe_period=%g\n", probe_period );
    fprintf( fp, "probe_num=%d\n", probe_num );
    fprintf( fp, "id\thop_s\thop_d\t send_s\t send_d\ttech\t  min_s\t  min_d\n" );
    
    // fprintf( stdout, "pih_mutex 15\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );

        fprintf( fp, "%2d\t%5d\t%5d\t%7.4lf\t%7.4lf\t%4d\t%7.2lf\t%7.2lf\n",
            probe_id, pi->ttl, ph->ttl_d, pi->time_offset, ph->time_offset_d, pi->technique,
            ph->min_s.value / 1000, ph->min_d.value / 1000 );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    
    HopInfo * hi;
    fprintf( fp, "\n# Hop information(forward path)\n" );
    fprintf( fp, "hop_num=%d\n", hop_num );
    fprintf( fp, "hop_s\t%15s\t%s\n", "ip", "host_name" );
    for ( int i=0; i < hop_num; i ++ )
    {
        hi = hop_info + i;
        fprintf( fp, "%5d\t%15s\t%s\n", hi->ttl, inet_ntoa( hi->ip ), hi->host_name );
    }
    fprintf( fp, "\n# Hop information(backward path)\n" );
    fprintf( fp, "hop_num_2=%d\n", hop_num_2 );
    fprintf( fp, "hop_d\t%15s\t%s\n", "ip", "host_name" );
    for ( int i=0; i < hop_num_2; i ++ )
    {
        hi = hop_info_2 + i;
        fprintf( fp, "%5d\t%15s\t%s\n", hi->ttl, inet_ntoa( hi->ip ), hi->host_name );
    }
}


void Pong::InitializeProbeConfiguration( void )
{
    int * complementary_d = new int[hop_num];
    int * technique = new int[hop_num];
    bool * used = new bool[hop_num_2];

    for ( int i = 0; i < hop_num; i ++ )
        technique[i] = TECH_NA;
    for ( int i = 0; i < hop_num_2; i ++ )
        used[i] = false;

    // Align from forward direction
    for ( int i = 1; i < hop_num && hop_num_2 - i >= 0; i ++ )
    {
        /*
        struct in_addr ip;
        ip.s_addr = Prefix( hop_info[i].ip );
        fprintf( stderr, "prefix1 = %15s\t", inet_ntoa( ip ) );
        ip.s_addr = Prefix( hop_info_2[hop_num_2 - i].ip );
        fprintf( stderr, "prefix2 = %15s\n", inet_ntoa( ip ) );
        */

        if ( hop_info[i].ip.s_addr != 0 && hop_info[i - 1].ip.s_addr != 0 && Prefix( hop_info[i].ip ) == Prefix( hop_info_2[hop_num_2 - i].ip ) )
        {
            technique[i - 1] = TECH_4P;
            complementary_d[i - 1] = hop_num_2 - i;
            used[hop_num_2 - i] = true;
        }
        else if ( hop_info[i].ip.s_addr != 0 && hop_info_2[hop_num_2 - i].ip.s_addr != 0 )
            break;
    }

    // Align from Backward direction
    for ( int i = 1; i < hop_num_2 && hop_num - i >= 0; i ++ )
    {
        if ( technique[hop_num - i] == TECH_NA )
        {
            /*
            struct in_addr ip;
            ip.s_addr = Prefix( hop_info_2[i].ip );
            fprintf( stderr, "prefix1 = %15s\t", inet_ntoa( ip ) );
            ip.s_addr = Prefix( hop_info[hop_num - i].ip );
            fprintf( stderr, "prefix2 = %15s\n", inet_ntoa( ip ) );
            */

            if ( hop_info_2[i].ip.s_addr != 0 && hop_info_2[i - 1].ip.s_addr != 0 && Prefix( hop_info_2[i].ip ) == Prefix( hop_info[hop_num - i].ip ) )
            {
                technique[hop_num - i] = TECH_4P;
                complementary_d[hop_num - i] = i - 1;
                used[i - 1] = true;
            }
            else if ( hop_info_2[i].ip.s_addr != 0 && hop_info[hop_num - i].ip.s_addr != 0 )
                break;
        }
    }
    
    probe_num = 0;
    for ( int i = 0; i < hop_num; i ++ )
    {
        if ( hop_info[i].ip.s_addr != 0 )
        {
            probe_num ++;
            if ( technique[i] == TECH_NA )
            {
                technique[i] = TECH_2P;
                if ( hop_num_2 - i - 1 >= 0 && ! used[hop_num_2 - i - 1] && hop_info_2[hop_num_2 - i - 1].ip.s_addr != 0 )
                    complementary_d[i] = hop_num_2 - i - 1;
                else
                    complementary_d[i] = -1;
            }
        }
    }

    probe_info = new ProbeInfo[probe_num];
    hrbuffer = new HalfRecordBuffer[probe_num];
    frbuffer = new FullRecordBuffer[probe_num];
    phistory = new ProbeHistory[probe_num];
    tdp_info = new TargetedDprobeInfo[probe_num];

    probe_period = 1.0;
    double probe_interval = PROBE_INTERVAL;
    if ( probe_num > probe_period / probe_interval )
        probe_interval = probe_period / probe_num;

    double time_offset = 0;
    for ( int i = 0, id = 0; i < hop_num; i ++ )
    {
        if ( hop_info[i].ip.s_addr != 0 )
        {
            probe_info[id].id = id;
            probe_info[id].ttl = hop_info[i].ttl;
            probe_info[id].time_offset = time_offset;
            probe_info[id].technique = technique[i];
            phistory[id].Initialize( id );
            if ( complementary_d[i] == -1 )
                phistory[id].ttl_d = -1;    // Might change to: phistory[id].ttl_d = TTL_TARGETED_D;
            else
                phistory[id].ttl_d = hop_info_2[complementary_d[i]].ttl;
            phistory[id].time_offset_d = time_offset;
            tdp_info[id].probe_id = id;

            hrbuffer[id].Initialize( id );
            frbuffer[id].Initialize( id );

            time_offset += probe_interval;
            id ++;
        }
    }
}


void Pong::InitializePairingHistory( void )
{
    bool valid_hop_d[64];
    for ( int i = 0; i < hop_num_2; i ++ )
    {
        HopInfo * hi = hop_info_2 + i;
        if ( hi->ip.s_addr != 0 )
            valid_hop_d[hi->ttl] = true;
        else
            valid_hop_d[hi->ttl] = false;
    }

    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeHistory * ph = FindProbeHistory( probe_info[i].id );
        assert( ph != NULL );
        int lower_bound, upper_bound;
        GetPairingRange( probe_info + i, lower_bound, upper_bound );
        ph->InitializePairingHistory( lower_bound, upper_bound, valid_hop_d );
    }
}


void Pong::PrintPairingHistory( bool compact )
{
    system( "clear" );
    fprintf( stdout, "******* Pairing history information *******\n" );
    fprintf( stdout, "id\tpairing history\n" );
    const char * pattern = ( compact ?  "%-18s" : "%-34s" );
    char format_string[30];
    sprintf( format_string, "  \t%s  %s  %s  ...\n", pattern, pattern, pattern );
    fprintf( stdout, format_string, "entry1", "entry2", "entry3" );
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeHistory * ph = phistory + i;

        fprintf( stdout, "%2d\t", ph->probe_id );
        int num = ph->pairing_history.size();
        for ( int j = 0; j < num; j ++ )
        {
            PairingHistory * pah = &(ph->pairing_history[j]);
            fprintf( stdout, "%2d(%-4d %-4d %4.2lf",
                pah->ttl_d, pah->times_tried, (int)(pah->average_age + 0.5), pah->average_QoM );
            if ( ! compact )
                fprintf( stdout, "<%4.2lf,%4.2lf,%4.2lf>", pah->QoM_4P, pah->QoM_fsd, pah->QoM_JP );
            fprintf( stdout, ")  " );
            if ( ( !compact && j % 4 == 3 || compact && j % 7 == 6 ) && j < num - 1 )
                fprintf( stdout, "\n  \t" );
        }
        fprintf( stdout, "\n" );
    }
    fprintf( stdout, "Each entry of 'pairing history' is in format of \"hop_d(times_tried average_age average_QoM" );
    if ( ! compact )
        fprintf( stdout, "<QoM_4P,QoM_fsd,QoM_JP>" );
    fprintf( stdout, ")\".\n" );
    fprintf( stdout, "*******************************************\n" );
    fflush( stdout );
}


unsigned int Prefix( struct in_addr ip )
{
    return ( ip.s_addr & 0xE0FFFFFF );
}


// Send probe configuration to receiver

void Pong::SendProbeConfiguration( void )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_CFG );
    
    ProbeConfigGlobal pcg;
    pcg.probe_period = probe_period;
    pcg.probe_num = probe_num;

    if ( write( sock_tcp, &pcg, sizeof( ProbeConfigGlobal ) ) != sizeof( ProbeConfigGlobal ) )
    {
        fprintf( stderr, "%s Send probe configuration failed\n", GetMeasurementTime() );
        exit(1);
    }

    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeConfig pcf;
        pcf.id = probe_info[i].id;
        pcf.technique = probe_info[i].technique;
        pcf.time_offset = phistory[i].time_offset_d;
        pcf.ttl = phistory[i].ttl_d;
        assert( probe_info[i].ttl >= 1 );
        HopInfo * hi = FindHopInfo( probe_info[i].ttl );
        assert( hi != NULL );
        pcf.ip_fwd = hi->ip;
        if ( write( sock_tcp, &pcf, sizeof( ProbeConfig ) ) != sizeof( ProbeConfig ) )
        {
            fprintf( stderr, "%s Send probe configuration failed\n", GetMeasurementTime() );
            exit(1);
        }
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive probe configuration from sender

void Pong::RecvProbeConfiguration( void )
{
    ProbeConfigGlobal pcg;
    if ( recv( sock_tcp, &pcg, sizeof( ProbeConfigGlobal ), MSG_WAITALL ) != sizeof( ProbeConfigGlobal ) )
    {
        fprintf( stderr, "%s Receive probe configuration failed\n", GetMeasurementTime() );
        exit(1);
    }
    probe_period = pcg.probe_period;
    probe_num = pcg.probe_num;

    probe_info = new ProbeInfo[probe_num];
    hrbuffer = new HalfRecordBuffer[probe_num];

    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeConfig pcf;
        if ( recv( sock_tcp, &pcf, sizeof( ProbeConfig ), MSG_WAITALL ) != sizeof( ProbeConfig ) )
        {
            fprintf( stderr, "%s Receive probe configuration failed\n", GetMeasurementTime() );
            exit(1);
        }
        probe_info[i].id = pcf.id;
        probe_info[i].technique = pcf.technique;
        probe_info[i].time_offset = pcf.time_offset;
        probe_info[i].ttl = pcf.ttl;
        probe_info[i].ip_fwd = pcf.ip_fwd;
        probe_info[i].tdp_state = TDP_ICMP;
        hrbuffer[i].Initialize( pcf.id );
    }
}


// Send pairing updates from sender to receiver

void Pong::SendPairingUpdates( void )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_PUD );

    int update_num = pairing_updates.size();
    if ( write( sock_tcp, &update_num, sizeof(int) ) != sizeof(int) )
    {
        fprintf( stderr, "%s Send pairing updates failed\n", GetMeasurementTime() );
        exit(1);
    }
    for ( int i = 0; i < update_num; i ++ )
    {
        if ( write( sock_tcp, &(pairing_updates[i]), sizeof( PairingUpdate ) ) != sizeof( PairingUpdate ) )
        {
            fprintf( stderr, "%s Send pairing updates failed\n", GetMeasurementTime() );
            exit(1);
        }
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive pairing updates from sender

void Pong::RecvPairingUpdates( void )
{
    int update_num;
    if ( recv( sock_tcp, &update_num, sizeof(int), MSG_WAITALL ) != sizeof(int) )
    {
        fprintf( stderr, "%s Receive pairing updates failed\n", GetMeasurementTime() );
        exit(1);
    }
    pairing_updates.resize( update_num );
    for ( int i = 0; i < update_num; i ++ )
    {
        if ( recv( sock_tcp, &(pairing_updates[i]), sizeof( PairingUpdate ), MSG_WAITALL ) != sizeof( PairingUpdate ) )
        {
            fprintf( stderr, "%s Receive pairing updates failed\n", GetMeasurementTime() );
            exit(1);
        }
    }
}


// Send technique updates from sender to receiver

void Pong::SendTechniqueUpdates( void )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_TUD );

    int update_num = technique_updates.size();
    if ( write( sock_tcp, &update_num, sizeof(int) ) != sizeof(int) )
    {
        fprintf( stderr, "%s Send technique updates failed\n", GetMeasurementTime() );
        exit(1);
    }
    for ( int i = 0; i < update_num; i ++ )
    {
        if ( write( sock_tcp, &(technique_updates[i]), sizeof( TechniqueUpdate ) ) != sizeof( TechniqueUpdate ) )
        {
            fprintf( stderr, "%s Send technique updates failed\n", GetMeasurementTime() );
            exit(1);
        }
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive technique updates from sender

void Pong::RecvTechniqueUpdates( void )
{
    int update_num;
    if ( recv( sock_tcp, &update_num, sizeof(int), MSG_WAITALL ) != sizeof(int) )
    {
        fprintf( stderr, "%s Receive technique updates failed\n", GetMeasurementTime() );
        exit(1);
    }
    technique_updates.resize( update_num );
    for ( int i = 0; i < update_num; i ++ )
    {
        if ( recv( sock_tcp, &(technique_updates[i]), sizeof( TechniqueUpdate ), MSG_WAITALL ) != sizeof( TechniqueUpdate ) )
        {
            fprintf( stderr, "%s Receive technique updates failed\n", GetMeasurementTime() );
            exit(1);
        }
    }
}


// Send hop information to the other side

void Pong::SendHopInformation( bool remote )
{
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );

    SendCtrCode( CC_HOI );

    int hop_num = ( remote ? hop_num_2 : this->hop_num );
    HopInfo * hop_info = ( remote ? hop_info_2 : this->hop_info );
    
    if ( write( sock_tcp, &hop_num, sizeof(int) ) != sizeof(int) )
    {
        fprintf( stderr, "%s Send hop information failed\n", GetMeasurementTime() );
        exit(1);
    }

    for ( int i = 0; i < hop_num; i ++ )
    {
        HopInfoNet hi;
        hi.ttl = hop_info[i].ttl;
        hi.ip = hop_info[i].ip;
        if ( write( sock_tcp, &hi, sizeof( HopInfoNet ) ) != sizeof( HopInfoNet ) )
        {
            fprintf( stderr, "%s Send hop information failed\n", GetMeasurementTime() );
            exit(1);
        }
    }

    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
}


// Receive hop information from receiver

void Pong::RecvHopInformation( bool remote )
{
    int &hop_num = ( remote ? hop_num_2 : this->hop_num );
    HopInfo* &hop_info = ( remote ? hop_info_2 : this->hop_info );

    if ( recv( sock_tcp, &hop_num, sizeof(int), MSG_WAITALL ) != sizeof(int) )
    {
        fprintf( stderr, "%s Receive hop information failed\n", GetMeasurementTime() );
        exit(1);
    }

    hop_info = new HopInfo[hop_num];
    for ( int i = 0; i < hop_num; i ++ )
    {
        HopInfoNet hi;
        if ( recv( sock_tcp, &hi, sizeof( HopInfoNet ), MSG_WAITALL ) != sizeof( HopInfoNet ) )
        {
            fprintf( stderr, "%s Receive hop information failed\n", GetMeasurementTime() );
            exit(1);
        }
        hop_info[i].ttl = hi.ttl;
        hop_info[i].ip = hi.ip;
        hop_info[i].host_name = ( hi.ip.s_addr != 0 ? strdup( GetName( &hi.ip ) ) : n_a );
    }
}


// Find the HalfRecordBuffer corresponding to <probe_id>

HalfRecordBuffer * Pong::FindHrBuffer( int probe_id )
{
    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( hrbuffer[i].probe_id == probe_id )
            return ( hrbuffer + i );
    }
    return NULL;
}


// Find the FullRecordBuffer corresponding to <probe_id>

FullRecordBuffer * Pong::FindFrBuffer( int probe_id )
{
    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( frbuffer[i].probe_id == probe_id )
            return ( frbuffer + i );
    }
    return NULL;
}


// Find the ProbeInfo corresponding to <probe_id>

ProbeInfo * Pong::FindProbeInfo( int probe_id )
{
    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( probe_info[i].id == probe_id )
            return ( probe_info + i );
    }
    return NULL;
}


// Find the ProbeInfo corresponding to <ttl>

ProbeInfo * Pong::FindProbeInfoByTTL( int ttl )
{
    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( probe_info[i].ttl == ttl )
            return ( probe_info + i );
    }
    return NULL;
}


// Find the ProbeHistory corresponding to <probe_id>

ProbeHistory * Pong::FindProbeHistory( int probe_id )
{
    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( phistory[i].probe_id == probe_id )
            return ( phistory + i );
    }
    return NULL;
}


// Find the TargetedDprobeInfo corresponding to <probe_id>

TargetedDprobeInfo * Pong::FindTdpInfo( int probe_id )
{
    for ( int i = 0; i < probe_num; i ++ )
    {
        if ( tdp_info[i].probe_id == probe_id )
            return ( tdp_info + i );
    }
    return NULL;
}


// Find the HopInfo corresponding to <ttl>

HopInfo * Pong::FindHopInfo( int ttl, bool remote_side )
{
    HopInfo * hop_info = this->hop_info;
    int hop_num = this->hop_num;
    if ( remote_side )
    {
        hop_info = hop_info_2;
        hop_num = hop_num_2;
    }
    for ( int i = 0; i < hop_num; i ++ )
    {
        if ( hop_info[i].ttl == ttl )
        {
            return ( hop_info + i );
        }
    }
    return NULL;
}


// usec = GetProbeTime( double time_offset )

double Pong::GetProbeTime( double time_offset )
{
    double usec = TimeInUsec( time_baseline );
    usec += ( round_id * probe_period + time_offset ) * 1000000.0;
    return usec;
}


// Send f or b probe

int Pong::SendUdpProbe( int round_id, int probe_id, double probe_time_scheduled )
{
    ProbeUdpPayload payload;
    payload.SetUdpProbe( round_id, probe_id, Timeval( probe_time_scheduled ) );
    if ( sendto( sock_udp, &payload, PROBE_PACKET_SIZE, 0, (struct sockaddr *)&target_udp_addr, sizeof( target_udp_addr ) ) < 0 )
        return -1;
    return 0;
}


// Receive f or b probe

int Pong::RecvUdpProbe( ProbeResult &pr )
{
    ProbeUdpPayload payload;
    int rc = recvfrom( sock_udp, &payload, PROBE_PACKET_SIZE, 0, (struct sockaddr *)NULL, (socklen_t *)NULL );
    if ( rc == -1 )
    {
        perror( "readfrom" );
        return -1;
    }
    pr.packet_size = rc;

    pr.probe_type = probe_fb_in;
    rc = ProcessUdpPayload( (char *)&payload, pr );
    if ( rc == PL_UDP && DebugOn( pr.probe_id ) )
        IF_DEBUG( 9, pr.ShowRecord() );

    return rc;
}


// Recevie s or d probe

int Pong::RecvRouterProbe( ProbeResult &pr )
{
    // NOTE: If RecvRouterProbe() will be called in multiple threads, 'recv_buf' must not be defined as "static"
    static char recv_buf[IP_MAXPACKET];
    int rc = read( socketIn, recv_buf, IP_MAXPACKET );
    if (  rc < 0 )
    {
        perror( "read" );
        return -1;
    }
    pr.packet_size = rc;
    
    rc = ProcessIcmpPacket( recv_buf, pr, true );
    if ( rc == 0 && DebugOn( pr.probe_id ) )
        IF_DEBUG( 9, pr.ShowRecord() );
    return rc;
}


// Send s or d probe

int Pong::SendRouterProbe( int round_id, int probe_id, int ttl, bool hop_probe )
{
    // Make up a UDP or ICMP packet to send out.  Start with an IP header.
    // See Section 25.3 of UNIX Network Programming, Second Edition,
    // for why we need to twiddle the byte-orders of some fields but
    // not others, depending on what OS we run.
    struct ip ipHeader;
    memset( &ipHeader, 0, sizeof( ipHeader ) );
#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeader.ip_vhl = ( sizeof( ip ) >> 2 ) | ( 4 << 4 );
#else
    ipHeader.ip_hl = sizeof( ip ) >> 2;
    ipHeader.ip_v = 4;
#endif /* __osf__ */
    ipHeader.ip_tos = 0;
    int ipPacketSize = sizeof( ip ) + sizeof( udphdr ) + PROBE_PACKET_SIZE;
    // NOTE: For targeted d probe(ICMP echo packet), the packet size is the same( sizeof( ip ) + ICMP_MINLEN + PROBE_PACKET_SIZE )
#ifdef linux
    ipHeader.ip_len = htons( ipPacketSize );
#else
    ipHeader.ip_len = ipPacketSize;
#endif /* linux */
    // Encode round_id into the high 10 bits, encode probe_id into the low 6 bits
    ipHeader.ip_id = htons( ( ( round_id & 0x3FF ) << 6 ) + ( probe_id & 0x3F ) );
#ifdef linux
    ipHeader.ip_off = htons( IP_DF );
#else
    ipHeader.ip_off = IP_DF;
#endif /* linux */

    ipHeader.ip_sum = 0;
    memcpy( &(ipHeader.ip_src), &originAddress, sizeof( struct in_addr ) );
    int tdp_state = TDP_UDP;
    if ( ttl == TTL_TARGETED_D )
    {
        // Targeted d probe
        assert( pthread_mutex_lock( &pih_mutex ) == 0 );
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL && pi->ip_fwd.s_addr != 0 );
        memcpy( &(ipHeader.ip_dst), &(pi->ip_fwd), sizeof( struct in_addr ) );
        tdp_state = pi->tdp_state;
        assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
        if ( tdp_state == TDP_NONE )
            return -1;
    }
    else
    {
        // TTL limited d probe
        memcpy( &(ipHeader.ip_dst), &targetAddress, sizeof( struct in_addr ) );
    }

    if ( ttl == TTL_TARGETED_D )
    {
        ipHeader.ip_ttl = ORIGIN_TTL;
        ipHeader.ip_p = ( tdp_state == TDP_ICMP ? IPPROTO_ICMP : IPPROTO_UDP );
    }
    else
    {
        ipHeader.ip_ttl = ttl;
        ipHeader.ip_p = IPPROTO_UDP;
    }

    // Make up UDP or ICMP header;
    struct udphdr udpHeader;
    struct icmp * icmpHeader = (icmp *)&udpHeader;
    if ( ttl == TTL_TARGETED_D && tdp_state == TDP_ICMP )
    {
           icmpHeader->icmp_type = ICMP_ECHO;
        icmpHeader->icmp_code = 0;
        icmpHeader->icmp_cksum = htons( 0 );
        icmpHeader->icmp_seq = ipHeader.ip_id;
        icmpHeader->icmp_id = htons( SD_UDP_PORT );
    }
    else
    {
        memset( &udpHeader, 0, sizeof( udphdr ) );
        udpHeader.uh_sport = htons( SD_UDP_PORT );
        udpHeader.uh_dport = htons( UDPRCV_PORT );
        udpHeader.uh_ulen = htons( PROBE_PACKET_SIZE + sizeof( udphdr ) );
        udpHeader.uh_sum = htons( 0 ); // Let the UDP checksum be 0
    }

    // UDP payload or ICMP data part
    ProbeUdpPayload udpPayload;
    if ( hop_probe )
        udpPayload.SetHopProbe();
    else
        udpPayload.SetRouterProbe();

    // Build the packet now.
    char ipPacket[MAX_SEND_BUF_SIZE];
    memcpy( ipPacket, &ipHeader, sizeof( ipHeader ) );
    memcpy( ipPacket + sizeof( ipHeader ), &udpHeader, sizeof( udphdr ) );
    memcpy( ipPacket + sizeof( ipHeader) + sizeof( udphdr ), &udpPayload, PROBE_PACKET_SIZE );

    if ( ttl == TTL_TARGETED_D && tdp_state == TDP_ICMP )
    {
        // Compute ICMP checksum.
        ( (icmp *)( ipPacket + sizeof( ipHeader ) ) )->icmp_cksum = InCksum( (u_short *)( ipPacket + sizeof( ipHeader ) ), ICMP_MINLEN + PROBE_PACKET_SIZE );
    }
    else
    {
        // Compute UDP checksum.  We're going to build this up part by part,
        // starting with the different components of the pseudo-header,
        // and then doing the UDP header and payload as a single block.
        // The idea for this comes from the libnet code, basically we build the
        // checksum for the pseudo-header without actually having to
        // store it anywhere.  Not real efficient but it should be OK for
        // our purposes.
        //
        // We note that the UDP checksum routine returns the 1s complement
        // of the 1s complement sum; we need to unroll this to do some
        // other manipulations before actually storing it.
        // We assemble some components to the checksum, and then re-do
        // the checksum carry and complement operations.
        u_int checksum;
        checksum = (u_short)~InCksum( (u_short *)&(ipHeader.ip_src), 2 * sizeof( struct in_addr ) );
        checksum += (u_short)htons( IPPROTO_UDP ) + (u_short)htons( sizeof( udphdr ) + PROBE_PACKET_SIZE );
        checksum += (u_short)~InCksum( (u_short *)( ipPacket + sizeof( ipHeader ) ), sizeof( udphdr ) + PROBE_PACKET_SIZE );
        while ( checksum >> 16 )
        {
            checksum = ( checksum & 0xffff ) + ( checksum >> 16 );
        }
        ( (udphdr *)( ipPacket + sizeof( ipHeader ) ) )->uh_sum = ~checksum;
    }

    // Send the packet
    if ( send( socketOut, ipPacket, ipPacketSize, 0 ) < 0 )
    {
        perror( "send" );
        return -1;
    }

    return 0;
}


void Pong::SetControlSocket_Rcv( void )
{
    /* 
    Create control channel: TCP connection
    */
    if ( ( sock_tcp = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_STREAM )" );
        exit(1);
    }
    int opt_value = 1;
    if ( setsockopt( sock_tcp, SOL_SOCKET, SO_REUSEADDR, (char *)&opt_value, sizeof( opt_value ) ) < 0 )
    {
        perror( "setsockopt( sock_tcp, SOL_SOCKET, SO_REUSEADDR )" );
        exit(1);
    }

    struct sockaddr_in snd_tcp_addr;
    bzero( (char *)&snd_tcp_addr, sizeof( snd_tcp_addr ) );
    snd_tcp_addr.sin_family = AF_INET;
    memcpy( (void *)&(snd_tcp_addr.sin_addr.s_addr), &targetAddress, sizeof( targetAddress ) );
    snd_tcp_addr.sin_port = htons( sender_control_port );
    if ( connect( sock_tcp, (struct sockaddr *)&snd_tcp_addr, sizeof( snd_tcp_addr ) ) < 0 )
    {
        perror( "Make sure that pong_snd runs at sender" );
        HeartBeat( "no_conn" );
        exit(1);
    }
    IF_DEBUG( 1, fprintf( stderr, "Connection established\n" ) );

    opt_value = 262142;
    socklen_t opt_len = sizeof( opt_value );
    setsockopt( sock_tcp, SOL_SOCKET, SO_SNDBUF, &opt_value, opt_len );
    getsockopt( sock_tcp, SOL_SOCKET, SO_SNDBUF, &opt_value, &opt_len );
    IF_DEBUG( 3, fprintf( stderr, "Sending buffer size is set to %d bytes\n", opt_value ) );

    SendReceiverName();
    ExchangeHostID();
    CheckAccessResult();
    CheckVersion();
}


void Pong::CheckAccessResult( void )
{
    assert( RecvCtrCode() == CC_ACC );
    if ( RecvIntValue() != 1 )
    {
        // Access deny
        fprintf( stderr, "Access is denied by Pong sender\n" );
        HeartBeat( "access_deny" );
        exit(1);
    }
}


void Pong::SetControlSocket_Snd( void )
{
    /* Control stream: TCP connection */
    if ( ( sock_tcp_listener = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_STREAM, 0 )" );
        exit(1);
    }
    int opt_value = 1;
    if ( setsockopt( sock_tcp_listener, SOL_SOCKET, SO_REUSEADDR, (char *)&opt_value, sizeof( opt_value ) ) < 0 )
    {
        perror( "setsockopt( SOL_SOCKET,SO_REUSEADDR )" );
        exit(1);
    }
    struct sockaddr_in snd_tcp_addr;
    bzero( (char *)&snd_tcp_addr, sizeof( snd_tcp_addr ) );
    snd_tcp_addr.sin_family = AF_INET;
    snd_tcp_addr.sin_addr.s_addr = htonl( INADDR_ANY );
    snd_tcp_addr.sin_port = htons( sender_control_port );
    if ( bind( sock_tcp_listener, (struct sockaddr *)&snd_tcp_addr, sizeof( snd_tcp_addr ) ) < 0 )
    {
        perror( "bind( sock_tcp_listener )" );
        exit(1);
    }
    if ( listen( sock_tcp_listener, 1 ) < 0 )
    {
        perror( "listen( sock_tcp_listener, 1 )" );
        exit(1);
    }
}


void Pong::WaitForReceiver( void )
{
    /* 
      Wait until receiver attempts to connect, starting new measurement cycle
    */
    int target_tcp_adrlen = sizeof( target_tcp_addr );

    bool receiver_contacted = false;
    while ( ! receiver_contacted )
    {
        IF_DEBUG( 1, fprintf( stderr, "Waiting for receiver to establish control channel --> " ) );
        
        sock_tcp = accept( sock_tcp_listener, (struct sockaddr *)&target_tcp_addr, (socklen_t *)&target_tcp_adrlen );
        if ( sock_tcp < 0 )
        {
           perror( "accept( sock_tcp_listener )" );
           exit(1);
        }
        IF_DEBUG( 1, fprintf( stderr, "OK\n" ) );

        int opt_value = 262142;
        socklen_t opt_len = sizeof( opt_value );
        setsockopt( sock_tcp, SOL_SOCKET, SO_SNDBUF, &opt_value, opt_len );
        getsockopt( sock_tcp, SOL_SOCKET, SO_SNDBUF, &opt_value, &opt_len );
        IF_DEBUG( 3, fprintf( stderr, "Sending buffer size is set to %d bytes\n", opt_value ) );

        RecvReceiverName();
        IF_DEBUG( 3, fprintf( stderr, "targetName = %s\n", targetName ) );

        ExchangeHostID();

        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_ACC );
        if ( receiver_id != VALUE_UNDEFINED && receiver_id != peer_host_id )
        {
            // Deny access
            SendIntValue( 0 );
            close( sock_tcp );
            IF_DEBUG( 3, fprintf( stderr, "Receiver host_id (%d) is mismatched (expected host_id = %d), deny access\n", peer_host_id, receiver_id ) );
        }
        else
        {
            // Accept access
            SendIntValue( 1 );
            receiver_contacted = true;
            close( sock_tcp_listener );
        }
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    }

    CheckVersion();

    /*
    struct hostent *host_rcv = gethostbyaddr( (char *)&(target_tcp_addr.sin_addr), sizeof( target_tcp_addr.sin_addr ), AF_INET );
    if ( host_rcv != NULL )
    {
        targetName = strdup( host_rcv->h_name );
        IF_DEBUG( 3, fprintf( stderr, "targetName = %s\n", targetName ) );
    }
    else
    {
        targetName = strdup( inet_ntoa( targetAddress ) );
        IF_DEBUG( 3, fprintf( stderr, "targetName = Unknown\n" ) );
    }
    */

    /*
      Form receiver's UDP address
    */
    bzero( (char *)&target_udp_addr, sizeof( target_udp_addr ) );
    target_udp_addr.sin_family = AF_INET;
    target_udp_addr.sin_addr.s_addr = target_tcp_addr.sin_addr.s_addr; 
    target_udp_addr.sin_port = htons( UDPRCV_PORT );
    
    targetAddress = target_udp_addr.sin_addr;
    IF_DEBUG( 3, fprintf( stderr, "targetAddress = %s\n", inet_ntoa( targetAddress ) ) );
    
    strcpy( probe_config_file, inet_ntoa( targetAddress ) );
    strcat( probe_config_file, ".pong" );

    if ( SetOriginName( OriginHost ) < 0 )
    {
        exit(1);
    }
    SetUdpSocket();
}


void Pong::SetupMonitorServer( void )
{
    /* Control stream: TCP connection */
    if ( ( sock_monitor_listener = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 )
    {
        perror( "socket( AF_INET, SOCK_STREAM, 0 )" );
        exit(1);
    }
    int opt_value = 1;
    if ( setsockopt( sock_monitor_listener, SOL_SOCKET, SO_REUSEADDR, (char *)&opt_value, sizeof( opt_value ) ) < 0 )
    {
        perror( "setsockopt( SOL_SOCKET,SO_REUSEADDR )" );
        exit(1);
    }
    struct sockaddr_in monitor_addr;
    bzero( (char *)&monitor_addr, sizeof( monitor_addr ) );
    monitor_addr.sin_family = AF_INET;
    monitor_addr.sin_addr.s_addr = htonl( INADDR_ANY );
    monitor_addr.sin_port = htons( MONITOR_PORT );
    if ( bind( sock_monitor_listener, (struct sockaddr *)&monitor_addr, sizeof( monitor_addr ) ) < 0 )
    {
        perror( "bind( sock_monitor_listener )" );
        exit(1);
    }
    if ( listen( sock_monitor_listener, 1 ) < 0 )
    {
        perror( "listen( sock_monitor_listener, 1 )" );
        exit(1);
    }
    sock_monitor = -1;
}


void Pong::AcceptMonitorClient( void )
{
    sock_monitor = accept( sock_monitor_listener, NULL, NULL );
    if ( sock_monitor < 0 )
    {
        perror( "accept( sock_monitor_listener )" );
        return;
    }
    //IF_DEBUG( 3, fprintf( stderr, "Monitor attached\n" ) );
    fclose( stdin );
    if ( dup2( sock_monitor, STDIN ) < 0 )
    {
        perror( "dup2( STDIN )" );
        close( sock_monitor );
        sock_monitor = -1;
        return;
    }
    stdin = fdopen( STDIN, "r" );
    if ( dup2( sock_monitor, STDOUT ) < 0 )
    {
        perror( "dup2( STDOUT )" );
        close( sock_monitor );
        sock_monitor = -1;
        return;
    }
    close( sock_monitor );
    enable_interactive_command = false;
}


void Pong::Initialize_Common_Pre( void )
{
    CheckProcess();
    PrintPID();

    ReadHostID();
    ReadVersion();
}


void Pong::Initialize_Common( void )
{
    // Get sockets
    if ( GetSocketOut() < 0 )
    {
        exit(1);
    }
    if ( GetSocketIn() < 0 )
    {
        exit(1);
    }

#ifdef HAVE_PCAP
    // If we're running with pcap enabled, set it up.
    SetupPCAP();
#endif // HAVE_PCAP

    SetupMonitorServer();

    ReportInitialize();

    SetupPacketSink();

    InitializeProbeInfo();
}


/*
Estimate round-trip time
*/

double Pong::EstimateRTT( void )
{
    double sum_rtt = 0;
    timeval first_time;
    for( int i = 0; i < 3; i ++ )
    {
        GetCurrentTime( first_time );
        assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
        SendCtrCode( CC_RTT );
        assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
        assert( RecvCtrCode() == CC_RTT );
        GetCurrentTime( current_time );
        if ( i > 0 )
            sum_rtt += delta_t( first_time, current_time );
    }
    return (double)sum_rtt / 2;
}


/* 
Bounce a number of empty messages back to the receiver, in order to measure RTT 
*/
void Pong::EstimateRTT_Respond( void )
{
    for ( int i = 0; i < 3; i ++ )
    {
      assert( RecvCtrCode() == CC_RTT );
      assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
      SendCtrCode( CC_RTT );
      assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    }
}


int Pong::ParsePacket( char *ipPacket, ProbeResult &pr )
{
    if ( pr.packet_size < sizeof( ip ) )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete IP packet of %d bytes\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }
    
    ip *ipHeader = (ip *)ipPacket;

    // Check protocol in IP header
    switch ( ipHeader->ip_p )
    {
    case IPPROTO_ICMP:
        return ProcessIcmpPacket( ipPacket, pr );

    case IPPROTO_UDP:
        return ProcessUdpPacket( ipPacket, pr );

    default:
        fprintf( stderr, "Received unknown protocol %d\n", ipHeader->ip_p );
        return -1;
    }

    return 0;
}


int Pong::ProcessIcmpPacket( char *ipPacket, ProbeResult &pr, bool silent )
{
    // Parse the ICMP packet, doing a little error checking along
    // the way.  By the end, we'll have ipHeader and icmpHeader
    // pointing to valid structures within the packet, and
    // ipHeader2 pointing to the IP header of the generating
    // IP packet.

    ip *ipHeader, *ipHeader2;
    icmp *icmpHeader;
    udphdr *udpHeader;
    unsigned int ipHeaderLength, ipHeaderLength2;

    ipHeader = (ip *)ipPacket;

#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeaderLength = ( ipHeader->ip_vhl & 0x0f ) << 2;
#else
    ipHeaderLength = ipHeader->ip_hl << 2;
#endif /* __osf__ */

    if ( pr.packet_size - ( 0 + ipHeaderLength ) < ICMP_MINLEN )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete ICMP packet of %d bytes\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }

    icmpHeader = (icmp *)( ((char *)ipHeader) + ipHeaderLength );

    pr.timestamp.tv_sec = pr.timestamp.tv_usec = 0;
    memcpy( &pr.icmpSourceAddress, &(ipHeader->ip_src), sizeof( struct in_addr ) );
    pr.icmpResult = GetAction( icmpHeader->icmp_type, icmpHeader->icmp_code );

    // Check ICMP type and format
    if ( ( icmpHeader->icmp_type == ICMP_ECHO || icmpHeader->icmp_type == ICMP_ECHOREPLY ) && ntohs( icmpHeader->icmp_id ) == SD_UDP_PORT && ProcessUdpPayload( (char *)icmpHeader + ICMP_MINLEN, pr ) == PL_ROUTER )
    {
        // Targeted d probe
        int id = ntohs( icmpHeader->icmp_seq );
        pr.round_id = RestoreRoundId( id >> 6 );
        pr.probe_id = id & 0x3F;
        pr.ttl = ipHeader->ip_ttl;
        if ( icmpHeader->icmp_type == ICMP_ECHO )
            pr.probe_type = probe_sd_out;
        else
            pr.probe_type = probe_sd_in;

        return 0;
    }
    else if ( ( icmpHeader->icmp_type != ICMP_TIMXCEED ) && ( icmpHeader->icmp_type != ICMP_UNREACH ) )
    {
        if ( ! silent )
        {
            IF_DEBUG( 5, fprintf( stderr, "%s Ignoring ICMP packet\n", GetMeasurementTime() ) );
            IF_DEBUG( 5, fprintf( stderr, "ICMP type = %d, code = %d, action = %d, src_ip = %s", icmpHeader->icmp_type, icmpHeader->icmp_code, pr.icmpResult, inet_ntoa( ipHeader->ip_src ) ) );
            IF_DEBUG( 5, fprintf( stderr, ", dest_ip = %s\n", inet_ntoa( ipHeader->ip_dst ) ) );
        }
        return -1;
    }
    
    if ( pr.packet_size - ( 0 + ipHeaderLength + ICMP_MINLEN ) < sizeof( ip ) )
    {
        if ( ! silent )
            IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete inner IP packet, %d bytes total\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }

    // Check for a valid (to us) IP header within the packet.
    // For "time exceeded" or "destination unreachable", this
    // header will be 8 bytes past the ICMP header.
    ipHeader2 = (ip *)( (char *)icmpHeader + ICMP_MINLEN );

#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeaderLength2 = (ipHeader2->ip_vhl & 0x0f) << 2;
#else
    ipHeaderLength2 = ipHeader2->ip_hl << 2;
#endif /* __osf__ */

    icmp * icmpHeader2 = (icmp *)( ((char *)ipHeader2) + ipHeaderLength2 );
    if ( ipHeader2->ip_p == IPPROTO_ICMP && icmpHeader2->icmp_type == ICMP_ECHO && ntohs( icmpHeader2->icmp_id ) == SD_UDP_PORT && pr.icmpResult == PongActionFiltered )
    {
        if ( ! silent )
        {
            int id = ntohs( icmpHeader2->icmp_seq );
            //pr.round_id = RestoreRoundId( id >> 6 );
            pr.probe_id = id & 0x3F;
            IF_DEBUG( 3, fprintf( stderr, "%s Targeted ICMP d probe has been blocked, try UDP instead, probe_id = %d\n", GetMeasurementTime(), pr.probe_id ) );
            assert( pthread_mutex_lock( &pih_mutex ) == 0 );
            ProbeInfo * pi = FindProbeInfo( pr.probe_id );
            assert( pi != NULL );
            pi->tdp_state = TDP_UDP;
            assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
            assert( pthread_mutex_lock( &hrb_mutex ) == 0 );
            HalfRecordBuffer * hrb = FindHrBuffer( pr.probe_id );
            assert( hrb != NULL );
            hrb->buffer.clear();
            assert( pthread_mutex_unlock( &hrb_mutex ) == 0 );
        }
        return -1;
    }
    // Additional checking here ... must be UDP
    else if ( ipHeader2->ip_p != IPPROTO_UDP )
    {
        if ( ! silent )
        {
            IF_DEBUG( 9, fprintf( stderr, "%s Ignoring ICMP packet for unknown non-UDP packet\n", GetMeasurementTime() ) );
            IF_DEBUG( 9, fprintf( stderr, "ICMP type = %d, code = %d, action = %d, src_ip = %s", icmpHeader->icmp_type, icmpHeader->icmp_code, pr.icmpResult, inet_ntoa( ipHeader->ip_src ) ) );
            IF_DEBUG( 9, fprintf( stderr, ", dest_ip = %s\n", inet_ntoa( ipHeader->ip_dst ) ) );
        }
        return -1;
    }

    // Align UDP header template.
    if ( pr.packet_size - ( 0 + ipHeaderLength + ICMP_MINLEN + ipHeaderLength2 ) < sizeof( udphdr ) )
    {
        if ( ! silent )
            IF_DEBUG( 3, fprintf( stderr, "%s Received incomplete inner UDP packet, %d bytes total\n", GetMeasurementTime(), pr.packet_size ) );
        return -1;
    }

    udpHeader = (udphdr *)( (char *)ipHeader2 + ipHeaderLength2 );

    // Check source and destination UDP port number and UDP
    // length (header+payload)
    if ( ( ntohs( udpHeader->uh_sport ) != SD_UDP_PORT ) || ( ntohs( udpHeader->uh_dport ) != UDPRCV_PORT ) || ( ntohs( udpHeader->uh_ulen ) != PROBE_PACKET_SIZE + sizeof( udphdr ) ) )
    {
        if ( ! silent )   
            IF_DEBUG( 9, fprintf( stderr, "%s Ignoring ICMP packet for unknown UDP packet\n", GetMeasurementTime() ) );
        return -1;
    }

    if ( pr.icmpResult == PongActionFiltered )
    {
        if ( ! silent )
        {
            assert( pthread_mutex_lock( &pih_mutex ) == 0 );
            ProbeInfo * pi = FindProbeInfo( pr.probe_id );
            assert( pi != NULL );
            bool is_tdp = false;
            if ( pi->ttl == TTL_TARGETED_D )
            {
                pi->tdp_state = TDP_NONE;
                is_tdp = true;
            }
            assert( pthread_mutex_unlock( &pih_mutex ) == 0 );

            if ( is_tdp )
            {
                assert( pthread_mutex_lock( &hrb_mutex ) == 0 );
                HalfRecordBuffer * hrb = FindHrBuffer( pr.probe_id );
                assert( hrb != NULL );
                hrb->buffer.clear();
                assert( pthread_mutex_unlock( &hrb_mutex ) == 0 );
                IF_DEBUG( 3, fprintf( stderr, "%s Targeted UDP d probe has been blocked, sending TDF, probe_id = %d\n", GetMeasurementTime(), pr.probe_id ) );
                SendTDF( pr.probe_id );
            }
            else if ( ipHeader2->ip_dst.s_addr == targetAddress.s_addr )
                IF_DEBUG( 3, fprintf( stderr, "%s TTL limited UDP probe has been blocked, probe_id = %d\n", GetMeasurementTime(), pr.probe_id ) );
        }
        return -1;
    }

    // Fill in return fields
    int ip_id = ntohs( ipHeader2->ip_id );
    pr.round_id = RestoreRoundId( ip_id >> 6 );
    pr.probe_id = ip_id & 0x3F;
    if ( ipHeader->ip_src.s_addr == targetAddress.s_addr )
        pr.probe_type = probe_icmp_error;
    else
        pr.probe_type = probe_sd_in;

    return 0;
}


int Pong::ProcessUdpPacket( char *ipPacket, ProbeResult &pr )
{
    ip *ipHeader;
    udphdr *udpHeader;
    unsigned int ipHeaderLength;

    ipHeader = (ip *)ipPacket;

#ifdef __osf__
    // Tru64 <netinet/ip.h> doesn't declar ip_hl if __STDC__ == 1
    ipHeaderLength = ( ipHeader->ip_vhl & 0x0f ) << 2;
#else
    ipHeaderLength = ipHeader->ip_hl << 2;
#endif /* __osf__ */

    udpHeader = (udphdr *)( (char *)ipHeader + ipHeaderLength );

    // Check source and destination UDP port number and UDP
    // length (header+payload)
    if ( ( ntohs( udpHeader->uh_sport ) != SD_UDP_PORT ) || ( ntohs( udpHeader->uh_dport ) != UDPRCV_PORT ) || ( ntohs( udpHeader->uh_ulen ) != PROBE_PACKET_SIZE + sizeof( udphdr ) ) )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Ignoring unknown(port) UDP packet\n", GetMeasurementTime() ) );
        return -1;
    }

    // Check source IP address and destination IP address
    if ( ipHeader->ip_src.s_addr == originAddress.s_addr )
    {
        // Outbound UDP traffic
        // NOTE: destination IP address might not necessarily be 'targetAddress' due to the introduction of targeted d probe
        if ( ipHeader->ip_ttl < 64 && ipHeader->ip_dst.s_addr == targetAddress.s_addr || ipHeader->ip_ttl == ORIGIN_TTL && ipHeader->ip_dst.s_addr != targetAddress.s_addr )
        {
            pr.probe_type = probe_sd_out;
        }
        else if ( ipHeader->ip_dst.s_addr == targetAddress.s_addr )
        {
            pr.probe_type = probe_fb_out;
        }
        else
        {
            IF_DEBUG( 3, fprintf( stderr, "%s Ignoring unknown(address) UDP packet\n", GetMeasurementTime() ) );
            return -1;
        }
    }
    else if ( ipHeader->ip_src.s_addr == targetAddress.s_addr && ipHeader->ip_dst.s_addr == originAddress.s_addr )
    {
        // Inbound UDP traffic
        pr.probe_type = probe_fb_in;
    }
    else
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Ignoring unknown(address) UDP packet\n", GetMeasurementTime() ) );
        return -1;
    }

    int rc = ProcessUdpPayload( (char *)udpHeader + sizeof( udphdr ), pr );
    if ( rc == -1 )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Invalid UDP payload received\n", GetMeasurementTime() ) );
        return -1;
    }
    if ( pr.probe_type == probe_sd_out )
    {
        assert( rc == PL_ROUTER );
        int ip_id = ntohs( ipHeader->ip_id );
        pr.round_id = RestoreRoundId( ip_id >> 6 );
        pr.probe_id = ip_id & 0x3F;
    }
    else if ( rc == PL_ROUTER )
    {
        IF_DEBUG( 3, fprintf( stderr, "%s Invalid UDP probe: this is a router-targeted probe, src_ip = %s, remain_ttl = %d", GetMeasurementTime(), inet_ntoa( ipHeader->ip_src ), ipHeader->ip_ttl ) );
        return -1;
    }
    else if ( rc == PL_HOP )
    {
        if ( pr.probe_type == probe_fb_out )
            pr.probe_type = probe_hop_out;
        else if ( pr.probe_type == probe_fb_in )
            pr.probe_type = probe_hop_in;
        else
            assert(0);
    }
    pr.ttl = ipHeader->ip_ttl;

    return 0;
}


int Pong::ProcessUdpPayload( char *udp_payload, ProbeResult &pr )
{
    ProbeUdpPayload * payload = (ProbeUdpPayload *)udp_payload;

    if ( payload->IsValid() )
    {
        pr.round_id = payload->round_id;
        pr.probe_id = payload->probe_id;
        pr.timestamp = payload->timestamp;
        if ( payload->type_code == PL_ROUTER )
            return PL_ROUTER;
        else if ( payload->type_code == PL_UDP )
            return PL_UDP;
        else if ( payload->type_code == PL_HOP )
            return PL_HOP;
        else
        {
            assert(0);
            return -1;
        }
    }
    else
        return -1;
}


bool DebugOn( int probe_id )
{
    return( ProbeToDebug == -1 || ProbeToDebug == probe_id ? true : false );
}


void Pong::SetupPacketSink()
{
    if ( pthread_create( &tid_sink, NULL, sink_process, this ) != 0 )
    {
        fprintf( stderr, "pthread_create(sink) failed\n" );
        exit(1);
    }
    IF_DEBUG( 3, fprintf( stderr, "Packet sink process installed\n" ) );
}


void * sink_process( void * arg )
{
    Pong *obj = (Pong *)arg;
    
    obj->ReceiveMessages();

    pthread_exit(NULL);
    
    return NULL;
}


void Pong::ProcessControlMessage()
{
    int ctr_code = RecvCtrCode();
    switch ( ctr_code )
    {
    case CC_HFR:
        {
            assert( is_sender );
            HalfRecord hr;
            RecvHalfRecord( hr );
            if ( DebugOn( hr.probe_id ) )
                IF_DEBUG( 7, fprintf( stderr, "HalfRecord received, probe_id=%d, round_id=%d\n", hr.probe_id, hr.round_id ) );
            // fprintf( stdout, "pih_mutex 16\n" );
            assert( pthread_mutex_lock( &pih_mutex ) == 0 );
            UpdateFullRecord( hr, false );
            assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
            break;
        }

    case CC_ATO:
        {
            assert( ! is_sender );
            int probe_id;
            double time_offset_increment;
            RecvATO( probe_id, time_offset_increment );
            if ( DebugOn( probe_id ) )
                IF_DEBUG( 5, fprintf( stderr, "ATO message received, probe_id=%d, time_offset_increment=%.3lf\n", probe_id, time_offset_increment * 1000 ) );
            AdjustTimeOffset( probe_id, time_offset_increment );
            SendATA( probe_id, round_id + 1 );
            break;
        }

    case CC_ATA:
        {
            assert( is_sender );
            int probe_id, round_id;
            RecvATA( probe_id, round_id );
            if ( DebugOn( probe_id ) )
                IF_DEBUG( 5, fprintf( stderr, "ATA message received, probe_id=%d, round_id=%d\n", probe_id, round_id ) );
            // fprintf( stdout, "pih_mutex 17\n" );
            assert( pthread_mutex_lock( &pih_mutex ) == 0 );
            ProbeHistory * ph = FindProbeHistory( probe_id );
            assert( ph != NULL && ph->ATO_pending );
            ph->ATA_round_id = round_id;
            ph->ATO_pending = false;
            assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
            break;
        }

    case CC_QUIT:
        {
            IF_DEBUG( 1, fprintf( stderr, "%s Quit command received\n", GetMeasurementTime() ) );
            SwitchDebug( true );
            IF_DEBUG( 1, fprintf( stderr, "Quitting ... \n" ) );
            done = true;
            break;
        }

    case CC_CNCL:
        {
            IF_DEBUG( 1, fprintf( stderr, "%s Cancel command received\n", GetMeasurementTime() ) );
            SwitchDebug( true );
            IF_DEBUG( 1, fprintf( stderr, "Quitting ... \n" ) );
            cancel_measurement = true;
            done = true;
            break;
        }

    case CC_TRA:
        {
            IF_DEBUG( 1, fprintf( stderr, "%s Trace RA command received\n", GetMeasurementTime() ) );
            SwitchDebug( true );
            IF_DEBUG( 1, fprintf( stderr, "Start tracing RA ... \n" ) );
            quit_reason = "route_alteration";
            phase_done = true;
            phase = PP_TRA;
            break;
        }

    case CC_TBL:
        {
            assert( is_sender );
            RecvTimeBaseline();
            OutputDataID();
            break;
        }

    case CC_PUD:
        {
            assert( ! is_sender );
            RecvPairingUpdates();
            pairing_updated = true;
            break;
        }

    case CC_PUA:
        {
            assert( is_sender );
            PUA_round_id = RecvIntValue();
            IF_DEBUG( 4, fprintf( stderr, "%s Acknowledge of pairing updates received, finalize pairing updates ... ", GetMeasurementTime() ) );
            assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
            UpdatePairings();
            pairing_updates.clear();
            pairing_update_pending = false;
            assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
            IF_DEBUG( 4, fprintf( stderr, "done.\n" ) );
            break;
        }

    case CC_TUD:
        {
            assert( ! is_sender );
            RecvTechniqueUpdates();
            technique_updated = true;
            break;
        }

    case CC_TUA:
        {
            assert( is_sender );
            TUA_round_id = RecvIntValue();
            IF_DEBUG( 4, fprintf( stderr, "%s Acknowledge of technique updates received, finalize technique updates ... ", GetMeasurementTime() ) );
            assert( pthread_mutex_lock( &tpt_mutex ) == 0 );
            assert( technique_update_pending && ! technique_updates.empty() );
            UpdateTechniques();
            technique_updates.clear();
            technique_update_pending = false;
            assert( pthread_mutex_unlock( &tpt_mutex ) == 0 );
            IF_DEBUG( 4, fprintf( stderr, "done.\n" ) );
            break;
        }

    case CC_TDF:
        {
            assert( is_sender );
            int probe_id = RecvIntValue();
            IF_DEBUG( 3, fprintf( stderr, "%s TDF message of probe %d recieved, disable targeted d probe for this hop\n", GetMeasurementTime(), probe_id ) );
            assert( pthread_mutex_lock( &pih_mutex ) == 0 );
            ProbeInfo * pi = FindProbeInfo( probe_id );
            assert( pi != NULL );
            ProbeHistory * ph = FindProbeHistory( probe_id );
            assert( ph != NULL );
            pi->technique = TECH_2P;
            ph->QoM_4P.age = ph->QoM_fsd.age = ph->QoM_fsb.age = TH_QoM_OLD_AGE + 100;
            ph->DeletePairingHistory( TTL_TARGETED_D );
            tdp_available[probe_id] = false;
            tdpa_updated = true;
            assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
            break;
        }

    default:
        {
            fprintf( stderr, "Unknown control code received, ctr_code=%d\n", ctr_code );
            exit(1);
        }
    }
}


// Temprarily turn off debug(by setting DebugLevel to 0) if <turn_on>=false
// Restore DebugLevel if <turn_on> = true

void SwitchDebug( bool turn_on )
{
    if ( turn_on )
    {
        if ( SavedDebugLevel != -99 )
        {
            DebugLevel = SavedDebugLevel;
            SavedDebugLevel = -99;
            fprintf( stdout, "(Note: Debug output is reenabled.)\n" );
            fflush( stdout );
        }
    }
    else
    {
        if ( enable_interactive_command && SavedDebugLevel == -99 )
        {
            SavedDebugLevel = DebugLevel;
            DebugLevel = 1;
        }
    }
}


void DebugOffWarning( void )
{
    if ( enable_interactive_command )
    {
        fprintf( stdout, "(Note: Debug output is disabled temporarily. Press <Enter> or issue 'd' command to reenable it.)\n" );
        fflush( stdout );
    }
}


void Pong::ToggleATPT( void )
{
    auto_tpt = ! auto_tpt;
    const char * status = ( auto_tpt ? "enabled" : "disabled" );
    fprintf( stdout, "Automatic tuning(of pairing and technique) is %s\n", status );
    fflush( stdout );
}


void Pong::ShowLCD( void )
{
    system( "clear" );
    fprintf( stdout, "***** Long-term congestion distribution *****\n" );
    fprintf( stdout, "   link   \tinst\t10sec\t30sec\t 1min\t 3min\t 10min\t 30min\t 1hour\t  3hour\t  6hour\t 12hour\t 24hour\n" );
    for ( int i = 0; i <= hop_num; i ++ )
    {
        fprintf( stdout, "hop: %2d-%-2d\t%4.1lf\t%5.1lf\t%5.1lf\t%5.1lf\t%5.1lf\t%6.1lf\t%6.1lf\t%6.1lf\t%7.1lf\t%7.1lf\t%7.1lf\t%7.1lf\n",
            i, i + 1, link_cc[i], link_cc_10sec[5][i], LCD_30sec[i], LCD_1min[i], LCD_3min[i], LCD_10min[i], LCD_30min[i], LCD_1hour[i], LCD_3hour[i], LCD_6hour[i], LCD_12hour[i], LCD_24hour[i] );
    }
    fprintf( stdout, "*********************************************\n" );
    fflush( stdout );
}


void Pong::HelpInteractiveCommand( void )
{
    fprintf( stdout, "***** Interactive commands helper *****\n" );
    if ( sock_monitor >= 0 )
    {
    fprintf( stdout, "x          - Exit monitor\n" );
    }
    fprintf( stdout, "quit       - Quit\n" );
    fprintf( stdout, "h          - Help\n" );
    fprintf( stdout, "hd,dh      - Help for debug output\n" );
    fprintf( stdout, "d          - Enable debug output if disabled (restore debug level)\n" );
    fprintf( stdout, "dd         - Disable debug output (by setting debug level to 1 temporarily)\n" );
    fprintf( stdout, "d#(#=1-9)  - Set debug level to #\n" );
    fprintf( stdout, "<Enter>    - Toggle debug output\n" );
    fprintf( stdout, "f#(#=0-30) - Set debug filter: only show information of probe id #\n" );
    fprintf( stdout, "f          - Disable debug filter\n" );
    if ( is_sender )
    {
    fprintf( stdout, "k          - Toggle debug output of clock skew information\n" );
    fprintf( stdout, "p          - Show runtime probe information (full record, compact format)\n" );
    fprintf( stdout, "pp         - Show runtime probe information (full record)\n" );
    fprintf( stdout, "c#[-#] #   - Change pairings of s and d probes, type 'c' to check the meaning of #\n" );
    fprintf( stdout, "t#[-#] @   - Change probing techniques, type 't' to check the meaning of # and @\n" );
    fprintf( stdout, "a          - Toggle automatic tuning(of pairing and technique)\n" );
    fprintf( stdout, "v          - Print automatic pairing history(compact format)\n" );
    fprintf( stdout, "vv         - Print automatic pairing history\n" );
    fprintf( stdout, "l          - Show long-term congestion distribution\n" );
    }
    fprintf( stdout, "ph         - Show runtime probe information (half record)\n" );
    fprintf( stdout, "pi         - Show path(hops) information\n" );
    if ( is_sender )
    {
    fprintf( stdout, "pr         - Show path(hops) information of remote side\n" );
    fprintf( stdout, "pb         - Show path(hops) information of both sides with pairing\n" );
    fprintf( stdout, "pm,'       - Show QoM information\n" );
    fprintf( stdout, "th         - Show threshold information\n" );
    fprintf( stdout, "r          - Show MT(measured time) distribution\n" );
    fprintf( stdout, "s          - Save probe configuration to file\n" );
    }
    fprintf( stdout, "***************************************\n" );
    fflush( stdout );
}


void Pong::HelpDebug()
{
    fprintf( stdout, "***** Debug output helper *****\n" );
    fprintf( stdout, "Time values in debug outputs are displayed in milliseconds.\n" );
    fprintf( stdout, "Abbreviations:\n" );
    fprintf( stdout, "  dsrt_fd - difference between the scheduled receiving times of f and d probes\n" );
    fprintf( stdout, "  dsrt_sb - difference between the scheduled receiving times of s and b probes\n" );
    fprintf( stdout, "   drt_fd - difference between the actual receiving times of f and d probes\n" );
    fprintf( stdout, "   drt_sb - difference between the actual receiving times of s and b probes\n" );
    fprintf( stdout, "     dt_f - difference between the scheduled receiving time and the actual receiving time of f probe\n" );
    fprintf( stdout, "     dt_b - difference between the scheduled receiving time and the actual receiving time of b probe\n" );
    fprintf( stdout, "      ATO - Adjust Time Offset\n" );
    fprintf( stdout, "      ATA - ATO Acknowledge\n" );
    fprintf( stdout, "      TDF - Targeted D probe Failure\n" );
    fprintf( stdout, "Debug levels:\n" );
    fprintf( stdout, "    9 - show received packet information\n" );
    fprintf( stdout, "  >=8 - show captured packet information\n" );
    fprintf( stdout, "  >=7 - show half records\n" );
    fprintf( stdout, "  >=6 - show full records\n" );
    fprintf( stdout, "  >=5 - show minor information(e.g., ATO, ATA, threshold, QoM messages)\n" );
    fprintf( stdout, "  >=4 - show medium information(e.g, technique/pairing updates)\n" );
    fprintf( stdout, "  >=3 - show minor anomalies\n" );
    fprintf( stdout, "  >=1 - show major information\n" );
    fprintf( stdout, "Current Debug level = %d\n", ( sock_monitor < 0 ? SavedDebugLevel : DebugLevel ) );
    if ( ProbeToDebug == -1 )
        fprintf( stdout, "Debug filter not enabled\n" );
    else
        fprintf( stdout, "Debug filter enabled, probe id = %d\n", ProbeToDebug );
    fprintf( stdout, "*******************************\n" );
    fflush( stdout );
}


void Pong::ProbeInformation( bool compact )
{
    system( "clear" );
    fprintf( stdout, "***** Runtime probe information (full record) *****\n" );
    double min_f = this->min_f.value / 1000;
    double min_b = this->min_b.value / 1000;
    double d_tbl = delta_t( time_baseline_receiver, time_baseline ) / 1000;
    fprintf( stdout, "Difference between time baselines of sender and receiver = %.2lf(ms)\n", d_tbl );
    fprintf( stdout, "probe_period=%.1lf, probe_num=%d\n", probe_period, probe_num );
    fprintf( stdout, "min_f = %.2lf, min_b = %.2lf (Minimum one way delay of f(b) probe. NOTE: These two values include clock offset between sender and receiver.)\n", min_f, min_b );
    fprintf( stdout, "id\thop_s\thop_d\ttech\t send_s\t send_d\t  min_s\t  min_d\tdsrt_fd\tdsrt_sb" );
    if ( ! compact )
        fprintf( stdout, "\trecv_f\t recv_d\t recv_s\t recv_b\tlast_rid\tbsize" );
    fprintf( stdout, "\n" );
    // fprintf( stdout, "pih_mutex 1\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );
        double send_s = pi->time_offset * 1000;
        double send_d = ph->time_offset_d * 1000;
        double min_s = ( pi->ttl >= 1 ? ph->min_s.value / 1000 : 8888.88 );
        double min_d = ( ph->ttl_d >= 1 ? ph->min_d.value / 1000 : 8888.88 );
        double recv_f = send_s + min_f;
        double recv_d = ( ph->ttl_d >= 1 ? send_d + min_d - d_tbl : 8888.88 );
        double recv_s = ( pi->ttl >= 1 ? send_s + min_s : 8888.88 );
        double recv_b = send_d + min_b - d_tbl;
        double dsrt_fd = ( ph->ttl_d >= 1 ? recv_f - recv_d : 8888.88 );
        double dsrt_sb = ( pi->ttl >= 1 ? recv_s - recv_b : 8888.88 );
        fprintf( stdout, "%2d\t%5d\t%5d\t%4s\t%7.2lf\t%7.2lf\t%7.2lf\t%7.2lf\t%7.2lf\t%7.2lf",
            probe_id, pi->ttl, ph->ttl_d, tech_code[pi->technique], send_s, send_d, min_s, min_d, dsrt_fd, dsrt_sb );
        if ( ! compact )
            fprintf( stdout, "\t%7.2lf\t%7.2lf\t%7.2lf\t%7.2lf\t%8d\t%5d", recv_f, recv_d, recv_s, recv_b, frb->last_round_id, frb->buffer.size() );
        fprintf( stdout, "\n" );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    fprintf( stdout, "(send_s -- scheduled sending time of s and f probes, send_d -- scheduled sending time of d and b probes"
                     ", min_s(d) -- minimum rtt of s(d) probe"
                     ", dsrt_fd -- difference between the scheduled receiving times of f and d probes" 
                     ", dsrt_sb -- difference between the scheduled receiving times of s and b probes" );
    if ( ! compact )
        fprintf( stdout, ", recv_f(d,s,b) -- scheduled receiving time of f,d,s,b probe"
                         ", last_rid -- last_round_id, bsize -- buffer size" ); 
    fprintf( stdout, ", 8888.88 -- n/a)\n" );
    fprintf( stdout, "***************************************************\n" );
    fflush( stdout );
}


void Pong::QoM_Information( void )
{
    system( "clear" );
    fprintf( stdout, "****************** QoM information ******************\n" );
    fprintf( stdout, "id\thop_s\thop_d\ttech\t  QoM_4P   age \t QoM_fsd   age \t QoM_fsb   age \t  QoM_JP   age \n" );
    // fprintf( stdout, "pih_mutex 18\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );

        double Q_JP = ( ph->QoM_JP.value != VALUE_UNDEFINED ? ph->QoM_JP.value : 8.88 );
        double Q_JP_l = ( ph->QoM_JP.value_l != VALUE_UNDEFINED ? ph->QoM_JP.value_l : 8.88 );
        double Q_4P = ( ph->QoM_4P.value != VALUE_UNDEFINED ? ph->QoM_4P.value : 8.88 );
        double Q_4P_l = ( ph->QoM_4P.value_l != VALUE_UNDEFINED ? ph->QoM_4P.value_l : 8.88 );
        double Q_fsd = ( ph->QoM_fsd.value != VALUE_UNDEFINED ? ph->QoM_fsd.value : 8.88 );
        double Q_fsd_l = ( ph->QoM_fsd.value_l != VALUE_UNDEFINED ? ph->QoM_fsd.value_l : 8.88 );
        double Q_fsb = ( ph->QoM_fsb.value != VALUE_UNDEFINED ? ph->QoM_fsb.value : 8.88 );
        double Q_fsb_l = ( ph->QoM_fsb.value_l != VALUE_UNDEFINED ? ph->QoM_fsb.value_l : 8.88 );

        /*
        fprintf( stdout, "%2d\t%5d\t%5d\t%4s\t%4.2lf(%4.2lf)\t%4.2lf(%4.2lf)\t%4.2lf(%4.2lf)\t%4.2lf(%4.2lf)\n",
            probe_id, pi->ttl, ph->ttl_d, tech_code[pi->technique], Q_4P, Q_4P_l, Q_fsd, Q_fsd_l, Q_fsb, Q_fsb_l, Q_JP, Q_JP_l );
        */
        fprintf( stdout, "%2d\t%5d\t%5d\t%4s\t%4.2lf(%4.2lf) %-4d\t%4.2lf(%4.2lf) %-4d\t%4.2lf(%4.2lf) %-4d\t%4.2lf(%4.2lf) %-4d\n",
            probe_id, pi->ttl, ph->ttl_d, tech_code[pi->technique], Q_4P, Q_4P_l, ph->QoM_4P.age, Q_fsd, Q_fsd_l, ph->QoM_fsd.age, Q_fsb, Q_fsb_l, ph->QoM_fsb.age, Q_JP, Q_JP_l, ph->QoM_JP.age );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    fprintf( stdout, "(QoM values in parentheses are the long term values.)\n" );
    fprintf( stdout, "*****************************************************\n" );
    fflush( stdout );
}


void Pong::ThresholdInformation( void )
{
    system( "clear" );
    fprintf( stdout, "****************** Threshold information ******************\n" );
    
    fprintf( stdout, "  Df threshold\t Df distribution\t  Db threshold\t Db distribution\n" );
    fprintf( stdout, "%6.2lf, %6.2lf\t%3d%%, %3d%%, %3d%%\t%6.2lf, %6.2lf\t%3d%%, %3d%%, %3d%%\n",
        th_f.small / 1000, th_f.big / 1000, (int)( th_f.percent_below * 100 + 0.5 ), (int)( th_f.percent_between * 100 + 0.5 ), (int)( th_f.percent_above * 100 + 0.5 ),
        th_b.small / 1000, th_b.big / 1000, (int)( th_b.percent_below * 100 + 0.5 ), (int)( th_b.percent_between * 100 + 0.5 ), (int)( th_b.percent_above * 100 + 0.5 ) );
    
    fprintf( stdout, "id\thop_s\thop_d\ttech\t  Ds threshold\t Ds distribution\t  Dd threshold\t Dd distribution\n" );
    // fprintf( stdout, "pih_mutex 2\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );
        
        fprintf( stdout, "%2d\t%5d\t%5d\t%4s\t%6.2lf, %6.2lf\t%3d%%, %3d%%, %3d%%\t%6.2lf, %6.2lf\t%3d%%, %3d%%, %3d%%\n",
            probe_id, pi->ttl, ph->ttl_d, tech_code[pi->technique],
            ph->th_s.small / 1000, ph->th_s.big / 1000, (int)( ph->th_s.percent_below * 100 + 0.5 ), (int)( ph->th_s.percent_between * 100 + 0.5 ), (int)( ph->th_s.percent_above * 100 + 0.5 ),
            ph->th_d.small / 1000, ph->th_d.big / 1000, (int)( ph->th_d.percent_below * 100 + 0.5 ), (int)( ph->th_d.percent_between * 100 + 0.5 ), (int)( ph->th_d.percent_above * 100 + 0.5 ) );

    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    fprintf( stdout, "(Thresholds are in format of \"small, big\". Distributions are in format of \"below, between, above\".)\n" );

    fprintf( stdout, "*****************************************************\n" );
    fflush( stdout );
}


void Pong::MT_Information( void )
{
    system( "clear" );
    fprintf( stdout, "*********** MT(measured time) distribution ***********\n" );
    
    fprintf( stdout, "--------------- Df distribution ------------------ \t--------------- Db distribution ------------------ \n" );
    fprintf( stdout, "     Below         Betweeen            Above       \t     Below         Betweeen            Above       \n" );
    fprintf( stdout, "pcnt( avg, dev) pcnt( avg, dev) pcnt(   avg,   dev)\tpcnt( avg, dev) pcnt( avg, dev) pcnt(   avg,   dev)\n" );
    fprintf( stdout, "%3d%%(%4.2lf,%4.2lf) %3d%%(%4.2lf,%4.2lf) %3d%%(%6.2lf,%6.2lf)\t%3d%%(%4.2lf,%4.2lf) %3d%%(%4.2lf,%4.2lf) %3d%%(%6.2lf,%6.2lf)\n",
        (int)( th_f.percent_below * 100 + 0.5 ), th_f.avg_below / 1000, th_f.dev_below / 1000,
        (int)( th_f.percent_between * 100 + 0.5 ), th_f.avg_betwn / 1000, th_f.dev_betwn / 1000,
        (int)( th_f.percent_above * 100 + 0.5 ), th_f.avg_above / 1000, th_f.dev_above / 1000,
        (int)( th_b.percent_below * 100 + 0.5 ), th_b.avg_below / 1000, th_b.dev_below / 1000,
        (int)( th_b.percent_between * 100 + 0.5 ), th_b.avg_betwn / 1000, th_b.dev_betwn / 1000,
        (int)( th_b.percent_above * 100 + 0.5 ), th_b.avg_above / 1000, th_b.dev_above / 1000 );
        
    fprintf( stdout, "id\thop_s\t--------------- Ds distribution ------------------ \thop_d\t--------------- Dd distribution ------------------ \n" );
    fprintf( stdout, "  \t     \t     Below         Betweeen            Above       \t     \t     Below         Betweeen            Above       \n" );
    fprintf( stdout, "  \t     \tpcnt( avg, dev) pcnt( avg, dev) pcnt(   avg,   dev)\t     \tpcnt( avg, dev) pcnt( avg, dev) pcnt(   avg,   dev)\n" );
    // fprintf( stdout, "pih_mutex 3\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );
        
        fprintf( stdout, "%2d\t%5d\t%3d%%(%4.2lf,%4.2lf) %3d%%(%4.2lf,%4.2lf) %3d%%(%6.2lf,%6.2lf)\t%5d\t%3d%%(%4.2lf,%4.2lf) %3d%%(%4.2lf,%4.2lf) %3d%%(%6.2lf,%6.2lf)\n",
            probe_id, pi->ttl,
            (int)( ph->th_s.percent_below * 100 + 0.5 ), ph->th_s.avg_below / 1000, ph->th_s.dev_below / 1000,
            (int)( ph->th_s.percent_between * 100 + 0.5 ), ph->th_s.avg_betwn / 1000, ph->th_s.dev_betwn / 1000,
            (int)( ph->th_s.percent_above * 100 + 0.5 ), ph->th_s.avg_above / 1000, ph->th_s.dev_above / 1000,
            ph->ttl_d,
            (int)( ph->th_d.percent_below * 100 + 0.5 ), ph->th_d.avg_below / 1000, ph->th_d.dev_below / 1000,
            (int)( ph->th_d.percent_between * 100 + 0.5 ), ph->th_d.avg_betwn / 1000, ph->th_d.dev_betwn / 1000,
            (int)( ph->th_d.percent_above * 100 + 0.5 ), ph->th_d.avg_above / 1000, ph->th_d.dev_above / 1000 );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );

    fprintf( stdout, "******************************************************\n" );
    fflush( stdout );
}


bool Pong::OutputMTLog( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "MT_log" );
    if ( ( fp = fopen( path, "a" ) ) == NULL )
        return false;

    double timestamp = (double)( current_time.tv_sec - time_baseline.tv_sec ) / 60;

    fprintf( fp, "%.1lf\t", timestamp );
    fprintf( fp, "%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\n",
        th_f.small,
        (int)( th_f.percent_below * 100 + 0.5 ), th_f.avg_below / 1000, th_f.dev_below / 1000,
        (int)( th_f.percent_between * 100 + 0.5 ), th_f.avg_betwn / 1000, th_f.dev_betwn / 1000,
        (int)( th_f.percent_above * 100 + 0.5 ), th_f.avg_above / 1000, th_f.dev_above / 1000,
        th_b.small,
        (int)( th_b.percent_below * 100 + 0.5 ), th_b.avg_below / 1000, th_b.dev_below / 1000,
        (int)( th_b.percent_between * 100 + 0.5 ), th_b.avg_betwn / 1000, th_b.dev_betwn / 1000,
        (int)( th_b.percent_above * 100 + 0.5 ), th_b.avg_above / 1000, th_b.dev_above / 1000 );
    
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );
        
        fprintf( fp, "%.1lf\t", timestamp );
        fprintf( fp, "%d\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\t%d\t%.2lf\t%.2lf\n",
            pi->ttl, ph->th_s.small,
            (int)( ph->th_s.percent_below * 100 + 0.5 ), ph->th_s.avg_below / 1000, ph->th_s.dev_below / 1000,
            (int)( ph->th_s.percent_between * 100 + 0.5 ), ph->th_s.avg_betwn / 1000, ph->th_s.dev_betwn / 1000,
            (int)( ph->th_s.percent_above * 100 + 0.5 ), ph->th_s.avg_above / 1000, ph->th_s.dev_above / 1000,
            ph->ttl_d, ph->th_d.small,
            (int)( ph->th_d.percent_below * 100 + 0.5 ), ph->th_d.avg_below / 1000, ph->th_d.dev_below / 1000,
            (int)( ph->th_d.percent_between * 100 + 0.5 ), ph->th_d.avg_betwn / 1000, ph->th_d.dev_betwn / 1000,
            (int)( ph->th_d.percent_above * 100 + 0.5 ), ph->th_d.avg_above / 1000, ph->th_d.dev_above / 1000 );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );

    fclose( fp );
    return true;
}


bool Pong::OutputQDBin( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "QDBin" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;

    fprintf( fp, "f\t%g\t", th_f.small );
    th_f.bin_count.Output( fp );
    fprintf( fp, "b\t%g\t", th_b.small );
    th_b.bin_count.Output( fp );

    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );
        fprintf( fp, "s%d\t%g\t", pi->ttl, ph->th_s.small );
        ph->th_s.bin_count.Output( fp );
        HopInfo *hi = FindHopInfo( ph->ttl_d, true );
        if ( hi != NULL )
        {
            hi->th_d_small = ph->th_d.small;
            hi->th_d_bin = ph->th_d.bin_count;
        }
    }
    for ( int i=0; i < hop_num_2; i ++ )
    {
        HopInfo *hi = hop_info_2 + i;
        fprintf( fp, "d%d\t%g\t", hi->ttl, hi->th_d_small );
        hi->th_d_bin.Output( fp );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );

    fclose( fp );
    return true;
}


void Pong::ProbeInformation_Half( void )
{
    system( "clear" );
    fprintf( stdout, "***** Runtime probe information (half record) *****\n" );
    fprintf( stdout, "id\thop\ttech\t send_%c\tlast_rid\tbsize", pn_sd );
    if ( ! is_sender )
        fprintf( stdout, "\tip_fwd         \ttdp_state" );
    fprintf( stdout, "\n" );
    // fprintf( stdout, "pih_mutex 4\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        HalfRecordBuffer * hrb = hrbuffer + i;
        int probe_id = hrb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        fprintf( stdout, "%2d\t%3d\t%4s\t%7.2lf\t%8d\t%5d", probe_id, pi->ttl, tech_code[pi->technique], pi->time_offset * 1000, hrb->last_round_id, hrb->buffer.size() );
        if ( ! is_sender )
            fprintf( stdout, "\t%-15s\t%d", inet_ntoa( pi->ip_fwd ), pi->tdp_state );
        fprintf( stdout, "\n" );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    fprintf( stdout, "(send_%c -- scheduled sending time of %c and %c probes, last_rid -- last_round_id, bsize -- buffer size, 8888.88 -- N/A)\n", pn_sd, pn_sd, pn_fb_out ); 
    fprintf( stdout, "***************************************************\n" );
    fflush( stdout );
}


void Pong::HopInformation( bool remote_side )
{
    HopInfo * hop_info = this->hop_info;
    int hop_num = this->hop_num;
    if ( remote_side )
    {
        hop_info = hop_info_2;
        hop_num = hop_num_2;
    }
    fprintf( stdout, "***** Path(hops) information *****\n" );
    if ( remote_side )
        fprintf( stdout, "From %s to %s:\n", targetName, originName );
    else
        fprintf( stdout, "From %s to %s:\n", originName, targetName );
    fprintf( stdout, "hop\tip             \thost_name\n" );
    for ( int i = 0; i < hop_num; i ++ )
    {
        fprintf( stdout, "%-3d\t%-15s\t%s\n", hop_info[i].ttl, inet_ntoa( hop_info[i].ip ), hop_info[i].host_name );
    }
    fprintf( stdout, "**********************************\n" );
    fflush( stdout );
}


bool Pong::OutputProbeConfiguration( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "probe_config" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
    {
        IF_DEBUG( 1, fprintf( stderr, "Error: OutputProbeConfiguration()\n" ) );
        return false;
    }

    WriteProbeConfiguration( fp );

    fclose( fp );

    return true;
}


bool Pong::OutputGlobalInfo( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "global_info" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
    {
        IF_DEBUG( 1, fprintf( stderr, "Error: OutputGlobalInfo()\n" ) );
        return false;
    }

    fprintf( fp, "hop_num\t%d\n", hop_num );
    fprintf( fp, "hop_num_2\t%d\n", hop_num_2 );
    fprintf( fp, "probe_num\t%d\n", probe_num );
    fprintf( fp, "sender_nID\t%d\n", host_id );
    fprintf( fp, "receiver_nID\t%d\n", peer_host_id );
    fclose( fp );

    return true;
}


void Pong::OutputPairing( FILE * fp, const char * title )
{
    // fprintf( stdout, "pih_mutex 5.1\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );

    fprintf( fp, "<TABLE ALIGN=\"center\" TABLE BORDER=3 STYLE=\"BORDER-COLLAPSE: COLLAPSE\" CLASS=\"style2\">\n" );
    fprintf( fp, "<TR><TD COLSPAN=%d BGCOLOR=#FFFF00 ALIGN=\"center\">%s</TD></TR>\n", probe_num + 1, title );
    fprintf( fp, "<TR><TD BGCOLOR=#DDDDDD ALIGN=\"right\">hop(forward path):&nbsp;</TD>\n" );
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        fprintf( fp, "<TD width=30 ALIGN=\"center\">%d</TD>\n", pi->ttl );
    }
    fprintf( fp, "</TR><TR><TD BGCOLOR=#DDDDDD ALIGN=\"right\">hop(backward path):&nbsp;</TD>\n" );
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        ProbeHistory * ph = FindProbeHistory( pi->id );
        assert( ph != NULL );
        fprintf( fp, "<TD ALIGN=\"center\">%d</TD>\n", ph->ttl_d );
    }
    fprintf( fp, "</TR><TR><TD BGCOLOR=#DDDDDD ALIGN=\"right\">technique used:&nbsp;</TD>\n" );
    for ( int i = 0; i < probe_num; i ++ )
    {
        ProbeInfo * pi = probe_info + i;
        fprintf( fp, "<TD ALIGN=\"center\">%s</TD>\n", tech_code[pi->technique] );
    }
    fprintf( fp, "</TR></TABLE>\n" );

    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
}


bool Pong::OutputHopInformation( void )
{
    char path[200];
    FILE * fp;

    strcpy( path, mmt_data_dir );
    strcat( path, "hop_info" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    /*
    fprintf( fp, "<HTML>\n"
                 "<HEAD>\n"
                 "<style type=\"text/css\">\n"
                 "<!--\n"
                 ".style2 {\n"
                 "font-size: 16px;\n"
                 "font-family: \"Times New Roman\", Times, serif;\n"
                 "}\n"
                 "-->\n"
                 "</style>\n"
                 "</HEAD>\n"
                 "<BODY>\n" );
                 */
    fprintf( fp, "<A NAME=\"hop_info\">\n" );
    fprintf( fp, "<TABLE ALIGN=\"center\"><TR VALIGN=\"top\"><TD><TABLE BORDER=3 STYLE=\"BORDER-COLLAPSE: COLLAPSE\" CLASS=\"style2\">\n" );
    fprintf( fp, "<TR ALIGN=\"center\" BGCOLOR=#FFFF00><TD COLSPAN=3>Forward path: From <U>%s</U> to <U>%s</U></TD></TR>\n", originName, targetName );
    fprintf( fp, "<TR ALIGN=\"center\" BGCOLOR=#DDDDDD><TD>hop</TD><TD>ip address</TD><TD>host_name</TD></TR>\n" );
    for ( int i = 0; i < hop_num; i ++ )
    {
        fprintf( fp, "<TR><TD>%d</TD><TD>%s</TD><TD>%s</TD></TR>\n", hop_info[i].ttl, inet_ntoa( hop_info[i].ip ), hop_info[i].host_name );
    }
    fprintf( fp, "</TABLE></TD><TD>\n" );
    
    fprintf( fp, "<TABLE BORDER=3 STYLE=\"BORDER-COLLAPSE: COLLAPSE\" ALIGN=\"center\" CLASS=\"style2\">\n" );
    fprintf( fp, "<TR ALIGN=\"center\" BGCOLOR=#FFFF00><TD COLSPAN=3>Backward path: From <U>%s</U> to <U>%s</U></TD></TR>\n", targetName, originName );
    fprintf( fp, "<TR ALIGN=\"center\" BGCOLOR=#DDDDDD><TD>hop</TD><TD>ip address</TD><TD>host_name</TD></TR>\n" );
    for ( int i = hop_num_2 - 1; i >= 0 ; i -- )
    {
        fprintf( fp, "<TR><TD>%d</TD><TD>%s</TD><TD>%s</TD></TR>\n", hop_info_2[i].ttl, inet_ntoa( hop_info_2[i].ip ), hop_info_2[i].host_name );
    }
    fprintf( fp, "</TABLE></TD></TR></TABLE>\n" );
    
    OutputPairing( fp, "Initial Hop Pairings" );
    /*
    fprintf( fp, "</BODY>\n"
                 "</HTML>\n" );
                 */
    fclose( fp );

    strcat( path, "_head" );
    if ( ( fp = fopen( path, "w" ) ) == NULL )
        return false;
    fprintf( fp, "<TD CLASS=\"style2\" BGCOLOR=#FFFF00 COLSPAN=2>From <U>%s</U> to <U>%s</U> (<A HREF=\"#hop_info\">Hop information</A>)</TD>\n", originName, targetName );
    fclose( fp );

    return true;
}


void Pong::HopInformation_Both( void )
{
    fprintf( stdout, "***** Path(hops) information of both sides*****\n" );
    fprintf( stdout, "From %s to %s:\n", originName, targetName );
    fprintf( stdout, "hop1\tip1            \thop2\tip2            \thost_name1, host_name2\n" );
    HopInfo h_n_a;  // n/a
    h_n_a.ttl = -1;
    h_n_a.ip.s_addr = 0;
    h_n_a.host_name = n_a;
    HopInfo h_tdp;  // targeted d probe
    h_tdp.ttl = TTL_TARGETED_D;
    // fprintf( stdout, "pih_mutex 5\n" );
    assert( pthread_mutex_lock( &pih_mutex ) == 0 );
    for ( int i = 0; i < probe_num; i ++ )
    {
        FullRecordBuffer * frb = frbuffer + i;
        int probe_id = frb->probe_id;
        ProbeInfo * pi = FindProbeInfo( probe_id );
        assert( pi != NULL );
        ProbeHistory * ph = FindProbeHistory( probe_id );
        assert( ph != NULL );
        HopInfo *local, *remote;
        if ( pi->ttl >= 1 )
        {
            local = FindHopInfo( pi->ttl );
            assert( local != NULL );
        }
        else
            local = &h_n_a;
        if ( ph->ttl_d >= 1 )
        {
            if ( ph->ttl_d == TTL_TARGETED_D )
            {
                remote = &h_tdp;
                assert( local != NULL );
                remote->ip = local->ip;
                remote->host_name = local->host_name;
            }
            else
            {
                remote = FindHopInfo( ph->ttl_d, true );
                assert( remote != NULL );
            }
        }
        else
            remote = &h_n_a;
        fprintf( stdout, "%-4d\t%-15s\t", local->ttl, inet_ntoa( local->ip ) );
        fprintf( stdout, "%-4d\t%-15s\t", remote->ttl, inet_ntoa( remote->ip ) );
        fprintf( stdout, "%s, %s\n", local->host_name, remote->host_name );
    }
    assert( pthread_mutex_unlock( &pih_mutex ) == 0 );
    fprintf( stdout, "***********************************************\n" );
    fflush( stdout );
}


bool IsRangeFormat( const char * input )
{
    bool range_format = false;
    int input_len = strlen( input );
    for ( int i = 0; i < input_len; i ++ )
    {
        if ( input[i] < '0' || input[i] > '9' )
        {
            if ( input[i] == '-' )
                range_format = true;
            break;
        }
    }
    return range_format;
}


void Pong::Quit( void )
{
    IF_DEBUG( 1, fprintf( stderr, "%s Send quit command to remote side\n", GetMeasurementTime() ) );
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
    SendCtrCode( CC_QUIT );
    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    IF_DEBUG( 1, fprintf( stderr, "%s Quitting ... \n", GetMeasurementTime() ) );
    done = true;
}


void Pong::StartTraceRA( void )
{
    IF_DEBUG( 1, fprintf( stderr, "%s Send trace RA command to receiver side\n", GetMeasurementTime() ) );
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
    SendCtrCode( CC_TRA );
    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    IF_DEBUG( 1, fprintf( stderr, "%s Start tracing RA ... \n", GetMeasurementTime() ) );

    InitializeTRA();
    phase_done = true;
    phase = PP_TRA;
}


void Pong::ResetRaCount( void )
{
    route_alteration_count = 0;
    for ( int i = 0; i < hop_num; i ++ )
        route_altered[i] = false;
    altered_hop_count = 0;
}


void Pong::InitializeTRA( void )
{
    hop_ips = new vector<int>[hop_num];
    /*
    for ( int i = 0; i < hop_num; i ++ )
        hop_ips[i].push_back( hop_info[i].ip.s_addr );
    */
    ResetRaCount();
}


void Pong::Cancel( void )
{
    IF_DEBUG( 1, fprintf( stderr, "%s Send cancel command to remote side\n", GetMeasurementTime() ) );
    assert( pthread_mutex_lock( &ctr_mutex ) == 0 );
    SendCtrCode( CC_CNCL );
    assert( pthread_mutex_unlock( &ctr_mutex ) == 0 );
    IF_DEBUG( 1, fprintf( stderr, "%s Quitting ... \n", GetMeasurementTime() ) );
    cancel_measurement = true;
    done = true;
}


void Pong::ProcessInteractiveCommand( void )
{
    char input[200];
    if( fgets( input, 200, stdin ) == NULL )
    {
        //IF_DEBUG( 3, fprintf( stderr, "Monitor detached\n" ) );
        sock_monitor = -1;
        return;
    }
    int input_len = strlen( input ) - 1;
    switch ( input[0] )
    {
    case 'q':
        if ( strcmp( input, "quit\n" ) == 0 )
            Quit();
        break;

    case 'x':
        TearDownClient();
        break;

    case 'h':
        SwitchDebug( false );
        if ( input_len > 1 && input[1] == 'd' )
            HelpDebug();
        else
            HelpInteractiveCommand();
        DebugOffWarning();
        break;
    
    case '\n':
        if ( SavedDebugLevel != -99 )
            SwitchDebug( true );
        else
        {
            SwitchDebug( false );
            DebugOffWarning();
        }
        break;

    case 'd':
        if ( input_len > 1 )
        {
            if ( input[1] == 'd' )
            {
                SwitchDebug( false );
                DebugOffWarning();
            }
            else if ( input[1] == 'h' )
            {
                SwitchDebug( false );
                HelpDebug();
                DebugOffWarning();
            }
            else if ( input[1] >= '1' && input[1] <= '9' )
            {
                SwitchDebug( true );
                DebugLevel = input[1] - '0';
            }
            else
            {
                fprintf( stdout, "Invalid debug level '%c'\n", input[1] );
                fflush( stdout );
            }
        }
        else
            SwitchDebug( true );
        break;

            
    case '\'':
        if ( is_sender )
        {
            SwitchDebug( false );
            QoM_Information();
            DebugOffWarning();
        }
        break;

    case 't':
        if ( input_len > 1 && input[1] == 'h' )
        {
            if ( is_sender )
            {
                SwitchDebug( false );
                ThresholdInformation();
                DebugOffWarning();
            }
        }
        else
        {
            if ( is_sender )
            {
                int hop_s_start, hop_s_end;
                char technique[200];
                hop_s_start = hop_s_end = 0;

                bool valid;
                if ( IsRangeFormat( input + 1 ) )
                    valid = ( sscanf( input + 1, "%d-%d %s", &hop_s_start, &hop_s_end, technique ) == 3 );
                else
                {
                    valid = ( sscanf( input + 1, "%d %s", &hop_s_start, technique ) == 2 );
                    hop_s_end = hop_s_start;
                }
                if ( ! valid )
                {
                    fprintf( stdout, "'t' command: Change probing techniques\n" );
                    fprintf( stdout, "     Syntax: t<hop_s_start>[-<hop_s_end>] <technique>\tPossible values of <technique>: 4P, 2P, fsd, fsb\n" );
                    fprintf( stdout, "     e.g.: \"t5 fsd\", \"t5-12 4P\"\n" );
                    fflush( stdout );
                }
                else
                {
                    valid = false;
                    int tech;
                    for ( tech = TECH_2P; tech <= TECH_4P; tech ++ )
                    {
                        if ( strcasecmp( technique, tech_code[tech] ) == 0 )
                        {
                            valid = true;
                            break;
                        }
                    }
                    if ( ! valid )
                    {
                        fprintf( stdout, "Error: invalid technique: %s\n", technique );
                        fflush( stdout );
                    }
                    else
                    {
                        if ( ChangeTechniques( hop_s_start, hop_s_end, tech ) )
                            SwitchDebug( true );
                    }
                }
            }
        }
        break;

    case 'r':
        if ( is_sender )
        {
            if ( input_len > 1 )
            {
                if ( phase == PP_PROBE && input[1] == 'a' )
                {
                    quit_reason = "route_alteration";
                    fprintf( stdout, "WARNING: Start testing route alteration tracing\n" );
                    fflush( stdout );
                    StartTraceRA();
                }
            }
            else
            {
                SwitchDebug( false );
                MT_Information();
                DebugOffWarning();
            }
        }
        break;

    case 'v':
        if ( input_len > 1 )
        {
            if ( input[1] == 'v' )
            {
                if ( is_sender )
                {
            
                    SwitchDebug( false );
                    PrintPairingHistory( false );
                    DebugOffWarning();
                }
            }
        }
        else if ( is_sender )
        {
    
            SwitchDebug( false );
            PrintPairingHistory( true );
            DebugOffWarning();
        }
        break;
        
    case 'p':
        if ( input_len > 1 )
        {
            if ( input[1] == 'h' )
            {
                SwitchDebug( false );
                ProbeInformation_Half();
                DebugOffWarning();
            }
            else if ( input[1] == 'p' )
            {
                if ( is_sender )
                {
                    SwitchDebug( false );
                    ProbeInformation( false );
                    DebugOffWarning();
                }
            }
            else if ( input[1] == 'm' )
            {
                if ( is_sender )
                {
                    SwitchDebug( false );
                    QoM_Information();
                    DebugOffWarning();
                }
            }
            else if ( input[1] == 'i' )
            {
                SwitchDebug( false );
                HopInformation();
                DebugOffWarning();
            }
            else if ( input[1] == 'b' )
            {
                if ( is_sender )
                {
                    SwitchDebug( false );
                    HopInformation_Both();
                    DebugOffWarning();
                }
            }
            else if ( input[1] == 'r' )
            {
                if ( is_sender )
                {
                    SwitchDebug( false );
                    HopInformation( true );
                    DebugOffWarning();
                }
            }
        }
        else if ( is_sender )
        {
            SwitchDebug( false );
            ProbeInformation( true );
            DebugOffWarning();
        }
        break;

    case 'c':
        if ( is_sender )
        {
            int hop_s_start, hop_s_end, hop_d_start;
            hop_s_start = hop_s_end = hop_d_start = 0;

            bool valid;
            if ( IsRangeFormat( input + 1 ) )
                valid = ( sscanf( input + 1, "%d-%d %d", &hop_s_start, &hop_s_end, &hop_d_start ) == 3 );
            else
            {
                valid = ( sscanf( input + 1, "%d %d", &hop_s_start, &hop_d_start ) == 2 );
                hop_s_end = hop_s_start;
            }
            if ( ! valid )
            {
                fprintf( stdout, "'c' command: Change pairings of s and d probes\n" );
                fprintf( stdout, "     Syntax: c<hop_s_start>[-<hop_s_end>] <hop_d_start>\n" );
                fprintf( stdout, "       e.g.: \"c5 14\", \"c5-12 14\"\n" );
                fflush( stdout );
            }
            else
            {
                if ( ChangePairings( hop_s_start, hop_s_end, hop_d_start ) )
                    SwitchDebug( true );
            }
        }
        break;

    case 'l':
        if ( is_sender )
        {
            SwitchDebug( false );
            ShowLCD();
            DebugOffWarning();
        }
        break;

    case 's':
        if ( is_sender )
        {
            SwitchDebug( false );
            SaveProbeConfiguration();
            DebugOffWarning();
        }
        break;

    case 'f':
        if ( input_len > 1 )
        {
            if ( input[1] >= '0' && input[1] <= '9' )
            {
                int probe_id;
                sscanf( input + 1, "%d", &probe_id );
                if ( probe_id >= 0 && probe_id < probe_num )
                    ProbeToDebug = probe_id;
                else
                {
                    fprintf( stdout, "Invalid probe id: %d, use 'ph' command to check valid probe ids\n", probe_id );
                    fflush( stdout );
                }
            }
        }
        else
            ProbeToDebug = -1;
        break;

    case 'a':
        if ( is_sender )
            ToggleATPT();
        break;

    case 'k':
        if ( is_sender )
        {
            suppress_clock_skew_info = ! suppress_clock_skew_info;
            if ( suppress_clock_skew_info )
            {
                fprintf( stdout, "Clock skew information is suppressed\n" );
                fflush( stdout );
            }
        }
        break;
    }
}


void Pong::TearDownClient( void )
{
    fputc( EOF, stdout );
    fflush( stdout );
    sock_monitor = -1;
}


void Pong::MeasureHopnum( void )
{
    if ( is_sender )
    {
        // Send hop number exploration probes
        for ( int i = 0; i < UDP_REPETITION; i ++ )
        {
            SendRouterProbe( 0, 0, ORIGIN_TTL, true );
            usleep( 50000 );    // sleep 50 milliseconds
        }
    }
    
    // Wait for response for a maximum of 6 * RTT
    timeval start_time;
    GetCurrentTime( start_time );
    do {
        usleep( 50000 );   // sleep 50 milliseconds
        if ( hop_num_2 > 0 )
            return;
        GetCurrentTime( current_time );
    } while ( delta_t( start_time, current_time ) < 6 * e2e_rtt );
    
    fprintf( stderr, "Fail to measure number of hops\n" );
    exit(1);
}


void Pong::MeasureHopnum_Respond( const ProbeResult &pr )
{
    assert( pr.probe_id == 99999 && pr.round_id == 99999 );
    if ( hop_num_2 == 0 )
        hop_num_2 = ORIGIN_TTL - pr.ttl;
    if ( ! is_sender )
        SendRouterProbe( 0, 0, ORIGIN_TTL, true );
}


void Pong::ExploreHopInfo( void )
{
    hop_info = new HopInfo[hop_num];
    for ( int i = 0; i < hop_num; i ++ )
    {
        hop_info[i].ttl = i + 1;
        hop_info[i].ip.s_addr = 0;
        hop_info[i].host_name = n_a;
    }
    
    for ( int i = 0; i < UDP_REPETITION; i ++ )
    {
        // Send hop info exploration probes
        for ( int j = 0; j < hop_num; j ++ )
        {
            if ( hop_info[j].ip.s_addr == 0 )
                SendRouterProbe( 0, j, j + 1 );
        }
        usleep( 500000 );  // sleep 500 milliseconds

        // Check exploration response
        int k;
        for ( k = 0; k < hop_num; k ++ )
        {
            if ( hop_info[k].ip.s_addr == 0 )
                break;
        }
        if ( k == hop_num )
            break;
    }

    for ( int i = 0; i < hop_num; i ++ )
    {
        if ( hop_info[i].ip.s_addr != 0 )
        {
            hop_info[i].host_name = strdup( GetName( &(hop_info[i].ip) ) );
            IF_DEBUG( 3, fprintf( stderr, "%d.", hop_info[i].ttl ) );
        }
    }
}


void Pong::ExploreHopInfo_Respond( const ProbeResult &pr )
{
    assert( pr.round_id == 0 && pr.probe_id >= 0 && pr.probe_id < hop_num );
    if ( hop_info[pr.probe_id].ip.s_addr == 0 )
    {
        hop_info[pr.probe_id].ip = pr.icmpSourceAddress;
        IF_DEBUG( 3, fprintf( stderr, "<%d>.", hop_info[pr.probe_id].ttl ) );
    }
}


ProbeUdpPayload::ProbeUdpPayload( void )
{
    // Set signature
    strncpy( signature, "PONG", 4 );
    bzero( padding, 4 );
}


bool ProbeUdpPayload::IsValid( void )
{
    if ( strncmp( signature, "PONG", 4 ) != 0 )
    {
        fprintf( stderr, "signature\n" );
        return false;
    }
    if ( type_code < PL_UDP || type_code > PL_ROUTER )
        return false;
    for ( int i = 0; i < 4; i ++ )
        if ( padding[i] )
            return false;
    return true;
}


void ProbeUdpPayload::SetRouterProbe( void )
{
    type_code = PL_ROUTER;
    round_id = 99999;
    probe_id = 99999;
    timestamp.tv_sec = 0;
    timestamp.tv_usec = 0;
}


void ProbeUdpPayload::SetHopProbe( void )
{
    type_code = PL_HOP;
    round_id = 99999;
    probe_id = 99999;
    timestamp.tv_sec = 0;
    timestamp.tv_usec = 0;
}


void ProbeUdpPayload::SetUdpProbe( int round_id, int probe_id, timeval timestamp )
{
    type_code = PL_UDP;
    this->round_id = round_id;
    this->probe_id = probe_id;
    this->timestamp = timestamp;
}


void ProcessUdpMessage( const char * packet )
{
    MsgHeader * msg_header = (MsgHeader * )packet;
    switch( msg_header->type )
    {
    }
}


// ************** PCAP related functions **************

#ifdef HAVE_PCAP

void Pong::SetupPCAP()
{
    int fileno;
    char PcapErrBuf[PCAP_ERRBUF_SIZE];

    if ( Interface == NULL )
    {
        Interface = pcap_lookupdev( PcapErrBuf );
        if ( Interface == NULL )
        {
            fprintf( stderr, "%s\n", PcapErrBuf );
            exit(1);
        }
    }

    IF_DEBUG( 3, fprintf( stderr, "PCAP interface = %s\n", Interface ) );

    pc = pcap_open_live( Interface, Mtu, 0, 3000, PcapErrBuf );
    if ( pc == NULL )
    {
        fprintf( stderr, "%s\n", PcapErrBuf );
        assert(0);
    }

    bpf_u_int32 netp, maskp;    // net and mask parameters

    if ( pcap_lookupnet( Interface, &netp, &maskp, PcapErrBuf ) < 0 )
    {
        fprintf( stderr, "%s\n", PcapErrBuf );
        assert(0);
    }

    fileno = pcap_fileno( pc );
    if ( fileno < 0 )
    {
        fprintf( stderr, "pcap_fileno() failed\n" );
        exit(1);
    }

#ifdef HAVE_BPF
    // Set "immediate" mode in BPF.  We need this to avoid libpcap
    // waiting the entire timeout period before passing any received
    // packets to us.
    u_int immediate = 1;
    if ( ioctl( fileno, BIOCIMMEDIATE, &immediate ) < 0 )
    {
        perror( "ioctl( BIOCIMMEDIATE )" );
        exit(1);
    }
#endif // HAVE_BPF

    // Initialize packet filter
    // 
    // Note that we will see both inbound and outbound packets
    // with this filter rule.  We *could* write up a new rule 
    // to add a src host predicate.

    char filter[200];
    //sprintf( filter, "(ip proto \\icmp) or (udp port %d and host %s)", UDPRCV_PORT, inet_ntoa( targetAddress ) );
    struct bpf_program fp;      // filter program
    sprintf( filter, "(ip proto \\icmp) or (udp port %d)", UDPRCV_PORT );
    if ( pcap_compile( pc, &fp, filter, 1, maskp ) < 0 )
    {
        fprintf( stderr, "pcap_compile failed\n" );
        exit(1);
    }
    
    if ( pcap_setfilter( pc, &fp ) < 0 )
    {
        fprintf( stderr, "pcap_setfilter failed\n" );
        exit(1);
    }

    // Set packet offset based on datalink type
    //
    // So far we only deals with Ethernet-like datalinks.  We need
    // something like what tcpdump's packet-printing does.
    packet_offset = 0;
    int link_type = pcap_datalink( pc );
    switch ( link_type )
    {
        case DLT_NULL:
            break;

        case DLT_EN10MB:
            packet_offset = sizeof( struct ether_header );
            break;

        default:
            fprintf( stderr, "Unknown datalink layer %d\n", link_type );
            exit(1);
    }

    if ( pthread_create( &tid_pcap, NULL, pcap_process, this ) != 0 )
    {
        fprintf( stderr, "pthread_create(pcap) failed\n" );
        exit(1);
    }

    IF_DEBUG( 3, fprintf( stderr, "Packet capture process installed\n" ) );
}


void * pcap_process( void * arg )
{
    Pong *obj = (Pong *)arg;

    // Run the capture process
    if ( pcap_loop( obj->pc, -1, pcap_callback, (u_char *)arg ) == -1 )
    // NOTE: Returning -2 is not an error. It is due to pcap_breakloop call. 
    {
        fprintf( stderr, "pcap_loop failed\n" );
        exit(1);
    }

    pthread_exit( NULL );

    return NULL;
}

//
// Pong::pcap_callback
//
// Inputs:  Object pointer, pcap_pkthdr pointer, pointer to packet
// data.
//
// Outputs:  None.
//
// pcap callback routine.
//
void pcap_callback( u_char *puc, const struct pcap_pkthdr *ph, const u_char *pd )
{
    Pong *obj = (Pong *)puc;
    unsigned int packetLength = ph->len - obj->packet_offset;
    char *packet = (char *)pd + obj->packet_offset;

    ProbeResult pr;
    pr.tv = ph->ts;
    pr.packet_size = packetLength;

    if ( obj->ParsePacket( packet, pr ) == 0 )
    {
        if ( obj->phase >= PP_PROBE )
        {
            bool is_probe = false;
            switch ( pr.probe_type )
            {
                case probe_fb_out:
                    if ( DebugOn( pr.probe_id ) )
                        IF_DEBUG( 8, fprintf( stderr, "%c(outbound) probe captured\n", obj->pn_fb_out ) );
                    is_probe = true;
                    break;
                case probe_fb_in:
                    if ( DebugOn( pr.probe_id ) )
                        IF_DEBUG( 8, fprintf( stderr, "%c(inbound) probe captured\n", obj->pn_fb_in ) );
                    is_probe = true;
                    break;
                case probe_sd_out:
                    if ( DebugOn( pr.probe_id ) )
                        IF_DEBUG( 8, fprintf( stderr, "%c(outbound) probe captured\n", obj->pn_sd ) );
                    is_probe = true;
                    break;
                case probe_sd_in:
                    if ( DebugOn( pr.probe_id ) )
                        IF_DEBUG( 8, fprintf( stderr, "%c(inbound) probe captured\n", obj->pn_sd ) );
                    is_probe = true;
                    break;
                case probe_icmp_error:
                    if ( DebugOn( pr.probe_id ) )
                        IF_DEBUG( 8, fprintf( stderr, "ICMP message from the other side captured\n" ) );
                    break;
                default:;
            }
            if ( DebugOn( pr.probe_id ) )
                IF_DEBUG( 8, pr.ShowRecord() );
            if ( is_probe )
            {
                if ( obj->phase == PP_PROBE )
                    obj->UpdateHalfRecord( pr );
                else if ( obj->phase == PP_TRA )
                    obj->UpdateRouteAlteration( pr );
            }
        }
        else
        {
            switch ( pr.probe_type )
            {
                case probe_hop_in:
                    obj->MeasureHopnum_Respond( pr );
                    break;

                case probe_sd_in:
                    obj->ExploreHopInfo_Respond( pr );
                    break;

                default:;
            }
        }
    }
}

#endif // HAVE_PCAP

// *********** End of PCAP related functions ***********
